;phantom boot 1.0, 4.4.1998sa
;upgraded to 1.1, 9.8.1998su2143

#include <asm/kernel.h>

[bits 16]
[org 0x7c00]

start:
cli
cld
xor     ax,ax
mov     ds,ax
mov     ax,SETUP_SEG
mov     es,ax
mov     ss,ax
xor     sp,sp			/* stack is long enough ... */

mov     si,BOOT_OFS
sub     di,di
mov     cx,0x80			/* size of bootsect / 4 */
rep     movsd

jmp     SETUP_SEG:@transfered

[org 0]
@transfered:
sti

push    es
pop     ds			/* restoring DS to SETUP_SEG */

mov     si,loading		/* alive message */
call    print

mov     ah,8
sub     dl,dl
int     0x13                    /* get drive parameters */

mov     [max_head],dh		/* decoding and storing ... */
push    cx
and     cl,111111b
mov     [max_sector],cl
pop     cx
and     cl,11000000b
xchg    cl,ch
mov     [max_track],cx

/* loading the whole setup */
mov     ax,SETUP_SEG+0x20
mov     es,ax                   /* destination address = 0x9020:0x0000 */
sub     bx,bx

mov     ah,0x02
mov	al,[setup_sects]	/* setup sector count */
sub     dx,dx                   /* boot drive (a:), head = 0 */
mov     cx,2                    /* track = 0, sector = 2 */
int     0x13			
jnc     @read_ok00
jmp     @read_error		/* long jump (check out!) */
				/* improve error handling!!! */
@read_ok00:
/*
note:	es:di now points to setup image size in sectors 
	ch = 0
	dl = 0
*/

mov	al,[setup_sects]
add	al,2			/* including bootsect + skip to next */
mov	[cur_sector],al		/* updating disk informations */

/* now load the kernel */
mov	ax,KERNEL_SEG
mov	es,ax
sub	bx,bx			/* new rally point (es:bx) */

call	read_it
call	kill_motor

push	ds
pop	es

mov	si,new_line
call	print

#ifdef BOOTDEBUG
xor	ah,ah
int	0x16
#endif

jmp     SETUP_SEG:0x200

/* kernel reader: input es:bx -- usually 0x1000:0 */
read_it:
cmp     word [kernel_sects],0
jnz     @still_reading
ret

@still_reading:
mov	cx,es
mov	dx,cx
and	cx,0xf000
add	cx,0x1000
sub	cx,dx
shr	cx,5			/* divide by 32 */

mov	al,[max_sector]
sub	al,[cur_sector]		/* sectors to be read in current track in al */
inc	al

cbw

cmp	cx,ax
jg	@phys_seg_ok

mov	ax,cx

@phys_seg_ok:

cmp	[kernel_sects],ax
jge	@normal_read

mov	al,[kernel_sects]

@normal_read:
sub	word [kernel_sects],ax
call	read_track
jmp	read_it

@read_error:
mov     si,read_error
call    print
#ifdef BOOTDEBUG
call	hex_byte		/* number of sectors */
jmp     halt
#endif

/* function for reading single track -- input: al / sector count */
read_track:

  pusha
  mov	ax,0x0e2e
  int	0x10
  popa
  
  mov     ah,[cur_sector]
  mov     cx,[cur_track]
  xchg    ch,cl
  or      cl,ah                   /* track and sector set */
  mov     dh,[cur_head]
  mov     ah,0x02
  xor	dl,dl
  pusha
  int     0x13
  popa
  jc      @read_error

#ifdef BOOTDEBUG
  pusha
  call hex_byte
  mov	al,[cur_sector]
  call hex_byte
  mov  al,[cur_head]
  call hex_byte
  mov  al,[cur_track]
  call hex_byte
  popa
#endif
  
  add     [cur_sector],al	/* updating current sector */
  cbw
  shl	  ax,5
  mov	  bx,es
  add	  bx,ax
  mov	  es,bx

#ifdef BOOTDEBUG
  mov	  al,bh
  call	 hex_byte
  mov	 al,bl
  call	hex_byte
#endif

  sub	  bx,bx
  mov     al,[cur_sector]
  cmp     al,[max_sector]
  jg     @inc_head
  ret

@inc_head:
  mov     byte [cur_sector],1   /* sector base is 1 !!! */
  inc     byte [cur_head]
  mov     al,[cur_head]
  cmp     al,[max_head]
  jg     @inc_track
  ret
 
@inc_track:
  mov     byte [cur_head],0
  inc     word [cur_track]
  mov     ax,[cur_track]
  cmp     ax,[max_track]
  jg     @too_long
  ret
    
@too_long:
  mov	si,disk_error
  call	print
  jmp	halt

cur_sector	db	0	/* will be modified ... */
cur_head        db      0	/* on-disk position keepers */
cur_track       dw      0

max_sector      db      0	/* maximal disk's values */
max_head        db      0
max_track       dw      0

loading         db      13,10,'loading phantom',0
read_error      db      13,10,'disk read error occured!',0
disk_error      db      13,10,'fatal disk error!',0
new_line	db	13,10,0

/* function for printing null-terminated string */
print:
        mov     ah,0eh          /* emulate teletype service */
@print_nextchar:
        lodsb
        or      al,al
        je      @print_quit
        int     0x10
        jmp     @print_nextchar
@print_quit:
        ret

putchar:
      mov	ah,0xe
      int	0x10
      ret
      
/* diskette motor killing (light off) */
kill_motor:
	cli
	mov	dx,0x03f2
	mov	al,0x04
	out	dx,al
	sti
	ret

halt:   jmp     halt

#ifdef BOOTDEBUG
/* input: al */
hex_byte:
push ax
shr al,4
call hexi
call putchar
pop ax
and al,0xf
call hexi
call putchar
mov al,' '
call putchar
ret

hexi:
cmp al,9
jg @letter
add al,'0'
ret
@letter:
add al,'a'-0xa
ret
#endif

padding:
        times   504-(padding-start) db 0

	setup_sects	dw	0		/* filled by remake */
	kernel_sects	dw	0
	root_dev	dw	0
        dw      0xAA55
