/*
  kernel/head.S
*/

#include <asm/kernel.h>
#include <asm/mm.h>
#include <asm/task.h>

#ifndef MONO
#define printd(a)	movb	$(a),0xB809E
#else
#define printd(a)	movb	$(a),0xB009E
#endif

.text    
.global _start

_start:
  movl	$(KERNEL_DS),%eax
  movw	%ax,%es
  movw	%ax,%ds
  movw	%ax,%fs
  movw	%ax,%gs
  movw	%ax,%ss
  movl	$(KERNEL_OFFSET),%esp

  printd('P')

  jmp	here
here:

printd('s');

movl	$_end, %eax
subl	$__bss_start, %eax
movl	%eax, %ecx			/* used for further zeroing */
addl	$4095, %eax
shrl	$12, %eax
addl	0x90210+KERNEL_TEXT_PAGES, %eax
movl	%eax, 0x90210+KERNEL_PAGES

/* keep bss space tidy */
movl	$__bss_start, %edi
addl	$0x03, %ecx
shrl	$2, %ecx
xorl	%eax, %eax
rep
stosl

movl	$0x90210,%esi			/* offset of sysinfo */
movl	$sysinfo,%edi
movl	$(SYSINFO_DWORDS),%ecx
rep
movsl					/* putting sysinfo out of the way */

printd('p');
call	setup_paging

printd('g');
lgdt	gdt_descriptor

printd('e');
lidt	idt_descriptor

ljmp	$(KERNEL_CS),$reset_queue
reset_queue:

printd('S');
sti
jmp	start_kernel		/* no return... */

/*
The page directory is situated at 0x1000.
One page table is created at 0x2000 for kernel purposes mapping 1st 4 megs.
Rest (as unmapping unwanted memory) is done by memory manager (mm/page.c).
*/

setup_paging:
movl	$(CR3),%edi
movl	$0x2007,%eax
movl	$0x400,%ecx		/* fill first and zero others */
stosl
xorl	%eax,%eax
rep
stosl

movl	$0x2000,%edi
movl	$7,%eax			/* present, rw, user... */
movl	$0x400,%ecx
next_pentry:
stosl
addl	$0x1000,%eax		/* next page */
loop	next_pentry

movl	$(CR3),%eax
movl	%eax,%cr3
movl	%cr0,%eax
orl	$0x80000000,%eax
movl	%eax,%cr0
ret

.align 4,0x00
null_idt_descriptor:
.word	0
.long	0

.global idt_descriptor
.align 4,0x00
idt_descriptor:
.word	256*8-1
.long	idt

.align 4,0x00
.global idt
idt:
.fill	256,8,0

.align 4,0x00
.word 0

gdt_descriptor:
    .word 8*(5+NR_TASK)-1
    .long gdt
    
.align 4,0x00
.global gdt

gdt:
	.quad	0x0000000000000000
	.quad	0x00CF9A000000FFFF		/* 0x08 code 4gig */
	.quad	0x00CF92000000FFFF		/* 0x10 data 4gig */
	.quad	0x00CFFA000000FFFF		/* 0x18 user code 3gig */
	.quad	0x00CFF2000000FFFF		/* 0x20 user data 3gig */
	.fill	2 * NR_TASK, 8, 0		/* tss0, ldt0, tss1, ... */
