/* todo: extended partitions */
#include <errno.h>
#include <phantom/fs.h>
#include <phantom/blk.h>
#include <phantom/mm.h>

struct dospart
{
  u8 boot, bhead;
  u16 bseccyl;
  u8 system;
  u8 ehead;
  u16 eseccyl;
  u32 begin, size;
};

void
partitions_probe (int dev, char *name)
{
/* blkdev[major].begin and bldev[major].size must be prepared by
device driver (128 * sizeof(int)) each */

  struct dospart *dospart;
  int p, digit_pos;
  int major = MAJOR (dev), minor = MINOR (dev);
  int drive;
  struct bhead *b;
  char tmp[8];

  if (minor != 0 && minor != NR_PARTITIONSPERDRIVE)
    {
      printf ("partitions_probe: invalid device number (minor) %02X:%02X\n",
	      major, minor);
      return;
    }

  strcpy (tmp, name);
  digit_pos = strlen (tmp);
  tmp[digit_pos + 1] = 0;	/* terminator */

  printk ("%s: ", name);
  b = bread (dev, 0);
  if (b)
    {
      dospart = (struct dospart *) (b->data + 0x1BE);
      printk ("%s: ", tmp);

      for (p = 1; p < 5; p++)
	{
	  if (dospart->size)
	    {
	      tmp[digit_pos] = '0' + p;

	      dfs_add (tmp, MKDEV (major, minor + p), S_IRWXU | S_IFBLK);

	      blkdev[major].begin[minor + p] = dospart->begin;
	      blkdev[major].size[minor + p] = dospart->size;

	      if (dospart->size > 20477952)
		printk ("%s:%dG ", tmp, dospart->size >> 21);
	      else
		{
		  if (dospart->size > 19998)
		    printk ("%s:%dM ", tmp, dospart->size >> 11);
		  else
		    printk ("%s:%dK ", tmp, dospart->size >> 1);
		}
	    }
	  dospart++;
	}
      brelse (b);
      printk ("\n");
      return;
    }

  printk ("%s: partition probe failed\n", name);
}
