/*
   dev/block.c
   Phantom block device handling.

   WRITEME:
   * add breada handling, update end_request...
   * implement breada...
   * add block and bunlock for block_read and block_write,
   'improve' returned values for damaged blocks
   (do not return always 0 but tcount - count or what...)
 */

#include <asm/system.h>
#include <phantom/sched.h>
#include <phantom/fs.h>
#include <phantom/blk.h>
#include <errno.h>
#include <conio.h>
#include <phantom/buffer.h>
#include <phantom/lock.h>

#define GETBLKSZ(dev)	(blkdev[MAJOR(dev)].block_size)

struct request request_list[MAX_BLKDEV][NR_REQUEST] = {
  {0,},
};

/* get device size */
int
get_dev_size (int dev)
{
  if (!blkdev[MAJOR (dev)].size)
    {
      printf ("get_dev_size: device size not specified\n");
      return (0);
    }
  return (blkdev[MAJOR (dev)].size[MINOR (dev)]);
}

/* try to run do_request sequence on a device if not already running */
void
start_request (int maj)
{
  if (blkdev[maj].request == NULL)
    return;

  if (blkdev[maj].do_request == NULL)
    {
      struct request *req = blkdev[maj].request;
      struct bhead *bhead = req->bhead;

      printk ("start_request: do_request NULL, dev %02X:%02X\n",
	      MAJOR (req->dev), MINOR (req->dev));

      if (bhead != NULL)
	{
	  bhead->fl_uptodate = 0;
	  bhead->fl_lock = 0;
	}

      return;
    }

  blkdev[maj].do_request (blkdev[maj].request);
}

/* add a new request and start working (if stopped) */
void
add_request (int maj, struct request *rq)
{
  CODE_LOCK (a);
  int i;
  struct request *req;

  SINGLE_CODE (a);

  for (;;)
    {
      for (i = 0; i < NR_REQUEST; i++)
	if (!request_list[maj][i].dev)
	  {			/* ok, found free place */
	    req = &request_list[maj][i];
	    *req = *rq;
	    req->next = NULL;

	    if (blkdev[maj].request_tail != NULL)
	      blkdev[maj].request_tail->next = req;

	    blkdev[maj].request_tail = req;

	    if (blkdev[maj].request != NULL)
	      {
		MULTI_CODE (a);
		return;
	      }

	    blkdev[maj].request = req;
	    MULTI_CODE (a);

	    start_request (maj);

	    return;
	  }
      printk ("Ooops, cannot find free request on %02X:%02X\n"
	      "Sleeping 5 seconds.\n", maj, MINOR (rq->dev));

      sys_sleep (5000);
    }
}

void
do_request (int dev, int blk, int nblk, char *data,
	    struct bhead *bhead, int cmd)
{
  int maj = MAJOR (dev), min = MINOR (dev);
  int r_nr;
  struct request req;

  req.dev = dev;
  req.sec = blk;

  if (blkdev[maj].begin)
    req.sec += blkdev[maj].begin[min];	/* logical device shift */
  else
    {
      if (min)
	{
	  printk
	    ("do_request: begin array not defined for device %02X:%02X\n",
	     MAJOR (dev), MINOR (dev));

	  if (bhead != NULL)
	    {
	      bhead->fl_uptodate = 0;
	      bhead->fl_lock = 0;
	      return;
	    }
	}
    }

  req.nsec = nblk;
  req.data = data;
  req.bhead = bhead;
  req.cmd = cmd;
  req.next = NULL;

  add_request (maj, &req);
}

/* Only ramdisk use this dirt. */
void
rblocks (int dev, int blk, int nblk, char *buff)
{
  do_request (dev, blk, nblk, buff, NULL, READ);
  while (blkdev[MAJOR (dev)].request != NULL);
}

int
block_read (struct file *file, char *buff, size_t count)
{
  struct bhead *b;
  int bsize, tmp0, tmp1, tcount, dev;

  tcount = count;
  dev = file->f_dev;
  bsize = GETBLKSZ (dev);
  tmp0 = file->f_pos % bsize;

#ifdef BLOCKDEBUG
  printf ("block_read: dev %02X:%02X  bsize %d  count %d\n",
	  MAJOR (dev), MINOR (dev), bsize, count);
#endif

  if (tmp0)
    {
#ifdef BLOCKDEBUG
      printf (" head ");
#endif
      b = bread (dev, file->f_pos / bsize);
      if (!b)
	return 0;

      tmp1 = bsize - tmp0;	/* head length */
      if (tmp1 > tcount)
	tmp1 = tcount;

      memcpy ((void *) buff, (void *) (b->data + tmp0), tmp1);
      brelse (b);
      file->f_pos += tmp1;
      if (tmp1 == tcount)
	return (tcount);
    }

  tmp1 = file->f_pos / bsize;
  while (tcount >= bsize)
    {
      b = bread (dev, tmp1++);
      if (!b)
	return 0;
      memcpy (buff, b->data, bsize);
      brelse (b);
      buff += bsize;
      tcount -= bsize;
      file->f_pos += bsize;
    }

  if (tcount)
    {
#ifdef BLOCKDEBUG
      printf ("tail\n");
#endif
      b = bread (dev, tmp1);
      if (!b)
	return 0;
      memcpy (buff, b->data, tcount);
      brelse (b);
      file->f_pos += tcount;
      tcount = 0;
    }

  return count;
}

int
block_write (struct file *file, char *buff, size_t count)
{
  struct bhead *b;
  int bsize, tmp0, tmp1, tcount, dev;

  tcount = count;
  dev = file->f_dev;
  bsize = GETBLKSZ (dev);
  tmp0 = file->f_pos % bsize;

#ifdef BLOCKDEBUG
  printf ("block_write: dev 0x%x  bsize %d  count %d\n", dev, bsize, count);
#endif

  if (tmp0)
    {
#ifdef BLOCKDEBUG
      printf (" head ");
#endif
      b = bread (dev, file->f_pos / bsize);
      if (!b)
	return 0;

      tmp1 = bsize - tmp1;	/* get head size */
      if (tmp1 > tcount)
	tmp1 = tcount;

      memcpy (b->data + tmp0, buff, tmp1);
      b->fl_dirty = 1;
      brelse (b);

      file->f_pos += tmp1;
      if (tcount == tmp1)
	return (tcount);

      buff += tmp1;
      tcount -= tmp1;
    }

  if (tcount >= bsize)
    {
#ifdef BLOCKDEBUG
      printf ("main write\n");
#endif
      b = bempty (dev, file->f_pos / bsize);
      if (!b)
	return 0;
      memcpy (b->data, buff, bsize);
      b->fl_dirty = 1;
      brelse (b);
      tmp0 = tcount - (tcount % bsize);	/* alignment */
      buff += tmp0;
      file->f_pos += tmp0;
      tcount -= tmp0;
    }

  if (tcount)
    {
#ifdef BLOCKDEBUG
      printf ("tail write\n");
#endif
      b = bread (dev, file->f_pos);
      if (!b)
	return 0;
/* block(b); */
      memcpy (b->data, buff, tcount);
      b->fl_dirty = 1;
/* bunlock(b); */
      brelse (b);
      file->f_pos += tcount;
      tcount = 0;
    }

  return (count - tcount);
}

int
block_open (struct file *file)
{
  int maj;

  if (file == NULL)
    {
      printk ("block_open: fatal, NULL file\n");
      return -EINVAL;
    }

  maj = MAJOR (file->f_dev);

  if (maj >= MAX_BLKDEV)
    return -EINVAL;

  if (!blkdev[maj].fops)
    return -ENODEV;

  return 0;
}

void
dev_printk (char *banner, int dev, int sec, int nsec)
{
  printk ("[%02X:%02X]: %s, sec=%d  nsec=%d\n",
	  MAJOR (dev), MINOR (dev), banner, sec, nsec);
}

void
end_request (int dev, int fl_error)
{
  int maj = MAJOR (dev);
  struct bhead *bhead;
  struct request *req = blkdev[maj].request, *req2;

  if (req == NULL)
    {
      printk ("end_request: %02X:%02X has NULL request\n", maj, MINOR (dev));
      return;
    }

#ifdef BLOCKDEBUG
  printk ("end_request: dev=%02X:%02X  fl_error=%d  errors=%d\n",
	  maj, MINOR (dev), fl_error, req->errors);
#endif

  if (fl_error)			/* if error occured skip it or what ? */
    dev_printk ("I/O error", dev, req->sec, req->nsec);

  req->errors = 0;

  bhead = req->bhead;

  if (bhead != NULL)
    {
      bhead->fl_uptodate = fl_error ? 0 : 1;
      bhead->fl_lock = 0;
    }

  req2 = req->next;
  req->dev = 0;			/* free current request */

  if (req2 == NULL)
    {
      blkdev[maj].request_tail = NULL;
      blkdev[maj].request = NULL;
      return;
    }

  blkdev[maj].request = req2;
  start_request (maj);
}
