/*
   dev/cs423x.c
   Phantom CS423X initializer.
 */

#include <asm/io.h>

#define REALLY_SLOW_IO
#define KEY_PORT	0x279

#define SB_BASE		0x220
#define WSS_BASE	0x534
#define FM_BASE		0x388
#define GAME_BASE	0x200
#define IRQ		9
#define DMA1		1
#define DMA2		4

#define CS_OUT(a)		outb (a, KEY_PORT);
#define CS_OUT2(a, b)		{CS_OUT(a);CS_OUT(b);}
#define CS_OUT3(a, b, c)	{CS_OUT(a);CS_OUT(b);CS_OUT(c);}

static unsigned char crystal_key[] =	/* magic initialization sequence */
{
  0x96, 0x35, 0x9a, 0xcd, 0xe6, 0xf3, 0x79, 0xbc,
  0x5e, 0xaf, 0x57, 0x2b, 0x15, 0x8a, 0xc5, 0xe2,
  0xf1, 0xf8, 0x7c, 0x3e, 0x9f, 0x4f, 0x27, 0x13,
  0x09, 0x84, 0x42, 0xa1, 0xd0, 0x68, 0x34, 0x1a
};

int
cs423x_init ()
{
  printf ("CS423X: sb[0x%x] wss[0x%x] irq[%d] dma1[%d] dma2[%d] fm[0x%x] joy[0x%x]\n",
  SB_BASE, WSS_BASE, IRQ, DMA1, DMA2, FM_BASE, GAME_BASE);

  CS_OUT2 (0x06, 0);

/* logical device 0 */
  CS_OUT2 (0x15, 0x00);
  CS_OUT3 (0x47, WSS_BASE >> 8, WSS_BASE & 0xFF);
  CS_OUT3 (0x48, FM_BASE >> 8, FM_BASE & 0xFF);

  CS_OUT3 (0x42, SB_BASE >> 8, SB_BASE & 0xFF);
  CS_OUT2 (0x22, IRQ);		/* SB+WSS IRQ */
  CS_OUT2 (0x2a, DMA1);		/* SB+WSS DMA */

  CS_OUT2 (0x25, DMA2)		/* WSS DMA2 */
  CS_OUT2 (0x33, 0x01);		/* activate device */

/* logical device 1 */
  CS_OUT2 (0x15, 0x01);
  CS_OUT3 (0x47, GAME_BASE >> 8, GAME_BASE & 0xFF);
  CS_OUT2 (0x33, 0x01);		/* activate device */

  CS_OUT (0x79);		/* total activation */
}
