/*
   fs/dev.c
 */

#include <string.h>
#include <errno.h>
#include <phantom/major.h>
#include <phantom/fs.h>
#include <phantom/blk.h>

struct chrdev_st chrdev[MAX_CHRDEV] = {
  {0,},
};

struct blkdev_st blkdev[MAX_BLKDEV] = {
  {0,},
};

int
device_info ()
{
  int i;

  printf ("Listing all registered chrdevs. (MAX_CHRDEV=%d)\n", MAX_CHRDEV);
  for (i = 0; i < MAX_CHRDEV; i++)
    if (chrdev[i].fops)
      printf ("%3d  %s\n", i, chrdev[i].name);

  printf ("Listing all registered blkdevs. (MAX_BLKDEV=%d)\n", MAX_BLKDEV);
  for (i = 0; i < MAX_BLKDEV; i++)
    if (blkdev[i].fops)
      printf ("%3d  %s\n", i, blkdev[i].name);

}

int
register_chrdev (unsigned int major, const char *name,
		 struct file_operations *fops)
{
/* find first hole when major == 0 */
  if (!major)
    {
      for (major = MAX_CHRDEV - 1; major > 0; major--)
	{
	  if (chrdev[major].fops == NULL)
	    {
	      chrdev[major].name = name;
	      chrdev[major].fops = fops;
	      return (major);
	    }
	}
      return (-EBUSY);
    }
  if (major >= MAX_CHRDEV)
    return (-EINVAL);
  if (chrdev[major].fops && chrdev[major].fops != fops)
    return (-EBUSY);
  chrdev[major].name = name;
  chrdev[major].fops = fops;
  return (0);
}

int
register_blkdev (unsigned int major, const char *name,
		 struct file_operations *fops, void (*do_request) (),
		 int block_size)
{
  if (major >= MAX_BLKDEV)
    return -EINVAL;
  if (blkdev[major].fops && blkdev[major].fops != fops)
    return -EBUSY;

  bzero(&blkdev[major], sizeof(blkdev[major]));

  blkdev[major].name = name;
  blkdev[major].fops = fops;
  blkdev[major].do_request = do_request;
  blkdev[major].block_size = block_size;

  return 0;
}

int
unregister_chrdev (unsigned int major, const char *name)
{
  if (major >= MAX_CHRDEV)
    return -EINVAL;
  if (!chrdev[major].fops)
    return -EINVAL;
  if (strcmp (chrdev[major].name, name))
    return -EINVAL;
  bzero ((void *) &chrdev[major], sizeof (struct chrdev_st));
  return 0;
}

int
unregister_blkdev (unsigned int major, const char *name)
{
  if (major >= MAX_BLKDEV)
    return -EINVAL;
  if (!blkdev[major].fops)
    return -EINVAL;
  if (strcmp (blkdev[major].name, name))
    return -EINVAL;
  bzero ((void *) &blkdev[major], sizeof (struct blkdev_st));
  return 0;
}

int
blkdev_open (struct file *file)
{
  file->f_op = blkdev[MAJOR (file->f_dev)].fops;
  if (file->f_op && file->f_op->open)
    return (file->f_op->open (file));
  if (!exist_blkdev (MAJOR (file->f_dev)))
    return (-ENODEV);
  return (0);
}

int
chrdev_open (struct file *file)
{
  file->f_op = chrdev[MAJOR (file->f_dev)].fops;
  if (file->f_op && file->f_op->open)
    return (file->f_op->open (file));
  if (!exist_chrdev (MAJOR (file->f_dev)))
    return (-ENODEV);
  return (0);
}
