/*
   dev/floppy.c

   floppy driver
   ought to work only with 3.5"
   too experimental...

   30051999su2045:  too dangerous, write support added
   support for 2 drives? (dfs)
 */

#include <asm/system.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/timer.h>
#include <string.h>
#include <errno.h>
#include <conio.h>
#include <phantom/major.h>
#include <phantom/fs.h>
#include <phantom/blk.h>
#include <phantom/dfs.h>

struct fd_st
{
  unsigned short size;		/* size in sectors */
  unsigned char sec;		/* sectors per track */
  unsigned char hd;		/* heads */
  unsigned char cyl;		/* cylinders/tracks */
  unsigned char gap, rate, spec0;
};

/* floppy type could be autodetected somehow ... */
#define NR_TYPES	2
struct fd_st floppy[] = {
  {2880, 18, 2, 80, 0x1B, 0, 0xCF},	/* 1.44  3.5" */
  {1440, 9, 2, 80, 0x23, 0x01, 0xDF}	/* 720kB 3.5" */
};

#define FLOPPY_LOGDEVS		2
int floppy_size[FLOPPY_LOGDEVS] = { 2880, 1440 };
int floppy_begin[FLOPPY_LOGDEVS] = { 0, };

#define NR_FD		2
#define NR_FD_RESULT	7

char ST[NR_FD_RESULT];

#define MAJOR_NR	FLOPPY_MAJOR
#define BASIC_DOR	(8|4)	/* dma and controller */

#define FD_IRQ		0x06
#define FD_DMA		0x02

#define FD_DOR		0x3f2	/* digital output reg. */
#define FD_STATUS	0x3f4
#define FD_DATA		0x3f5
#define FD_FDR		0x3f6
#define FD_CONTROL	0x3f7
#define FD_DIR		0x3f7

#define STATUS_BUSY	0x10
#define STATUS_DMA	0x20
#define STATUS_DIR	0x40
#define STATUS_READY	0x80

#define CTRL_MOTOR	0x10

#define CMD_READ		0xE6
#define CMD_WRITE		0xC5
#define CMD_SPECIFY		0x03
#define CMD_RECALIBRATE		0x07
#define CMD_SEEK		0x0F
#define CMD_SENSEI		0x08

/* ST flags */
#define ST0_HNAR	0x04	/* head number at interrupt */
#define ST0_NOTREADY	0x08	/* drive not ready */
#define ST0_ECOAE	0x10
#define ST0_SEEKCOMP	0x20	/* seek completed */
#define ST0_ST0		0x80
#define ST0_ST1		0x40

#define ST1_IAMNF	0x01	/* ID address mark not found */
#define ST1_WP		0x02	/* write protected */
#define ST1_SINF	0x04	/* sector ID not found */
#define ST1_OVERRUN	0x10	/* overrun */
#define ST1_CRCERR	0x20	/* crc error in ID or datafield */
#define ST1_EOC		0x80	/* end of cylinder, read behind ... */

#define ST2_MDAM	0x01	/* missing data address mark */
#define ST2_BCIDNF	0x02	/* bad cylinder, ID not found */
#define ST2_SNF		0x04	/* scan command failed, sector not found */
#define ST2_WC		0x10	/* wrong cylinder - seek error */
#define ST2_CRCERR	0x20	/* data CRC error */
#define ST2_DDAMD	0x40	/* deleted data address mark detected */

int fd_getresult ();

#define MOTOR_KEEP_ALIVE_TIME	3000	/* miliseconds */
#define MOTOR_STOPPED		0
#define MOTOR_RUNNING		(-1)
#define MOTOR_PROC_PERIOD	250	/* refresh rate in ms */

static char *dma_buffer;

volatile int fd_interrupt = 0;		/* interrupt semaphore */
int c_spec0 = -1;
int c_rate = -1;
int c_cyl = -1;
int c_drv = 0;
int change_mask = 0;

volatile int mrun[NR_FD] = { 0, };

static void
fd_killmotor ()
{
  int i;

  for (i = 0; i < NR_FD; i++)
    {
      if (mrun[i] <= 0)
	continue;
      mrun[i] -= MOTOR_PROC_PERIOD;
      if (mrun[i] <= 0)
	{
	  mrun[i] = 0;
	  outb_p (BASIC_DOR | i, FD_DOR);
	  c_drv = i;
	}
    }
}

/* get floppy disk drive */
int
getdrv (int dev)
{
  return (MINOR (dev) / 64);
}

/* get floppy type */
int
getflp (int dev)
{
  return (MINOR (dev) % 64);
}

int
fd_open (struct file *file)
{
  int drv, flp;

  drv = getdrv (file->f_dev);
  flp = getflp (file->f_dev);

  if (drv > 1)
    return (-EIO);

  if (flp >= NR_TYPES)
    return (-EIO);

  return (0);
}

struct file_operations fd_fops = {
  NULL,				/* lseek */
  block_read,
  block_write,
  NULL,				/* readdir */
  NULL,				/* ioctl */
  fd_open,			/* open */
  NULL				/* release */
};

int
fd_outb (char c)
{
  int ret = 100000;
  int status;

  while (ret--)
    {
      status = inb_p (FD_STATUS) & (STATUS_DIR | STATUS_READY);
      if (status == STATUS_READY)
	{
	  outb_p (c, FD_DATA);
	  return (0);
	}
    }
  printf ("floppy.c: write timeout\n");
  return (-1);
}

static void
unexpect_int (int irq_nr, struct regdump *regs)
{
  printf ("floppy.c: floppy unexpected interrupt\n");
  fd_interrupt = 0;
}

static void
rw_int (int irq_nr, struct regdump *regs)
{
#ifdef FLOPPYDEBUG
  printf ("floppy read/write interrupt\n");
#endif
  fd_interrupt = 0;
}

static void
seek_int (int irq_nr, struct regdump *regs)
{
#ifdef FLOPPYDEBUG
  printf ("floppy.c: seek interrupt\n");
#endif
  fd_outb (CMD_SENSEI);
  if (fd_getresult () != 2 || (ST[0] & 0xF8) != ST0_SEEKCOMP)
    {
      printf ("floppy.c: seek failed, trying to recalibrate\n");
/* somehow ... */
    }
  fd_interrupt = 0;
}

int
floppy_access (int drv, int flp, int block, int cmd)
{
  int cyl, hd, sec;

#ifdef FLOPPYDEBUG
  printf ("floppy_access();  block=0x%x\n", block);
#endif

  if (block >= floppy[flp].size)
    return (-EIO);

  sec = (block % floppy[flp].sec) + 1;
  block /= floppy[flp].sec;
  hd = block % floppy[flp].hd;
  block /= floppy[flp].hd;
  cyl = block;

  if (cyl != c_cyl)
    {
      fd_interrupt = 1;
      irq_adjust (FD_IRQ, IRQ_FAST, &seek_int);
      fd_outb (CMD_SEEK);
      fd_outb (hd << 2 | drv);
      fd_outb (cyl);
      while (fd_interrupt)
        schedule();
	
      c_cyl = cyl;
      irq_adjust (FD_IRQ, IRQ_FAST, &rw_int);
    }

  fd_interrupt = 1;

#ifdef FLOPPYDEBUG
  printf ("cyl=%d  hd=%d  sec=%d\n", cyl, hd, sec);
#endif
  cli ();
  dma_setuptransfer (FD_DMA, cmd == CMD_READ ? DMA_READ : DMA_WRITE, 512);
  sti ();

  fd_outb (cmd);
  fd_outb (hd << 2 | drv);	/* hd << 2 | drv */
  fd_outb (cyl);		/* cyl */
  fd_outb (hd);			/* hd */
  fd_outb (sec);		/* sec */
  fd_outb (0x02);		/* sector size = 512 */
  fd_outb (floppy[flp].sec);
  fd_outb (floppy[flp].gap);
  fd_outb (0xFF);		/* count, sector size */

  return (0);
}


int
fd_getresult ()
{
  int ret = 100000, status, i;

  i = 0;
  while (ret--)
    {
      status = inb_p (FD_STATUS) & (STATUS_READY | STATUS_BUSY | STATUS_DIR);
      if (status == STATUS_READY)
	break;
      if (status == (STATUS_DIR | STATUS_BUSY | STATUS_READY))
	{
	  if (i >= NR_FD_RESULT)
	    break;
	  ST[i++] = inb_p (FD_DATA);
	}
    }

  if (!ret)
    {
      printf ("floppy.c: read timeout!\n");
      return (-1);
    }
#ifdef FLOPPYDEBUG
  else
    printf ("ret=%d  i=%d\n", ret, i);
#endif
  return (i);
}

static void
reset_int (int irq_nr, struct regdump *regs)
{
  int i;

  for (i = 0; i < 4; i++)
    {
      fd_outb (CMD_SENSEI);
      (void) fd_getresult ();
    }
}

void
fd_reset ()
{
  int i;

  c_spec0 = -1;
  c_cyl = -1;

  irq_adjust (FD_IRQ, IRQ_FAST, &reset_int);
  cli ();
  outb_p (BASIC_DOR & ~0x04, FD_DOR);
  for (i = 0; i < 1000; i++)
    __asm__ ("nop");
  outb_p (BASIC_DOR, FD_DOR);
  c_drv = 0;

  sti ();
  irq_adjust (FD_IRQ, IRQ_FAST, &unexpect_int);
}

static void
recal_int ()
{
  fd_outb (CMD_SENSEI);
  if (fd_getresult () != 2)
    {
      printf ("floppy.c: recalibration fails\n");
    }
}

/* FIXME ? */
void
fd_recalibrate (int drv, int hd)
{
  irq_adjust (FD_IRQ, IRQ_FAST, &recal_int);
  fd_outb (CMD_RECALIBRATE);
  fd_outb (hd << 2 | drv);
  irq_adjust (FD_IRQ, IRQ_FAST, &unexpect_int);
}

int
fd_changed (u16 dev)
{
  u8 drv = getdrv (dev);

  if (change_mask & (1 << dev))
    return (1);

  return (0);
}

void
fd_dorequest ()
{
  int drv;
  int flp;
  int ecode;
  int i, nr_arg;
  int dev, sec, nsec;
  struct request *req = blkdev[MAJOR_NR].request;

  dev = req->dev;
  sec = req->sec;
  nsec = req->nsec;

  if (nsec > 1)
    {
      printf ("fd_dorequest: nsec=%d>1 not supported yet...\n", nsec);
      end_request (dev, 1);
      return;
    }

  drv = getdrv (dev);
  flp = getflp (dev);

#ifdef FLOPPYDEBUG
  printf ("drive=%d  type=%d  sec=%d  nsec=%d\n", drv, flp, sec, nsec);
#endif

  if (req->cmd != READ && req->cmd != WRITE)
    {
      printk ("fd_dorequest: invalid cmd=%d\n", req->cmd);
      end_request (dev, 1);
      return;
    }

  ecode = 0;

  if (drv > NR_FD)
    {
      printk ("floppy.c: invalid drive!\n");
      end_request (dev, 1);
      return;
    }

  if (floppy[flp].rate != c_rate)
    {
#ifdef FLOPPYDEBUG
      printf ("changing drive rate\n");
#endif
      outb_p (floppy[flp].rate, FD_CONTROL);
      c_rate = floppy[flp].rate;
    }

  if (floppy[flp].spec0 != c_spec0)
    {
#ifdef FLOPPYDEBUG
      printf ("changing spec0\n");
#endif
      c_spec0 = floppy[flp].spec0;
      fd_outb (CMD_SPECIFY);
      fd_outb (c_spec0);
      fd_outb (6);		/* spec1 - HUT 6ms */
    }

  mrun[drv] = MOTOR_RUNNING;
  outb_p (CTRL_MOTOR | BASIC_DOR | drv, FD_DOR);	/* select the floppy */
  c_drv = drv;

  if (inb_p (FD_DIR) & 0x80)
    change_mask |= 1 << drv;
  else
    change_mask &= ~(1 << drv);

  irq_adjust (FD_IRQ, IRQ_FAST, &rw_int);

  while (nsec)
    {
      fd_interrupt = 1;

      if (req->cmd == WRITE)
	memcpy ((void *) dma_buffer, (void *) req->data, 512);

      floppy_access (drv, flp, sec, req->cmd == READ ? CMD_READ : CMD_WRITE);

      while (fd_interrupt)
        schedule();
	
      nr_arg = fd_getresult ();
      switch (ST[0] >> 6)
	{
	case 1:		/* error during execution */
	  if (ST[1] & ST1_WP)	/* drive is write protected */
	    {
	      printf ("floppy.c: write protected!\n");
	      nsec = 1;
	      ecode = -EACCES;
	    }
	  else
	    {
	      if (ST[1] & ST1_CRCERR)
		printf ("floppy.c: CRC error\n");
	      else if (ST[2] & ST2_CRCERR)
		printf ("floppy.c: data CRC error\n");
	      else if (ST[1] & ST1_IAMNF)
		printf ("floppy.c: ID address mark not found\n");
	      else if (ST[1] & ST1_SINF)
		printf ("floppy.c: sector ID not found\n");
	      else if (ST[2] & ST2_WC)
		printf ("floppy.c: wrong cylinder, seek error\n");
	      else if (ST[1] & ST1_EOC)
		printf ("floppy.c: trying to read behind the disk\n");
	      else if (ST[1] & ST1_OVERRUN)
		printf ("floppy.c: overrun\n");
	      else
		{
		  printf ("floppy.c: unmapped error\n  ");
		  for (i = 0; i < nr_arg; i++)
		    printf ("ST%d=0x02%X ", i, ST[i]);
		}
	      nsec = 1;
	      ecode = -EIO;
	    }
	  break;
	case 2:		/* invalid command */
	  printf ("floppy.c: invalid FDC command\n");
	  break;
	case 3:		/* terminated abnormally by change in ready signal */
	  printf
	    ("floppy.c: terminated abnormally by change in ready signal\n");
	  break;
	}

      if (req->cmd == READ)
	memcpy ((void *) req->data, (void *) dma_buffer, 512);

      sec++;
      nsec--;
    }

  mrun[flp] = MOTOR_KEEP_ALIVE_TIME;
  if (ecode)
    {
      printk ("fd_dorequest: ecode=%d\n", ecode);
      end_request (dev, 1);
    }
  else
    end_request (dev, 0);
}

void
floppy_init (void)
{
  int i;
  char name[8] = "fd ";

  if (register_blkdev (MAJOR_NR, "floppy", &fd_fops, fd_dorequest, 512))
    {
      printk ("floppy_init: cannot register\n");
      return;
    }

  dma_buffer = dma_alloc (FD_DMA, 1024);

  if (!dma_buffer)
    {
      printk ("floppy_init: failed, cannot get dma\n");
      return;
    }

  setattr (ATTR (GREEN, BLACK));
  printf ("Floppy driver, dma at 0x%p\n", dma_buffer);
  setattr (DEFAULT_ATTR);

  blkdev[MAJOR_NR].logdevs = FLOPPY_LOGDEVS;
  blkdev[MAJOR_NR].size = floppy_size;
  blkdev[MAJOR_NR].begin = floppy_begin;

  for (i = 0; i < NR_TYPES; i++)
    {
      name[2] = '0' + i;
      dfs_add (name, MKDEV (MAJOR_NR, i), S_IRWXU | S_IFBLK);
    }

  irq_request (FD_IRQ, IRQ_FAST, &unexpect_int, "floppy");
  add_timer (TIMER_FLOPPY, &fd_killmotor, MOTOR_PROC_PERIOD);
}

int
floppy_done ()
{
  int i;

  for (i = 0; i < NR_FD; i++)	/* stop motors */
    outb_p (BASIC_DOR | i, FD_DOR);

  dma_free (FD_DMA);
  irq_free (FD_IRQ);

  return (0);
}
