#include <phantom/ide.h>
#include <sys/types.h>
#include <sys/cdio.h>
#include <errno.h>

#define CAP	ide->u.cd.cap

#define X(a)	cmd.pkt_data[(a)]

int
cd_attach(IDE *ide, IDECTRL *ctrl)
{
  int i, count;
  IDECMD cmd=ZEROCMD;
  
  ide->flags = F_MEDIA_CHANGED;

  X(0)=ATAPI_MODE_SENSE;
  X(2)=ATAPI_CDROM_CAP_PAGE;
  X(7)=sizeof(cdp->cap)>>8;
  X(8)=sizeof(cdp->cap);

    cmd.count=sizeof(CAP);
    cmd.buff=(char *)&CAP;

    for (count = 0 ; count < 5 ; count++) 
        result = ide_issue_immediate(ide, ctrl, cmd);

    /* Some drives have shorter capabilities page. */
    if (cmd.code == 0) 
      {
    	CAP.max_speed = ntohs(CAP.max_speed);
    	CAP.max_vol_levels = ntohs(CAP.max_vol_levels);
    	CAP.buf_size = ntohs(CAP.buf_size);
    	CAP.cur_speed = ntohs(CAP.cur_speed);

        cd_describe(ide);

	return 0;
      }

  return -EIO;
}

void 
cd_describe()
{
    int comma;
    char *mechanism;

    printk("%s: drive speed ", ide->name);
    if (CAP.cur_speed != CAP.max_speed)
        printk("%d - ", CAP.cur_speed * 1000 / 1024);
    printk("%dKB/sec", CAP.max_speed * 1000 / 1024);
    if (CAP.buf_size)
        printk(", %dKB cache\n", CAP.buf_size);

    printk("%s: supported read types:", ide->name);
    comma = 0;
    if (CAP.read_cdr) {
        printk(" CD-R"); comma = 1;
    }
    if (CAP.read_cdrw) {
        printk("%s CD-RW", comma ? "," : ""); comma = 1;
    }
    if (CAP.cd_da) {
        printk("%s CD-DA", comma ? "," : ""); comma = 1;
    }
    if (CAP.method2)
        printk("%s packet track", comma ? "," : "");
    if (CAP.write_cdr || CAP.write_cdrw) {
    	printk("\n%s: supported write types:", ide->name);
        comma = 0;
    	if (CAP.write_cdr) {
            printk(" CD-R" ); comma = 1;
	}
    	if (CAP.write_cdrw) {
            printk("%s CD-RW", comma ? "," : ""); comma = 1;
	}
    	if (cdp->cap.test_write) {
            printk("%s test write", comma ? "," : ""); comma = 1;
	}
    }
    if (CAP.audio_play) {
    	printk("\nwcd%d: Audio: ", ide->name);
    	if (CAP.audio_play)
            printk("play");
    	if (CAP.max_vol_levels)
            printk(", %d volume levels", CAP.max_vol_levels);
    }

    printk("\n%s: Mechanism: ", ide->name);
    switch (cdp->cap.mech) {
    case MST_MECH_CADDY:
        mechanism = "caddy"; break;
    case MST_MECH_TRAY:
        mechanism = "tray"; break;
    case MST_MECH_POPUP:
        mechanism = "popup"; break;
    case MST_MECH_CHANGER:
        mechanism = "changer"; break;
    case MST_MECH_CARTRIDGE:
        mechanism = "cartridge"; break;
    default:
        mechanism = 0; break;
    }
    if (mechanism)
        printk("%s%s", CAP.eject ? "ejectable " : "", mechanism);
    else if (CAP.eject)
        printk("ejectable");

    if (cdp->cap.mech != MST_MECH_CHANGER) {
        printk("\nwcd%d: Medium: ", cdp->lun);
        switch (cdp->cap.medium_type & MST_TYPE_MASK_HIGH) {
        case MST_CDROM:
            printk("CD-ROM "); break;
        case MST_CDR:
            printk("CD-R "); break;
        case MST_CDRW:
            printk("CD-RW "); break;
        case MST_DOOR_OPEN:
            printk("door open"); break;
        case MST_NO_DISC:
            printk("no/blank disc inside"); break;
        case MST_FMT_ERROR:
            printk("medium format error"); break;
	}
        if ((CAP.medium_type & MST_TYPE_MASK_HIGH) < MST_TYPE_MASK_HIGH) 
	  {
            switch (CAP.medium_type & MST_TYPE_MASK_LOW) {
            case MST_DATA_120:
                printk("120mm data disc loaded"); break;
            case MST_AUDIO_120:
                printk("120mm audio disc loaded"); break;
            case MST_COMB_120:
                printk("120mm data/audio disc loaded"); break;
            case MST_PHOTO_120:
                printk("120mm photo disc loaded"); break;
            case MST_DATA_80:
                printk("80mm data disc loaded"); break;
            case MST_AUDIO_80:
                printk("80mm audio disc loaded"); break;
            case MST_COMB_80:
                printk("80mm data/audio disc loaded"); break;
            case MST_PHOTO_80:
                printk("80mm photo disc loaded"); break;
            case MST_FMT_NONE:
                switch (CAP.medium_type & MST_TYPE_MASK_HIGH) {
	        case MST_CDROM:
		    printk("unknown medium"); break;
                case MST_CDR:
                case MST_CDRW:
                    printk("blank medium"); break;
	        }
                break;
            default:
                printk("unknown type=0x%x", CAP.medium_type); break;
            }
	}
    }
    if (CAP.lock)
        printk(CAP.locked ? ", locked" : ", unlocked");
    if (CAP.prevent)
        printk(", lock protected");
    printk("\n");
}

static int
acdopen(dev_t dev, int flags, int fmt, struct proc *p)
{
    int lun = dkunit(dev);
    struct acd *cdp;

    if (lun >= acdnlun || !atapi_request_immediate)
        return ENXIO;
    cdp = acdtab[lun];

    if (!(cdp->flags & F_BOPEN) && !cdp->refcnt) {
        /* Prevent user eject */
        acd_request_wait(cdp, ATAPI_PREVENT_ALLOW,
            0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0);
        cdp->flags |= F_LOCKED;
    }
    if (fmt == S_IFBLK)
        cdp->flags |= F_BOPEN;
    else
        ++cdp->refcnt;
    dev->si_bsize_phys = cdp->block_size;
    if (!(flags & O_NONBLOCK) && acd_read_toc(cdp) && !(flags & FWRITE))
        printf("acd%d: read_toc failed\n", lun);
    return 0;
}

int 
cd_close(IDE *ide, IDECTRL *ctrl)
{
    if (fmt == S_IFBLK)
    	cdp->flags &= ~F_BOPEN;
    else
    	--cdp->refcnt;

    /* Are we the last open ?? */
    if (!(cdp->flags & F_BOPEN) && !cdp->refcnt) {
	/* Yup, do we need to close any written tracks */
        if ((flags & FWRITE) != 0) {
            if ((cdp->flags & F_TRACK_PREPED) != 0) {
                acd_close_track(cdp);
                cdp->flags &= ~(F_TRACK_PREPED | F_TRACK_PREP);
            }
        }
	/* Allow the user eject */
        acd_request_wait(cdp, ATAPI_PREVENT_ALLOW,
        		 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    }
    cdp->flags &= ~F_LOCKED;
    return 0;
}

int cd_docmd(IDE *ide, IDECTRL *ctrl, IDECMD *cmd)
{
    u32 lba, blocks;
    int cmd;
    int count;

    if(cmd->req->cmd==WRITE)
      {
      printk("%s: cannot write...\n", ide->name);
      return 0;
      }
      
    /* Should reject all queued entries if media have changed. */
    if (ide->flags & F_MEDIA_CHANGED)
      {
        sync_dev(ide->dev);
        return;
      }

    if (bp->b_flags & B_READ)
    	lba = bp->b_blkno / (cdp->block_size / DEV_BSIZE);

    blocks = (bp->b_bcount + (cdp->block_size - 1)) / cdp->block_size;
        count = bp->b_bcount;

    XX(0) = ATAPI_READ_BIG;
    XX(2) = lba >> 24;
    XX(3) = lba >> 16;
    XX(4) = lba >> 8;
    XX(5) = lba;
    XX(7) = nblk >> 8;
    XX(8) = nblk;

    return ide_issue(ide, ctrl, cmd);
}

static void 
acd_done(struct acd *cdp, struct buf *bp, int resid, struct atapires result)
{

    if (result.code) {
        atapi_error(cdp->ata, cdp->unit, result);
        bp->b_error = EIO;
        bp->b_flags |= B_ERROR;
    } else {
        bp->b_resid = resid;
        if ((bp->b_flags & B_READ) == B_WRITE)
            cdp->flags |= F_WRITTEN;
    }
    devstat_end_transaction_buf(cdp->device_stats, bp);
    biodone(bp);
    acd_start(cdp);
}

int
cd_issue(IDECMD *cmd)
{
    ide_issue(cmd->ide, cmd->ctrl, cmd);
    if (cmd.code) 
      {
        cd_error(cdp->ata, cdp->unit, result);
        return EIO;
       }
       
    return 0;
}

static __inline void 
lba2msf(int lba, u_char *m, u_char *s, u_char *f)
{
    lba += 150;
    lba &= 0xffffff;
    *m = lba / (60 * 75);
    lba %= (60 * 75);
    *s = lba / 75;
    *f = lba % 75;
}

static __inline int 
msf2lba(u_char m, u_char s, u_char f)
{
    return (m * 60 + s) * 75 + f - 150;
}

int 
acdioctl(dev_t dev, u_long cmd, caddr_t addr, int flag, struct proc *p)
{
    int lun = dkunit(dev);
    struct acd *cdp = acdtab[lun];
    int error = 0;

    if (cdp->flags & F_MEDIA_CHANGED)
        switch (cmd) {
        case CDIOCRESET:
            break;
        default:
            acd_read_toc(cdp);
            acd_request_wait(cdp, ATAPI_PREVENT_ALLOW,
            		     0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0);
            cdp->flags |= F_LOCKED;
            break;
        }
    switch (cmd) {
/*
    case CDIOCRESUME:
        bzero(cdb);
        cdb->cmd = ATAPI_PAUSE;
        cdb->b8 = 0x01;
        return atapi_cmd_wait(cdp->ata, cdp->unit, cdb, 0, 0, timout, 0);
*/
    case CDIOCRESUME:
        return acd_request_wait(cdp, ATAPI_PAUSE, 
				0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0);

    case CDIOCPAUSE:
        return acd_request_wait(cdp, ATAPI_PAUSE, 
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

    case CDIOCSTART:
        return acd_request_wait(cdp, ATAPI_START_STOP,
				1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0);

    case CDIOCSTOP:
        return acd_request_wait(cdp, ATAPI_START_STOP,
				1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

    case CDIOCALLOW:
        acd_select_slot(cdp);
        cdp->flags &= ~F_LOCKED;
        return acd_request_wait(cdp, ATAPI_PREVENT_ALLOW,
            			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

    case CDIOCPREVENT:
        acd_select_slot(cdp);
        cdp->flags |= F_LOCKED;
        return acd_request_wait(cdp, ATAPI_PREVENT_ALLOW,
            			0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0);

    case CDIOCRESET:
        error = suser(p);
        if (error)
            return (error);
        return acd_request_wait(cdp, ATAPI_TEST_UNIT_READY,
            			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

    case CDIOCEJECT:
        if ((cdp->flags & F_BOPEN) && cdp->refcnt)
            return EBUSY;
        return acd_eject(cdp, 0);

    case CDIOCCLOSE:
        if ((cdp->flags & F_BOPEN) && cdp->refcnt)
            return 0;
        return acd_eject(cdp, 1);

    case CDIOREADTOCHEADER:
        if (!cdp->toc.hdr.ending_track)
            return EIO;
        bcopy(&cdp->toc.hdr, addr, sizeof(cdp->toc.hdr));
        break;

    case CDIOREADTOCENTRYS:
	{
            struct ioc_read_toc_entry *te = (struct ioc_read_toc_entry *)addr;
            struct toc *toc = &cdp->toc;
            struct toc buf;
            u_long len;
            u_char starting_track = te->starting_track;

            if (!cdp->toc.hdr.ending_track)
                return EIO;

            if (te->data_len < sizeof(toc->tab[0]) || 
		(te->data_len % sizeof(toc->tab[0])) != 0 || 
		(te->address_format != CD_MSF_FORMAT &&
		te->address_format != CD_LBA_FORMAT))
                return EINVAL;

            if (!starting_track)
                starting_track = toc->hdr.starting_track;
            else if (starting_track == 170) 
                starting_track = toc->hdr.ending_track + 1;
            else if (starting_track < toc->hdr.starting_track ||
                     starting_track > toc->hdr.ending_track + 1)
                return EINVAL;

            len = ((toc->hdr.ending_track + 1 - starting_track) + 1) *
		  sizeof(toc->tab[0]);
            if (te->data_len < len)
                len = te->data_len;
            if (len > sizeof(toc->tab))
                return EINVAL;

            if (te->address_format == CD_MSF_FORMAT) {
                struct cd_toc_entry *entry;

                buf = cdp->toc;
                toc = &buf;
                entry = toc->tab + (toc->hdr.ending_track + 1 -
                    	toc->hdr.starting_track) + 1;
                while (--entry >= toc->tab)
                    lba2msf(ntohl(entry->addr.lba), &entry->addr.msf.minute,
                            &entry->addr.msf.second, &entry->addr.msf.frame);
            }
            return copyout(toc->tab + starting_track - toc->hdr.starting_track,
			   te->data, len);
        }

    case CDIOREADTOCENTRY:
	{
            struct ioc_read_toc_single_entry *te =
            	(struct ioc_read_toc_single_entry *)addr;
            struct toc *toc = &cdp->toc;
            struct toc buf;
            u_char track = te->track;

            if (!cdp->toc.hdr.ending_track)
                return EIO;

            if (te->address_format != CD_MSF_FORMAT && 
		te->address_format != CD_LBA_FORMAT)
                return EINVAL;

            if (!track)
                track = toc->hdr.starting_track;
            else if (track == 170)
                track = toc->hdr.ending_track + 1;
            else if (track < toc->hdr.starting_track ||
                     track > toc->hdr.ending_track + 1)
                return EINVAL;

            if (te->address_format == CD_MSF_FORMAT) {
                struct cd_toc_entry *entry;

                buf = cdp->toc;
                toc = &buf;
                entry = toc->tab + (track - toc->hdr.starting_track);
                lba2msf(ntohl(entry->addr.lba), &entry->addr.msf.minute,
                        &entry->addr.msf.second, &entry->addr.msf.frame);
            }
            bcopy(toc->tab + track - toc->hdr.starting_track,
                  &te->entry, sizeof(struct cd_toc_entry));
        }
	break;

    case CDIOCREADSUBCHANNEL:
	{
            struct ioc_read_subchannel *args =
            	(struct ioc_read_subchannel *)addr;
            struct cd_sub_channel_info data;
            u_long len = args->data_len;
            int abslba, rellba;

            if (len > sizeof(data) ||
                len < sizeof(struct cd_sub_channel_header))
                return EINVAL;

            if (acd_request_wait(cdp, ATAPI_READ_SUBCHANNEL,
				 0, 0x40, 1, 0, 0, 0, 
				 sizeof(cdp->subchan)>>8, sizeof(cdp->subchan),
                		 0,
				 (char *)&cdp->subchan, 
				 sizeof(cdp->subchan)) != 0)
                return EIO;
            if (cdp->flags & F_DEBUG)
                atapi_dump(cdp->ata->ctrlr, cdp->lun, "subchan", &cdp->subchan, 
			   sizeof(cdp->subchan));

            abslba = cdp->subchan.abslba;
            rellba = cdp->subchan.rellba;
            if (args->address_format == CD_MSF_FORMAT) {
                lba2msf(ntohl(abslba),
                    &data.what.position.absaddr.msf.minute,
                    &data.what.position.absaddr.msf.second,
                    &data.what.position.absaddr.msf.frame);
                lba2msf(ntohl(rellba),
                    &data.what.position.reladdr.msf.minute,
                    &data.what.position.reladdr.msf.second,
                    &data.what.position.reladdr.msf.frame);
            } else {
                data.what.position.absaddr.lba = abslba;
                data.what.position.reladdr.lba = rellba;
            }
            data.header.audio_status = cdp->subchan.audio_status;
            data.what.position.control = cdp->subchan.control & 0xf;
            data.what.position.addr_type = cdp->subchan.control >> 4;
            data.what.position.track_number = cdp->subchan.track;
            data.what.position.index_number = cdp->subchan.indx;
            return copyout(&data, args->data, len);
        }

    case CDIOCPLAYMSF:
	{
            struct ioc_play_msf *args = (struct ioc_play_msf *)addr;

            return acd_request_wait(cdp, ATAPI_PLAY_MSF, 0, 0,
                		    args->start_m, args->start_s, args->start_f,
                		    args->end_m, args->end_s, args->end_f,
				    0, 0, 0);
        }

    case CDIOCPLAYBLOCKS:
	{
            struct ioc_play_blocks *args = (struct ioc_play_blocks *)addr;

            return acd_request_wait(cdp, ATAPI_PLAY_BIG, 0,
                		    args->blk>>24 & 0xff, args->blk>>16 & 0xff,
                		    args->blk>>8 & 0xff, args->blk & 0xff,
                		    args->len>>24 & 0xff, args->len>>16 & 0xff,
                		    args->len>>8 & 0xff, args->len & 0xff,
				    0, 0);
        }

    case CDIOCPLAYTRACKS:
	{
            struct ioc_play_track *args = (struct ioc_play_track *)addr;
            u_long start, len;
            int t1, t2;

            if (!cdp->toc.hdr.ending_track)
                return EIO;

            if (args->end_track < cdp->toc.hdr.ending_track + 1)
                ++args->end_track;
            if (args->end_track > cdp->toc.hdr.ending_track + 1)
                args->end_track = cdp->toc.hdr.ending_track + 1;
            t1 = args->start_track - cdp->toc.hdr.starting_track;
            t2 = args->end_track - cdp->toc.hdr.starting_track;
            if (t1 < 0 || t2 < 0)
                return EINVAL;
            start = ntohl(cdp->toc.tab[t1].addr.lba);
            len = ntohl(cdp->toc.tab[t2].addr.lba) - start;

            return acd_request_wait(cdp, ATAPI_PLAY_BIG, 0,
                		    start>>24 & 0xff, start>>16 & 0xff,
                		    start>>8 & 0xff, start & 0xff,
                		    len>>24 & 0xff, len>>16 & 0xff,
                		    len>>8 & 0xff, len & 0xff, 0, 0);
        }

    case CDIOCREADAUDIO:
	{
	    struct ioc_read_audio* args = (struct ioc_read_audio*) addr;
	    int lba, frames, result = 0;
	    u_char *buffer, *ubuf = args->buffer;

	    if (!cdp->toc.hdr.ending_track)
		return EIO;
		
	    if ((frames = args->nframes) < 0)
		return EINVAL;

	    if (args->address_format == CD_LBA_FORMAT)
		lba = args->address.lba;
	    else if (args->address_format == CD_MSF_FORMAT)
	        lba = msf2lba(args->address.msf.minute,
			     args->address.msf.second,
			     args->address.msf.frame);
	    else
		return EINVAL;
#ifndef CD_BUFFER_BLOCKS
#define CD_BUFFER_BLOCKS 8
#endif
            if (!(buffer = malloc(CD_BUFFER_BLOCKS * 2352, M_TEMP, M_NOWAIT)))
                return ENOMEM;

            while (frames > 0) {
                u_char blocks;
                int size;

                blocks = (frames>CD_BUFFER_BLOCKS) ? CD_BUFFER_BLOCKS : frames;
                size = blocks * 2352;

                result = acd_request_wait(cdp, ATAPI_READ_CD, 4,
                                          lba>>24, (lba>>16)&0xff,
                                          (lba>>8)&0xff, lba&0xff, 0, 0,
                                          blocks, 0xf0, buffer, size);
                if (result != 0)
                    break;

                result = copyout(buffer, ubuf, size);
                if (result != 0)
                    break;
                    
                ubuf += size;
                frames -= blocks;
                lba += blocks;
            }

            free(buffer, M_TEMP);
            return result;
        }

    case CDIOCGETVOL:
	{
            struct ioc_vol *arg = (struct ioc_vol *)addr;

            error = acd_request_wait(cdp, ATAPI_MODE_SENSE, 0, CDROM_AUDIO_PAGE,
                		     0, 0, 0, 0, 
				     sizeof(cdp->au)>>8, sizeof(cdp->au), 0,
                		     (char *)&cdp->au, sizeof(cdp->au));
            if (error)
                return error;
            if (cdp->flags & F_DEBUG)
                atapi_dump(cdp->ata->ctrlr, cdp->lun, "au", &cdp->au,
			   sizeof(cdp->au));
            if (cdp->au.page_code != CDROM_AUDIO_PAGE)
                return EIO;
            arg->vol[0] = cdp->au.port[0].volume;
            arg->vol[1] = cdp->au.port[1].volume;
            arg->vol[2] = cdp->au.port[2].volume;
            arg->vol[3] = cdp->au.port[3].volume;
        }
        break;

    case CDIOCSETVOL:
	{
            struct ioc_vol *arg = (struct ioc_vol *)addr;

            error = acd_request_wait(cdp, ATAPI_MODE_SENSE, 0, CDROM_AUDIO_PAGE,
                		     0, 0, 0, 0, 
				     sizeof(cdp->au)>>8, sizeof(cdp->au), 0,
                		     (char *)&cdp->au, sizeof(cdp->au));
            if (error)
                return error;
            if (cdp->flags & F_DEBUG)
                atapi_dump(cdp->ata->ctrlr, cdp->lun, "au", &cdp->au, 
			   sizeof(cdp->au));
            if (cdp->au.page_code != CDROM_AUDIO_PAGE)
                return EIO;

            error = acd_request_wait(cdp, ATAPI_MODE_SENSE, 0, 
				     CDROM_AUDIO_PAGE_MASK, 0, 0, 0, 0, 
				     sizeof(cdp->aumask)>>8,sizeof(cdp->aumask),
				     0,
				     (char *)&cdp->aumask, sizeof(cdp->aumask));
            if (error)
                return error;
            if (cdp->flags & F_DEBUG)
                atapi_dump(cdp->ata->ctrlr, cdp->lun, "mask", &cdp->aumask, 
			   sizeof(cdp->aumask));

            cdp->au.data_length = 0;
            cdp->au.port[0].channels = CHANNEL_0;
            cdp->au.port[1].channels = CHANNEL_1;
            cdp->au.port[0].volume = arg->vol[0] & cdp->aumask.port[0].volume;
            cdp->au.port[1].volume = arg->vol[1] & cdp->aumask.port[1].volume;
            cdp->au.port[2].volume = arg->vol[2] & cdp->aumask.port[2].volume;
            cdp->au.port[3].volume = arg->vol[3] & cdp->aumask.port[3].volume;
            return acd_request_wait(cdp, ATAPI_MODE_SELECT, 0x10,
                		    0, 0, 0, 0, 0, 
				    sizeof(cdp->au)>>8, sizeof(cdp->au),
                		    0, (char *)&cdp->au, -sizeof(cdp->au));
        }

    case CDIOCSETPATCH:
	{
            struct ioc_patch *arg = (struct ioc_patch *)addr;

            return acd_setchan(cdp, arg->patch[0], arg->patch[1],
                	       arg->patch[2], arg->patch[3]);
        }

    case CDIOCSETMONO:
        return acd_setchan(cdp, CHANNEL_0|CHANNEL_1, CHANNEL_0|CHANNEL_1, 0, 0);

    case CDIOCSETSTEREO:
        return acd_setchan(cdp, CHANNEL_0, CHANNEL_1, 0, 0);

    case CDIOCSETMUTE:
        return acd_setchan(cdp, 0, 0, 0, 0);

    case CDIOCSETLEFT:
        return acd_setchan(cdp, CHANNEL_0, CHANNEL_0, 0, 0);

    case CDIOCSETRIGHT:
        return acd_setchan(cdp, CHANNEL_1, CHANNEL_1, 0, 0);

    case CDRIOCNEXTWRITEABLEADDR:
	{
	    struct acd_track_info track_info;

	    if ((error = acd_read_track_info(cdp, 0xff, &track_info)))
		break;
	    if (!track_info.nwa_valid)
		return EINVAL;
	    cdp->next_writeable_lba = track_info.next_writeable_addr;
	    *(int*)addr = track_info.next_writeable_addr;
	}
	break;
 
    case WORMIOCPREPDISK:
        {
            struct wormio_prepare_disk *w = (struct wormio_prepare_disk *)addr;

            if (w->dummy != 0 && w->dummy != 1)
                error = EINVAL;
            else {
                error = acd_open_disk(cdp, w->dummy);
                if (error == 0) {
                    cdp->flags |= F_DISK_PREPED;
                    cdp->dummy = w->dummy;
                    cdp->speed = w->speed;
                }
            }
        }
        break;

    case WORMIOCPREPTRACK:
        {
            struct wormio_prepare_track *w =(struct wormio_prepare_track *)addr;

            if (w->audio != 0 && w->audio != 1)
                error = EINVAL;
            else if (w->audio == 0 && w->preemp)
                error = EINVAL;
            else if ((cdp->flags & F_DISK_PREPED) == 0) {
                error = EINVAL;
                printf("wcd%d: sequence error (PREP_TRACK)\n", cdp->lun);
            } else {
                cdp->flags |= F_TRACK_PREP;
                cdp->preptrack = *w;
            }
        }
        break;

    case WORMIOCFINISHTRACK:
        if ((cdp->flags & F_TRACK_PREPED) != 0)
            error = acd_close_track(cdp);
        cdp->flags &= ~(F_TRACK_PREPED | F_TRACK_PREP);
        break;

    case WORMIOCFIXATION:
        {
            struct wormio_fixation *w =
            (struct wormio_fixation *)addr;

            if ((cdp->flags & F_WRITTEN) == 0)
                error = EINVAL;
            else if (w->toc_type < 0 /* WORM_TOC_TYPE_AUDIO */ ||
                w->toc_type > 4 /* WORM_TOC_TYPE_CDI */ )
                error = EINVAL;
            else if (w->onp != 0 && w->onp != 1)
                error = EINVAL;
            else {
                /* no fixation needed if dummy write */
                if (cdp->dummy == 0)
                    error = acd_close_disk(cdp);
                cdp->flags &=
                    ~(F_WRITTEN|F_DISK_PREPED|F_TRACK_PREP|F_TRACK_PREPED);
            }
        }
        break;

    case CDRIOCBLANK:
        return acd_blank_disk(cdp);

    default:
        return ENOTTY;
    }
    return error;
}

static int 
acd_read_toc(struct acd *cdp)
{
    int ntracks, len;
    struct atapires result;

    bzero(&cdp->toc, sizeof(cdp->toc));
    bzero(&cdp->info, sizeof(cdp->info));

    acd_select_slot(cdp);

    result = atapi_request_wait(cdp->ata, cdp->unit, ATAPI_TEST_UNIT_READY,
        			0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0);

    if (result.code == RES_ERR &&
        (result.error & AER_SKEY) == AER_SK_UNIT_ATTENTION) {
        cdp->flags |= F_MEDIA_CHANGED;
    	cdp->flags &= ~(F_WRITTEN|F_TRACK_PREP|F_TRACK_PREPED);
        result = atapi_request_wait(cdp->ata, cdp->unit, ATAPI_TEST_UNIT_READY,
				    0, 0, 0, 0, 0, 0, 0, 0,
				    0, 0, 0, 0, 0, 0, 0, 0, 0);
    }

    if (result.code) {
        atapi_error(cdp->ata, cdp->unit, result);
        return EIO;
    }

    cdp->flags &= ~F_MEDIA_CHANGED;

    len = sizeof(struct ioc_toc_header) + sizeof(struct cd_toc_entry);
    if (acd_request_wait(cdp, ATAPI_READ_TOC, 0, 0, 0, 0, 0, 0,
        		 len>>8, len & 0xff, 0, (char *)&cdp->toc, len) != 0) {
        bzero(&cdp->toc, sizeof(cdp->toc));
        return 0;
    }
    ntracks = cdp->toc.hdr.ending_track - cdp->toc.hdr.starting_track + 1;
    if (ntracks <= 0 || ntracks > MAXTRK) {
        bzero(&cdp->toc, sizeof(cdp->toc));
        return 0;
    }

    len = sizeof(struct ioc_toc_header) + ntracks * sizeof(struct cd_toc_entry);
    if (acd_request_wait(cdp, ATAPI_READ_TOC, 0, 0, 0, 0, 0, 0,
        		 len>>8, len & 0xff, 0, (char *)&cdp->toc, len) & 0xff){
        bzero(&cdp->toc, sizeof(cdp->toc));
        return 0;
    }

    cdp->toc.hdr.len = ntohs(cdp->toc.hdr.len);

    if (acd_request_wait(cdp, ATAPI_READ_CAPACITY, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			 (char *)&cdp->info, sizeof(cdp->info)) != 0)
        bzero(&cdp->info, sizeof(cdp->info));

    cdp->toc.tab[ntracks].control = cdp->toc.tab[ntracks - 1].control;
    cdp->toc.tab[ntracks].addr_type = cdp->toc.tab[ntracks - 1].addr_type;
    cdp->toc.tab[ntracks].track = 170;
    cdp->toc.tab[ntracks].addr.lba = cdp->info.volsize;

    cdp->info.volsize = ntohl(cdp->info.volsize);
    cdp->info.blksize = ntohl(cdp->info.blksize);

    if (cdp->info.volsize && cdp->toc.hdr.ending_track
        && (cdp->flags & F_DEBUG)) {
        printf("wcd%d: ", cdp->lun);
        if (cdp->toc.tab[0].control & 4)
            printf("%ldMB ", cdp->info.volsize / 512);
        else
            printf("%ld:%ld audio ", cdp->info.volsize / 75 / 60,
                cdp->info.volsize / 75 % 60);
        printf("(%ld sectors (%ld bytes)), %d tracks\n", 
	    cdp->info.volsize, cdp->info.blksize,
            cdp->toc.hdr.ending_track - cdp->toc.hdr.starting_track + 1);
    }
    return 0;
}

/*
 * Set up the audio channel masks.
 */
static int 
acd_setchan(struct acd *cdp, u_char c0, u_char c1, u_char c2, u_char c3)
{
    int error;

    error = acd_request_wait(cdp, ATAPI_MODE_SENSE, 0, CDROM_AUDIO_PAGE,
        		     0, 0, 0, 0, 
			     sizeof(cdp->au)>>8, sizeof(cdp->au), 0,
        		     (char *)&cdp->au, sizeof(cdp->au));
    if (error)
        return error;
    if (cdp->flags & F_DEBUG)
        atapi_dump(cdp->ata->ctrlr, cdp->lun, "au", &cdp->au, sizeof(cdp->au));
    if (cdp->au.page_code != CDROM_AUDIO_PAGE)
        return EIO;

    cdp->au.data_length = 0;
    cdp->au.port[0].channels = c0;
    cdp->au.port[1].channels = c1;
    cdp->au.port[2].channels = c2;
    cdp->au.port[3].channels = c3;
    return acd_request_wait(cdp, ATAPI_MODE_SELECT, 0x10,
        		    0, 0, 0, 0, 0, 
			    sizeof(cdp->au)>>8, sizeof(cdp->au), 0,
			    (char *)&cdp->au, -sizeof(cdp->au));
}

static int 
acd_eject(struct acd *cdp, int close)
{
    struct atapires result;

    acd_select_slot(cdp);

    result = atapi_request_wait(cdp->ata, cdp->unit, ATAPI_START_STOP, 1,
			        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

    if (result.code == RES_ERR &&
        ((result.error & AER_SKEY) == AER_SK_NOT_READY ||
        (result.error & AER_SKEY) == AER_SK_UNIT_ATTENTION)) {
        int err;

        if (!close)
            return 0;
        err = acd_request_wait(cdp, ATAPI_START_STOP, 0, 0, 0, 3,
			       0, 0, 0, 0, 0, 0, 0);
        if (err)
            return err;

        acd_read_toc(cdp);

        acd_request_wait(cdp, ATAPI_PREVENT_ALLOW, 0, 0, 0, 1,
			 0, 0, 0, 0, 0, 0, 0);
        cdp->flags |= F_LOCKED;
        return 0;
    }
    if (result.code) {
        atapi_error(cdp->ata, cdp->unit, result);
        return EIO;
    }
    if (close)
        return 0;

    tsleep((caddr_t) &lbolt, PRIBIO, "wcdej1", 0);
    tsleep((caddr_t) &lbolt, PRIBIO, "wcdej2", 0);

    acd_request_wait(cdp, ATAPI_PREVENT_ALLOW, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    cdp->flags &= ~F_LOCKED;

    cdp->flags |= F_MEDIA_CHANGED;
    cdp->flags &= ~(F_WRITTEN|F_TRACK_PREP|F_TRACK_PREPED);
    return acd_request_wait(cdp, ATAPI_START_STOP, 0, 0, 0, 2,
			    0, 0, 0, 0, 0, 0, 0);
}

static void
acd_select_slot(struct acd *cdp)
{
    if (cdp->slot < 0 || cdp->changer_info->current_slot == cdp->slot)
        return;

    /* Unlock (might not be needed but its cheaper than asking) */
    acd_request_wait(cdp, ATAPI_PREVENT_ALLOW, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

    /* Unload the current media from player */
    acd_request_wait(cdp, ATAPI_LOAD_UNLOAD, 0, 0, 0, 2,
		     0, 0, 0, cdp->changer_info->current_slot, 0, 0, 0);

    /* load the wanted slot */
    acd_request_wait(cdp, ATAPI_LOAD_UNLOAD, 0, 0, 0, 3,
		     0, 0, 0, cdp->slot, 0, 0, 0);

    cdp->changer_info->current_slot = cdp->slot;

    /* Lock the media if needed */
    if (cdp->flags & F_LOCKED) {
        acd_request_wait(cdp, ATAPI_PREVENT_ALLOW, 0, 0, 0, 1,
			 0, 0, 0, 0, 0, 0, 0);
    }
}

static int
acd_open_disk(struct acd *cdp, int test)
{
    cdp->next_writeable_lba = 0;
    return 0;
}

static int
acd_close_disk(struct acd *cdp)
{
    return acd_request_wait(cdp, ATAPI_CLOSE_TRACK, 0x00,
        		    0x02, 0, 0, 0/*track*/, 0, 0, 0, 0, 0, 0);
}

static int
acd_open_track(struct acd *cdp, struct wormio_prepare_track *ptp)
{
    struct write_param param;
    struct atapires result;

    result = atapi_request_wait(cdp->ata, cdp->unit, ATAPI_MODE_SENSE,
        			0, 0x05, 0, 0, 0, 0, 
				sizeof(param)>>8, sizeof(param),
        			0, 0, 0, 0, 0, 0, 0,
        			(char *)&param, sizeof(param));

    if (cdp->flags & F_DEBUG)
        atapi_dump(cdp->ata->ctrlr, cdp->lun, "0x05", &param, sizeof(param));

    if (result.code == RES_UNDERRUN)
        result.code = 0;

    if (result.code) {
        atapi_error(cdp->ata, cdp->unit, result);
        return EIO;
    }
    param.page_code = 0x05;
    param.page_length = 0x32;
    param.test_write = cdp->dummy ? 1 : 0;
    param.write_type = CDR_WTYPE_TRACK;

    switch (ptp->audio) {
/*    switch (data_type) { */

    case 0:
/*    case CDR_DATA: */
	cdp->block_size = 2048;
    	param.track_mode = CDR_TMODE_DATA;
    	param.data_block_type = CDR_DB_ROM_MODE1;
    	param.session_format = CDR_SESS_CDROM;
	break;

    default:
/*    case CDR_AUDIO: */
	cdp->block_size = 2352;
	if (ptp->preemp)
    	    param.track_mode = CDR_TMODE_AUDIO;
	else
    	    param.track_mode = 0;
    	param.data_block_type = CDR_DB_RAW;
    	param.session_format = CDR_SESS_CDROM;
	break;

/*
    case CDR_MODE2:
    	param.track_mode = CDR_TMODE_DATA;
    	param.data_block_type = CDR_DB_ROM_MODE2;
    	param.session_format = CDR_SESS_CDROM;
	break;

    case CDR_XA1:
    	param.track_mode = CDR_TMODE_DATA;
    	param.data_block_type = CDR_DB_XA_MODE1;
    	param.session_format = CDR_SESS_CDROM_XA;
	break;

    case CDR_XA2:
    	param.track_mode = CDR_TMODE_DATA;
    	param.data_block_type = CDR_DB_XA_MODE2_F1;
    	param.session_format = CDR_SESS_CDROM_XA;
	break;

    case CDR_CDI:
    	param.track_mode = CDR_TMODE_DATA;
    	param.data_block_type = CDR_DB_XA_MODE2_F1;
    	param.session_format = CDR_SESS_CDI;
	break;
*/
    }

    param.multi_session = CDR_MSES_NONE;
    param.fp = 0;
    param.packet_size = 0;

    if (cdp->flags & F_DEBUG)
        atapi_dump(cdp->ata->ctrlr, cdp->lun, "0x05", &param, sizeof(param));

    result = atapi_request_wait(cdp->ata, cdp->unit, ATAPI_MODE_SELECT,
        			0x10, 0, 0, 0, 0, 0, 
				sizeof(param)>>8, sizeof(param),
        			0, 0, 0, 0, 0, 0, 0,
        			(char *)&param, -sizeof(param));

    if (result.code == RES_UNDERRUN)
        result.code = 0;

    if (result.code) {
        atapi_error(cdp->ata, cdp->unit, result);
        return EIO;
    }
    return 0;
}

static int
acd_close_track(struct acd *cdp)
{
    return acd_request_wait(cdp, ATAPI_SYNCHRONIZE_CACHE, 0,
        		    0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
}

static int
acd_read_track_info(struct acd *cdp, int lba, struct acd_track_info *info)
{
    int error;

    error = acd_request_wait(cdp, ATAPI_READ_TRACK_INFO, 0x01,
                             lba>>24, (lba>>16)&0xff,
                             (lba>>8)&0xff, lba&0xff,
		 	     0, 
			     sizeof(*info)>>8, sizeof(*info), 0,
			     (char *)info, sizeof(*info));
    if (error)
	return error;
    info->track_start_addr = ntohl(info->track_start_addr);
    info->next_writeable_addr = ntohl(info->next_writeable_addr);
    info->free_blocks = ntohl(info->free_blocks);
    info->fixed_packet_size = ntohl(info->fixed_packet_size);
    info->track_length = ntohl(info->track_length);
    return 0;
}

int
cd_blank_disk(IDE *ide, IDECTRL *ctrl)
{
    IDECMD *cmd=ZEROCMD;

    cmd.ide=ide;
    cmd.ctrl=ctrl;

    X(0)=0xA1;
    X(1)=0x01;
        
    ide->flags |= F_MEDIA_CHANGED;
    ide->flags &= ~(F_WRITTEN|F_TRACK_PREP|F_TRACK_PREPED);

    return cd_issue(&cmd);
}

void
cd_error(IDE *ide, IDECTRL *ctrl, IDECMD *cmd)
{
    if (cmd.code != RES_ERR) {
        printk("%s: ERROR %d, status=0x%x, error=0x%x\n", 
	       ide->name, result.code, result.status, result.error);
        return;
    }

    switch (cmd.error & AER_SKEY) {
    case AER_SK_NOT_READY:
            printk("%s: not ready\n", ide->name);
        break;

    case AER_SK_BLANK_CHECK:
            printk("%s: blank check\n", ide->name);
        break;

    case AER_SK_MEDIUM_ERROR:
            printk("%s: medium error\n", ide->name);
        break;

    case AER_SK_HARDWARE_ERROR:
        printk("%s: hardware error\n", ide->name);
        break;

    case AER_SK_ILLEGAL_REQUEST:
        printf("%s: illegal request\n", ide->name);
        break;

    case AER_SK_UNIT_ATTENTION:
            printk("%s: unit attention\n", ide->name);
        break;

    case AER_SK_DATA_PROTECT:
            printk("%s: reading protected data\n", ide->name);
        break;

    case AER_SK_ABORTED_COMMAND:
            printk("%s: command aborted\n", ide->name);
        break;

    case AER_SK_MISCOMPARE:
            printk("%s: data don't match medium\n", ide->name);
        break;

    default:
            printk("%s: unknown error, status=0x%x, error=0x%x\n", 
	       	   ide->name, result.status, result.error);
    }
}
