/* TODO: make all cmds dynamic */

#include <phantom/ide.h>
#include <sys/types.h>
#include <sys/cdio.h>
#include <errno.h>

#define X(a)  cmd.pkt_data[(a)]
#define XX(a)  cmd->pkt_data[(a)]
#define CAP ide->u.floppy.cap

void zip_error (IDE * ide, IDECTRL * ctrl, IDECMD * cmd);

void
zip_callback (IDECMD * cmd)
{
  printk ("zip_callback: cmd=0x%x\n", cmd);
  if (cmd->code != RES_OK)
    {
      zip_error (cmd->ide, cmd->ctrl, cmd);
      ide_end_request (cmd, 1);

      return;
    }

  printk ("zip_callback: request successfully ended\n");
  ide_end_request (cmd, 0);
}

void
zip_prevent (IDE * ide, int fl_prevent)
{
  IDECMD *cmd;

  cmd = ide_newcmd (ide, zip_callback);

  if (cmd == NULL)
    return;

  cmd->retries = 3;

  XX (0) = ATAPI_PREVENT_ALLOW;
  XX (4) = fl_prevent ? 1 : 0;

  ide_runcmd (cmd);
}

void
zip_describe (IDE * ide)
{
  CAP.cyls = ntohs (CAP.cyls);
  CAP.sector_size = ntohs (CAP.sector_size);
  CAP.rpm = ntohs (CAP.rpm);
  CAP.transfer_rate = ntohs (CAP.transfer_rate);

  printk ("%s: %d/%d/%d  bsize=%d  rpm=%d\n",
	  ide->name, CAP.cyls, CAP.heads, CAP.sectors,
	  CAP.sector_size, CAP.rpm);
}

/* boot-time function, should be universalised... */
void
zip_attach (IDE * ide, IDECTRL * ctrl)
{
  IDECMD cmd = ZEROCMD;

  X (0) = ATAPI_MODE_SENSE;
  X (2) = CAP_PAGE;
  X (7) = sizeof (ide->u.floppy.cap) >> 8;
  X (8) = sizeof (ide->u.floppy.cap);

  cmd.ide = ide;
  cmd.ctrl = ctrl;
  cmd.buff = (char *) &ide->u.floppy.cap;
  cmd.count = sizeof (ide->u.floppy.cap);

  ide_issue_immediate (ide, ctrl, &cmd);
  ide_issue_immediate (ide, ctrl, &cmd);
  ide_issue_immediate (ide, ctrl, &cmd);

  zip_describe (ide);
}

void
zip_dorequest (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  u32 blk;
  REQUEST *req = cmd->req;

  cmd->buff = req->data;

  switch (req->cmd)
    {
    case READ:
    case WRITE:
      blk = req->sec;

      bzero (cmd->pkt_data, sizeof (cmd->pkt_data));

      XX (0) = req->cmd == READ ? ATAPI_READ_BIG : ATAPI_WRITE_BIG;
      XX (2) = blk >> 24;
      XX (3) = blk >> 16;
      XX (4) = blk >> 8;
      XX (5) = blk;
      XX (7) = req->nsec >> 8;
      XX (8) = req->nsec;

      cmd->count = req->nsec * 512;
      if (req->cmd == WRITE)
	cmd->count = -cmd->count;	/* set data direction */

      cmd->callback = zip_callback;

      ide_execcmd (cmd);
      break;

    default:
      printk ("zip_docmd: invalid command %d\n", req->cmd);
      break;
    }

  return;
}

void
zip_error (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  if (cmd->code != RES_ERR)
    return;

  switch (cmd->error & AER_SKEY)
    {
    case AER_SK_NOT_READY:
      if (cmd->error & ~AER_SKEY)	/* Not Ready */
	{
	  printk ("zip_error: %s not ready\n", ide->name);
	  return;
	}
      /* Tray open. */
      if (!(ide->flags & F_MEDIA_CHANGED))
	printk ("%s: tray open\n", ide->name);
      ide->flags |= F_MEDIA_CHANGED;
      return;

    case AER_SK_UNIT_ATTENTION:	/* Media changed. */
      printk ("zip_error: %s media changed, syncing\n", ide->name);

      sync_dev (ide->dev);
      dfs_removechilds(ide->dev);
      
      if (!(ide->flags & F_MEDIA_CHANGED))
	printk ("%s: media changed\n", ide->name);
      ide->flags |= F_MEDIA_CHANGED;


      return;

    case AER_SK_ILLEGAL_REQUEST:
      /* Unknown command or invalid command arguments. */
      printk ("zip_error: %s invalid command\n", ide->name);
      return;
    }

  printk ("%s: i/o error, status=0x%x, error=0x%x\n", ide->name,
	  cmd->status, cmd->error);
}

void
zip_eject (IDE * ide)
{
  IDECMD *cmd;

  cmd = ide_newcmd (ide, zip_callback);
  if (cmd == NULL)
    return;

  cmd->retries = 3;

  XX (0) = ATAPI_START_STOP;
  XX (1) = 1;

  ide_runcmd (cmd);

  zip_prevent (ide, 0);

  cmd = ide_newcmd (ide, zip_callback);
  if (cmd == NULL)
    return;

  cmd->retries = 3;

  XX (0) = ATAPI_START_STOP;	/* eject media */
  XX (1) = 0;
  XX (4) = 2;

  ide_runcmd (cmd);
}

int
zip_ioctl (IDE * ide, int cmd, int arg)
{
  switch (cmd)
    {
    case CDIOCEJECT:
#ifdef ZIPDEBUG
      printk ("zip: ejecting %s\n", ide->name);
#endif
      zip_eject (ide);
      return 0;

    default:
      printk ("zip_ioctl: unknown command 0x%x\n", cmd);
      return -EINVAL;
    }
}
