/*
   dev/ide_probe.c
   ATA(PI) devices probing.
 */

#include <asm/regdump.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <phantom/hdregs.h>
#include <phantom/major.h>
#include <errno.h>
#include <phantom/dfs.h>
#include <phantom/blk.h>
#include <phantom/ide.h>
#include <phantom/fs.h>

extern int hd_attach (IDE * ide, struct drive_info *);
int ide_ioctl (struct file *file, unsigned int cmd, unsigned int arg);

struct file_operations ide_fops =
{
  NULL,				/* lseek */
  block_read,			/* read */
  block_write,			/* write */
  NULL,				/* readdir */
  ide_ioctl,			/* ioctl */
  block_open,			/* open */
  NULL				/* close */
};

IDECTRL ctrl_table[NR_IDE_CTRL] =
{
  {DEFAULT_IDE0_IRQ, DEFAULT_IDE0_PORT, IDE0_MAJOR, "ide0", NULL, 0},
  {DEFAULT_IDE1_IRQ, DEFAULT_IDE1_PORT, IDE1_MAJOR, "ide1", NULL, 0}
};

IDE ide_table[NR_IDE] =
{
  {
    MKDEV (IDE0_MAJOR, 0), "da", &ctrl_table[0],
    ARD_DRIVE0,
    0,},
  {
    MKDEV (IDE0_MAJOR, NR_PARTITIONSPERDRIVE), "db", &ctrl_table[0],
    ARD_DRIVE1,
    0,},
  {
    MKDEV (IDE1_MAJOR, 0), "dc", &ctrl_table[1],
    ARD_DRIVE0,
    0,},
  {
    MKDEV (IDE1_MAJOR, NR_PARTITIONSPERDRIVE), "dd", &ctrl_table[1],
    ARD_DRIVE1,
    0,}
};

int
ide_attach (int drive)
{
  IDE *ide = &ide_table[drive];
  IDECTRL *ctrl = ide->ctrl;
  struct drive_info id;
  int fl_ata, fl_atapi, ret;

  fl_ata = 0;
  fl_atapi = 0;

  ide = &ide_table[drive];

  fl_atapi = !ide_probe (&id, ide, WIN_PIDENTIFY);
  if (!fl_atapi)
    fl_ata = !ide_probe (&id, ide, WIN_IDENTIFY);

  if (!fl_ata && !fl_atapi)
    {
      printk ("%s: not available\n", ide->name);
      return EIO;
    }

  printk ("%s: %s %s", ide->name, id.model, id.fw_rev);

  dfs_add (ide->name, ide->dev, S_IRWXU | S_IFBLK);

  if (!fl_atapi)
    {
      if (id.lbaflag)
	{
	  ide->fl_lba = 1;
	  printk (" LBA\n");
	}

      if (!hd_attach (ide, &id))
	{
	  register_blkdev (ctrl->major, ide->name, &ide_fops, ide_dorequest, 512);
	  ide->fl_attached = 1;
	  ide->dev_type = IDE_HD;
	}
      return 0;
    }

  if (id.removable)
    printk (" removable");

  /* packet command size */
  switch (id.cmdsz)
    {
    case AT_PSIZE_12:
      printk (" cmd12");
      break;
    case AT_PSIZE_16:
      printk (" cmd16");
      ide->fl_cmd16 = 1;
      break;
    }

  /* DRQ type */
  switch (id.drqtype)
    {
    case AT_DRQT_MPROC:
      printk (" slow");
      ide->fl_slow = 1;
      break;
    case AT_DRQT_INTR:
      printk (" intr");
      ide->fl_intrcmd = 1;
      break;
    case AT_DRQT_ACCEL:
      printk (" accel");
      break;
    }
  /* when 'slow' is set, clear 'intrcmd' */
  if (ide->fl_slow)
    ide->fl_intrcmd = 0;

  /* overlap operation supported */
  if (id.ovlapflag)
    printk (" ovlap");

  /* interleaved DMA supported */
  if (id.idmaflag)
    printk (" idma");
  /* DMA supported */
  else if (id.dmaflag)
    printk (" dma");

  /* IORDY can be disabled */
  if (id.iordydis)
    printk (" iordis");
  /* IORDY supported */
  else if (id.iordyflag)
    printk (" iordy");

  printk ("\n");

  switch (id.devtype)
    {
    case AT_TYPE_DIRECT:
      ide->dev_type = IDE_FLOPPY;
      zip_attach (ide, ctrl);
      break;

    case AT_TYPE_CDROM:
      ide->dev_type = IDE_CD;
/*      cd_attach (ide, ctrl); */
      break;

    case AT_TYPE_TAPE:
      ide->dev_type = IDE_TAPE;
      printk ("Streamer devices not supported.\n");
      break;

    case AT_TYPE_OPTICAL:
      ide->dev_type = IDE_OPTICAL;
      printk ("optical disk (assume floppy driver?)\n");
      break;

    default:
      ide->dev_type = IDE_UNKNOWN;
      printk ("unknown ide device, devtype=0x%x\n", id.devtype);
      break;
    }

  ide->fl_attached = 1;
  return 0;
}

/* swap if fl_swap and remove extra spaces */
void
str_fix (char *str, int len, char fl_swap)
{
  int i, j, fl_space;
  char sw;

  if (fl_swap)
    for (i = 0; i < len; i += 2)
      {
	sw = str[i + 1];
	str[i + 1] = str[i];
	str[i] = sw;
      }

  for (i = 0; i < len; i++)
    if (!str[i])
      str[i] = ' ';

  i = len - 1;
  while (i && str[i] == ' ')
    str[i--] = 0;

  i = 0;
  fl_space = 0;
  for (j = 0; j < len; j++)
    {
      if (!str[j])
	break;
      if (str[j] == ' ')
	{
	  if (fl_space)
	    continue;
	  else
	    fl_space = 1;
	}
      else
	fl_space = 0;
      if (i != j)
	str[i] = str[j];
      i++;
    }

  if (i >= len)
    i = len - 1;
  str[i] = 0;
}

int
ide_probe (struct drive_info *id, IDE * ide, int cmd)
{
  int ret, i;
  char fl_swap = 1;
  int port = ide->ctrl->port;

  outb (ide->select_mask, port + AR_DRIVE);

  if (inb (port + AR_STATUS) == 0xff)
    return EIO;

  ret = 2000;
  while ((inb_p (port + AR_STATUS) & BUSY_STAT) && ret--);

  if (!ret)
    {
      printk ("ide_probe: still busy...\n");
      return EIO;
    }

  outb (ide->select_mask, port + AR_DRIVE);
  outb (cmd, port + AR_COMMAND);

#define STAT_MASK (BUSY_STAT | READY_STAT | WRERR_STAT | SEEK_STAT | ERR_STAT)
#define STAT_OK (READY_STAT | SEEK_STAT)

  ret = 20000;
  while (ret--)
    {
      i = inb_p (port + AR_STATUS);
      if (i & BUSY_STAT)
	continue;
      if ((i & STAT_MASK) != STAT_OK)
	break;
      if (i & DRQ_STAT)
	goto read_ready;
    }

#ifdef IDEPROBEDEBUG
  printk ("ide_probe: no device here...\n");
#endif
  return EIO;

read_ready:
#ifdef IDE32
  insl (port + AR_DATA, (char *) id, sizeof (struct drive_info) >> 2);
#else
  insw (port + AR_DATA, (char *) id, sizeof (struct drive_info) >> 1);
#endif

  /* shuffle string byte order, Mitsumi, NEC and Pioneer drives 
     don't need this */
  if (cmd == WIN_PIDENTIFY)
    if (((id->model[0] == 'N' && id->model[1] == 'E') ||
	 (id->model[0] == 'F' && id->model[1] == 'X')) ||
	(id->model[0] == 'P' && id->model[1] == 'i'))
      fl_swap = 0;

  str_fix (id->model, sizeof (id->model), fl_swap);
  str_fix (id->serial, sizeof (id->serial), fl_swap);
  str_fix (id->fw_rev, sizeof (id->fw_rev), fl_swap);

  return 0;
}

void
ide_intr (int irq_nr, struct regdump *regdump)
{
  IDE *ide;
  IDECTRL *ctrl;
  IDECMD *cmd;

  switch (irq_nr)
    {
    case DEFAULT_IDE0_IRQ:
      if (ctrl_table[0].handler != NULL)
	{
	  if ((cmd = ctrl_table[0].head) != NULL)
	    {
	      ide = cmd->ide;
	      ctrl = cmd->ctrl;
	      ctrl_table[0].handler (ide, ctrl, cmd);
	    }
	  else
	    printk ("ide_intr: IDE0, NULL cmd head!\n");
	}
      else
	printk ("ide_intr: IDE0, noone wants me...\n");
      break;

    case DEFAULT_IDE1_IRQ:
      if (ctrl_table[1].handler != NULL)
	{
	  if ((cmd = ctrl_table[1].head) != NULL)
	    {
	      ide = cmd->ide;
	      ctrl = cmd->ctrl;
	      ctrl_table[1].handler (ide, ctrl, cmd);
	    }
	  else
	    printk ("ide_intr: IDE1, NULL cmd head!\n");
	}
      else
	printk ("ide_intr: IDE1, noone wants me...\n");
      break;

    default:
      printk ("ide_intr: invalid interrupt 0x%x arrived..\n");
      break;
    }
}

void
ide_init ()
{
  int ctrl, drv, i;

  for (drv = 0; drv < 4; drv++)
    ide_attach (drv);

  for (ctrl = 0; ctrl < NR_IDE_CTRL; ctrl++)
    if ((ide_table[ctrl << 1].fl_attached ||
	 ide_table[(ctrl << 1) + 1].fl_attached))
      {
	if (irq_request (ctrl_table[ctrl].irq, IRQ_FAST, &ide_intr, "idex"))
	  {
	    printk ("ide_init: irq_request failed (ctrl=%d)\n", ctrl);
	    ide_table[ctrl << 1].fl_attached = 0;
	    ide_table[(ctrl << 1) + 1].fl_attached = 0;
	  }
	else
	  {
	    if ((blkdev[ctrl_table[ctrl].major].begin =
		 (int *) kmalloc (sizeof (int) * 2 *
				  NR_PARTITIONSPERDRIVE)) == NULL)
	        return;
	    bzero (blkdev[ctrl_table[ctrl].major].begin,
		   sizeof (int) * 2 * NR_PARTITIONSPERDRIVE);

	    if ((blkdev[ctrl_table[ctrl].major].size =
		 (int *) kmalloc (sizeof (int) * 2 *
				  NR_PARTITIONSPERDRIVE)) == NULL)
	        return;
	    bzero (blkdev[ctrl_table[ctrl].major].size,
		   sizeof (int) * 2 * NR_PARTITIONSPERDRIVE);

	    blkdev[ctrl_table[ctrl].major].do_request = &ide_dorequest;

	    if (ide_table[ctrl << 1].fl_attached)
	      {
		i = ide_table[ctrl << 1].dev_type;
		if (i == IDE_HD || i == IDE_FLOPPY)
		  partitions_probe (MKDEV (ctrl_table[ctrl].major, 0),
				    ide_table[ctrl << 1].name);
	      }

	    if (ide_table[(ctrl << 1) + 1].fl_attached)
	      {
		i = ide_table[(ctrl << 1) + 1].dev_type;
		if (i == IDE_HD || i == IDE_FLOPPY)
		  partitions_probe (MKDEV (ctrl_table[ctrl].major,
					   NR_PARTITIONSPERDRIVE),
				    ide_table[(ctrl << 1) + 1].name);
	      }
	  }
      }

/* all should be done, from this moment on we can do regular job... */
}

IDE *
ide_get_ide (int dev)
{
  int maj = MAJOR (dev), min = MINOR (dev);
  int res = 0;

  switch (maj)
    {
    case IDE0_MAJOR:
      res = 0;
      break;
    case IDE1_MAJOR:
      res = 2;
      break;
    default:
      printk ("ide_get_ide: device %02X:%02X out of bounds...\n", maj, min);
      return NULL;
    }

  if (min >= NR_PARTITIONSPERDRIVE)
    res++;

  return &ide_table[res];
}

int
ide_ioctl (struct file *file, unsigned int cmd, unsigned int arg)
{
  IDE *ide;

  ide = ide_get_ide (file->f_dev);
  if (ide == NULL || !ide->fl_attached)
    {
      printk ("ide_ioctl: invalid device or not attached...\n");
      return -EINVAL;
    }

  switch (ide->dev_type)
    {
    case IDE_FLOPPY:
      return zip_ioctl (ide, cmd, arg);

    default:
      printk ("ide_ioctl: dev_type %d does not support ioctl...\n");
      return -EINVAL;
    }
}
