/*
   dev/keyboard.c
   Phantom keyboard driver.

   key magic:
   B    buffer list
   F    flush all buffers
   R    instant reboot
   V    vnode list
   M    memory info
   T	task list
   D	device list
   DEL  system down  
 */

#define REALLY_SLOW_IO
#define SLOW_IO_BY_JUMPING
#include <asm/regdump.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/system.h>
#include <phantom/keyboard.h>
#include <phantom/tty.h>
#include <phantom/kb.h>
#include <phantom/keyb.h>
#include <phantom/major.h>
#include <phantom/fs.h>
#include <phantom/sched.h>
#include <conio.h>
#include <phantom/k_unistd.h>
#include <phantom/ext.h>
#include <rg/kernel.h>

#define KB_IRQ		0x01

volatile int fl_ack = 0;
volatile int fl_resend = 0;
volatile int fl_reply = 0;
volatile int kb_leds = -1;
volatile int prev_sc = 0;

/* kb_action flags */
#define KBA_BLIST	1
#define KBA_VLIST	2
#define KBA_MLIST	3
#define KBA_FLUSH	4
#define KBA_TLIST	5
#define KBA_QBOOT	6
#define KBA_SDOWN	7
#define KBA_MLIST2	8
#define KBA_SWINDOW	9
#define KBA_DLIST	10

int kb_action = 0;

volatile char key[NR_KEY] =
{0,};				/* global keymap */

#ifdef KHL
int khl[NR_KEY];
char *keynames[] =
{
  "<0>", "escape",
  "1", "2", "3", "4", "5", "6", "7", "8", "9", "0", "-", "=",
  "backspace", "tab", "q", "w", "e", "r", "t", "y", "u", "i", "o", "p",
"[", "]", "enter", "ctrl", "a", "s", "d", "f", "g", "h", "j", "k", "l", ";",
  "'", "`", "left shift", "\\",
  "z", "x", "c", "v", "b", "n", "m", ",", ".", "/",
  "right shift", "print screen", "alt", "space", "capslock",
  "f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8", "f9", "f10",
  "numlock", "scrolllock", "home", "up arrow", "page up", "keypad -",
  "left arrow", "keypad 5", "right arrow", "keypad +", "end", "down arrow",
  "page down", "insert", "delete", "alt print screen",
  "<55>", "<56>", "f11", "f12",
};
#endif

short key_map[] =
{
  0, 27, '1', '2', '3', '4', '5', '6',
  '7', '8', '9', '0', '-', '=', 8, 9,
  'q', 'w', 'e', 'r', 't', 'y', 'u', 'i',
  'o', 'p', '[', ']', 10, 0, 'a', 's',
  'd', 'f', 'g', 'h', 'j', 'k', 'l', ';',
  '\'', '`', 0, '\\', 'z', 'x', 'c', 'v',
  'b', 'n', 'm', ',', '.', '/', 0, '*',
  0, 32, 0, KB_F1, KB_F2, KB_F3, KB_F4, KB_F5,
  KB_F6, KB_F7, KB_F8, KB_F9, 0, 0, 0, 0,
  1, 0, '-', 3, 0, 4, '+', 0,
  2, 0, 0, 0, KB_DELETE, 0, '<', 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0
};

short shift_map[] =
{
  0, 27, '!', '@', '#', '$', '%', '^',
  '&', '*', '(', ')', '_', '+', 8, KB_RTAB,
  'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I',
  'O', 'P', '{', '}', 10, 0, 'A', 'S',
  'D', 'F', 'G', 'H', 'J', 'K', 'L', ':',
  '"', '~', '0', '|', 'Z', 'X', 'C', 'V',
  'B', 'N', 'M', '<', '>', '?', 0, '*',
  0, 32, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  KB_UP, 0, '-', KB_LEFT, 0, KB_RIGHT, '+', 0,
  KB_DOWN, 0, 0, 0, 0, 0, '>', 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0
};

short altscroll_map[] =
{
  0, 0, 0, '@', 0, '$', 0, 0,
  '{', '[', ']', '}', '\\', 0, 8, KB_RTAB,
  0xC9, 0xCD, 0xBB, 0xDA, 0xC4, 0xBF, 0, 0,
  0, 0, 0, '~', 10, 0, 0xBA, 0,
  0xBA, 0xB3, 0, 0xB3, 0, 0, 0, 0,
  0, 0, 0, 0, 0xC8, 0xCD, 0xBC, 0xC0,
  0xC4, 0xD9, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  1, 0, 0, 3, 0, 4, 0, 0,
  2, 0, 0, 0, 0, 0, '|', 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0
};

short alt_map[] =
{
  0, 0, 0, '@', 0, '$', 0, 0,
  '{', '[', ']', '}', '\\', 0, 8, KB_RTAB,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, '~', 10, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  1, 0, 0, 3, 0, 4, 0, 0,
  2, 0, 0, 0, 0, 0, '|', 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0
};

u8 e0_keys[128] =
{
  0, 0, 0, 0, 0, 0, 0, 0,	/* 0x00-0x07 */
  0, 0, 0, 0, 0, 0, 0, 0,	/* 0x08-0x0f */
  0, 0, 0, 0, 0, 0, 0, 0,	/* 0x10-0x17 */
  0, 0, 0, 0, K_ENTER, K_RCONTROL, 0, 0,	/* 0x18-0x1f */
  0, 0, 0, 0, 0, 0, 0, 0,	/* 0x20-0x27 */
  0, 0, 0, 0, 0, 0, 0, 0,	/* 0x28-0x2f */
  0, 0, 0, 0, 0, K_SLASH, 0, K_PRSCR,	/* 0x30-0x37 */
  K_RALT, 0, 0, 0, 0, 0, 0, 0,	/* 0x38-0x3f */
  0, 0, 0, 0, 0, 0, K_BREAK, K_HOME,	/* 0x40-0x47 */
  K_UP, K_PGUP, 0, K_LEFT, 0, K_RIGHT, 0, K_END,	/* 0x48-0x4f */
  K_DOWN, K_PGDN, K_INSERT, K_DELETE, 0, 0, 0, 0,	/* 0x50-0x57 */
  0, 0, 0, K_LWIN, K_RWIN, K_TMWIN, K_POWER, K_SLEEP,	/* 0x58-0x5f */
  0, 0, 0, K_WAKE, 0, 0, 0, 0,	/* 0x60-0x67 */
  0, 0, 0, 0, 0, 0, 0, 0,	/* 0x68-0x6f */
  0, 0, 0, 0, 0, 0, 0, 0,	/* 0x70-0x77 */
  0, 0, 0, 0, 0, 0, 0, 0	/* 0x78-0x7f */
};

int getch ();
int getkey (int);
void kb_addbuffer (TTY *, int, int);
int __kbhit (TTY *);
int kbhit ();
void kb_wait ();
void kb_setleds ();

void
kb_wait ()
{
  int i = 0xFFFF;
  while (i--)
    if ((inb_p (KB_STATUS_REG) & KB_STAT_IBF) == 0)
      return;
  printk ("kb_wait: timeout\n");
}

void
kb_send_cmd (u8 cmd)
{
  kb_wait ();
  outb_p (cmd, KB_STATUS_REG);
}

int
kb_send_data (u8 data)
{
  int ret, ret2 = 0x10;

  fl_ack = 0;
  fl_resend = 0;
  fl_reply = 1;

  while (ret2--)
    {
      kb_wait ();
      outb_p (data, KB_DATA_REG);
      ret = 0x500;
      while (ret--)
	{
	  inb_p (KB_STATUS_REG);	/* smaller delay */
	  if (fl_ack)
	    return 0;
	  if (fl_resend)
	    break;
	}
    }

  printk ("kb_send_data: failed, data=0x%02x  fl_resend=%d\n",
	  data, fl_resend);
  return -1;
}

void
kb_irq (int irq_nr, struct regdump *regs)
{
  int i, sc;
  TTY *t = current_tty;

  kb_send_cmd (KB_CMD_DISABLE);
  kb_wait ();
  sc = inb_p (KB_DATA_REG);

  if (fl_reply)
    {
      fl_reply = 0;
      if (sc == 0xFA)
	{
	  fl_ack = 1;
	  goto kb_end_irq;
	}
      if (sc == 0xFE)
	{
	  fl_resend = 1;
	  goto kb_end_irq;
	}
      printk ("kb_irq: fl_reply set, got scancode 0x%02x\n", sc);
    }

  switch (sc)
    {
    case 0xE0:
      prev_sc = 0xE0;
    case 0xE1:			/* ignore pause */
      goto kb_end_irq;
    case 0x00:
    case 0xFF:			/* errors */
      prev_sc = 0;
      goto kb_end_irq;
    }

  if (sc < 0x80)
    {
      if (prev_sc == 0xE0)
	{
	  prev_sc = 0;
	  sc = e0_keys[sc];
	}

      switch (sc)
	{
	case K_RSHIFT:
	  if (key[K_RSHIFT])
	    break;
	  t->kb_flags |= FL_SHIFT;
	  t->kb_flags ^= FL_CAPS;
	  break;
	case K_LSHIFT:
	  if (key[K_LSHIFT])
	    break;
	  t->kb_flags |= FL_SHIFT;
	  t->kb_flags ^= FL_CAPS;
	  break;
	case K_CAPSLOCK:
	  if (key[K_CAPSLOCK])
	    break;
	  t->kb_flags ^= FL_CAPS;
	  if (t->kb_flags & FL_CAPS)
	    t->kb_leds |= LED_CAPS;
	  else
	    t->kb_leds &= ~LED_CAPS;
	  ext_run (KB_EXT);
	  break;
	case K_NUMLOCK:
	  if (key[K_NUMLOCK])
	    break;
	  t->kb_leds ^= LED_NUM;
	  ext_run (KB_EXT);
	  break;
	case K_SCRLOCK:
	  if (key[K_SCRLOCK])
	    break;
	  t->kb_flags ^= FL_SCROLL;
	  if (t->kb_flags & FL_SCROLL)
	    t->kb_leds |= LED_SCROLL;
	  else
	    t->kb_leds &= ~LED_SCROLL;
	  ext_run (KB_EXT);
	  break;
	case K_CONTROL:
	  if (key[K_CONTROL])
	    break;
	  t->kb_flags |= FL_CTRL;
	  break;
	case K_ALT:
	  if (key[K_ALT])
	    break;
	  t->kb_flags |= FL_ALT;
	  break;
	default:		/* this should be sorted somehow */
	  if (t->kb_flags == FL_NONE)
	    {
	      kb_addbuffer (t, key_map[sc], sc);
	      break;
	    }

	  if (t->kb_flags & FL_ALT)
	    {
	      if ((t->kb_flags & FL_CTRL))
		{
		  switch (sc)
		    {
		    case K_BACKSPACE:	/* kills X :] */
		    case K_Z:
		      /* no pain for printregdump to be called in an interrupt */
		      printregdump ("\nUser requested dump", regs);
		      break;
		    case K_DELETE:	/* system terminator */
		      kb_action = KBA_SDOWN;
		      ext_run (KB_EXT);
		      break;
		    case K_B:	/* buffer list */
		      kb_action = KBA_BLIST;
		      ext_run (KB_EXT);
		      break;
		    case K_F:	/* flush buffers */
		      kb_action = KBA_FLUSH;
		      ext_run (KB_EXT);
		      break;
		    case K_R:	/* remove all buffers,
				   could be used as instant
				   reboot later */
		      kb_action = KBA_QBOOT;
		      ext_run (KB_EXT);
		      break;
		    case K_V:	/* vnode list */
		      kb_action = KBA_VLIST;
		      ext_run (KB_EXT);
		      break;
		    case K_M:	/* memory manager list */
		      kb_action = KBA_MLIST;
		      ext_run (KB_EXT);
		      break;
		    case K_N:	/* dump memory map */
		      kb_action = KBA_MLIST2;
		      ext_run (KB_EXT);
		      break;
		    case K_T:
		      kb_action = KBA_TLIST;
		      ext_run (KB_EXT);
		      break;
		    case K_D:
		      kb_action = KBA_DLIST;
		      ext_run (KB_EXT);
		      break;
		    }
		}
	      else
		{
		  if (sc == K_TAB)
		    {
		      kb_action = KBA_SWINDOW;
		      ext_run (KB_EXT);
		      break;
		    }

		  if (t->kb_flags & FL_SCROLL)
		    kb_addbuffer (t, altscroll_map[sc], sc);
		  else
		    kb_addbuffer (t, alt_map[sc], sc);
		  break;
		}
	    }
	  if (t->kb_flags & FL_SHIFT)
	    {
	      i = shift_map[sc];
	      if ((!(t->kb_flags & FL_CAPS)) && (i >= 'A' && i <= 'Z'))
		i += 'a' - 'A';
	      kb_addbuffer (t, i, sc);
	      break;
	    }

	  if (t->kb_flags & FL_CAPS)
	    {
	      i = shift_map[sc];
	      if (i >= 'A' && i <= 'Z')
		kb_addbuffer (t, i, sc);
	      else
		kb_addbuffer (t, key_map[sc], sc);
	      break;
	    }

	  if (t->kb_flags & FL_CTRL)
	    {
	      if (sc <= 0x44 && sc >= 0x3B)
		tty_switch (sc - 0x3B);
	      else
		kb_addbuffer (t, key_map[sc], sc);
	      break;
	    }
	}

#ifdef KHL
      if (key[sc] == 0)
	khl[sc]++;
#endif

      key[sc] = 1;
    }
  else
    {
      sc ^= 0x80;
      if (prev_sc == 0xE0)
	{
	  prev_sc = 0;
	  sc = e0_keys[sc];
	}

      key[sc] = 0;
      switch (sc)
	{
	case K_RSHIFT:
	case K_LSHIFT:
	  t->kb_flags &= ~FL_SHIFT;
	  t->kb_flags ^= FL_CAPS;
	  break;
	case K_CONTROL:
	  t->kb_flags &= ~FL_CTRL;
	  break;
	case K_ALT:
	  t->kb_flags &= ~FL_ALT;
	  break;
	}
    }
  sc = inb_p (0x61);
  outb_p (sc | 0x80, 0x61);
  outb_p (sc, 0x61);
kb_end_irq:
  kb_send_cmd (KB_CMD_ENABLE);
}

void
kb_addbuffer (TTY * t, int s, int sc)
{
  if (t->kb_buff_i < KB_BUFFSIZE)
    {
      t->kb_buff[t->kb_buff_i] = s;
      t->raw_buff[t->kb_buff_i] = sc;
      t->kb_buff_i++;

      wake_up (&t->key_wait);
    }
}

int
getkey (int k)
{
  if (k >= 0 && k < NR_KEY)
    return (key[k]);
  else
    return (0);			/* non-existent key not pressed... */
}

void
kb_setleds ()
{
  ext_run (KB_EXT);
}

void
kb_done ()
{
#ifdef KHL
  int i, j, k;

  setattr (ATTR (LIGHTGREEN, BLACK));
  printf ("KHL's top ten\n");
  printf ("hits                  keyname   scancode\n");
  setattr (ATTR (GREEN, BLACK));
  for (i = 0; i < 10; i++)
    {
      k = 1;
      for (j = 2; j < NR_KEY; j++)
	if (khl[j] > khl[k])
	  k = j;
      printf ("  %-3d   %20s     %-3d\n", khl[k], keynames[k], k);
      khl[k] = -1;
    }
  setattr (DEFAULT_ATTR);
#endif
}

#define KB_NO_DATA	(-1)	/* No data */
#define KB_BAD_DATA	(-2)	/* Parity or other error */

int
kb_read ()
{
  int ret = KB_NO_DATA;
  unsigned char status;

  status = inb (KB_STATUS_REG);
  if (status & KB_STAT_OBF)
    {
      unsigned char data = inb (KB_DATA_REG);

      ret = data;
      if (status & (KB_STAT_GTO | KB_STAT_PERR))
	ret = KB_BAD_DATA;
    }
  return (ret);
}

void
kb_flush ()
{
  int ret = 256;

  while (ret--)
    if (kb_read () == KB_NO_DATA)
      return;

  printk ("kb_flush: unflushable...\n");
}

void kb_pcsetleds(u32 leds)
{
 if (kb_send_data (KB_DTA_LEDS) || kb_send_data (leds))
        {
      kb_send_data (KB_DTA_REENABLE);
      return;
    }
}

void
kb_ext (void)
{
  switch (kb_action)
    {
    case KBA_BLIST:
      blist ();
      break;
    case KBA_VLIST:
      vlist ();
      break;
    case KBA_MLIST:
      mm_info ();
      break;
    case KBA_FLUSH:
      sys_sync ();
      break;
    case KBA_QBOOT:
      sys_reboot (SE_QUICK_BOOT);
      break;
    case KBA_SDOWN:
      sys_reboot (SE_SAFE_BOOT);
      break;
    case KBA_SWINDOW:
      rg_switch ();
      break;
    case KBA_TLIST:
      task_info();
      break;
    case KBA_DLIST:
      device_info();
      break;
    }

  kb_action = 0;

  if (kb_leds != current_tty->kb_leds)
    {
      kb_pcsetleds(current_tty->kb_leds);
      kb_leds = current_tty->kb_leds;
    }
}

void
kb_init ()
{
  int i;

  setattr (ATTR (GREEN, BLACK));
  printf ("Keyboard\n");
  setattr (DEFAULT_ATTR);

  kb_flush ();

#ifdef KHL
  for (i = 0; i < NR_KEY; i++)
    khl[i] = 0;
#endif

  irq_request (KB_IRQ, IRQ_COMPLEX, kb_irq, "keyboard");
  ext_enable (KB_EXT);
  ext_table[KB_EXT] = kb_ext;

  i = inb_p (0x61);
  outb_p (i | 0x80, 0x61);
  outb_p (i, 0x61);

  kb_setleds ();
  kb_leds = current_tty->kb_leds;
}
