/*
   dev/lpt.c
   Phantom printer port driver.
 */

#include <asm/regdump.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <phantom/major.h>
#include <errno.h>
#include <conio.h>
#include <phantom/fs.h>

#define MAJOR_NR	LPT_MAJOR

#define MAX_BUFF	256

#define LPT_IRQ		0x07
#define LPT_DATA	0x378
#define LPT_BASE	LPT_DATA

#define LPT_CONTROL	(LPT_BASE + 2)
#define LPT_STATUS	(LPT_BASE + 1)

#define LPT_INIT	0x04
#define LPT_SELECT	0x08

#define LPT_SLEEP	10	/* miliseconds */

char lpt_buff[MAX_BUFF];
int lpt_buff_i = 0;

void 
lpt_int (int irq_nr, struct regdump *regs)
{
  if (lpt_buff_i < MAX_BUFF)
    lpt_buff[lpt_buff_i++] = inb_p (LPT_DATA);
  else
    printf ("lpt.c: buffer exhausted\n");
}

int 
lpt_reset (int port)
{
  outb_p (0, LPT_CONTROL);
  sys_sleep (LPT_SLEEP);
  outb_p (LPT_SELECT | LPT_INIT, LPT_CONTROL);

  return inb_p (LPT_STATUS);
}

int 
lpt_read (struct file *file, char *buff, size_t count)
{
  int tcount = count;
  int i;

  while (count && lpt_buff_i)
    {
      *buff++ = lpt_buff[0];
      for (i = 1; i < lpt_buff_i; i++)
	lpt_buff[i - 1] = lpt_buff[i];
      lpt_buff_i--;
      count--;
    }

  return (tcount - count);
}

int 
lpt_write (struct file *file, char *buff, size_t count)
{
  int tcount = count;

  while (count--)
    outb_p (*buff++, LPT_DATA);

  return (tcount);
}

int 
lpt_open (struct file *file)
{
  if (MINOR (file->f_dev) != 0)
    return (-EIO);
  return (0);
}

struct file_operations lpt_fops =
{
  NULL,				/* lseek */
  lpt_read,
  lpt_write,
  NULL,				/* readdir */
  NULL,				/* ioctl */
  lpt_open,
  NULL				/* release */
};

void
lpt_init ()
{
  int i;

  setattr (ATTR (GREEN, BLACK));
  printf ("Printer port driver\n");
  setattr (DEFAULT_ATTR);

  register_chrdev (MAJOR_NR, "parallel", &lpt_fops);
  dfs_add ("lp0", MKDEV (MAJOR_NR, 0), S_IRWXU | S_IFCHR);

  i = inb_p (LPT_BASE);

  if (i != 0xFF)
    lpt_reset (LPT_BASE);

  irq_request(LPT_IRQ, IRQ_FAST, &lpt_int, "parallel");
}
