/*
   dev/ramdisk.c
   Ramdisk driver.

   NOTE: Here will be no problem with memory allocation
         since dev/buffer.c works with blocksize=pagesize...
 */

#include <conio.h>
#include <string.h>
#include <asm/sysinfo.h>
#include <phantom/blk.h>
#include <phantom/major.h>
#include <phantom/fs.h>
#include <errno.h>
#include <phantom/time.h>
#include <phantom/mm.h>
#include <phantom/buffer.h>

#define MAJOR_NR		RAMDISK_MAJOR
#define BLOCK_SIZE		512

int ramdisk_block, ramdisk_blocks, ramdisk_size, ramdisk_pages;
char *ramdisk_base;

int
ramdisk_open (struct file *file)
{
  if (MINOR (file->f_dev))
    return (-EIO);

  return (0);
}

void
ramdisk_dorequest ()
{
  struct request *req = blkdev[MAJOR_NR].request;

  if (!ramdisk_blocks)
    {
      printk ("ramdisk: no ramdisk\n");
      end_request (req->dev, 1);
      return;
    }

  switch (req->cmd)
    {
    case READ:
      memcpy ((void *) req->data,
	      (void *) (ramdisk_base + req->sec * BLOCK_SIZE),
	      req->nsec * BLOCK_SIZE);
      end_request (req->dev, 0);
      break;
    case WRITE:
      memcpy ((void *) (ramdisk_base + req->sec * BLOCK_SIZE),
	      (void *) req->data, req->nsec * BLOCK_SIZE);
      end_request (req->dev, 0);
      break;
    default:
      printk ("ramdisk_dorequest: unknown command\n");
      end_request (req->dev, 1);
      break;
    }
}

struct file_operations ramdisk_fops = {
  NULL,
  block_read,
  block_write,
  NULL,
  NULL,
  ramdisk_open,
  NULL
};

void
ramdisk_init ()
{
  int i;
  u32 start_time, stop_time;

  ramdisk_blocks = sysinfo.ramdisk_blocks;
  if (!ramdisk_blocks)
    return;

  ramdisk_block = sysinfo.ramdisk_block;
  ramdisk_size = ramdisk_blocks >> 1;	/* in KBs */
  ramdisk_pages = ramdisk_blocks >> 3;	/* in pages */

  setattr (ATTR (GREEN, BLACK));
  printf ("Ramdisk (%d blocks, %d pages) %s\n", ramdisk_blocks, ramdisk_pages,
  sysinfo.ramdisk_base ? "already loaded" : "");
  setattr (DEFAULT_ATTR);

  register_blkdev (MAJOR_NR, "rd", &ramdisk_fops, ramdisk_dorequest, 512);
  blkdev[MAJOR_NR].size = &ramdisk_size;

/* ramdisk not loaded yet */
  if(!sysinfo.ramdisk_base)
    {
  if ((ramdisk_base = (char *) pages_alloc (ramdisk_pages))==NULL)
    {
      printk ("Ooops, not enough memory to load ramdisk.\n");
      return;
    }

/* it should not be problem to get this continuos one... */
    printf ("Loading image...");
    start_time = CURRENT_TIME;
    for (i = 0; i < ramdisk_blocks; i++)
      rblocks (sysinfo.root_dev, ramdisk_block + i, 1, ramdisk_base + (i << 9));

    stop_time = CURRENT_TIME;

    printf (" %d Kbit/s\n", (sysinfo.ramdisk_blocks * 50 * 8)
  	    / (stop_time - start_time));
    } else {	/* ramdisk already loaded */
	    ramdisk_base = (char *)sysinfo.ramdisk_base;
	   }

    sysinfo.root_dev = MKDEV (MAJOR_NR, 0);

  return;
}
