/*
   dev/serial.c
   Phantom serial port/mouse driver.

   TODO: make better port initialization, etc. (locking)
 */

#include <asm/irq.h>
#include <asm/io.h>
#include <phantom/major.h>
#include <errno.h>
#include <phantom/fs.h>
#include <phantom/fifo.h>
#include <phantom/mouse_io.h>
#include <phantom/mouse.h>
#include <conio.h>

#define MAJOR_NR	SERIAL_MAJOR
#define MOUSE_FLAG	0x10

#define NR_COM		2
#define BUFF_MAX	512

#define BASEPORT(a)	com[(a)].port

#define COM_LINE(a)	(BASEPORT(a)+0x0B)
#define COM_MLINE(a)	(BASEPORT(a)+0x0C)
#define COM_LOBAUD(a)	(BASEPORT(a)+0x08)
#define COM_HIBAUD(a)	(BASEPORT(a)+0x09)
#define COM_IER(a)	(BASEPORT(a)+0x09)
#define COM_DATA(a)	(BASEPORT(a)+0x08)
#define COM_STATUS(a)	(BASEPORT(a)+0x0D)

#define getcom(a)	MINOR(a)

struct com_struct
  {
    int port;
    int irq;
    unsigned char buffer[BUFF_MAX];
    int buffer_i;    
    char on;
  };

struct mouse_struct
  {
    int b, x, y;
    int ratio_x, ratio_y;
    int limit_x, limit_y;
    char bytenum;
    char on;
    unsigned char buffer[3];
  };

struct com_struct com[NR_COM] =
{
  {
    0x03F0,
    0x04,
    {0,},
    0,
    1,
  },
  {
    0x02F0,
    0x03,
    {0,},
    0,
    1
  }
};

struct mouse_struct mouse[NR_COM] =
{
  {
    0, 0, 0,
    1, 1,
    640,400,
    0,
    0,
    {0,}
  },
  {
    0, 0, 0,
    1, 1,
    640,400,
    0,
    0,
    {0,}
  }
};

/* only first com now... */
void
serial_int (int irq_nr, struct regdump *regs)
{
  int cn = 0;
  int c;
  int dx, dy;

  c = inb (COM_DATA (cn));
  fifo_putc (c, com[cn].buffer, &com[cn].buffer_i, BUFF_MAX, FIFO_ROLL);

/* mouse section */
  if (mouse[cn].on)
    {
      if (c & 0x80==0)
	mouse[cn].bytenum = 0;	/* synchronization */

      mouse[cn].buffer[mouse[cn].bytenum++] = c;

      if (mouse[cn].bytenum >= 3)
	{
	  mouse[cn].bytenum=0;

	  dx = ((mouse[cn].buffer[0] & 0x03) << 6) | (mouse[cn].buffer[1]&0x3F);
	  dy = ((mouse[cn].buffer[0] & 0x0C) << 4) | (mouse[cn].buffer[2]&0x3F);
	  mouse[cn].b = mouse[cn].buffer[0] & (MOUSE_BL | MOUSE_BR);
	  if (dx >= 0x80)
	    dx -= 0x100;
	  if (dy >= 0x80)
	    dy -= 0x100;
	  mouse[cn].x += dx * mouse[cn].ratio_x;
	  mouse[cn].y += dy * mouse[cn].ratio_y;
	  if (mouse[cn].x < 0)
	    mouse[cn].x = 0;
	  if (mouse[cn].y < 0)
	    mouse[cn].y = 0;
	  if (mouse[cn].x >= mouse[cn].limit_x)
	    mouse[cn].x = mouse[cn].limit_x-1;
	  if (mouse[cn].y >= mouse[cn].limit_y)
	    mouse[cn].y = mouse[cn].limit_y-1;
       }
    }
}

int
serial_read (struct file *file, char *buff, size_t count)
{
  int cn = getcom (file->f_dev);
  int tcount;
  int i;

  tcount = count;
  while (count)
    {
      if ((i = fifo_getc (com[cn].buffer, &com[cn].buffer_i)) != -1)
	{
	  *buff++ = i;
	  count--;
	}
      else
	return (tcount - count);
    }

  return (tcount);
}

int
serial_write (struct file *file, char *buff, size_t count)
{
  int cn = getcom (file->f_dev);
  int tcount;

  tcount = count;
  while (count)
    {
/*  while(inb_p(COM_STATUS(com))&0x20==0); */
      outb_p (*buff++, COM_DATA (cn));
      count--;
    }

  return (tcount);
}

int
serial_ioctl (struct file *file, unsigned int cmd, unsigned arg)
{
  int cn = getcom (file->f_dev);

  switch (cmd)
    {
    case MIOC_GETX:
      return (mouse[cn].x);
    case MIOC_GETY:
      return (mouse[cn].y);
    case MIOC_GETB:
      return (mouse[cn].b);
    case MIOC_ON:			/* and reset the mouse */
      mouse[cn].x=0;
      mouse[cn].y=0;
      mouse[cn].b=0;
      mouse[cn].bytenum=0;
      mouse[cn].on = 1;
      break;
    case MIOC_OFF:
      mouse[cn].on = 0;
      break;
    case MIOC_SETRATIO:
      mouse[cn].ratio_y = arg;
    case MIOC_SETRATIOX:
      mouse[cn].ratio_x = arg;
      break;
    case MIOC_SETRATIOY:
      mouse[cn].ratio_y = arg;
      break;
    case MIOC_SETLIMIT:
      mouse[cn].limit_y=arg;
    case MIOC_SETLIMITX:
      mouse[cn].limit_x=arg;
      break;
    case MIOC_SETLIMITY:
      mouse[cn].limit_y=arg;
      break;
    default:
      return (-EINVAL);
    }

  return (0);
}

int
serial_open (struct file *file)
{
  if (MINOR (file->f_dev) != 0)
    return (-EIO);		/* only com0 */
  return (0);
}

struct file_operations serial_fops =
{
  NULL,				/* lseek */
  serial_read,
  serial_write,
  NULL,				/* readdir */
  serial_ioctl,
  serial_open,
  NULL				/* close */
};

/* set up all com ports to 9600 */
void
serial_on ()
{
  int i;

  for (i = 0; i < NR_COM; i++)
    {
      outb (0x80, COM_LINE (i));
      outb (0x00, COM_HIBAUD (i));
      outb (0x60, COM_LOBAUD (i));	/* 1200 baud, good for mouse */
      outb (0x03, COM_LINE (i));	/* 8 bits, one stop, no parity */
      outb (0x0b, COM_MLINE (i));
      outb (0x01, COM_IER (i));		/* interrupt, when data home */

      irq_request(com[i].irq, IRQ_FAST, &serial_int, "serial port");
      (void) inb (COM_DATA (i));
    }
}

void
serial_init ()
{
  setattr(ATTR(GREEN,BLACK));
  printf ("Serial/Mouse driver\n");
  setattr(DEFAULT_ATTR);
  
  register_chrdev (MAJOR_NR, "serial", &serial_fops);
  
  dfs_add ("sp0", MKDEV (MAJOR_NR, 0), S_IRWXU | S_IFCHR);
  dfs_add ("sp1", MKDEV (MAJOR_NR, 1), S_IRWXU | S_IFCHR);
  dfs_add ("sm0", MKDEV (MAJOR_NR, 0 | MOUSE_FLAG), S_IRWXU | S_IFCHR);
  dfs_add ("sm1", MKDEV (MAJOR_NR, 1 | MOUSE_FLAG), S_IRWXU | S_IFCHR);
  
/*  serial_on ();*/
}
