/*
   dev/ss.c
   Sound System Interface 1.0, 14012000fr1648
 */

#include <asm/system.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <phantom/fs.h>
#include <errno.h>
#include <phantom/major.h>
#include <string.h>
#include <phantom/audio.h>

#define SB_IO		0x220
#define SB_IRQ		0x05
#define SB_DMA		0x01
#define SB_INT		SB_IRQ+IRQ_HOLE

#define DSP_RESET	(SB_IO+0x06)
#define DSP_STATUS	(SB_IO+0x0C)
#define DSP_IN		(SB_IO+0x0A)
#define DSP_OUT		(SB_IO+0x0C)
#define DSP_DATAAVAIL	(SB_IO+0x0E)
#define DSP_DATAAVAIL16	(SB_IO+0x0F)

#define CMD_SPEAKERON	0xD1
#define CMD_SPEAKEROFF  0xD3
#define CMD_DMARUN	0xD4
#define CMD_DMASTOP	0xD0
#define CMD_STC		0x40
#define CMD_GETDSPVER	0xE1

#define MAJOR_NR	SOUND_MAJOR

#define SS_RATES_NR	16

int ss_rates[SS_RATES_NR] = {8000, 5513, 16000, 11025, 27429,
18900, 32000, 22050, 0, 37800, 0, 44100, 48000, 33075, 9600,
6615};

word ss_bases[]={0x530, 0x604, 0xE80, 0xF40};

int sb_rate = 11025;
int sb_dsp = 0;
int sb_mode = 0;
int sb_initialized = 0;

int
dsp_write (unsigned char c)
{
  int ret;

  for (ret = 0; ret < 0xFFFF; ret++)
    if (!(inb_p (DSP_STATUS) & 0x80))
      {
	outb_p (c, DSP_OUT);
	return (0);
      }

  printk ("sb_write: time out!\n");
  return (-1);
}

int
dsp_read ()
{
  int ret;

  for (ret = 0; ret < 0xFFFF; ret++)
    if (inb_p (DSP_DATAAVAIL) & 0x80)
      return (inb_p (DSP_IN));

  printk ("dsp_read: time out!\n");
  return (-1);
}

void
sb_mono_int ()
{
  dsp_write (0x14);
  dsp_write (0xFF);
  dsp_write (0xFF);
  (void) inb (DSP_DATAAVAIL);

  IRQ_END
}

void
sb_monoauto_int ()
{
  if (sb_mode & MODE_16BIT)
    (void) inb (DSP_DATAAVAIL16);
  else
    (void) inb (DSP_DATAAVAIL);

  IRQ_END
}

int
sb_setup (int rate, int mode, unsigned char **buff, int *buff_size)
{
#ifdef SBDEBUG
  printf ("sb_setup(%d, %d, ...);\n", rate, mode);
#endif

  if (sb_initialized == 0)
    {
      outb_p (0x01, DSP_RESET);
      sys_sleep (10);
      outb_p (0x00, DSP_RESET);

      if (dsp_read () != 0xAA)
	{
	  printf ("sound blaster not found\n");
	  return (-1);
	}

#ifdef SBDEBUG
      printf ("dsp reset ok\n");
#endif

      buff = dma_alloc (SB_DMA, 65536);
      *buff_size = 65536;

      dsp_write (CMD_GETDSPVER);
      sb_dsp = dsp_read () << 8;
      sb_dsp |= dsp_read ();

      printf ("  SB, IO=0x%x  IRQ=%d  DMA=%d  DSP=%x.%02x\n",
	      SB_IO, SB_IRQ, SB_DMA, sb_dsp >> 8, sb_dsp & 0xFF);

      dsp_write (CMD_SPEAKERON);

      enable_irq (SB_IRQ);

      sb_initialized = 1;
    }

  sb_mode = mode;
  sb_rate = rate;

#ifdef SBDEBUG
  printf ("sb.c: time constant=%d\n", 256 - (1000000 / sb_rate));
#endif

  dma_setuptransfer (SB_DMA, DMA_WRITE_AI, 65536);

  dsp_write (CMD_STC);		/* set time constant */
  dsp_write (256 - (1000000 / sb_rate));

  if (sb_dsp < 0x150)		/* use autoinit? */
    {
#ifdef SBDEBUG
      printf ("sb.c: using non-autoinit interrupt\n");
#endif

      set_int_gate (SB_INT, &sb_mono_int);
      dsp_write (0x14);		/* data type, mono, 8-bit */
    }
  else
    {
#ifdef SBDEBUG
      printf ("sb.c: auto-init interrupt in progress\n");
#endif

      set_int_gate (SB_INT, &sb_monoauto_int);
      dsp_write (0x48);		/* set block length */
    }

  dsp_write (0xFF);
  dsp_write (0xFF);

  if (sb_dsp >= 0x150)
    {
      if (sb_rate <= 22000)
	dsp_write (0x1C);	/* start PCM output */
      else
	dsp_write (0x90);	/* high-speed */
    }

  return (0);
}

int
sb_getpos ()
{
  return (dma_getpos (SB_DMA));
}

int 
ss_close ()
{
  dma_free (SB_DMA);

  return (0);
}

struct audio_driver sb_driver =
{
  "Sound System (Crystal/Analog CODEC) Driver 1.0",
  ss_setup,
  ss_close,
  ss_getpos
};
