/* TODO: local or whole task locking... */
#include <phantom/major.h>
#include <phantom/buffer.h>
#include <phantom/dfs.h>
#include <phantom/fs.h>
#include <phantom/sched.h>
#include <errno.h>

#define DFSTASKDEBUG

/* v_num = dir_index, tsk dir index, dir/file type */
#define T_ROOTDIR	0
#define T_TASKDIR	1
#define T_FDDIR		2
#define T_FLAGSBITS	4
#define T_FLAGSMASK	((1 << T_FLAGSBITS) - 1)
#define T_TASKBITS	10	/* -> NR_TASK <= 1024 */
#define T_TASKMASK	((1 << T_TASKBITS) - 1)
#define T_DIRBITS	10	/* second level dir max. entries */
#define T_DIRMASK	((1 << T_DIRBITS) - 1)
/* 24 bits total */

#define T_ISROOT(a)		(((a) & T_FLAGSMASK) == T_ROOTDIR)
#define T_ISTASK(a)		(((a) & T_FLAGSMASK) == T_TASKDIR)
#define T_ISFD(a)		(((a) & T_FLAGSMASK) == T_FDDIR)
#define T_GETTASK(a)		(((a) >> T_FLAGSBITS) & T_TASKMASK)
#define T_GETSECOND(a)		((a) >> (T_FLAGSBITS + T_TASKBITS))
#define T_VNUM(dir,tsk,sec)	(((sec) << (T_TASKBITS + T_DIRBITS)) | ((tsk) << T_FLAGSBITS) | (dir))

struct super_block *task_super;

int task_read_vnode (struct vnode *vnode);
int task_lookup (struct vnode **dst, struct vnode *src, char *name, int len);
int task_readdir (struct file *file, struct dirent *dirent, int count);

struct super_operations task_super_operations =
{
  task_read_vnode,		/* read vnode */
  NULL,				/* write vnode */
  NULL				/* release super */
};

struct vnode_operations task_vnode_operations =
{
  NULL,				/* create */
  NULL,				/* erase */
  task_lookup,			/* lookup */
  NULL,				/* rename */
  NULL,				/* truncate */
  NULL,				/* mkdir */
  NULL				/* rmdir */
};

struct file_operations task_file_operations =
{
  NULL,				/* default seek */
  NULL,				/* read */
  NULL,				/* write */
  task_readdir,			/* readdir */
  NULL,				/* ioctl */
  NULL,				/* open */
  NULL				/* close */
};


struct dirnames_st
  {
    char *str;
    int reclen;
  };

#define NR_TASKENTRIES		3

struct dirnames_st dirnames[] =
{
  {"fd", 2},			/* 0 */
  {"info", 4},			/* 1 */
  {"cwd", 3}			/* 2 */
};

/* FIXME: codelock, task could exit meanwhile */
int
task_lookup (struct vnode **dst, struct vnode *src, char *name, int len)
{
  int task_nr, vnum = src->v_num;

#ifdef DFSTASKDEBUG
  printk ("task_lookup: src=0x%x  name=%s  len=%d\n", src, name, len);
#endif

  if (vnum == DFS_ROOT_VNODE_NUMBER)	/* root dir */
    {
      task_nr = atoi (name);
      if (task_nr < 0 || task_nr >= NR_TASK || task_table[task_nr] == NULL)
	return -ENOENT;

      *dst = vnode_get (task_super, T_VNUM (T_ROOTDIR, task_nr, 0));

      return 0;
    }

  task_nr = T_GETTASK (vnum);

  if (T_ISROOT (vnum))		/* task dir */
    {
      int i;

/* dot case should be handled in /fs/vnode.c */
      if (len == 2 && name[0] == '.' && name[1] == '.')		/* dotdot */
	{
	  *dst = vnode_get (task_super, DFS_ROOT_VNODE_NUMBER);
	  return 0;
	}

      for (i = 0; i < NR_TASKENTRIES; i++)
	{
	  if (dirnames[i].reclen == len && !strncmp (dirnames[i].str, name, len))
	    {
	      *dst = vnode_get (task_super, T_VNUM (T_TASKDIR, task_nr, i));
	      return 0;
	    }
	}

#ifdef DFSTASKDEBUG
      printk ("task_lookup: not found, somewhere in task, task_nr=%d\n",
	      task_nr, i);
#endif

      return -ENOENT;
    }

  if (T_ISTASK (vnum))		/* inside task/nr/??/ */
    {
      int i;

      if (len == 2 && name[0] == '.' && name[1] == '.')		/* dotdot */
	{
	  *dst = vnode_get (task_super, T_VNUM (T_ROOTDIR, task_nr, 0));
	  return 0;
	}

      i = T_GETSECOND (src->v_num);

      switch (i)
	{
	case 0:		/* fd */
	  {
	    int fd_nr = atoi (name);

	    if (fd_nr < 0 || fd_nr >= NR_OPEN)
	      return -ENOENT;

	    *dst = vnode_get (task_super, T_VNUM (T_FDDIR, task_nr, fd_nr));
	    return 0;
	  }

	default:
#ifdef DFSTASKDEBUG
	  printk ("task_lookup: not found, somewhere in task, task_nr=%d, second=%d\n",
		  task_nr, i);
#endif

	  return -ENOENT;
	}
    }

#ifdef DFSTASKDEBUG
  printk ("task_lookup: not found ][");
#endif

  return -ENOENT;
}

int
task_read_vnode (struct vnode *vnode)
{
  int vnum = vnode->v_num;

  vnode->v_op = &task_vnode_operations;
  vnode->f_op = &task_file_operations;
  vnode->v_mount = NULL;
  vnode->v_dev = 0;
  vnode->v_atime = 0;
  vnode->v_ctime = 0;
  vnode->v_mtime = 0;

  if (vnum == DFS_ROOT_VNODE_NUMBER || T_ISROOT (vnum)
      || (T_ISTASK (vnum) && T_GETSECOND(vnum)==0))
    {
      vnode->v_mode =
	S_IFDIR | S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
    }
  else if (T_ISFD (vnum) || (T_ISTASK(vnum) && T_GETSECOND(vnum)!=0))
    vnode->v_mode = S_IFREG | S_IRUSR | S_IRGRP | S_IROTH;
  else
    {
      printk ("task_read_vnode: strange, unknown type, vnum=0x%x\n", vnum);
      vnode->v_mode = S_IFREG | S_IRUSR | S_IRGRP | S_IROTH;
    }

  return 0;
}

#define EDOTS 		2	/* dotdot and dot entries */
int
do_dots (struct file *file, struct dirent *dirent)
{
  switch (file->f_pos)
    {
    case 0:			/* dotdot */
      strcpy (dirent->d_name, "..");
      dirent->d_reclen = 2;
      file->f_pos++;
      return 1;

    case 1:
      strcpy (dirent->d_name, ".");
      dirent->d_reclen = 1;
      file->f_pos++;
      return 1;
    }
}

int
task_readdir (struct file *file, struct dirent *dirent, int count)
{
  char numero[16];
  struct vnode *vnode = file->f_vnode;
  int i, len, vnum = vnode->v_num, task_nr;

#ifdef DFSTASKDEBUG
  printk ("task_readdir: file=0x%x  pos=%d  dirent=0x%x  count=%d\n",
	  file, file->f_pos, dirent, count);
#endif

  if (vnum == DFS_ROOT_VNODE_NUMBER)	/* root dir */
    {
      if (file->f_pos >= NR_TASK + EDOTS)
	return 0;

      if (do_dots (file, dirent) == 1)
	return 1;

      for (i = file->f_pos - EDOTS; i < NR_TASK; i++)
	{
	  if (task_table[i] != NULL)
	    {
	      len = sprintf (numero, "%d", i);
	      strcpy (dirent->d_name, numero);
	      dirent->d_reclen = len;
	      file->f_pos = i + 1 + EDOTS;	/* switch to next! */
	      return 1;
	    }
	}
      file->f_pos = NR_TASK + EDOTS;

      return 0;
    }

  task_nr = T_GETTASK (vnum);

  if (T_ISROOT (vnum))		/* object in root dir ==> task dir */
    {
      if (file->f_pos >= NR_TASKENTRIES + EDOTS)
	return 0;

      if (do_dots (file, dirent) == 1)
	return 1;

      strcpy (dirent->d_name, dirnames[file->f_pos - EDOTS].str);
      dirent->d_reclen = dirnames[file->f_pos - EDOTS].reclen;
      file->f_pos++;

      return 1;
    }

  if (T_ISTASK (vnum))		/* in task dir, fd dir (the only one there) */
    {
      if (file->f_pos >= NR_OPEN + EDOTS)
	return 0;

      if (do_dots (file, dirent) == 1)
	return 1;

      for (i = file->f_pos - EDOTS; i < NR_OPEN; i++)
	{
	  if (task_table[task_nr]->filp[i] != NULL)
	    {
	      len = sprintf (numero, "%d", i);
	      strcpy (dirent->d_name, numero);
	      dirent->d_reclen = len;
	      file->f_pos = i + 1 + EDOTS;	/* switch to next! */
	      return 1;
	    }
	}

      file->f_pos = NR_OPEN + EDOTS;

      return 0;
    }

  return 0;
}

int
task_read_super (struct super_block *sb)
{
  struct vnode *vnode;

#ifdef DFS_DEBUG
  printk ("task_readsuper\n");
#endif

  sb->s_dev = 0;
  sb->s_op = &task_super_operations;
  vnode = vnode_get (sb, DFS_ROOT_VNODE_NUMBER);
  vnode->v_size = 0;
  vnode->v_dev = 0;
  vnode->v_mode = S_IFDIR | S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
  vnode->v_cover = dfs_super->s_mount;
  sb->s_mount = vnode;
  task_super = sb;

  return 0;
}
