/*
   fs/fat/super.c
   FAT 12/16/32 mounting.
 */

#include <phantom/fs.h>
#include <phantom/fat.h>
#include <phantom/major.h>
#include <errno.h>
#include <phantom/buffer.h>
#include <conio.h>
#include <phantom/mm.h>

extern int fat_read_vnode (struct vnode *vnode);	/* vnode.c */
int fat_release_super (struct super_block *sb);

struct super_operations fat_super_operations = {
  fat_read_vnode,		/* read vnode */
  NULL,				/* write vnode */
  fat_release_super		/* umount */
};

/* fat dirtyness byte-field will be used...
   good idea */
int
fat_release_super (struct super_block *sb)
{
  int blk, tbl;
  struct bhead *b, *bb;

  printf ("fat_release_super on %02X:%02X.\n",
	  MAJOR (sb->s_dev), MINOR (sb->s_dev));

/* releasing other fats (if more than one) */
  if (sb->u.fat.fatcount > 1)
    {
      for (blk = 0; blk < sb->u.fat.fatsectors; blk++)
	{
	  if (sb->u.fat.table_dirt[blk])
	    {
	      b = bread (sb->s_dev, sb->u.fat.reserved + blk);
	      for (tbl = 1; tbl < sb->u.fat.fatcount; tbl++)
		{
		  bb =
		    bempty (sb->s_dev,
			    sb->u.fat.reserved + tbl * sb->u.fat.fatsectors +
			    blk);
		  block(bb);

		  memcpy (bb->data, b->data, 512);

		  bunlock(bb);
		  brelse (bb);
		}
	      brelse (b);
	    }
	}
      kfree(sb->u.fat.table_dirt);
    }

  vnode_release (sb->s_mount);

  bzero (sb, sizeof (sb));	/* superblock release */
  return (0);
}

/* this is really annoying */
int
fat_read_super (struct super_block *sb)
{
  int i, j;
  struct vnode *vnode;
  struct bhead *bh;
  char *buff;

  bh = bread (sb->s_dev, 0);
  if (!bh)
    return -EIO;
  buff = bh->data;

  if (buff[0x36 + 3] == '1')
    {
      if (buff[0x36 + 4] == '6')
	{
	  printk ("FAT16 found.\n");
	  sb->u.fat.type = 16;
	}
      else if (buff[0x36 + 4] == '2')
	{
	  printk ("FAT12 found.\n");
	  sb->u.fat.type = 12;
	}
    }
  else if (buff[0x55] == '3' && buff[0x56] == '2')
    {
      printk ("FAT32 found.\n");
      sb->u.fat.type = 32;
    }
  else
    {
      printf ("FAT not found at all.\n");
      brelse (bh);
      return (-EIO);
    }

  sb->s_blocksize = 512;
  sb->u.fat.clusterblocks = *(unsigned char *) (buff + 0xd);
  sb->u.fat.fatcount = *(unsigned char *) (buff + 0x10);
  sb->u.fat.reserved = *(unsigned short *) (buff + 0xe);

  if (sb->u.fat.type == 32)
    {
      sb->u.fat.rootentries = 0;
      sb->u.fat.fatsectors = *(unsigned int *) (buff + 0x24);
      sb->u.fat.rootdircluster = *(unsigned int *) (buff + 0x2C);
    }
  else
    {
      sb->u.fat.rootentries =
	*((unsigned short *) (buff + 0x11)) * ITEMSIZE / sb->s_blocksize;
      sb->u.fat.fatsectors = *(unsigned short *) (buff + 0x16);
      sb->u.fat.rootdircluster = 0;
    }

  sb->u.fat.hidden = *(unsigned short *) (buff + 0x1c);
  sb->u.fat.size = *(unsigned int *) (buff + 0x20);
  sb->u.fat.rootentrypos = sb->u.fat.fatcount * sb->u.fat.fatsectors +
    sb->u.fat.reserved;
  sb->u.fat.datastart = sb->u.fat.rootentrypos +
    sb->u.fat.rootentries - 2 * sb->u.fat.clusterblocks;
  sb->u.fat.itemcount = sb->s_blocksize / ITEMSIZE;

  i = sb->u.fat.fatsectors * 512;
  switch (sb->u.fat.type)
    {
    case 12:
      i *= 3;
      i >>= 1;
      break;
    case 16:
      i >>= 1;
      break;
    case 32:
      i >>= 2;
      break;
    }
  brelse (bh);
  sb->u.fat.fatentries = i;

#ifdef SFATDEBUG
  printk ("datastart=%d  reserved=%d  fatsectors=%d\n", sb->u.fat.datastart,
	  sb->u.fat.reserved, sb->u.fat.fatsectors);
  printk ("rootentrypos=%d  itemcount=%d\n", sb->u.fat.rootentrypos,
	  sb->u.fat.itemcount);
  printk ("fatentries=%d  clusterblocks=%d\n", sb->u.fat.fatentries,
	  sb->u.fat.clusterblocks);
#endif

/*  if (sb->u.fat.fatentries + 2 > get_dev_size (sb->s_dev) * 2)
    sb->u.fat.fatentries =
      get_dev_size (sb->s_dev) * 2 - sb->u.fat.datastart + 2;

  if (sb->u.fat.fatentries <= 0)
    {
      printf ("invalid fat boot!\n");
      return (-EIO);
    }
*/

/* zeroed by default */
  sb->u.fat.table_dirt =
    (unsigned char *) kmalloc (sb->u.fat.fatsectors, PG_KERNEL);

  sb->s_op = &fat_super_operations;
  vnode = vnode_get (sb, FAT_ROOT_VNODE_NUMBER);

  if (sb->u.fat.type != 32)
    {
      vnode->u.fat.cluster = 0;
      vnode->v_size = sb->u.fat.rootentries * 512;
    }
  else
    {
      vnode->u.fat.cluster = sb->u.fat.rootdircluster;
      vnode->v_size = fat_getsize (sb, sb->u.fat.rootdircluster)
	* 512 * sb->u.fat.clusterblocks;
#ifdef F32DEBUG
      printk ("  root has %d bytes\n", vnode->v_size);
      printk ("  cluster = %d\n", vnode->u.fat.cluster);

#endif
    }

  vnode->v_mode = S_IFDIR | S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;

  sb->s_mount = vnode;

  return 0;
}
