/*
   fs/fs.c
   Phantom File System ]I[.

   WRITEME: fs recognition
 */

#include <phantom/major.h>
#include <string.h>
#include <phantom/fs.h>
#include <phantom/dfs.h>
#include <phantom/sched.h>
#include <conio.h>
#include <asm/sysinfo.h>
#include <asm/system.h>

extern int fat_read_super (struct super_block *);

struct super_block super_table[NR_SUPER] = { {0,}, };

struct file_system_type fs_table[] = {
  {&fat_read_super, "fat"},
  {&dfs_read_super, "dev"},
  {&task_read_super, "task"},
/*  {&sys_read_super, "sys"},*/
  {NULL,}
};

struct file_system_type *
get_fs_type (char *name)
{
  struct file_system_type *tmp = fs_table;

  while (tmp->name)
    {
      if (!strcmp (name, tmp->name))
	return tmp;
      tmp++;
    }
  return NULL;
}

struct super_block *
get_free_super ()
{
  ulg fl;
  struct super_block *super;

  get_flags (fl);
  cli ();
  for (super = super_table; super < super_table + NR_SUPER; super++)
    if (!super->s_used)
      {
	super->s_used = 1;
	set_flags (fl);
	return super;
      }

  set_flags (fl);
  printf ("get_free_super: cannot get free super\n");
  return NULL;
}

int
sys_mount (char *devname, char *dir, char *fsname, unsigned int flags)
{
  struct super_block *super;
  struct file_system_type *fs;
  struct vnode *vnode;
  struct stat st;
  int ret, dev;

  if (!fsname)
    {
      printk ("sys_mount: no fsname given...\n");
      return -EINVAL;
    }

  fs = get_fs_type (fsname);
  if (!fs)
    {
      printk ("sys_mount: unknown file system \"%s\"\n", fsname);
      return -EINVAL;
    }

  if (ret = sys_stat (devname, &st))
    {
      printk ("sys_mount: stat fails, ret=%d\n", ret);
      return ret;
    }

  dev = st.st_dev;
  printk ("sys_mount: recognized device %02X:%02X\n",
	  MAJOR (dev), MINOR (dev));

  for (super = super_table; super < super_table + NR_SUPER; super++)
    if (super->s_used && super->s_dev == dev)
      {
	printk ("sys_mount: device %02X:%02X already mounted\n",
		MAJOR (dev), MINOR (dev));
	return -EIO;
      }

  if (!(super = get_free_super ()))
    {
      printk ("sys_mount: cannot get free super\n");
      return -EIO;
    }

  if (ret = vnode_open (&vnode, dir, O_RDONLY))
    {
      printk ("sys_mount: cannot access mount destination (ret=%d)\n", ret);
      return ret;
    }

  if (vnode->v_mount || vnode == ctask->root)
    {
      printk ("sys_mount: vnode already act as mount-point (or root)...\n");
      vnode_release (vnode);
      return -EBUSY;
    }

  super->s_dev = dev;
  if (ret = fs->read_super (super))
    {
      printk ("sys_mount: read super failed\n");
      super->s_dev = 0;
      vnode_release (vnode);
      return ret;
    }

  vnode->v_mount = super->s_mount;
  vnode->v_mount->v_cover = vnode;

  return 0;
}

int
sys_umount (char *str)
{
  struct super_block *sb;
  struct vnode *vnode;
  struct stat st;
  int vnodes;
  int ret;

  if (ret = sys_stat (str, &st))
    {
      printk ("sys_umount: cannot stat %s\n", str);
      return ret;
    }

  for (sb = super_table; sb < super_table + NR_SUPER; sb++)
    if (sb->s_dev == st.st_dev)
      break;

  if (sb == super_table + NR_SUPER)
    {
      printf ("sys_mount: nothing mounted on %02X:%02X\n",
	      MAJOR (st.st_dev), MINOR (st.st_dev));
      return -EINVAL;
    }

  vnodes = 0;
  for (vnode = vnode_table; vnode < vnode_table + NR_VNODE; vnode++)
    if (vnode->v_used && vnode->v_sb == sb && vnode != sb->s_mount)
      vnodes++;			/* count 'extra' vnodes */

  if (vnodes)
    {
      printk ("sys_umount: fatal, %d active vnodes found\n", vnodes);
      return -EAGAIN;
    }

  if (sb->s_mount->v_used != 1)
    {
      printk ("sys_umount: mount vnode has big v_used=%d\n",
	      sb->s_mount->v_used);
      return -EAGAIN;
    }

  if (sb->s_mount->v_cover)
    vnode_release (sb->s_mount->v_cover);

  vnode_release (sb->s_mount);
  bzero (sb, sizeof (struct super_block));

  printk ("sys_umount: unmounted!\n");

  return 0;
}

int
mount_root (int dev, struct super_block *sb)
{
  struct vnode *vnode;
  struct file_system_type *tmp = &fs_table[0];

  sb->s_dev = dev;

  if (MAJOR (dev) == FLOPPY_MAJOR)
    {
      printf ("Insert root floppy and press any key.\n");
      getch ();
    }

  if (!tmp->read_super (sb))
    {
      vnode = sb->s_mount;
      ctask->pwd = vnode;
      ctask->root = vnode;
      ctask->pwd_str[0] = '/';
      ctask->pwd_str[1] = 0;
      vnode->v_used += 2;	/* this should be 3,
				   1 for being mount-point, +2
				   for being current directory and root */
#ifdef FSDEBUG
      printf ("root vnode=0x%p\n", vnode);
#endif
      return (0);
    }

  printf ("mount_root: failed\n");

  sb->s_used = 0;
  return (-1);
}

/* fs initialization, root mounting */
void
fs_init ()
{
  int k, root_dev;
  struct super_block *sb;
  char tmp[256], name[256];
  struct stat st;

  setattr (ATTR (GREEN, BLACK));
  printf ("File System Interface\n");
  setattr (DEFAULT_ATTR);

/* dfs: special mount, vnode nr. 0 */
/* DO NOT SWAP ORDER! */
  sb = get_free_super ();
  dfs_read_super (sb);
  sb = get_free_super ();
  task_read_super (sb);
/*  sb = get_free_super ();
  sys_read_super(sb);*/

  root_dev = sysinfo.root_dev;
  if (!root_dev)
    {
      printf ("Root-offline mode no longer supported.\nSystem halted.\n");
      for (;;);
/*      ctask->pwd_str[0] = '/';
      ctask->pwd_str[1] = 0;*/
    }

  sb = get_free_super ();

  while (mount_root (root_dev, sb))
    {
      sync_dev (root_dev);
      printf ("Cannot mount root on device %02X:%02X\n" \
	      "Press ESC to enter other boot device\n" \
	      "or any other key to flush cache and try again.\n",
	      MAJOR (root_dev), MINOR (root_dev));
      k = getch ();
      if (k == 0x1B)
	{
	  printf ("Device name: ");
	  gets (tmp);		/* FIXME: do this safer... */
	  putchar ('\n');
	  strcpy (name, "/" DFS_NAME "/");
	  strcat (name, tmp);
	  if (!sys_stat (name, &st))
	    {
	      root_dev = st.st_dev;
	      printf ("Device check ok, %02X:%02X, trying to mount\n",
		      MAJOR (root_dev), MINOR (root_dev));
	    }
	  else
	    printf ("Invalid device given.\n");
	}
    }
}

int
fs_done ()
{
  struct super_block *tmp;

  setattr (ATTR (BLACK, GREEN));
  printf ("File System Interface\n");
  setattr (DEFAULT_ATTR);

  for (tmp = super_table; tmp < super_table + NR_SUPER; tmp++)
    if (tmp->s_op && tmp->s_op->release_super)
      tmp->s_op->release_super (tmp);

  return (0);
}
