#include <phantom/fs.h>
#include <errno.h>
#include <phantom/sched.h>
#include <phantom/dfs.h>

int
sys_lseek (unsigned int fd, off_t offset, int whence)
{
  struct file *file;
  long tmp;

  if (fd >= NR_OPEN || !(file = ctask->filp[fd]))
    return (-EBADF);

  if (file->f_used && file->f_op && file->f_op->lseek)
    return (file->f_op->lseek (file, offset, whence));

  switch (whence)
    {
    case SEEK_SET:
      tmp = offset;
      break;
    case SEEK_CUR:
      tmp = offset + file->f_pos;
      break;
    case SEEK_END:
      tmp = file->f_vnode->v_size + offset;
      break;
    default:
      return (-EINVAL);
    }
  if (tmp < 0)
    return (-EINVAL);
  file->f_pos = tmp;
  return (tmp);
}

int
sys_getdents (unsigned int fd, struct dirent *dirent, unsigned int count)
{
  int ret;
  struct file *file;

  if (fd >= NR_OPEN || !(file = ctask->filp[fd]))
    return (-EBADF);

/* dfs root offline mode patch */
  if (!file->f_vnode)
    {
#ifdef FS_DEBUG
      printf ("readdir: offline, file->f_pos=%d\n", file->f_pos);
#endif

      if (!file->f_pos)
	{
	  file->f_pos++;
	  strncpy (dirent->d_name, DFS_NAME, PATH_MAX);
	  dirent->d_reclen = DFS_NAME_LEN;
	  return 1;
	}

      return 0;			/* end of dir mark */
    }

#ifdef FS_DEBUG
  printf ("sys_readdir: ordinary vnode (0x%x)\n", file->f_vnode);
#endif

/* zerotask root == absolute root */
/* TODO: hiding of original 'dev' dir on filesystem, if it exists there :) */
  if(file->f_vnode == ztask->root)		/* show "dev" */
    {
  if (!file->f_pos)		/* show dfs as first */
    {
      file->f_pos++;
      strncpy (dirent->d_name, DFS_NAME, PATH_MAX);
      dirent->d_reclen = DFS_NAME_LEN;
      return 1;
    }

  if (file->f_used && file->f_op && file->f_op->readdir)
    {
      file->f_pos--;		/* dfs hack */
      ret = file->f_op->readdir (file, dirent, count);
      file->f_pos++;
      return ret;
    }
   } else {
  if (file->f_used && file->f_op && file->f_op->readdir)
      return file->f_op->readdir (file, dirent, count);
    }
   
  return -ENOTDIR;
}

int
  sys_read (unsigned int fd, char *buff, size_t count)
{
  struct file *file;

  if (fd >= NR_OPEN || !(file = ctask->filp[fd]))
      return -EBADF;

  if (!count)
      return 0;
/*  if (!(file->f_mode &))
   return (-EBADF); */

  if (file->f_used && file->f_op && file->f_op->read)
      return file->f_op->read (file, buff, count);

    return -EINVAL;
}

int
  sys_write (unsigned int fd, char *buff, size_t count)
{
  struct file *file;

  if (fd >= NR_OPEN || !(file = ctask->filp[fd]))
      return -EBADF;

  if (!count)
      return 0;

/*  if (!(file->f_mode &))
   return (-EBADF); */

  if (file->f_used && file->f_op && file->f_op->write)
      return file->f_op->write (file, buff, count);

    return -EINVAL;
}

int
  sys_ioctl (unsigned int fd, unsigned int cmd, unsigned int arg)
{
  struct file *file;

  if (fd >= NR_OPEN || !(file = ctask->filp[fd]))
      return (-EBADF);

  if (file->f_used && file->f_op && file->f_op->ioctl)
      return (file->f_op->ioctl (file, cmd, arg));
    return (-EINVAL);
}
