#ifndef I386_IDTBUILD_H
#define I386_IDTBUILD_H

#include <asm/kernel.h>

#define __STR(n)		#n
#define STR(n)			__STR(n)

#define QUICK_SAVE \
	"cld\n" \
	"push %es\n" \
	"push %ds\n" \
	"pushl %eax\n" \
	"pushl %edx\n" \
	"pushl %ecx\n" \
	"movl $" STR(KERNEL_DS) ",%edx\n" \
	"mov %dx,%ds\n" \
	"mov %dx,%es\n"

#define QUICK_LOAD \
	"popl %ecx\n" \
	"popl %edx\n" \
	"popl %eax\n" \
	"pop %ds\n" \
	"pop %es\n"

#define TOTAL_LOAD \
	"popl %ebx\n" \
	"popl %ecx\n" \
	"popl %edx\n" \
	"popl %esi\n" \
	"popl %edi\n" \
	"popl %ebp\n" \
	"popl %eax\n" \
	"pop %ds\n" \
	"pop %es\n" \
	"pop %fs\n" \
	"pop %gs\n"

/* order of this (and that above) cannot be changed because of kernel/syscall.S... */

#define TOTAL_SAVE \
	"cld\n" \
	"push %gs\n" \
	"push %fs\n" \
	"push %es\n" \
	"push %ds\n" \
	"pushl %eax\n" \
	"pushl %ebp\n" \
	"pushl %edi\n" \
	"pushl %esi\n" \
	"pushl %edx\n" \
	"pushl %ecx\n" \
	"pushl %ebx\n" \
	"movl $" STR(KERNEL_DS) ",%edx\n" \
	"mov %dx,%ds\n" \
	"mov %dx,%es\n" \
	"mov %dx,%fs\n" \
	"mov %dx,%gs\n"

#define ASM_TOTAL_LOAD \
	popl %ebx; \
	popl %ecx; \
	popl %edx; \
	popl %esi; \
	popl %edi; \
	popl %ebp; \
	popl %eax; \
	pop %ds; \
	pop %es; \
	pop %fs; \
	pop %gs;

#define ASM_TOTAL_SAVE \
	cld; \
	push %gs; \
	push %fs; \
	push %es; \
	push %ds; \
	pushl %eax; \
	pushl %ebp; \
	pushl %edi; \
	pushl %esi; \
	pushl %edx; \
	pushl %ecx; \
	pushl %ebx; \
	movl $KERNEL_DS,%edx; \
	mov %dx,%ds; \
	mov %dx,%es; \
	mov %dx,%fs; \
	mov %dx,%gs;

#define LCK_1(mask) \
	"inb $0x21,%al\n" \
	"orb $" #mask ",%al\n" \
	"movb %al,port_21\n" \
	"outb %al,$0x21\n" \
	"jmp 1f\n" \
	"1: jmp 1f\n" \
	"1: movb $0x20,%al\n" \
	"outb %al,$0x20\n"

#define LCK_2(mask) \
	"inb $0xA1,%al\n" \
	"orb $" #mask ",%al\n" \
	"movb %al,port_A1\n" \
	"outb %al,$0xA1\n" \
	"jmp 1f\n" \
	"1: jmp 1f\n" \
	"1: movb $0x20,%al\n" \
	"outb %al,$0xA0\n" \
	"jmp 1f\n" \
	"1: jmp 1f\n" \
	"1: outb %al,$0x20\n"

#define ULCK_1(mask) \
	"inb $0x21,%al\n" \
	"andb $~("#mask"),%al\n" \
	"movb %al,port_21\n" \
	"outb %al,$0x21\n"

#define ULCK_2(mask) \
	"inb $0xA1,%al\n" \
	"andb $~("#mask"),%al\n" \
	"movb %al,port_A1\n" \
	"outb %al,$0xA1\n"

#define BUILD_IRQ(nr, chip, mask, emask) \
void complex_IRQ_##nr (); \
void fast_IRQ_##nr (); \
void wrong_IRQ_##nr (); \
__asm__( \
".align 4\n" \
"complex_IRQ_"#nr":\n" \
	"push $0x0\n" \
	TOTAL_SAVE \
	LCK_##chip(mask) \
	"incl nr_running_irq\n"\
	"sti\n" \
	"pushl $"#nr"\n" \
	"call complex_IRQ_handler\n" \
	"addl $0x04,%esp\n" \
       	"cli\n" \
	ULCK_##chip(mask) \
	"decl nr_running_irq\n" \
	"jmp irq_syscall_return\n" \
	".align 4\n" \
"fast_IRQ_"#nr":\n" \
	QUICK_SAVE \
	LCK_##chip(mask) \
	"pushl $" #nr "\n" \
	"call fast_IRQ_handler\n" \
	"addl $0x04,%esp\n" \
	"cli\n" \
	ULCK_##chip(mask) \
	QUICK_LOAD \
	"iret\n" \
".align 4\n" \
"wrong_IRQ_" #nr ":\n" \
	QUICK_SAVE \
	LCK_##chip(mask) \
	QUICK_LOAD \
	"iret\n" \
	);

#define BUILD_TRAP(nr, signal, name, banner, task) \
void name (struct regdump *regdump) \
{ \
task->trap_nr = nr; \
printregdump (banner, regdump); \
send_sig (signal, task); \
}

/* 
Default trap used for exceptions without error_code.
Space for error_code is simply filled in with zero.
*/
#define DEFAULT_TRAP(name) \
void trap_##name (); \
__asm__( \
".align 4\n" \
"trap_"#name":\n" \
"pushl $0x00\n" \
TOTAL_SAVE \
"movl %esp,%ebx\n" \
"pushl %ebx\n" \
"call "#name"\n" \
"addl $0x04,%esp\n" \
TOTAL_LOAD \
"addl $0x04,%esp\n" \
"iret\n" \
)

/* 
A bit simpler case. Exceptions with error_code.
Error code should be 32-bites wide, although
only 16 bits are mostly used (reserved exception, etc.)
ATHELP lied...
*/

#define ERROR_CODE_TRAP(name) \
void trap_##name (); \
__asm__( \
".align 4\n" \
"trap_"#name":\n" \
TOTAL_SAVE \
"movl %esp,%ebx\n" \
"pushl %ebx\n" \
"call "#name"\n" \
"addl $0x04,%esp\n" \
TOTAL_LOAD \
"addl $0x04,%esp\n" \
"iret\n" \
)

#define set_gate(gate_addr,type,dpl,addr) \
do  { \
  int __d0, __d1; \
  __asm__ __volatile__ ("movw %%dx,%%ax\n\t" \
	"movw %4,%%dx\n\t" \
	"movl %%eax,%0\n\t" \
	"movl %%edx,%1" \
	:"=m" (*((long *) (gate_addr))), \
	 "=m" (*(1+(long *) (gate_addr))), "=&a" (__d0), "=&d" (__d1) \
	:"i" ((short) (0x8000+(dpl<<13)+(type<<8))), \
	 "3" ((char *) (addr)),"2" (KERNEL_CS << 16)); \
} while(0)

#define set_int_gate(n,addr) \
	set_gate((u32 *)&idt[n],14,0,(u32)addr)

#define set_trap_gate(n,addr) \
	set_gate((u32 *)&idt[n],15,0,(u32)addr)

#define set_system_gate(n,addr) \
	set_gate((u32 *)&idt[n],15,3,(u32)addr)

#define set_call_gate(a,addr) \
	set_gate((u32 *)a,12,3,(u32)addr)

#endif
