#ifndef ATACD_H
#define ATACD_H

#include <types.h>

struct cd_cap_st
  {
    word data_length;
    byte medium_type;
#define MDT_UNKNOWN     0x00
#define MDT_DATA_120    0x01
#define MDT_AUDIO_120   0x02
#define MDT_COMB_120    0x03
#define MDT_PHOTO_120   0x04
#define MDT_DATA_80     0x05
#define MDT_AUDIO_80    0x06
#define MDT_COMB_80     0x07
#define MDT_PHOTO_80    0x08
#define MDT_NO_DISC     0x70
#define MDT_DOOR_OPEN   0x71
#define MDT_FMT_ERROR   0x72
    byte reserved1[5];

    byte page_code;
#define CAP_PAGE        0x2a
    byte param_len;
    byte reserved2[2];

    byte audio_play:1;		/* audio play supported */
    byte composite:1;		/* composite audio/video supported */
    byte dport1:1;		/* digital audio on port 1 */
    byte dport2:1;		/* digital audio on port 2 */
    byte mode2_form1:1;		/* mode 2 form 1 (XA) read */
    byte mode2_form2:1;		/* mode 2 form 2 format */
    byte multisession:1;	/* multi-session photo-CD */
      byte:1;
    byte cd_da:1;		/* audio-CD read supported */
    byte cd_da_stream:1;	/* CD-DA streaming */
    byte rw:1;			/* combined R-W subchannels */
    byte rw_corr:1;		/* R-W subchannel data corrected */
    byte c2:1;			/* C2 error pointers supported */
    byte isrc:1;		/* can return the ISRC info */
    byte upc:1;			/* can return the catalog number UPC */
      byte:1;
    byte lock:1;		/* could be locked */
    byte locked:1;		/* current lock state */
    byte prevent:1;		/* prevent jumper installed */
    byte eject:1;		/* can eject */
      byte:1;
    byte mech:3;		/* loading mechanism type */
#define MECH_CADDY      0
#define MECH_TRAY       1
#define MECH_POPUP      2
#define MECH_CHANGER    4
#define MECH_CARTRIDGE  5
    byte sep_vol:1;		/* independent volume of channels */
    byte sep_mute:1;		/* independent mute of channels */
      byte:6;

    word max_speed;		/* max raw data rate in bytes/1000 */
    word max_vol_levels;	/* number of discrete volume levels */
    word buf_size;		/* internal buffer size in bytes/1024 */
    word cur_speed;		/* current data rate in bytes/1000  */

    /* Digital drive output format description (optional?) */
    byte reserved3;
    byte bckf:1;		/* data valid on failing edge of BCK */
    byte rch:1;			/* high LRCK indicates left channel */
    byte lsbf:1;		/* set if LSB first */
    byte dlen:2;
#define DLEN_32         0	/* 32 BCKs */
#define DLEN_16         1	/* 16 BCKs */
#define DLEN_24         2	/* 24 BCKs */
#define DLEN_24_I2S     3	/* 24 BCKs (I2S) */
      byte:3;
    byte reserved4[2];
  };

struct cd_info
  {
    int flags;			/* device state flags */
    struct cd_cap_st cap;	/* capabilities */
  };

#endif
