#ifndef _BLK_H
#define _BLK_H

#include <phantom/fs.h>
#include <phantom/major.h>		/* MAX_BLK_DEV */
#include <sys/types.h>

#define NR_REQUEST	32
#define CURRENT		(blk_dev[MAJOR_NR].request)

#define BLK_SIZE	512
#define BLK_BITS	9

#define NR_PARTITIONSPERDRIVE		64

struct chrdev_st
  {
    const char *name;
    struct file_operations *fops;
  };

extern struct chrdev_st chrdev[MAX_CHRDEV];
extern struct blkdev_st blkdev[MAX_BLKDEV];

#define exist_blkdev(major)	(blkdev[(major)].fops)
#define exist_chrdev(major)	(chrdev[(major)].fops)

#define READ		0		/* read command */
#define WRITE		1		/* write command */

#define REQUEST	struct request

REQUEST
  {
    int dev;			/* dev == 0 -> not used */
    int cmd;			/* READ, WRITE, READA, WRITEA */
    unsigned int sec;		/* starting block */
    unsigned int nsec;		/* block count */
    int errors;			/* number of errors occurred */
    char *data;			/* source / target */
    struct bhead *bhead;	/* request's buffer head */
    REQUEST *next;		/* next request */    
  };

struct blkdev_st
  {
    const char *name;
    struct file_operations *fops;
    void (*do_request) (REQUEST *);	/* request proc */
    REQUEST *request;			/* current request */
    REQUEST *request_tail;		/* request tail */
    int block_size;			/* block size in bytes */
    int *size;				/* sizes in kilobytes of all subdevices */
    int *begin;				/* and where they 'begin' (partitions) */
    int logdevs;			/* logical devices count */
  };

extern int block_read (struct file *file, char *buff, size_t count);
extern int block_write (struct file *file, char *buff, size_t count);
extern int block_open (struct file *file);
extern void end_request (int dev, int fl_error);

#endif
