#ifndef FAT_H
#define FAT_H

#define DELETED			0xE5
#define A_RONLY			0x01
#define A_HIDDEN		0x02
#define A_SYSTEM		0x04
#define A_LABEL			0x08
#define A_DIR			0x10
#define A_ARCHIVE		0x20

#define FAT_ROOT_VNODE_NUMBER	0

#define FAT_ISDIR(attr)		(((attr) & A_DIR) && !((attr) & A_LABEL))
#define FAT_ISREG(attr)		(!((attr) & A_LABEL) && !((attr) & A_DIR))
#define FAT_ISROOT(vnode)	((vnode)->v_num==FAT_ROOT_VNODE_NUMBER)

#define FAT_READ		0x01
#define FAT_WRITE		0x02
#define FAT_CMDMASK		0x03
#define FAT_DIR			0x04

#define FAT_EOF12		0x0FF8
#define FAT_EOF16		0xFFF8
#define FAT_EOF32		0xFFFFFFF8
#define FAT_EOF			FAT_EOF32

extern struct diritem *fat_getdiritem (char *buff, struct vnode *vnode);

struct diritem
{
  u8 name[11];
  u8 attr;
  u8 lcase;
  u8 ctime_ms;
  u16 ctime;
  u16 cdate;
  u16 adate;
  u16 cluster_hi;  
  u16 time;
  u16 date;
  u16 cluster;
  u32 size;
};

#define FAT

#define ITEMCOUNT	0x10
#define ITEMCOUNTBITS	0x04
#define ITEMCOUNTMASK	0x0F

/* ITEMCOUNT * ITEMSIZE = 512, ok */
#define ITEMSIZE		0x20
#define ITEMSIZEBITS		0x05
#define ITEMSIZEMASK		0x1F

#define VNUM2SEC(vnum)		((vnum) >> ITEMCOUNTBITS)
#define VNUM2SUBSEC(vnum)	((vnum) & ITEMCOUNTMASK)
#define SS2VNUM(sec,subsec)	(((sec) << ITEMCOUNTBITS) | (subsec))

#define BLOCKSIZE		512
#define BLOCKBITS		9
#define TOBLOCK(a)		((a)>>BLOCKBITS)
#define TOBYTES(a)		((a)<<BLOCKBITS)

int fat_toblock (struct super_block *sb, int cluster);
int fat_reserveclusters (struct super_block *sb, int cluster, int count);
int fat_cluster (struct super_block *sb, int *dst, int cluster, int n);
int fat_setfatcell (struct super_block *sb, int cluster, u32 next);
u32 fat_getfatcell (struct super_block *sb, int cluster);
int fat_getsize (struct super_block *sb, int cluster);
void fat_unparse (char *dst, const char *src, int len);
void fat_parse (char *dst, const char *src);
int fat_blockseek (struct super_block *sb, int *dstcluster, int *dstsec,
		   int cluster, int sec, int whence);

#endif
