#ifndef HDREGS_H
#define HDREGS_H

struct hd_info
{
  int cyl;
  int hd;
  int sec;
  int bpi;			/* blocks per interrupt */
};


#define	HD_LBA	0x40		/* use LBA */

#define HD_BLOCKSIZE	0x200

#define HD_IRQ 14		/* the standard disk interrupt */

#define HD_DATA		0x0	/* _CTL when writing */
#define HD_ERROR	0x1	/* see err-bits */
#define HD_NSECTOR	0x2	/* nr of sectors to read/write */
#define HD_SECTOR	0x3	/* starting sector */
#define HD_LCYL		0x4	/* starting cylinder */
#define HD_HCYL		0x5	/* high byte of starting cyl */
#define HD_CURRENT	0x6	/* 101dhhhh , d=drive, hhhh=head */
#define HD_STATUS	0x7	/* see status-bits */
#define HD_FEATURE HD_ERROR	/* same io address, read=error, write=feature */
#define HD_PRECOMP HD_FEATURE	/* obsolete use of this port - predates IDE */
#define HD_COMMAND HD_STATUS	/* same io address, read=status, write=cmd */

#define HD_CMD		0x3f6	/* used for resets */
#define HD_ALTSTATUS	0x3f6	/* same as HD_STATUS but doesn't clear irq */

/* Bits of HD_STATUS */
#define ERR_STAT	0x01
#define INDEX_STAT	0x02
#define ECC_STAT	0x04	/* Corrected error */
#define DRQ_STAT	0x08
#define SEEK_STAT	0x10
#define WRERR_STAT	0x20
#define READY_STAT	0x40
#define BUSY_STAT	0x80

/* Values for HD_COMMAND */
#define WIN_RESTORE		0x10
#define WIN_READ		0x20
#define WIN_WRITE		0x30
#define WIN_VERIFY		0x40
#define WIN_FORMAT		0x50
#define WIN_INIT		0x60
#define WIN_SEEK 		0x70
#define WIN_DIAGNOSE		0x90
#define WIN_SPECIFY		0x91	/* set drive geometry translation */
#define WIN_SETIDLE1		0xE3
#define WIN_SETIDLE2		0x97

#define WIN_DOORLOCK		0xde	/* lock door on removable drives */
#define WIN_DOORUNLOCK		0xdf	/* unlock door on removable drives */

#define WIN_MULTREAD		0xC4	/* read sectors using multiple mode */
#define WIN_MULTWRITE		0xC5	/* write sectors using multiple mode */
#define WIN_SETMULT		0xC6	/* enable/disable multiple mode */
#define WIN_IDENTIFY		0xEC	/* ask drive to identify itself   */
#define WIN_SETFEATURES		0xEF	/* set special drive features */
#define WIN_READDMA		0xc8	/* read sectors using DMA transfers */
#define WIN_WRITEDMA		0xca	/* write sectors using DMA transfers */

/* Additional drive command codes used by ATAPI devices. */
#define WIN_PIDENTIFY		0xA1	/* identify ATAPI device  */
#define WIN_SRST		0x08	/* ATAPI soft reset command */
#define WIN_PACKETCMD		0xa0	/* Send a packet command. */

/* Bits for HD_ERROR */
#define MARK_ERR	0x01	/* Bad address mark */
#define TRK0_ERR	0x02	/* couldn't find track 0 */
#define ABRT_ERR	0x04	/* Command aborted */
#define ID_ERR		0x10	/* ID field not found */
#define ECC_ERR		0x40	/* Uncorrectable ECC error */
#define	BBD_ERR		0x80	/* block marked bad */

struct drive_info
{
  unsigned cmdsz:2;		/* packet command size */
#define AT_PSIZE_12     0	/* 12 bytes */
#define AT_PSIZE_16     1	/* 16 bytes */
  unsigned:3;
  unsigned drqtype:2;		/* DRQ type */
#define AT_DRQT_MPROC   0	/* microprocessor DRQ - 3 msec delay */
#define AT_DRQT_INTR    1	/* interrupt DRQ - 10 msec delay */
#define AT_DRQT_ACCEL   2	/* accelerated DRQ - 50 usec delay */
  unsigned removable:1;		/* device is removable */

  unsigned devtype:5;		/* device type */
#define AT_TYPE_DIRECT  0	/* direct-access (magnetic disk) */
#define AT_TYPE_TAPE    1	/* streaming tape (QIC-121 model) */
#define AT_TYPE_CDROM   5	/* CD-ROM device */
#define AT_TYPE_OPTICAL 7	/* optical disk */
  unsigned:1;
  unsigned proto:2;		/* command protocol */
#define AT_PROTO_ATAPI  2

  unsigned short cyls;		/* "physical" cyls */
  unsigned short reserved2;	/* reserved (word 2) */
  unsigned short heads;		/* "physical" heads */
  unsigned short track_bytes;	/* unformatted bytes per track */
  unsigned short sector_bytes;	/* unformatted bytes per sector */
  unsigned short sectors;	/* "physical" sectors per track */
  unsigned short vendor0;	/* vendor unique */
  unsigned short vendor1;	/* vendor unique */
  unsigned short vendor2;	/* vendor unique */
  unsigned char serial[20];	/* 0 = not_specified */
  unsigned short buf_type;
  unsigned short buf_size;	/* 512 byte increments; 0 = not_specified */
  unsigned short ecc_bytes;	/* for r/w long cmds; 0 = not_specified */
  unsigned char fw_rev[8];	/* 0 = not_specified */
  unsigned char model[40];	/* 0 = not_specified */
  unsigned char max_multsect;	/* 0=not_implemented */
  unsigned char vendor3;	/* vendor unique */
  unsigned short dword_io;	/* 0=not_implemented; 1=implemented */
  unsigned char vendor4;	/* vendor unique */

  unsigned dmaflag:1;		/* DMA supported */
  unsigned lbaflag:1;		/* LBA supported - always 1 */
  unsigned iordydis:1;		/* IORDY can be disabled */
  unsigned iordyflag:1;		/* IORDY supported */
  unsigned:1;
  unsigned ovlapflag:1;		/* overlap operation supported */
  unsigned:1;
  unsigned idmaflag:1;		/* interleaved DMA supported */

  unsigned short reserved50;	/* reserved (word 50) */
  unsigned char vendor5;	/* vendor unique */
  unsigned char tPIO;		/* 0=slow, 1=medium, 2=fast */
  unsigned char vendor6;	/* vendor unique */
  unsigned char tDMA;		/* 0=slow, 1=medium, 2=fast */
  unsigned short field_valid;	/* bits 0:cur_ok 1:eide_ok */
  unsigned short cur_cyls;	/* logical cylinders */
  unsigned short cur_heads;	/* logical heads */
  unsigned short cur_sectors;	/* logical sectors per track */
  unsigned short cur_capacity0;	/* logical total sectors on drive */
  unsigned short cur_capacity1;	/*  (2 words, misaligned int)     */
  unsigned char multsect;	/* current multiple sector count */
  unsigned char multsect_valid;	/* when (bit0==1) multsect is ok */
  unsigned int lba_capacity;	/* total number of sectors */
  unsigned short dma_1word;	/* single-word dma info */
  unsigned short dma_mword;	/* multiple-word dma info */
  unsigned short eide_pio_modes;	/* bits 0:mode3 1:mode4 */
  unsigned short eide_dma_min;	/* min mword dma cycle time (ns) */
  unsigned short eide_dma_time;	/* recommended mword dma cycle time (ns) */
  unsigned short eide_pio;	/* min cycle time (ns), no IORDY  */
  unsigned short eide_pio_iordy;	/* min cycle time (ns), with IORDY */
  unsigned short word69;	/* reserved (word 69) */
  unsigned short word70;	/* reserved (word 70) */
  /* HDIO_GET_IDENTITY currently returns only words 0 through 70 */
  unsigned short word71;	/* reserved (word 71) */
  unsigned short word72;	/* reserved (word 72) */
  unsigned short word73;	/* reserved (word 73) */
  unsigned short word74;	/* reserved (word 74) */
  unsigned short queue_depth;	/*  */
  unsigned short word76;	/* reserved (word 76) */
  unsigned short word77;	/* reserved (word 77) */
  unsigned short word78;	/* reserved (word 78) */
  unsigned short word79;	/* reserved (word 79) */
  unsigned short major_rev_num;	/*  */
  unsigned short minor_rev_num;	/*  */
  unsigned short command_set_1;	/* bits 0:Smart 1:Security 2:Removable 3:PM */
  unsigned short command_set_2;	/* bits 14:Smart Enabled 13:0 zero */
  unsigned short cfsse;		/* command set-feature supported extensions */
  unsigned short cfs_enable_1;	/* command set-feature enabled */
  unsigned short cfs_enable_2;	/* command set-feature enabled */
  unsigned short csf_default;	/* command set-feature default */
  unsigned short dma_ultra;	/*  */
  unsigned short word89;	/* reserved (word 89) */
  unsigned short word90;	/* reserved (word 90) */
  unsigned short CurAPMvalues;	/* current APM values */
  unsigned short word92;	/* reserved (word 92) */
  unsigned short hw_config;	/* hardware config */
  unsigned short word94;	/* reserved (word 94) */
  unsigned short word95;	/* reserved (word 95) */
  unsigned short word96;	/* reserved (word 96) */
  unsigned short word97;	/* reserved (word 97) */
  unsigned short word98;	/* reserved (word 98) */
  unsigned short word99;	/* reserved (word 99) */
  unsigned short word100;	/* reserved (word 100) */
  unsigned short word101;	/* reserved (word 101) */
  unsigned short word102;	/* reserved (word 102) */
  unsigned short word103;	/* reserved (word 103) */
  unsigned short word104;	/* reserved (word 104) */
  unsigned short word105;	/* reserved (word 105) */
  unsigned short word106;	/* reserved (word 106) */
  unsigned short word107;	/* reserved (word 107) */
  unsigned short word108;	/* reserved (word 108) */
  unsigned short word109;	/* reserved (word 109) */
  unsigned short word110;	/* reserved (word 110) */
  unsigned short word111;	/* reserved (word 111) */
  unsigned short word112;	/* reserved (word 112) */
  unsigned short word113;	/* reserved (word 113) */
  unsigned short word114;	/* reserved (word 114) */
  unsigned short word115;	/* reserved (word 115) */
  unsigned short word116;	/* reserved (word 116) */
  unsigned short word117;	/* reserved (word 117) */
  unsigned short word118;	/* reserved (word 118) */
  unsigned short word119;	/* reserved (word 119) */
  unsigned short word120;	/* reserved (word 120) */
  unsigned short word121;	/* reserved (word 121) */
  unsigned short word122;	/* reserved (word 122) */
  unsigned short word123;	/* reserved (word 123) */
  unsigned short word124;	/* reserved (word 124) */
  unsigned short word125;	/* reserved (word 125) */
  unsigned short word126;	/* reserved (word 126) */
  unsigned short word127;	/* reserved (word 127) */
  unsigned short dlf;		/* device lock function
				 * 15:9       reserved
				 * 8    security level 1:max 0:high
				 * 7:6  reserved
				 * 5    enhanced erase
				 * 4    expire
				 * 3    frozen
				 * 2    locked
				 * 1    en/disabled
				 * 0    capability
				 */
  unsigned short csfo;		/* current set features options
				 * 15:4       reserved
				 * 3    auto reassign
				 * 2    reverting
				 * 1    read-look-ahead
				 * 0    write cache
				 */
  unsigned short reserved[126];
};

#endif
