#ifndef IDE_H
#define IDE_H

#include <phantom/blk.h>
#include <phantom/hdregs.h>
#include <sys/types.h>
#include <phantom/hdregs.h>

#define MAXRETRIES	16	/* this must be enough.. */

#define NR_IDE		4	/* drives count */
#define NR_IDE_CTRL	2	/* controllers count */

#define DEFAULT_IDE0_IRQ	14
#define DEFAULT_IDE1_IRQ	15

#define DEFAULT_IDE0_PORT	0x01F0
#define DEFAULT_IDE1_PORT	0x0170

#define AR_DATA         0x0	/* RW - data register (16 bits) */
#define AR_ERROR        0x1	/*  R - error register */
#define AR_FEATURES     0x1	/*  W - features */
#define AR_IREASON      0x2	/* RW - interrupt reason */
#define AR_TAG          0x3	/*    - reserved for SAM TAG byte */
#define AR_CNTLO        0x4	/* RW - byte count, low byte */
#define AR_CNTHI        0x5	/* RW - byte count, high byte */
#define AR_DRIVE        0x6	/* RW - drive select */
#define AR_COMMAND      0x7	/*  W - command register */
#define AR_STATUS       0x7	/*  R - immediate status */

/* status register bits */
#define ARS_ERROR       0x01	/* error occured, see sense key/code */
#define ARS_INDEX	0x02
#define ARS_ECC         0x04	/* correctable error occured */
#define ARS_DRQ         0x08	/* data request / ireason valid */
#define ARS_SEEK        0x10	/* immediate operation completed */
#define ARS_WRERR       0x20
#define ARS_READY       0x40	/* ready to get command */
#define ARS_BUSY        0x80	/* registers busy */
/* overlap mode: */
#define ARS_SERVICE     0x10	/* service is requested */
#define ARS_DMARDY      0x20	/* ready to start a DMA transfer */

/*
 * Status register bits
 */
#define ARS_CHECK       0x01	/* error occured, see sense key/code */
					/* bit 0x02 reserved */
#define ARS_CORR        0x04	/* correctable error occured */
#define ARS_DRQ         0x08	/* data request / ireason valid */
#define ARS_DSC         0x10	/* immediate operation completed */
#define ARS_DF          0x20	/* drive fault */
#define ARS_DRDY        0x40	/* ready to get command */
#define ARS_BSY         0x80	/* registers busy */
					/* for overlap mode only: */
#define ARS_SERVICE     0x10	/* service is requested */
#define ARS_DMARDY      0x20	/* ready to start a DMA transfer */
#define ARS_BITS        "\20\010busy\7ready\6fault\5opdone\4drq\3corr\1check"

/* error register bits */
#define AER_ILI         0x01	/* illegal length indication */
#define AER_EOM         0x02	/* end of media detected */
#define AER_ABRT        0x04	/* command aborted */
#define AER_MCR         0x08	/* media change requested */
#define AER_SKEY        0xf0	/* sense key mask */
#define AER_SK_NO_SENSE         0x00	/* no specific sense key info */
#define AER_SK_RECOVERED_ERROR  0x10	/* command succeeded, data recovered */
#define AER_SK_NOT_READY        0x20	/* no access to drive */
#define AER_SK_MEDIUM_ERROR     0x30	/* non-recovered data error */
#define AER_SK_HARDWARE_ERROR   0x40	/* non-recoverable hardware failure */
#define AER_SK_ILLEGAL_REQUEST  0x50	/* invalid command parameter(s) */
#define AER_SK_UNIT_ATTENTION   0x60	/* media changed */
#define AER_SK_DATA_PROTECT     0x70	/* reading read-protected sector */
#define AER_SK_ABORTED_COMMAND  0xb0	/* command aborted, try again */
#define AER_SK_MISCOMPARE       0xe0	/* data did not match the medium */
#define AER_BITS        "\20\4mchg\3abort\2eom\1ili"

/* feature register bits */
#define ARF_DMA         0x01	/* transfer data via DMA */
#define ARF_OVERLAP     0x02	/* release the bus until completion */

/* interrupt reason register bits */
#define ARI_CMD         0x01	/* command(1) or data(0) */
#define ARI_IN          0x02	/* transfer to(1) or from(0) the host */
#define ARI_RELEASE     0x04	/* bus released until completion */

#define ARD_DRIVE0      0xa0	/* drive 0 selected */
#define ARD_DRIVE1      0xb0	/* drive 1 selected */

#define ATAPIC_IDENTIFY         0xa1	/* get drive parameters */
#define ATAPIC_PACKET           0xa0	/* execute packet command */

#define ATAPI_TEST_UNIT_READY   0x00	/* check if the device is ready */
#define ATAPI_REQUEST_SENSE     0x03	/* get sense data */
#define ATAPI_FORMAT_UNIT	0x04	/* format block */
#define ATAPI_START_STOP        0x1b	/* start/stop the media */
#define ATAPI_PREVENT_ALLOW     0x1e	/* prevent/allow media removal */
#define ATAPI_READ_CAPACITY     0x25	/* get volume capacity */
#define ATAPI_READ_BIG          0x28	/* read data */
#define ATAPI_WRITE_BIG		0x2a	/* write data */
#define ATAPI_READ_TOC          0x43	/* get table of contents */
#define ATAPI_READ_SUBCHANNEL   0x42	/* get subchannel info */
#define ATAPI_MODE_SELECT_BIG   0x55	/* set device parameters */
#define ATAPI_MODE_SENSE        0x5a	/* get device parameters */
#define ATAPI_PLAY_CD           0xb4	/* universal play command */

/* optional packet commands */
#define ATAPI_PLAY_MSF          0x47	/* play by MSF address */
#define ATAPI_PAUSE             0x4b	/* stop/start audio operation */

/* nonstandard packet commands */
#define ATAPI_PLAY_TRACK        0x48	/* play by track number */
#define ATAPI_PLAY_BIG          0xa5	/* play by logical block address */

#define RES_OK          0	/* i/o done */
#define RES_ERR         1	/* i/o finished with error */
#define RES_NOTRDY      2	/* controller not ready */
#define RES_NODRQ       3	/* no data request */
#define RES_INVDIR      4	/* invalid bus phase direction */
#define RES_OVERRUN     5	/* data overrun */
#define RES_UNDERRUN    6	/* data underrun */

typedef struct ide IDE;
typedef struct ide_ctrl IDECTRL;
typedef struct ide_cmd IDECMD;

/* device types */
#define IDE_UNKNOWN		0x00
#define IDE_HD			0x01
#define IDE_CD			0x02
#define IDE_FLOPPY		0x03
#define IDE_TAPE		0x04
#define IDE_OPTICAL		0x05

struct floppy_cappage
{
  /* Mode data header */
  u16 data_length;
  u8 medium_type;
#define MDT_UNKNOWN     0x00
#define MDT_NO_DISC     0x70
#define MDT_DOOR_OPEN   0x71
#define MDT_FMT_ERROR   0x72

#define MDT_2DD_UN	0x10
#define MDT_2DD		0x11
#define MDT_2HD_UN	0x20
#define MDT_2HD_12_98	0x22
#define MDT_2HD_12	0x23
#define MDT_2HD_144	0x24
#define MDT_LS120	0x31

  unsigned reserved0:7;
  unsigned wp:1;		/* Write protect */
  u8 reserved1[4];

  /* Capabilities page */
  unsigned page_code:6;		/* Page code - Should be 0x5 */
#define CAP_PAGE        0x05
  unsigned reserved1_6:1;	/* Reserved */
  unsigned ps:1;		/* The device is capable of saving the page */
  u8 page_length;		/* Page Length - Should be 0x1e */
  u16 transfer_rate;		/* In kilobits per second */
  u8 heads, sectors;		/* Number of heads, Number of sectors per track */
  u16 sector_size;		/* Byes per sector */
  u16 cyls;			/* Number of cylinders */
  u8 reserved10[10];
  u8 motor_delay;		/* Motor off delay */
  u8 reserved21[7];
  u16 rpm;			/* Rotations per minute */
  u8 reserved30[2];
};

struct floppy_info
{
  struct floppy_cappage cap;
};


struct ide			/* ATAPI device structure */
{
  int dev;			/* device number */
  char *name;			/* device name da, db, etc. */
  IDECTRL *ctrl;		/* controller */
  u8 select_mask;		/* selection mask (primary/secondary) */
  char fl_lba;			/* lba flag */
  char fl_cmd16;		/* 16-byte command flag */
  char fl_intrcmd;		/* interrupt before cmd flag */
  char fl_ack;			/* acknowledged when fl_intrcmd */
  char fl_slow;			/* slow reaction device */
  char fl_attached;		/* attached device flag */

  int dev_type;			/* device type */
  union
  {
    struct hd_info hd;
    struct floppy_info floppy;
/*    struct cd_info cd;*/
  }
  u;

  int flags;
};

#define F_OPEN          0x0001	/* The block device is opened */
#define F_MEDIA_CHANGED 0x0002	/* The media have changed since open */
#define F_DEBUG         0x0004	/* Print debug info */
#define F_LOCKED        0x0008	/* This unit is locked (or should be) */
#define F_TRACK_PREP    0x0010	/* Track should be prep'ed */
#define F_TRACK_PREPED  0x0020	/* Track has been prep'ed */
#define F_DISK_PREPED   0x0040	/* Disk has been prep'ed */
#define F_WRITTEN   	0x0080	/* The medium has been written to */

#define IDE_HOOK_IRQ(i, irq_name)	{ \
					(i)->handler=(irq_name); \
					}
#define IDE_UNHOOK_IRQ(i)		{ \
					(i)->handler=NULL; \
					}
#define IDE_IRQ_ACK_RESET(i)		{ \
					(i)->ctrl->ack=0; \
					}

struct ide_cmd
{
  REQUEST *req;
  struct ide_cmd *next;
  IDE *ide;
  IDECTRL *ctrl;

  u8 code;			/* result code */
#define RES_OK          0	/* i/o done */
#define RES_ERR         1	/* i/o finished with error */
#define RES_NOTRDY      2	/* controller not ready */
#define RES_NODRQ       3	/* no data request */
#define RES_INVDIR      4	/* invalid bus phase direction */
#define RES_OVERRUN     5	/* data overrun */
#define RES_UNDERRUN    6	/* data underrun */
  u8 status;			/* status register contents */
  u8 error;			/* error register contents */

  int count;
  char *buff;

  int retries;

  void (*callback) (IDECMD *);

  char pkt_data[12];

/* these are visible only to ide.c :^) */
  char *orig_buff;
  int orig_count;
};

#define ZEROCMD		{0,}
#define ZERO_CMD	ZEROCMD

struct ide_ctrl			/* atapi controller structure */
{
  int irq;			/* irq */
  int port;			/* io base */
  int major;			/* major number */
  char *name;			/* name /debug purpose/ */
  void (*handler) (IDE *, IDECTRL *, IDECMD *);	/* interrupt handler */
  int ack;			/* acknowledgement flag/counter */
  int busy;			/* already in dorequest loop ? */
  IDECMD *head;			/* ide request list head */
  IDECMD *tail;			/* ide request list tail */
};

IDE *ide_get_ide (int dev);

void ide_end_request (IDECMD *, int);
void ide_dorequest (REQUEST * req);
void ide_start_processing (IDECTRL * ctrl);
void hd_dorequest (IDE * ide, IDECTRL * ctrl, IDECMD * cmd);
void zip_dorequest (IDE * ide, IDECTRL * ctrl, IDECMD * cmd);
void zip_attach (IDE * ide, IDECTRL * ctrl);
int zip_ioctl (IDE * ide, int cmd, int arg);
void ide_runcmd (IDECMD * cmd);
void ide_execcmd (IDECMD * cmd);
IDECMD *ide_newcmd (IDE * ide, void (*callback) (IDECMD *));

#endif
