#ifndef MM_H
#define MM_H

#include <asm/mm.h>
#include <sys/types.h>
#include <string.h>

#define PG_NOTHING		0x00
#define PG_FREE			0x00
#define PG_KERNEL		0x01
#define PG_SHARE		0x04
#define PG_USED			PG_KERNEL
#define PG_DMA			0x08
#define PG_BUFFER		0x10

#define FREE_PAGE_ENTRY		0x00
#define PAGE_ENTRIES		0x400

#define ADDR_MASK		0xFFFFF000
#define FLAGS_MASK		0x00000FFF

#define TOPAGE(v_addr)		(((int)(v_addr))>>PAGE_BITS)
#define TOADDR(page)		(((int)(page))<<PAGE_BITS)

#define invalidate() \
__asm__ __volatile__("movl %%cr3,%%eax\nmovl %%eax,%%cr3": : :"ax")

#define setpd(pd) \
__asm__ __volatile__("movl %%eax,%%cr3": : "ax" (pd))

/*#define copy_page(to,from) \
__asm__("cld ; rep ; movsl": :"S" (from),"D" (to),"c" (1024):"cx","di","si")*/
#define copy_page(to,from)	memcpy((void *)to,(void *)from,4096)

#define MAX_KMALLOC_SIZE	131056

#define OBJ_FREE		0x01
#define OBJ_USED		0x02
#define OBJ_DMA			0x04

#define PGS_PRESENT		0x0001		/* present in phys. mem */
#define PGS_WRITEABLE		0x0002		/* user write */
#define PGS_USER		0x0004		/* user access */
#define PGS_ACESS		0x0020		/* page accessed */
#define PGS_DIRTY		0x0040		/* page modified */
#define PGS_AVL0		0x0200
#define PGS_AVL1		0x0400
#define PGS_AVL2		0x0800
#define PGS_USED		PGS_AVL0
#define PGS_COW			PGS_AVL1
#define PGS_TABLELOCKED		PGS_AVL0

#define SET_USED(e)		(e) |= PGS_USED
#define SET_FREE(e)		(e) &= ~PGS_USED
#define SET_COW(e)		(e) |= PGS_COW
#define UNSET_COW(e)		(e) &= ~PGS_COW
#define S_ISCOW(e)		(((e) & PGS_COW) == PGS_COW)
#define S_ISFREE(e)		(((e) & PGS_USED) == 0)

#define FIRST_KERNEL_PAGE	0x10

#define INPAGES(bts)		((bts)+PAGE_SIZE-1)/PAGE_SIZE

#define PAGE2TABLE(page)	((page) >> 10)
#define PAGE2ENTRY(page)	((page) & 0x03FF)
#define TE2PAGE(table,entry)	(((table) << 10) | (entry))

#define CPDE(tsk,tbl)		((tsk)->tss.cr3[(tbl)])
#define PDE(tbl)		CPDE(ctask,tbl)

#define CPT(tsk,tbl)		((u32 *)((u32)(tsk)->tss.cr3[(tbl)] & ADDR_MASK))
#define PT(tbl)			CPT(ctask,tbl)

#define KERNEL_HPT		0x100	/* highest page table of kernel (+1) */
#define USER_AREA		0x40000	/* user area first page */
#define USER_LOWEST_PAGE	USER_AREA

extern void *kmalloc (int size, int k_flag);
extern void kfree (void *);

#define NO_PAGE			(-1)

/* used in psh.c -- sysinfo */
extern int nr_pages;
extern int nr_free_pages;
extern int nr_dma_pages;
extern int nr_free_dma_pages;

extern ulg zero_page_dir_map[];
extern ulg* zero_page_dir;

void free_pages (int, int);

void set_page(int, int, int);

int __get_free_pages (int count, int);
#define page_alloc()		TOADDR(__get_free_pages(1, PG_NOTHING))
#define pages_alloc(a)		TOADDR(__get_free_pages((a), PG_NOTHING))
#define dma_page_alloc()	TOADDR(__get_free_pages(1, PG_DMA))
#define dma_pages_alloc(a)	TOADDR(__get_free_pages((a), PG_DMA))

#define page_free(a)		set_map((a),1,PG_FREE)

#define gfp()		__get_free_pages(1, PG_NOTHING)
#define gfps(a)		__get_free_pages((a), PG_NOTHING)

#endif
