#ifndef NETDEV_H
#define NETDEV_H

#include <sys/types.h>
#include <phantom/if_ether.h>

#define HWADDR_LEN		6
#define IPADDR_LEN		4

#define STR_DUMP_HW		"%02X:%02X:%02X:%02X:%02X:%02X"
#define VAR_DUMP_HW(hw)		(hw)[0],(hw)[1],(hw)[2],(hw)[3],(hw)[4],(hw)[5]
#define STR_DUMP_IP		"%d.%d.%d.%d"
#define VAR_DUMP_IP(ip)		((u8*)ip)[0],((u8*)ip)[1],((u8*)ip)[2],((u8*)ip)[3]

struct pkt_head
{
  struct pkt_head *next;
  struct pkt_head *prev;

  struct pkt_head **l_head;		/* list head and tail */
  struct pkt_head **l_tail;

  struct device *dev;

  void *data;			/* start of 'real' packet */
  int data_len;			/* 'real' packet length */

  u8 hw_data[0];		/* data start 'mark', hw part included */
};

struct ne_info
{
  int open:1;
  int word16:1;			/* We have the 16-bit (vs 8-bit) version of the card. */
  int txing:1;			/* Transmit Active */
  int dmaing:2;			/* Remote DMA Active */
  int irqlock:1;		/* 8390's intrs disabled when '1'. */
  int interrupt:1;		/* in interrupt flag */
  int tbusy:1;

  unsigned char tx_start_page, rx_start_page, stop_page;
  unsigned char current_page;	/* Read pointer in buffer  */
  unsigned char interface_num;	/* Net port (AUI, 10bT.) to use. */
  unsigned char txqueue;	/* Tx Packet buffer queue length. */
  short tx1, tx2;		/* Packet lengths for ping-pong tx. */
  short lasttx;			/* Alpha version consistency check. */
  unsigned char reg0;		/* Register '0' in a WD8013 */
  unsigned char reg5;		/* Register '5' in a WD8013 */

  int trans_start;		/* last transaction start time */

  struct enet_statistics stat;
};


struct device
{
  char *name;			/* interface name */

  int irq;			/* irq */
  int base_addr;		/* io base */

  u8 ip_addr[IPADDR_LEN];	/* ip address */
  u8 hw_addr[HWADDR_LEN];	/* hw address */
  u8 broad_addr[HWADDR_LEN];	/* broadcast address */

  int hw_hdr_len;		/* size of hardware header */
  int mtu;			/* interface mtu value */

  struct device *next;		/* next device in queue */
  struct device *prev;		/* previous device in queue */

  int (*pkt_tx) (struct pkt_head *);

  union
  {
    struct ne_info *ne;
  }
  u;
};

void pkt_rx(struct pkt_head *pkt);

struct pkt_head *pkt_alloc(struct device *dev, int pkt_size);
void pkt_free(struct pkt_head *);				
void pkt_send(struct pkt_head *);
void pkt_dequeue(struct pkt_head *pkt);
void pkt_hw_header(struct pkt_head *pkt, u8 *src, u8*dst, u16 type);
void pkt_dump (struct pkt_head *pkt);

#endif
