#ifndef SCHED_H
#define SCHED_H

#include <sys/types.h>
#include <asm/task.h>
#include <phantom/fs.h>
#include <asm/kernel.h>
#include <rg/window.h>
#include <errno.h>
#include <signal.h>

extern unsigned int kernel_stack[];

#define IO_BITMAP_SIZE	32
/* from beginning of TSS */
#define IO_PERM_OFFSET	104

struct tss_struct
{
  u16 back_link, __blh;
  u32 esp0;
  u16 ss0, __ss0h;
  u32 esp1;
  u16 ss1, __ss1h;
  u32 esp2;
  u16 ss2, __ss2h;
  u32 *cr3;
  u32 eip;
  u32 eflags;
  u32 eax;
  u32 ecx;
  u32 edx;
  u32 ebx;
  u32 esp;
  u32 ebp;
  u32 esi;
  u32 edi;
  u16 es, __esh;
  u16 cs, __csh;
  u16 ss, __ssh;
  u16 ds, __dsh;
  u16 fs, __fsh;
  u16 gs, __gsh;
  u16 ldt, __ldth;
  u16 trace, io_perm_offset;
  u32 io_bitmap[IO_BITMAP_SIZE + 1];
};

#define TASK_STOPPED	0x00
#define TASK_RUNNING	0x01
/* terminated task is waiting for parent's wait() */
#define TASK_ZOMBIE	0x02
/* task is being spawned by fork() */
#define TASK_SPAWNING	0x03

#define TASK struct task

TASK
{
  int status;
  int beats;
  u32 sigqueuestat;		/* signal queue status */
  u32 blockedsig;		/* blocked signals */
  u32 signal;			/* first pending signal */
  struct task *thread_lock;	/* free */
  struct task *thread_next;
  int pid;
  int exit_code;
  int trap_nr;
  int error_code;
  int fl_dump, fl_lock;
/* parent, child, next sibling, previous sibling */
  TASK *parent, *child, *nsbl, *psbl;
/* physical addresses */
  u32 code_pos, code_end, data_end;
  int code_pages;		/* code pages (to set ro) */
  int nr_pages;			/* pages of the code as file + bss */
  int nr_userpages;		/* pages owned by user */
  int tty;			/* which tty to use (-1 = no tty i/o) */
  WINDOW *win;			/* which window to use (NULL = no window) */
  struct vnode *pwd;		/* work directory */
  struct vnode *root;		/* root directory */
  struct vnode *vnode;		/* exe-file vnode */
  struct file *filp[NR_OPEN];	/* open filedes lookup table */
  struct tss_struct tss;	/* tss descriptor (see above) */
  char pwd_str[PATH_MAX];	/* FIXME: path string (rip this off!) */
  TASK *next;			/* scheduler chain, next task */
  TASK *prev;			/* previous task */
  unsigned long tr;		/* task register */
  u8 ldt[8];			/* default task stack descriptor place */
  int kernel_stack_page;	/* kernel stack page */
  void *user_stack;		/* user stack (page) */
  int kernel_stack_pages;	/* this should be fixed value */
  int user_stack_pages;		/* floating value... */
  struct sigaction sigaction[NR_SIGNAL]; 
  u8 sigpool[NR_SIGNAL];	/* signal trampoline */
  int lsig;			/* last pending signal */
};

/* this ought to be synchronized with boot/head.S */
#define FIRST_TSS	5
#define FIRST_LDT	6
#define TSS(n) 		((FIRST_TSS << 3) + ((ulg)(n) << 4))
#define LDT(n)		((FIRST_LDT << 3) + ((ulg)(n) << 4))

#define DATA_SEG_0	0x92
#define DATA_SEG_3	0xF2
#define CODE_SEG_0	0x9A
#define CODE_SEG_3	0xFA

#define load_tr(n) __asm__("ltr %%ax"::"a" (n))
#define clear_nt() __asm__("pushfl ; andl $0xffffbfff,(%esp) ; popfl")

#define set_tssldt_desc(n,addr,limit, type) \
__asm__ __volatile__ ("movw $" #limit ",%1\n" \
	"movw %%ax,%2\n" \
	"rorl $16,%%eax\n" \
	"movb %%al,%3\n" \
	"movb $" #type ",%4\n" \
	"movb $0x00,%5\n" \
	"movb %%ah,%6\n" \
	"rorl $16,%%eax\n" \
	: /* no output */ \
	:"a" (addr), "m" (*(n)), "m" (*(n+2)), "m" (*(n+4)), \
	 "m" (*(n+5)), "m" (*(n+6)), "m" (*(n+7)) \
	)

#define set_tss_desc(n,addr) set_tssldt_desc(((char *) (n)),((ulg)(addr)), 0xFF, 0x89)
#define set_ldt_desc(n,addr,nr_des) set_tssldt_desc(((char *) (n)),((ulg)(addr)), ((nr_des)<<3)-1, 0x82)

extern struct task *task_table[];	/* psh purposes */
extern struct task *ctask;
extern TASK *ztask;

#define switch_task(tsk) \
__asm__("cli \n" \
	"movl %%ecx,ctask\n" \
	"ljmp %0\n" \
	"sti \n" \
	"clts \n" \
	: /* no output */ \
	: "m" (*(((char *)&tsk->tr)-4)), "c" (tsk)\
	)

int send_sig (int sig_nr, struct task *task);
void kill_task(struct task *task);

#endif
