#ifndef SOCKET_H
#define SOCKET_H

#include <sys/types.h>
#include <phantom/netdev.h>

/* Socket types. */
#define SOCK_STREAM	1	/* stream (connection) socket   */
#define SOCK_DGRAM	2	/* datagram (conn.less) socket  */
#define SOCK_RAW	3	/* raw socket                   */
#define SOCK_RDM	4	/* reliably-delivered message   */
#define SOCK_SEQPACKET	5	/* sequential packet socket     */
#define SOCK_PACKET	10	/* hw level packets */

/* Supported address families. */
#define AF_UNSPEC	0
#define AF_UNIX		1	/* Unix domain sockets          */
#define AF_INET		2	/* Internet IP Protocol         */
#define AF_AX25		3	/* Amateur Radio AX.25          */
#define AF_IPX		4	/* Novell IPX                   */
#define AF_APPLETALK	5	/* Appletalk DDP                */
#define	AF_NETROM	6	/* Amateur radio NetROM         */
#define AF_BRIDGE	7	/* Multiprotocol bridge         */
#define AF_AAL5		8	/* Reserved for Werner's ATM    */
#define AF_X25		9	/* Reserved for X.25 project    */

/* Protocol families, same as address families. */
#define PF_UNSPEC	AF_UNSPEC
#define PF_UNIX		AF_UNIX
#define PF_INET		AF_INET
#define PF_AX25		AF_AX25
#define PF_IPX		AF_IPX
#define PF_APPLETALK	AF_APPLETALK
#define	PF_NETROM	AF_NETROM
#define PF_BRIDGE	AF_BRIDGE
#define PF_AAL5		AF_AAL5
#define PF_X25		AF_X25

#define PP_UDP		0x01
#define PP_TCP		0x02
#define PP_PACKET	0x03

struct sockaddr
{
  int size;			/* structure size (addr len) */
  u16 sin_family;		/* protocol family */
  u16 sin_port;			/* connection port */
  u8 sin_addr[6];		/* ipv6 ready */
};

struct socket
{
  struct sockaddr src;		/* src addr (port) */
  struct sockaddr dst;		/* destination */

  struct device *dev;		/* this should be removed someday */

  struct proto *proto;		/* protocol and it's operations */

  struct socket *next;		/* next socket */
  struct socket *prev;		/* prev socket */
};

struct proto
{
  char *name;			/* protocol name */

  int domain;			/* id */
  int type;
  int protocol;

  void (*interpret)(struct pkt_head *pkt);

/* protocol socket operations */
  int (*create) (struct socket *);
  int (*release) (struct socket *);
  int (*connect) (struct socket *, struct sockaddr *);
  int (*send) (struct socket *, char *, int, int);
  int (*recv) (struct socket *, char *, int, int);
  int (*ioctl) (struct socket *, unsigned int, unsigned long);
};

struct proto *get_proto (int domain, int type, int protocol);
extern struct proto udp_proto;

#endif
