#ifndef	AUDIOIO_H
#define	AUDIOIO_H

#include <sys/ioccom.h>

typedef struct audio_prinfo
{
  unsigned sample_rate;		/* samples per second */
  unsigned channels;		/* # of channels (interleaved) */
  unsigned precision;		/* sample size in bits */
  unsigned encoding;		/* encoding method used */

  unsigned gain;		/* volume level: 0 - 255 */
  unsigned port;		/* input/output device */
  unsigned _fill1[4];

  unsigned samples;		/* samples played */
  unsigned eof;			/* ?!? */
  unsigned char pause;		/* !=0 pause, ==0 continue */
  unsigned char error;		/* !=0 if overflow/underflow */
  unsigned char waiting;	/* !=0 if others wants access */
  unsigned char _fill2[3];

  unsigned char open;		/* is device open */
  unsigned char active;		/* !=0 if sound hardware is active */
}
audio_prinfo_t;

typedef struct audio_info
{
  audio_prinfo_t play;
  audio_prinfo_t record;
  unsigned monitor_gain;
  unsigned _fill[4];
}
audio_info_t;

#define	AUDIO_ENCODING_ULAW	(1)	/* u-law encoding */
#define	AUDIO_ENCODING_ALAW	(2)	/* A-law encoding */
#define	AUDIO_ENCODING_RAW	(3)	/* linear encoding */

#define	AUDIO_MIN_GAIN		(0)	/* minimum volume value */
#define	AUDIO_MAX_GAIN		(255)	/* maximum volume value */

#define	AUDIO_INITINFO(i)	memset((void*)i, 0xff, sizeof(audio_info_t))

#define	AUDIO_GETINFO		_IOR('A', 1, audio_info_t)
#define	AUDIO_SETINFO		_IORW('A', 2, audio_info_t)
#define	AUDIO_DRAIN		_IO('A', 3)
#define	AUDIO_FLUSH		_IO('A', 4)

/* compatibility to /dev/audio */
#define AUDIO_COMPAT_DRAIN	_IO('P', 1)
#define AUDIO_COMPAT_FLUSH	_IO('P', 0)

#endif
