#ifndef TWINS_FORM_H
#define TWINS_FORM_H

#include <twins/misc.h>

#define CELL struct cell_head
#define FORM struct form_st

#define ATTR_FORM           ATTR(WHITE,LIGHTGRAY)
#define ATTR_SCROLLBAR	    ATTR(LIGHTGRAY,BLUE)

#define WINFRAME_NONE           0
#define WINFRAME_DOUBLE         1

/* cell types */
#define CELL_BUTTON        1
#define CELL_MEMO          2
#define CELL_EDIT	   3
#define CELL_TABLE	   4
#define CELL_CHECK         5
#define CELL_RADIO         6
#define CELL_CANVAS	   7

struct cell_head
 {
  int x,y;
  unsigned activated:1;
  unsigned disabled:1;
  unsigned passive:1;
  int type;			/* cell type */
  struct cell_head *next;
  struct cell_head *prev;
  FORM *form;
  WINDOW *window;
  void (*event_handler)(CELL *, unsigned int, unsigned int);

  union 
    {
    BUTTON *button;
    MEMO *memo;
    EDIT *edit;
    TABLE *table;
    CHECK *check;
    RADIO *radio;
    CANVAS *canvas;
    void *any;
    } u;
 };

void unlink_cell(CELL *cell);
CELL *new_cell (int x, int y, int cell_type, void *cell_data, int passive, void (*handler)(CELL*, unsigned int, unsigned int), CELL *next);

#define new_button(x,y,b,next) new_cell(x,y,CELL_BUTTON,(void*)b,0,&button_handler,next)
#define new_memo(x,y,m,next) new_cell(x,y,CELL_MEMO,(void*)m,1,&memo_handler,next)
#define new_edit(x,y,e,next) new_cell(x,y,CELL_EDIT,(void*)e,0,&edit_handler,next)
#define new_table(x,y,t,next) new_cell(x,y,CELL_TABLE,(void*)t,0,&table_handler,next)
#define new_check(x,y,c,next) new_cell(x,y,CELL_CHECK,(void*)c,0,&check_handler,next)
#define new_radio(x,y,r,next) new_cell(x,y,CELL_RADIO,(void*)r,0,&radio_handler,next)
#define new_canvas(x,y,c,next) new_cell(x,y,CELL_CANVAS,(void*)c,0,&null_handler,next)

struct form_st
{
void (*event_handler)(CELL *, unsigned int, unsigned int);         /* event handler */
WINDOW *window;			       /* draw window */
int winframe;                          /* frame type */
char *caption;			       /* window caption */
int x,y;                               /* coordinates */
int lx,ly;	         	       /* sizes */
CELL *cell_queue;                      /* cell list */
CELL *cell_tail;                       /* tail of cell list */
CELL *current_cell;                    /* current (active) cell */
char *screen;     		       /* "screen behind" backup */
int fl_close;                          /* form close flag */
int exit_code;                         /* exit code */
CELL *default_cell;                    /* default (enter) cell */
CELL *cancel_cell;                     /* cancel (esc) cell */
};

#define DEFAULT_FORM    {\
                        NULL,                   /* event handler */\
			NULL,			/* window */ \
                        WINFRAME_DOUBLE,        /* winframe */\
                        NULL,                   /* caption */\
                        -1,-1,                  /* x,y */\
                        30,10,                  /* lx,ly */\
                        NULL,                   /* cell_queue */\
                        NULL,                   /* cell_tail */\
                        NULL,                   /* current_cell */\
                        NULL,                   /* screen */\
                        0,                      /* fl_close */\
                        0,                      /* exit_code */\
                        NULL,                   /* default cell */\
                        NULL,                   /* cancel cell */\
                        }

void form_handler (FORM *f, unsigned int event_type);

extern EVENT event;		/* form.c */

#endif
