/*
   kernel/irq.c
   Hardware interrupt handlers.

  TODO: split irq and ext...
 */

#include <asm/regdump.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/head.h>
#include <sys/types.h>
#include <errno.h>
#include <phantom/ext.h>

struct irq_action
{
  void (*handler) (int, struct regdump *);
  int flags;
  const char *name;
};

static struct irq_action irq_table[NR_IRQ] = {
  {NULL,},
};

u32 ext_mask = 0;
u32 ext_active = 0;
u32 nr_running_irq = 0;

u32 flg_global;			/* single_code, multi_code routines */

void (*ext_table[NR_EXT]) (void) =
{
NULL,};

/* "cached" ports */
static int port_21;
static int port_A1;

BUILD_IRQ (0x00, 1, 0x01, 0x0001);
BUILD_IRQ (0x01, 1, 0x02, 0x0002);
BUILD_IRQ (0x02, 1, 0x04, 0x0004);
BUILD_IRQ (0x03, 1, 0x08, 0x0008);
BUILD_IRQ (0x04, 1, 0x10, 0x0010);
BUILD_IRQ (0x05, 1, 0x20, 0x0020);
BUILD_IRQ (0x06, 1, 0x40, 0x0040);
BUILD_IRQ (0x07, 1, 0x80, 0x0080);
BUILD_IRQ (0x08, 2, 0x01, 0x0100);
BUILD_IRQ (0x09, 2, 0x02, 0x0200);
BUILD_IRQ (0x0A, 2, 0x04, 0x0400);
BUILD_IRQ (0x0B, 2, 0x08, 0x0800);
BUILD_IRQ (0x0C, 2, 0x10, 0x1000);
BUILD_IRQ (0x0D, 2, 0x20, 0x2000);
BUILD_IRQ (0x0E, 2, 0x40, 0x4000);
BUILD_IRQ (0x0F, 2, 0x80, 0x8000);

static void (*complex_irq[NR_IRQ]) (void) =
{
complex_IRQ_0x00, complex_IRQ_0x01, complex_IRQ_0x02, complex_IRQ_0x03,
    complex_IRQ_0x04, complex_IRQ_0x05, complex_IRQ_0x06, complex_IRQ_0x07,
    complex_IRQ_0x08, complex_IRQ_0x09, complex_IRQ_0x0A, complex_IRQ_0x0B,
    complex_IRQ_0x0C, complex_IRQ_0x0D, complex_IRQ_0x0E, complex_IRQ_0x0F};

static void (*fast_irq[NR_IRQ]) (void) =
{
fast_IRQ_0x00, fast_IRQ_0x01, fast_IRQ_0x02, fast_IRQ_0x03,
    fast_IRQ_0x04, fast_IRQ_0x05, fast_IRQ_0x06, fast_IRQ_0x07,
    fast_IRQ_0x08, fast_IRQ_0x09, fast_IRQ_0x0A, fast_IRQ_0x0B,
    fast_IRQ_0x0C, fast_IRQ_0x0D, fast_IRQ_0x0E, fast_IRQ_0x0F};

static void (*wrong_irq[NR_IRQ]) (void) =
{
wrong_IRQ_0x00, wrong_IRQ_0x01, wrong_IRQ_0x02, wrong_IRQ_0x03,
    wrong_IRQ_0x04, wrong_IRQ_0x05, wrong_IRQ_0x06, wrong_IRQ_0x07,
    wrong_IRQ_0x08, wrong_IRQ_0x09, wrong_IRQ_0x0A, wrong_IRQ_0x0B,
    wrong_IRQ_0x0C, wrong_IRQ_0x0D, wrong_IRQ_0x0E, wrong_IRQ_0x0F};
/*
void
set_gate (u32 * gate_addr, u32 type, u32 dpl, u32 addr)
{
  gate_addr[0] = (KERNEL_CS << 16) | (addr & 0xFFFF);
  gate_addr[1] = (addr & 0xFFFF0000) | 0x8000 | (dpl << 13) | (type << 8);

  printk("gaddr=0x%p  type=0x%x  dpl=0x%x  addr=0x%x\n",
	    gate_addr, type, dpl, addr);
  printk("gate_addr[0] = 0x%p\n", gate_addr[0]);
  printk("gate_addr[1] = 0x%p\n", gate_addr[1]);
}
*/
void
irq_disable (unsigned int i)
{
  unsigned long fl;
  unsigned char mask;
  get_flags (fl);
  cli ();
  mask = 1 << (i & 7);
  if (i < 8)
    outb (inb (0x21) | mask, 0x21);
  else
    outb (inb (0xA1) | mask, 0xA1);
  set_flags (fl);
}

void
irq_enable (unsigned int i)
{
  unsigned long fl;
  unsigned char mask;
  get_flags (fl);
  cli ();
  mask = ~(1 << (i & 7));
  if (i < 8)
    outb (inb (0x21) & mask, 0x21);
  else
    outb (inb (0xA1) & mask, 0xA1);
  set_flags (fl);
}

void
complex_IRQ_handler (int irq_nr, struct regdump regs)
{
  struct irq_action *irq = irq_table + irq_nr;
  irq->handler (irq_nr, &regs);
}

void
ext_handler ()
{
  int i, k;
  u32 mask = ext_mask & ext_active;

  for (i = 0; i < NR_IRQ; i++)
    {
      k = mask & (1 << i);
      if (k)
	{
	  ext_active &= ~k;
	  if (ext_table[i])
	    ext_table[i] ();
	}
    }
}

void
fast_IRQ_handler (int irq_nr)
{
  struct irq_action *irq = irq_table + irq_nr;
  irq->handler (irq_nr, NULL);
}

int
irq_request (int irq_nr, int flags, void (*handler) (int, struct regdump *),
	     const char *name)
{
  struct irq_action *irq;

  if (irq_nr >= NR_IRQ)
    return -EINVAL;

  irq = irq_table + irq_nr;
  if (irq->handler)
    {
      printk ("irq_request: irq %d already used by %s\n", irq_nr, irq->name);
      return -EBUSY;
    }

  irq->handler = handler;
  irq->flags = flags;
  irq->name = name;

  if (flags & IRQ_FAST)
    set_int_gate (irq_nr + IRQ_HOLE, fast_irq[irq_nr]);
  else
    set_int_gate (irq_nr + IRQ_HOLE, complex_irq[irq_nr]);

  irq_enable (irq_nr);

  return 0;
}

int
irq_adjust (int irq_nr, int flags, void (*handler) (int, struct regdump *))
{
  struct irq_action *irq;
  ulg fl;

  if (irq_nr >= NR_IRQ)
    return -EINVAL;

  irq = irq_table + irq_nr;

  get_flags (fl);
  cli ();
  irq->flags = flags;
  irq->handler = handler;

  if (flags & IRQ_FAST)
    set_int_gate (irq_nr + IRQ_HOLE, fast_irq[irq_nr]);
  else
    set_int_gate (irq_nr + IRQ_HOLE, complex_irq[irq_nr]);

  set_flags (fl);
  return 0;
}

int
irq_free (int irq_nr)
{
  struct irq_action *irq = irq_table + irq_nr;

  if (irq_nr >= NR_IRQ)
    return -EINVAL;

  bzero ((void *) irq, sizeof (struct irq_action));

  irq_disable (irq_nr);
  set_int_gate (irq_nr + IRQ_HOLE, wrong_irq[irq_nr]);
  return 0;
}

void
irq_init ()
{
  int i;

  for (i = 0; i < NR_IRQ; i++)
    set_int_gate (IRQ_HOLE + i, wrong_irq[i]);

  port_21 = inb (0x21);
  port_A1 = inb (0xA1);

  nr_running_irq = 0;		/* just to be sure */
}

#define IRQPROBESTART		3	/* starting irq for probe */

char irq_probe_table[NR_IRQ] = { 0, };

void
irq_probe_int (int irq, struct regdump *rd)
{
#ifdef IRQPROBEDEBUG
  printk ("irq_probe_int: int=0x%x\n", irq);
#endif
  irq_probe_table[irq]++;
}

int
irq_probe_end ()
{
  int i, res = -1, max = 0;

  for (i = IRQPROBESTART; i < NR_IRQ; i++)
    {
      if (irq_table[i].handler == &irq_probe_int && irq_probe_table[i])
	{
	  irq_free (i);

	  if (res != -1)
	    printk ("irq_probe_end: probe collision, irq=0x%x\n", i);

	  if (irq_probe_table[i] >= max)
	    {
	      max = irq_probe_table[i];
	      res = i;
	    }
	}
    }

  return res;
}

void
irq_probe_start ()
{
  int i;

  bzero (irq_probe_table, sizeof (irq_probe_table));

  for (i = IRQPROBESTART; i < NR_IRQ; i++)
    {
      if (!irq_table[i].handler)
	irq_request (i, IRQ_COMPLEX, &irq_probe_int, "irq-probe");
    }
}
