/*
   kernel/main.c
 */

#include <asm/sysinfo.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/timer.h>
#include <phantom/version.h>
#include <string.h>
#include <phantom/major.h>
#include <conio.h>
#include <phantom/fs.h>
#include <phantom/tty.h>
#include <phantom/sys_enum.h>
#include <asm/head.h>
#include <phantom/mm.h>
#include <phantom/k_unistd.h>

struct sysinfo sysinfo;		/* head.S will pass here info */

#define BREAKPOINT()		setattr(ATTR(LIGHTRED, BLACK)); \
				printf("breakpoint reached,"\
				" any key to continue\n"); \
				setattr(DEFAULT_ATTR); \
				getch();

unsigned short null_idt[3] =
{0,};

/* call done for all except for keyboard, used for SE_ACK_REBOOT */
int fl_system_exit = SE_REBOOT;

void
done_all ()
{
  buffer_done ();
  timer_done ();
  kb_done ();
/*  fs_done ();
   buffer_done ();
   block_done (); */
/* mixer_done(); */
/* mm_done */
}

int fl_exit = 0;

void
sys_reboot (int method)
{
  if (fl_exit)
    return;

  fl_exit = 1;

  switch (method)
    {
    case SE_QUICK_BOOT:
      printk ("SE_QUICK_BOOT n/a yet\n");
      break;

    case SE_SAFE_HALT:
      sys_sync ();
    case SE_HALT:
      for (;;);
      break;			/* ;-) */

    case SE_SAFE_BOOT:
      sys_sync ();
    case SE_POWER_DOWN:
    case SE_REBOOT:
    case SE_ACK_REBOOT:
      if (method == SE_REBOOT)
	printf ("Reboot in progress.\n");
      else
	printf ("Halt in progress.\n");
      sti ();

      if (method == SE_ACK_REBOOT)
	{
	  flush_keyboard ();
	  printf ("Press any key to reboot.\n");
	  getch ();
	}

      printf ("Rebooting...\n");
      cli ();
      __asm__ __volatile__ ("lidt %0 ; int $0"::"m" (null_idt));
      for (;;);

      break;
    }
}

int
fork ()
{
  int ret;
__asm__ ("int $0x80": "=a" (ret):"a" (__NR_fork));
  return ret;
}

int
execve (char *name, char **argv, char **envp)
{
  int ret;
__asm__ ("int $0x80": "=a" (ret):"a" (__NR_execve),
	   "b" (name), "c" (argv), "d" (envp));
  return ret;
}

void start_services()
{
  trap_init ();
  irq_init ();
  video_init ();

  tty_init ();

  mm_init ();

  tty_init2 ();

  sched_init ();
  timer_init ();
  kb_init ();

  pci_init ();

/* some char devices... */
  null_init ();
/*  serial_init ();
   lpt_init (); */

/* drivers which will use kmalloc, etc. */
  floppy_init ();
  ide_init ();

  ramdisk_init ();
  sched_init ();

/* mixer does not work, probably not working alloc_dma_al_pages... */
/*  mixer_init ();      */

  fs_init ();			/* mount root and misc... */
  net_init ();

  tty_init3 ();

  net_init ();
  rg_init ();

  printk ("Going into user mode, running /bin/init...\n");

/* create user stack for zero task */
  map_pages (USER_STACK_PAGE - NR_USER_STACK_PAGES, NR_USER_STACK_PAGES);
}

/* maybe with zero code optimalization could work in one function :-) */
void
start_kernel ()			/* no return function */
{
/*  char *envp[] = { "HOME=/", NULL };
   char *argv[] = { "/", NULL }; */

  start_services ();
/* Note: Optimalization -O2 did "addl 0xc,esp" after switch_to_user
(after fork call to minimize addl x,esp usage),
which already set user stack to beginning (0) -> collision */
  
  switch_to_user ();		/* jump to user mode */

  if (!fork ())
    execve ("/bin/init", NULL, NULL /*argv, envp */ );

  for (;;);			/* idle loop */  
  
/* no return */
}
