/*
   kernel/sched.c
   Phantom scheduler stuff.
 */

#include <asm/system.h>
#include <asm/head.h>
#include <asm/regdump.h>
#include <phantom/k_unistd.h>
#include <conio.h>
#include <sys/types.h>
#include <phantom/sched.h>
#include <phantom/tty.h>
#include <phantom/mm.h>
#include <phantom/fs.h>
#include <phantom/lock.h>
#include <sys/wait.h>

#define DEFAULT_BEATS	3

extern struct file *get_free_filp ();

void syscall_return (void);

struct task zero_task = {
  TASK_RUNNING,			/* status */
  DEFAULT_BEATS,		/* (heart) beats */
  0,				/* sigqueuestat */
  0,				/* signal */
  0,				/* blockedsig */
  NULL,				/* thread_lock */
  NULL,				/* thread_next */
  0,				/* PID */
  0,				/* exit_code */
  0,				/* trap_nr */
  0,				/* error_code */
  1,				/* fl_dump */
  0,				/* fl_lock */
  NULL,				/* parent task */
  NULL,				/* child task */
  NULL,				/* next sibling task */
  NULL,				/* previous sibling task */
  0x00000000,			/* code_pos */
  0x40000000,			/* code_end */
  0x40000000,			/* data_end */
  0,				/* code pages */
  0,				/* nr_pages */
  0,				/* nr_userpages */
  0,				/* tty number */
  NULL,				/* no window by default */
  NULL,				/* pwd */
  NULL,				/* root */
  NULL,				/* exe-file vnode */
  {NULL,},			/* file descriptor table */
  {				/* TSS */
   0, 0,			/* back-link */
   FIRST_KERNEL_PAGE * PAGESIZE,	/* esp0 */
   KERNEL_DS, 0,		/* ss0 */
   0,				/* esp1 */
   0, 0,			/* ss1 */
   0,				/* esp2 */
   0, 0,			/* ss2 */
   (u32 *) CR3,			/* cr3 */
   0,				/* eip */
   0,				/* eflags */
   0,				/* eax */
   0,				/* ecx */
   0,				/* edx */
   0,				/* ebx */
   0,				/* esp */
   0,				/* ebp */
   0,				/* esi */
   0,				/* edi */
   0, 0,			/* es */
   0, 0,			/* cs */
   USER_DS, 0,			/* ss */
   0, 0,			/* ds */
   0, 0,			/* fs */
   0, 0,			/* gs */
   0, 0,			/* ldt */
   0, IO_PERM_OFFSET,		/* trace, io_perm_offset */
   {~0,}			/* ioperm */
   },
  "",				/* pwd_str */
  &zero_task,			/* next task (cycled list) */
  &zero_task,			/* previous task */
  TSS (0),			/* tss selector */
  {0,},				/* ldt selector */
  FIRST_KERNEL_PAGE,		/* kernel stack page */
  NULL,				/* user stack */
  NR_KERNEL_STACK_PAGES,	/* kernel stack pages */
  NR_USER_STACK_PAGES,		/* user stack pages */
  {0,},				/* sigaction */
  {0,},				/* sigpool */
  0				/* lsig */
};

TASK *task_table[NR_TASK] = { &zero_task, NULL, };
TASK *ctask = &zero_task;
TASK *ztask = &zero_task;	/* constant */

//#define SCHEDULEDEBUG

/* find task via pid */
TASK *find_task (int pid);
void task_info ();

void
schedule ()
{
  struct task *task = ctask;

#ifdef SCHEDULEDEBUG
//  printf ("schedule: seeking task to switch, ctask=0x%p, pid=%d\n", 
  //  ctask, ctask->pid);
#endif

  for (;;)
    {
      task = task->next;

      if (task == ctask)
	{
#ifdef SCHEDULEDEBUG
//        printf ("schedule: giving up, single task\n");
#endif
	  return;
	}
#ifdef SCHEDULEDEBUG
      else
	printf ("schedule: next task in queue=0x%p, pid=%d\n",
		task, task->pid);
#endif

      if (task->status == TASK_RUNNING)
	{
#ifdef SCHEDULEDEBUG
	  printf ("schedule: task switched, task=0x%p, pid=%d\n",
		  task, task->pid);
#endif
	  task->beats = DEFAULT_BEATS;
	  switch_task (task);
	  return;
	}
    }
}

void
set_desc (u32 dst, u32 base, u32 pages, u8 type)
{
  ((u16 *) dst)[0] = pages & 0xFFFF;
  ((u16 *) dst)[1] = base & 0xFFFF;
  ((u16 *) dst)[2] = ((base >> 16) & 0xFF) | (type << 8);
  ((u16 *) dst)[3] = ((base >> 16) & 0xFF00) | 0xC0 | ((pages >> 16) & 0x0F);
}

/* Had to create set_page_desc because of messing with
   numbers bigger than integer. Thanx, GCC. */
/* I cannot realize why this does not work properly
   with gdt as destination!!!
   Descriptor was slightly degenerated... */
void
set_page_desc (u32 dst, u32 base, u32 pages, u8 type)
{
  ((u16 *) dst)[0] = pages & 0xFFFF;
  ((u16 *) dst)[1] = (base & 0x0F) << 12;
  ((u16 *) dst)[2] = ((base >> 4) & 0xFF) | (type << 8);
  ((u16 *) dst)[3] = ((base >> 4) & 0xFF00) | ((pages >> 16) & 0x0F) | 0xC0;
}

void
lock_threads ()
{
  struct task *thread;
/*
   ctask->status = TASK_[
   thread=ctask->thread
 */
}

void
unlock_threads ()
{
}

int
copy_fd (struct task *task)
{
  int i;
  struct file *file;

  lock_threads ();

  for (i = 0; i < NR_OPEN; i++)
    {
      if (ctask->filp[i])
	{
	  if (!(file = get_free_filp ()))
	    return -EMFILE;

	  if (ctask->filp[i]->f_vnode != NULL)
	    vnode_gain (ctask->filp[i]->f_vnode);
	  memcpy (file, ctask->filp[i], sizeof (struct file));
	  task->filp[i] = file;
	}
    }
  unlock_threads ();

  return 0;
}

/* should add some kind of lock when messing with threads... */
void
copy_pd (struct task *task)
{
  u32 *pt, *pg;
  int i, table, entry, count;

/* copy kernel area map -- this really must be done */
  task->tss.cr3 = (u32 *) page_alloc ();
  bzero (task->tss.cr3, PAGE_SIZE);

  for (i = 0; i < 0x100; i++)
    CPDE (task, i) = PDE (i);

  if (!task->nr_pages)
    {
      printk ("copy_pd: warning, nr_pages = 0\n");
      return;
    }

  table = 0x100;
  entry = 0;
  count = task->nr_pages + task->nr_userpages;

/* set all tables cow (NOT STACKS!) and not codelock! */
/* 0x300 should be enough and should not damage stacks ;) */
  for (table = 0x100; table < 0x300; table++)
    {
      pt = PT (table);
      if (pt == NULL)
	continue;

      pg = (u32 *) page_alloc ();	/* page dir could be cowish, too */
      bzero (pg, PAGE_SIZE);

      CPDE (task, table) = (u32) pg | PGS_WRITEABLE | PGS_USER | PGS_PRESENT;

      for (entry = 0; entry < 1024; entry++)
	{
	  if (pt[entry] & ADDR_MASK)
	    {
	      pg[entry] =
		(pt[entry] & ADDR_MASK) | PGS_COW | PGS_PRESENT | PGS_USER;
	      pt[entry] = pg[entry];	/* set cow here, too! */
	      set_map (TOPAGE (pt[entry]), 1, PG_SHARE);

	      count--;
	    }

	  if (!count)
	    return;
	}
    }

  printk ("copy_pd: pretty strange, the nr_userpages may be wrong (%d)\n",
	  ctask->nr_userpages);
}

int
sys_fork (struct regdump rd)
{
  int i, nr_tables;
  TASK *task;
  struct regdump *newdump;
  CODE_LOCK (a);

  if (!(task = (TASK *) kmalloc (sizeof (struct task), PG_KERNEL)))
    return -ENOMEM;

  SINGLE_CODE (a);

  for (i = 0; i < NR_TASK; i++)
    if (task_table[i] == NULL)
      break;

  if (i == NR_TASK)
    {
      MULTI_CODE (a);
      kfree (task);

      return -ENOMEM;
    }

  task->status = TASK_SPAWNING;
  task_table[i] = task;
  MULTI_CODE (a);

  task->pid = i;
  task->root = ctask->root;
  task->pwd = ctask->pwd;
  strncpy (task->pwd_str, ctask->pwd_str, PATH_MAX);

  task->parent = ctask;
  task->tty = ctask->tty;

  task->nr_pages = ctask->nr_pages;
  task->nr_userpages = ctask->nr_userpages;

/* building kernel stack */
  task->kernel_stack_page = gfps (NR_KERNEL_STACK_PAGES);
  task->kernel_stack_page += NR_KERNEL_STACK_PAGES;

  copy_fd (task);
  copy_pd (task);

  task->tss.cs = KERNEL_CS;
  task->tss.ds = KERNEL_DS;
  task->tss.es = KERNEL_DS;
  task->tss.fs = KERNEL_DS;
  task->tss.gs = KERNEL_DS;
  task->tss.ss = KERNEL_DS;
  task->tss.ss0 = KERNEL_DS;
  task->tss.esp0 = TOADDR (task->kernel_stack_page);
  task->tss.eflags = 0x202;	/* interrupts enabled */
  task->tr = TSS (task->pid);
  task->tss.ldt = LDT (task->pid);

  vnode_gain (task->root);
  vnode_gain (task->pwd);

  set_tss_desc ((u32) gdt + TSS (task->pid), &task->tss);
  set_ldt_desc ((u32) gdt + LDT (task->pid), task->ldt, 1);

/* copy user stack */
  {
    u32 *pt;
    u32 addr;
    int i;

    if (CPDE (task, 1023))
      {
	printf ("task_make_stacks: cr3[1023] defined!\n");
	return -1;
      }

    pt = (u32 *) page_alloc ();
    bzero (pt, PAGE_SIZE);
    CPDE (task, 1023) = (u32) pt | PGS_USER | PGS_WRITEABLE | PGS_PRESENT;

/* IMPORTANT: last entry of that table is used for a stack trap... */

    for (i = 0; i < NR_USER_STACK_PAGES; i++)
      {
	addr = page_alloc ();
/* copy */
	memcpy ((void *) addr,
		(void *) TOADDR (USER_STACK_PAGE - i - 1), PAGE_SIZE);
/* and map */
	pt[1023 - i] = addr | PGS_PRESENT | PGS_USER | PGS_WRITEABLE;
      }
  }

/* fork specific */
  newdump = ((struct regdump *) (TOADDR (task->kernel_stack_page))) - 1;
  task->tss.esp = (u32) newdump;
  *newdump = rd;		/* copy kernel stack */
  newdump->eax = 0;		/* child's return value */

  task->tss.eip = (u32) & syscall_return;

  task->status = TASK_RUNNING;

  task->parent = ctask;
  task->psbl = NULL;
  task->nsbl = ctask->child;

  task->next = ctask->next;
  task->prev = ctask;

  ctask->child = task;

  SINGLE_CODE (a);
  if (task->nsbl != NULL)
    task->nsbl->psbl = task;

  ctask->next->prev = task;
  ctask->next = task;
  MULTI_CODE (a);

  return task->pid;		/* father returns pid */
}

void
count_str (char **s, int *a_count, int *a_size)
{
  int count = 0, size = 0;

  while (s[count])
    {
      size += strlen (s[count]) + 1;	/* +terminator */
      count++;
    }

  *a_count = count;
  *a_size = size;
}

void
copy_args (void *dst, char **src, int count)
{
  char **ptr = (char **) dst;
  char *last = (char *) (dst + (count + 1) * sizeof (u32));

  while (count--)
    {
      *ptr++ = last;

      strcpy (last, *src);
      last += strlen (*src) + 1;	/* extra one for terminator */

      src++;			/* next entry */
    }

  *ptr = NULL;
}

/* FIXME: check whether we have enough memory or load and then free
   ctask... */
int
sys_execve (char *fname, char *argv[], char *envp[])
{
  int i, ret;
  u32 flg, *pt;
  u32 *stack;
  int argc = 0, envc = 0, arg_size, env_size, arg_page;
  TASK *tmp;
  struct regdump *rd = (struct regdump *) &fname;

  if (!ctask->pid)
    {
      printf ("execve: zero task cannot do execve!\n");
      return -EPERM;
    }

  if ((tmp = kmalloc (sizeof (TASK), 0)) == NULL)
    {
      printk ("execve: oom\n");
      return -ENOMEM;
    }

  if (read_header (fname, tmp))
    {
      printk ("load_task: invalid executable file format\n");
      return -EINVAL;
    }

  if (argv != NULL)
    {
      count_str (argv, &argc, &arg_size);
      if (arg_size + sizeof (u32) * argc > PAGESIZE)
	{
	  printk ("execve: argument list too long (argc=%d, arg_size=%d)\n",
		  argc, arg_size);
	  return -ENOMEM;
	}

      arg_page = gfp ();

      copy_args ((void *) TOADDR (arg_page), argv, argc);

      {
	char **x;
	int i;
	x = (char **) TOADDR (arg_page);
	printk ("  %d arguments\n", argc);
	for (i = 0; i <= argc; i++)
	  printk ("%d.) %s\n", i, x[i]);
      }
    }

/* remove all pages (which may be shared) but keep stacks */
  ctask_cleanup ();

  ctask->vnode = tmp->vnode;
  ctask->code_pos = tmp->code_pos;
  ctask->code_end = tmp->code_end;
  ctask->data_end = tmp->data_end;
  ctask->code_pages = tmp->code_pages;
  ctask->nr_pages = tmp->nr_pages;

  kfree (tmp);

  map_pages (FIRST_USER_PAGE, ctask->nr_pages);

  if (argv != NULL)
    link_page (arg_page, FIRST_USER_PAGE + ctask->nr_pages);

  ro_pages (FIRST_USER_PAGE, ctask->code_pages);

#ifdef CODELOCKDEBUG
  printk ("load_task: %d pages of code were locked (set ro)\n",
	  ctask->code_pages);
  printk ("code_pos=0x%x\n", ctask->code_pos);
#endif

/* We do not need map_tables and free_tables, since 1:1 mapping.
   read_exec routine will be only slightly complicated. */
  if (read_exec (ctask))
    {
      printf ("load_task: too bad, cannot read file\n");
      return -1;
    }

  ctask->nr_userpages = 0;

  ctask->tss.esp = TOADDR (ctask->kernel_stack_page);
  ctask->tss.esp0 = TOADDR (ctask->kernel_stack_page);
  ctask->tss.eflags = 0x202;	/* interrupts enabled */
  ctask->tss.io_perm_offset = IO_PERM_OFFSET;
  for (i = 0; i < IO_BITMAP_SIZE; i++)
    ctask->tss.io_bitmap[i] = 0xFFFFFFFF;

/* free user stack */
/* WRITEME: add arguments */
  stack = (u32 *) (TOADDR (USER_STACK_PAGE) - 8);
  if (argv != NULL)
    {
      stack[0] = argc;
      stack[1] = TOADDR (FIRST_USER_PAGE + ctask->nr_pages);
    }
  else
    {
      stack[0] = 0;
      stack[1] = 0;
    }
  rd->esp = (u32) stack;
/* FIXME: user stack should be bzeroed */

  rd->eip = ctask->code_pos;

  return 0;
}

/* do final drop of the task, all used resorces are freed */
void
task_drop (TASK * t)
{
  int i, task_idx = -1, has = 0;
  TASK *parent = NULL, *lsbl = NULL, *rsbl = NULL, *tmp;
  CODE_LOCK (a);

  printk ("task_drop: pid=%d (by pid %d)\n", t->pid, ctask->pid);

  SINGLE_CODE (a);
/* unlink from task circullar buffer */
  t->prev->next = t->next;
  t->next->prev = t->prev;

  for (i = 0; i < NR_TASK; i++)
    {
      if ((tmp = task_table[i]) == NULL)
	continue;

/* find the task itself */
      if (tmp == t)
	task_idx = i;

/* it's parent */
      if (tmp->parent == t)
	parent = t;

/* it's left/previous sibling */
      if (tmp->nsbl == t)
	lsbl = t;

/* it's right/next sibling */
      if (tmp->psbl == t)
	rsbl = t;
    }

  if (lsbl != NULL)
    {
      if (parent != NULL)
	{
	  parent->child = lsbl;
	  parent = NULL;
	}
      lsbl->nsbl = t->nsbl;
    }

  if (rsbl != NULL)
    {
      if (parent != NULL)
	{
	  parent->child = rsbl;
	  parent = NULL;
	}
      rsbl->psbl = t->psbl;
    }

  if (parent != NULL)
    parent->child = NULL;

  if (task_idx != -1)
    task_table[task_idx] = NULL;
  else
    printk ("task_drop: Ooops, cannot find task in task_table.\n");

  MULTI_CODE (a);

  task_free (t);		/* free ALL used memory */
  kfree (t);
}

/* put task in zombie mode */
void
kill_task (TASK * task)
{
  int i;

#ifdef TASKDEBUG
  printk ("kill_task: being called, task=0x%p, ec=0x%x\n", task, ec);
#endif

  if (!task->pid)
    {
      printk ("kill_task: cannot exit zero task!\nSystem halted.");
      for (;;);
    }

  task->status = TASK_ZOMBIE;

/* close all opened files */
  for (i = 0; i < NR_OPEN; i++)
    if (task->filp[i])
      sys_close (i);

/* drop current dir and root vnodes */
  vnode_release (task->pwd);
  vnode_release (task->root);

  if (task->parent != NULL)
    send_sig (SIGCHLD, task->parent);

  schedule ();
}

void
sys_exit (int ec)		/* what to do with ec now... ? */
{
  kill_task (ctask);

  printk ("sys_exit: cannot get here.. ctask=0x%x  pid=%d\n",
	  ctask, ctask->pid);
  sti ();
  for (;;);			/* we should never get here */
}

/* ugh, rewrite me somehow someday... */
int
sys_waitpid (int pid, int *status, int options)
{
  TASK *t;
  int ret, fl_found;

  printk ("sys_waitpid: pid=%d  status=0x%x  options=%d\n",
	  pid, status, options);

  for (;;)
    {
      fl_found = 0;
/* pid > 0 .. wait for specified pid */
/* stroll thru all siblings */
      for (t = ctask->child; t != NULL; t = t->nsbl)
	{
	  if (pid > 0 && pid != t->pid)
	    continue;

	  switch (t->status)
	    {
	    case TASK_ZOMBIE:
	      if (status != NULL)
		*status = t->exit_code;

	      pid = t->pid;	/* pid could be -1 */
	      task_drop (t);
	      return pid;

	    case TASK_STOPPED:
	      if (options & WUNTRACED)
		return t->pid;	/* else continue to fl_found = 1 */

	    default:
	      fl_found = 1;	/* at least one searched task living */
	      break;
	    }
	}

      if (!fl_found)
	{
	  printk ("s_wp: !fl_found\n");
	  return -ECHILD;
	}

      if (options & WNOHANG)
	return 0;

      schedule ();
    }
}

int
sys_wait (int *status)
{
  return sys_waitpid (-1, status, 0);
}

/* prepare user stack for signal call, modify eip in kernel stack */
volatile void
prepare_signal (struct regdump rd)
{
  struct sigaction *sa;
  u32 sa_handler;
  u32 *us_ptr;
  u32 old_eip;

  sa = ctask->sigaction + ctask->signal;
  sa_handler = (u32) sa->sa_handler;

  printk ("ESP=0x%x  EIP=0x%x  (&arg=0x%x)\n", rd.esp, rd.eip, &rd);

  old_eip = rd.eip;
  rd.eip = sa_handler;

  us_ptr = (u32 *) rd.esp;
  us_ptr -= 16;			/* 16 * 4 = 64 bytes */
  rd.esp = (u32) us_ptr;

/* return address (signal handler return) 
   and argument - signal number - 8 bytes */
  us_ptr[0] = (u32) (us_ptr + 12);
  printk ("prepare_signal: sig_nr = %d\n", ctask->signal);
  printk ("return address: 0x%x\n", us_ptr + 12);
  us_ptr[1] = ctask->signal;

/* pusha - 32 bytes */
  us_ptr[2] = rd.edi;
  us_ptr[3] = rd.esi;
  us_ptr[4] = rd.ebp;
  us_ptr[5] = 0;		/* popa ignores this value (esp is not modified) */
  us_ptr[6] = rd.ebx;
  us_ptr[7] = rd.edx;
  us_ptr[8] = rd.ecx;
  us_ptr[9] = rd.eax;

/* till here it will be freed with popa */

/* eflags - 4 bytes */
  us_ptr[10] = rd.eflags;	/* this will be freed by popf */

/* original return address - 4 bytes */
  us_ptr[11] = old_eip;

/* recovery code - 16 bytes */
/* 
58		pop eax
61 		popad
9D		popf
50		push eax
b820000000	mov 0x20,eax
CD80		int 0x80
58		pop eax
90		nop
C2 10 00	ret 16		(skip recovery code)
*/
  us_ptr[12] = 0x509D6158;
  us_ptr[13] = 0x000020B8;
  us_ptr[14] = 0x5880CD00;
  us_ptr[15] = 0x0010C290;
/* and this one will be freed by final 'ret' */

  ctask->sigqueuestat = SQUEUE_RUNNING;	/* ready to run the signal */
}

void
sys_sigret (struct regdump rd)
{
  int tmp;
  CODE_LOCK (a);

#ifdef SIGRETDEBUG
  printk ("sys_sigret: being called!\n");
#endif

  if (!ctask->signal)
    {
      printk ("sys_sigret: oops, signal == 0 (pid=%d)\n", ctask->pid);
      return;
    }

  SINGLE_CODE (a);
  tmp = ctask->sigpool[ctask->signal];
  ctask->sigpool[ctask->signal] = 0;
  ctask->signal = tmp;		/* do next signal or 0 for nop */

  if (!tmp)
    {
#ifdef SIGRETDEBUG
      printk ("last signal in queue\n");
#endif
      ctask->sigqueuestat = SQUEUE_EMPTY;	/* entirely empty signal queue */
      ctask->lsig = 0;
    }
  else
    {
#ifdef SIGRETDEBUG
      printk ("gearing next signal\n");
#endif
      ctask->sigqueuestat = SQUEUE_PENDING;	/* waiting for ignition */
    }

#ifdef SIGRETDEBUG
  printk ("sys_sigret: next signal = %d\n", tmp);
#endif

  MULTI_CODE (a);
}

__sighandler_t
sys_signal (int nr, void (*sighandler) (int))
{
  __sighandler_t sh;

  if (nr < 0 || nr >= NR_SIGNAL || nr == SIGKILL || nr == SIGSTOP)
    return SIG_ERR;

  sh = (__sighandler_t) ctask->sigaction[nr].sa_handler;
  ctask->sigaction[nr].sa_handler = sighandler;

  return sh;
}

int
send_sig (int sig_nr, TASK * task)
{
  if (task == NULL)
    {
      printk ("send_sig: warning, NULL task given...\n");
      return -1;
    }

#ifdef SIGNALDEBUG
  printk ("send_sig: pid %d sends signal %d to pid %d\n",
	  ctask->pid, sig_nr, task->pid);
#endif

  switch (sig_nr)
    {
    case SIGKILL:
#ifdef SIGNALDEBUG
      printk ("send_sig: KILLING pid %d\n", task->pid);
#endif

      printf ("Killing task (pid=%d).\n", task->pid);
      printk ("Killing task (pid=%d).\n", task->pid);

      kill_task (task);
      break;

    case SIGSTOP:
      printk ("send_sig: stop signal caught (pid=%d)\n", task->pid);
      task->status = TASK_STOPPED;
      schedule ();
      break;

    default:
      {
	CODE_LOCK (a);

	if (sig_nr < 0 || sig_nr >= NR_SIGNAL)
	  {
	    printk ("send_sig: invalid signal number\n");
	    return -EINVAL;
	  }

/* signal blocked, cannot allow */
	if (task->blockedsig & (1 << sig_nr))
	  {
	    printk ("send_sig: tried to call blocked signal\n");
	    return 0;
	  }

/* check whether signal is pending */
	if (task->lsig == sig_nr || task->sigpool[sig_nr])
	  {
#ifdef SIGNALDEBUG
	    printk ("send_sig: signal %d already pending (pid=%d)\n", sig_nr,
		    task->pid);
#endif
	    return 0;
	  }

/* signal not defined or ignored */
	if (task->sigaction[sig_nr].sa_handler == SIG_IGN)
	  {
	    switch (sig_nr)
	      {

#define task_kill(t) kill_task(t)
/* 'A' terminate */
	      case SIGHUP:
		printf ("SIGHUP (pid %d)\n", task->pid);
		task_kill (task);
		break;
	      case SIGINT:
		printf ("SIGINT (pid %d)\n", task->pid);
		task_kill (task);
		break;
	      case SIGPIPE:
		printf ("SIGPIPE (pid %d)\n", task->pid);
		task_kill (task);
		break;
	      case SIGALRM:
		printf ("SIGALRM (pid %d)\n", task->pid);
		task_kill (task);
		break;
	      case SIGTERM:
		printf ("SIGTERM (pid %d)\n", task->pid);
		task_kill (task);
		break;
/* 'C' terminate and dump core */
	      case SIGQUIT:
		printf ("SIGQUIT (pid %d)\n", task->pid);
		task_kill (task);
		break;
	      case SIGILL:
		printf ("SIGILL (pid %d)\n", task->pid);
		task_kill (task);
		break;
	      case SIGABRT:
		printf ("SIGABRT (pid %d)\n", task->pid);
		task_kill (task);
		break;
	      case SIGFPE:
		printf ("SIGFPE (pid %d)\n", task->pid);
		task_kill (task);
		break;
	      case SIGSEGV:
		printf ("SIGSEGV (pid %d)\n", task->pid);
		task_kill (task);
		break;

	      default:
		printk("pid %d ignores signal %d\n", task->pid, sig_nr);
	        return;
	      }
	  }

	SINGLE_CODE (a);
/* first signal to run (no signal running now) */
	if (task->sigqueuestat == SQUEUE_EMPTY)
	  task->sigqueuestat = SQUEUE_PENDING;

#ifdef SENDSIGDEBUG
	printk ("send_sig(%d)\n", sig_nr);
#endif
	if (!task->signal)
	  {
#ifdef SENDSIGDEBUG
	    printk ("first in queue\n");
#endif
	    task->signal = sig_nr;
	  }
	else
	  {
#ifdef SENDSIGDEBUG
	    printk ("next in queue\n");
#endif
	    task->sigpool[task->lsig] = sig_nr;
	    task->sigpool[sig_nr] = 0;	/* just to be sure */
	  }

	task->lsig = sig_nr;

	MULTI_CODE (a);
      }
      break;
    }

  return 0;
}

/* find task via pid */
TASK *
find_task (int pid)
{
  int i;

  for (i = 0; i < NR_TASK; i++)
    {
      if (task_table[i] != NULL && task_table[i]->pid == pid)
	return task_table[i];
    }

  return NULL;
}

int
sys_kill (int pid, int nr)
{
  TASK *tsk;

  tsk = find_task (pid);
  return send_sig (nr, tsk);
}

/* raise signal for ctask */
int
sys_raise (int nr)
{
  printk ("sys_raise: ESP=0x%x  EIP=0x%x (&arg=0x%x)\n",
	  ((struct regdump *) &nr)->esp, ((struct regdump *) &nr)->eip, &nr);

  return send_sig (nr, ctask);
}

void
task_info ()
{
  TASK *tsk;
  int i, nr = 0;

  for (i = 0; i < NR_TASK; i++)
    if ((tsk = task_table[i]) != NULL)
      {
	char status;

	nr++;

	switch (tsk->status)
	  {
	  case TASK_STOPPED:
	    status = 'S';
	    break;
	  case TASK_RUNNING:
	    status = 'R';
	    break;
	  case TASK_ZOMBIE:
	    status = 'Z';
	    break;
	  }

	printf ("i: %-2d pid: %-3d  status: %c  p: 0x%x  n: 0x%x  t: 0x%x\n"
		"p: 0x%x  c: 0x%x  ns: 0x%x  ps: 0x%x\n",
		i, tsk->pid, status, tsk->prev, tsk->next, tsk,
		tsk->parent, tsk->child, tsk->nsbl, tsk->psbl);
      }

  printf ("=====%d tasks total=====\n", nr);
}

void
sched_init ()
{
  setattr (ATTR (GREEN, BLACK));
  printf ("Task Manager (sizeof(TASK)=%d)\n", sizeof (TASK));
  setattr (DEFAULT_ATTR);
  set_tss_desc (gdt + FIRST_TSS, &(zero_task.tss));
  set_ldt_desc (gdt + FIRST_LDT, zero_task.ldt, 1);
  clear_nt ();
  load_tr (TSS (0));
}
