/*
   kernel/syscall.S
 */

#include <asm/idtbuild.h>
#include <asm/kernel.h>

CF_MASK = 0x00000001
IF_MASK = 0x00000200
NT_MASK = 0x00004000
VM_MASK = 0x00020000

/* offsets in struct task (TASK) */
STATUS = 0
BEATS  = 4
SQUEUE = 8

/* offset in stack save structure, the same for pushad/popad... */
EAX = 0x18

.align 4,0x90
syscall_table:

#include <asm/syscalls.h>

.globl syscall_interrupt
.align 4,0x90
syscall_interrupt:
  subl $4,%esp				/* faked error-code */

  ASM_TOTAL_SAVE

  cmpl $NR_SYSCALL, %eax
  jge syscall_warning

  call *syscall_table(,%eax,4)

  movl %eax,EAX(%esp)			/* keep return value alive :) */

  movl (ctask),%eax

  movl SQUEUE(%eax),%ebx
  cmpl $SQUEUE_PENDING,%ebx
  jne no_signal

  call prepare_signal

no_signal:

.global irq_syscall_return
.global syscall_return

irq_syscall_return:			/* still clied by complex_irq */
  cmpl $0,nr_running_irq
  jne skip_schedule
 
  movl ext_mask,%eax
  andl ext_active,%eax
  jz syscall_return
  
  incl nr_running_irq
  pushfl
  sti
  call ext_handler
  popfl
  decl nr_running_irq

syscall_return:
  movl (ctask),%eax			/* prepare_signal damages eax */
  cmpl $0,STATUS(%eax)			/* TASK_STOPPED ? */
  jz run_schedule
  cmpl $0,BEATS(%eax)
  jnz skip_schedule

run_schedule:
  call schedule

skip_schedule:  
  ASM_TOTAL_LOAD
  addl $4,%esp				/* skip error code.. */

  iret

syscall_warning:
  pushl %eax
  pushl $NR_SYSCALL
  pushl $warn
  call printf
  addl $8,%esp
  jmp syscall_return
  
warn:
.ascii "NR_SYSCALL (0x%x) exceeded, eax=0x%x\n"
.byte 0
