/*
   kernel/trap.c
   WRITEME:     *fix esp
   *make reserved numberings, etc..
 */

#include <asm/regdump.h>
#include <asm/head.h>
#include <signal.h>
#include <phantom/sched.h>

extern void page_fault (struct regdump *, int);

void dostacktrace(u32 *bp, int nr_level)
{
while(nr_level--)
  {
  if((u32)bp<4096)
    break;
    
  printk("0x%X ", bp[1]);
  bp=(u32 *)bp[0];
  }  
}

void printcodetrace(struct regdump *rd)
{
printk("code trace: 0x%p\n", *(u32*)rd->eip);
}

void printstacktrace(struct regdump *rd)
{
int i;

printk("user stack trace: ");
dostacktrace((u32 *)rd->ebp, 4);
printk("\n");

printk("kernel stack trace: ");
{
u32 ebp;
__asm__ __volatile__ ("movl %%ebp,%0" : "=g" (ebp) : /* no input */ : "memory");
dostacktrace((u32 *)ebp, 4);
}
printk("\n");
}

void
printregdump (const char *title, struct regdump *regs)
{
  if (title)
    printk ("%s\n", title);

  printk ("pid: %d  eip: 0x%p  efl: 0x%p  error_code: 0x%p\n" \
	  "eax: 0x%p  ebx: 0x%p  ecx: 0x%p  edx: 0x%p\n" \
	  "esi: 0x%p  edi: 0x%p  ebp: 0x%p  kesp: 0x%p\n" \
	  "ss: 0x%04x  esp: 0x%p\n" \
	  "cs: 0x%04x  ds: 0x%04x  es: 0x%04x  fs: 0x%04x  gs: 0x%04x\n",
	  ctask->pid, regs->eip, regs->eflags, regs->error_code,
	  regs->eax, regs->ebx, regs->ecx, regs->edx,
	  regs->esi, regs->edi, regs->ebp, (u32) & regs,
	  regs->ss, regs->esp,
	  regs->cs, regs->ds, regs->es, regs->fs, regs->gs);

//  printcodetrace(regs);
//  printstacktrace(regs);
}

void
print_tss (struct tss_struct *tss)
{
  printk ("tss: 0x%p\n" \
	  "eip: 0x%p  efl: 0x%p  cr3: 0x%p\n" \
	  "eax: 0x%p  ebx: 0x%p  ecx: 0x%p  edx: 0x%p\n" \
	  "esi: 0x%p  edi: 0x%p  ebp: 0x%p  esp: 0x%p\n" \
	  "ss: 0x%04x  ss0: 0x%04x  ss1:0x%04x  ss2: 0x%04x\n" \
	  "cs: 0x%04x  ds: 0x%04x  es: 0x%04x  fs: 0x%04x  gs: 0x%04x\n",
	  tss,
	  tss->eip, tss->eflags, tss->cr3,
	  tss->eax, tss->ebx, tss->ecx, tss->edx,
	  tss->esi, tss->edi, tss->ebp, tss->esp,
	  tss->ss, tss->ss0, tss->ss1, tss->ss2,
	  tss->cs, tss->ds, tss->es, tss->fs, tss->gs);
}

BUILD_TRAP (0x00, SIGFPE, divide_error, "0 Divide error", ctask);
BUILD_TRAP (0x01, SIGSEGV, trace, "1 Trace", ctask);
BUILD_TRAP (0x02, SIGSEGV, non_maskable_interrupt, "2 NMI", ctask);
BUILD_TRAP (0x03, SIGTRAP, int3, "3 INT3", ctask);
BUILD_TRAP (0x04, SIGSEGV, overflow, "4 Overflow", ctask);
BUILD_TRAP (0x05, SIGSEGV, bounds, "5 Bounds", ctask);
BUILD_TRAP (0x06, SIGILL, invalid_operand, "6 Invalid operand", ctask);
BUILD_TRAP (0x07, SIGSEGV, device_not_available, "7 Device not available",
	    ctask);
BUILD_TRAP (0x08, SIGSEGV, double_fault, "8 Double fault", ctask);
BUILD_TRAP (0x09, SIGFPE, coprocessor_segment_overrun,
	    "9 Coprocessor segment overrun", ctask);
BUILD_TRAP (0x0A, SIGSEGV, invalid_tss, "A Invalid TSS", ctask);
BUILD_TRAP (0x0B, SIGBUS, segment_not_present, "B Segment not present",
	    ctask);
BUILD_TRAP (0x0C, SIGBUS, stack_segment, "C Stack segment", ctask);
BUILD_TRAP (0x0D, SIGSEGV, general, "D General protection", ctask);
/* page fault */
BUILD_TRAP (0x0F, SIGSEGV, reserved, "F (reserved)", ctask);
BUILD_TRAP (0x11, SIGSEGV, alignment_check, "E Alignment check", ctask);

DEFAULT_TRAP (divide_error);
DEFAULT_TRAP (trace);
DEFAULT_TRAP (non_maskable_interrupt);
DEFAULT_TRAP (int3);
DEFAULT_TRAP (overflow);
DEFAULT_TRAP (bounds);
DEFAULT_TRAP (invalid_operand);
DEFAULT_TRAP (device_not_available);
DEFAULT_TRAP (double_fault);
DEFAULT_TRAP (coprocessor_segment_overrun);
DEFAULT_TRAP (invalid_tss);
DEFAULT_TRAP (segment_not_present);
DEFAULT_TRAP (stack_segment);
DEFAULT_TRAP (alignment_check);
DEFAULT_TRAP (reserved);

ERROR_CODE_TRAP (general);
ERROR_CODE_TRAP (page_fault);

void *trap_handler[32] = {
  trap_divide_error, trap_trace, trap_non_maskable_interrupt,
  trap_int3, trap_overflow, trap_bounds, trap_invalid_operand,
  trap_device_not_available, trap_double_fault,
  trap_coprocessor_segment_overrun, trap_invalid_tss,
  trap_segment_not_present, trap_stack_segment, trap_general,
  trap_page_fault, trap_reserved, trap_reserved,
  trap_alignment_check, trap_reserved,
};

void
trap_init ()
{
  int i;

  for (i = 0; i < 32; i++)
    set_trap_gate (i, trap_handler[i]);

  set_system_gate (0x80, &syscall_interrupt);
}
