/*
   mm/page.c
   Memory manager core routines.

   WRITEME: add some 'reserveds' for set_map...
   FIXME: respect PGS_TABLE_LOCKED!
 */

#include <asm/regdump.h>
#include <asm/sysinfo.h>
#include <asm/system.h>
#include <phantom/sched.h>
#include <sys/types.h>
#include <errno.h>
#include <phantom/mm.h>
#include <conio.h>
#include <phantom/lock.h>

#define DMA_ABLE_PAGES		4096
#define IS_DMA_ABLE(p_page)	((p_page)<DMA_ABLE_PAGES)

u16 *mem_map = NULL;

/* physical memory adittional information */
int nr_pages;
int nr_free_pages;
int nr_dma_pages;
int nr_free_dma_pages;
int nr_min_free_pages;
int non_dma_start;

void
set_map (int page, int count, int flags)
{
  u32 *pt;
  int i;

  if (!count)
    return;

  if (page + count > nr_pages)
    {
      printf
	("set_map: page (0x%x) + count (0x%x) > nr_pages (0x%x) giving up\n",
	 page, count, nr_pages);
      return;
    }

  for (i = page; i < page + count; i++)
    {
      switch (flags)
	{
	case PG_USED:
	  if (mem_map[i])
	    printk ("set_map: page 0x%x already being used\n", i);
	  mem_map[i] = 1;
	  break;

	case PG_FREE:
	  if (!mem_map[i])
	    printk ("set_map: freeing free page 0x%x\n", page);
	  else
	    mem_map[i]--;
	  break;

	case PG_SHARE:
	  if (!mem_map[i])
	    printk ("set_map: page 0x%x not used (wanna be shared...)\n",
		    page);
	  if (mem_map[i] == 65535)
	    printk ("set_map: fatal, page 0x%x shared too many times\n",
		    page);
	  else
	    mem_map[i]++;
	  break;
	}
    }
}

/* used for linking argument page */
void
link_page (int src_pg, int dst_pg)
{
  int table, entry;
  u32 *pt;

  table = PAGE2TABLE (dst_pg);
  entry = PAGE2ENTRY (dst_pg);

  pt = PT (table);
  if (pt == NULL)
    {
      printk ("link_page: table %d NULL (src %d, dst %d)\n",
	      table, src_pg, dst_pg);
      return;
    }

  pt[entry] = TOADDR (src_pg) | PGS_USER | PGS_WRITEABLE | PGS_PRESENT;

  invalidate ();
}

/* 
   opossite to free_pages, it maps pages to desired location 
   used by mmap
 */
int
map_pages (int start_page, int count)
{
  u32 *pt;
  int table, entry, tmp_count;

#ifdef BRKDEBUG
  printk ("map_pages: start_page=0x%p  count=%d\n", start_page, count);
#endif

  if (count <= 0 || start_page < nr_pages)
    {
      printk ("map_pages: error, count=%d  start_page=0x%x\n",
	      count, start_page);
      return -EINVAL;
    }

  table = PAGE2TABLE (start_page);
  entry = PAGE2ENTRY (start_page);

  tmp_count = count;

/* Do not use count here, save it... */
  while (tmp_count > 0)
    {
      pt = PT (table);
      if (!pt)
	{
#ifdef MMDEBUG
	  printk ("map_pages: table 0x%x does not exist, ok\n", table);
#endif
	  if (tmp_count > 1024)
	    tmp_count -= 1024;
	  else
	    tmp_count = 0;	/* don't like negative values */

	  table++;
	  continue;
	}
      while (tmp_count > 0 && entry < 1024)
	{
	  if (pt[entry])
	    {
	      printk
		("map_pages: something already here\n"
		 "(table=0x%x, entry=0x%x, value=0x%p)!\n",
		 table, entry, pt[entry]);
	      return -1;
	    }
	  entry++;
	  tmp_count--;
	}
      entry = 0;
      table++;
    }

/* ok, all is free, check passed */
  table = PAGE2TABLE (start_page);
  entry = PAGE2ENTRY (start_page);

  for (;;)
    {
      if (!PDE (table))
	{
/* missing table, oughta create it */
#ifdef MMDEBUG
	  printk ("map_pages: creating missing table 0x%x\n", table);
#endif
	  pt = (u32 *) page_alloc ();
	  bzero (pt, PAGE_SIZE);
	  PDE (table) = (u32) pt | PGS_USER | PGS_WRITEABLE | PGS_PRESENT;
	  invalidate ();
	}
      else
	pt = PT (table);

      while (entry < 1024)
	{
	  pt[entry] = page_alloc () | PGS_USER | PGS_WRITEABLE | PGS_PRESENT;
	  entry++;
	  count--;
	  if (!count)
	    {
	      invalidate ();
	      return 0;
	    }
	}
      entry = 0;
      table++;
    }
}

/* Zeromaps count of pages, starting at v_page.
   This can be only used in user's area.
   invalidate() not needed. */
void
zeromap_pages (int v_page, int count)
{
  int table, entry;
  u32 *pt;

  if (v_page < USER_AREA)
    {
      printf ("zeromap_pages: invalid starting page given (0x%x)\n", v_page);
      return;
    }

  table = PAGE2TABLE (v_page);
  entry = PAGE2ENTRY (v_page);

  for (;;)
    {
      if (!(pt = PT (table)))
	{
	  printf ("zeromap_pages: pt 0x%x does not exist\n", table);
	  return;
	}

      while (entry < 1024)
	{
	  pt[entry++] = 0;	/* free status for user */
	  count--;
	  if (!count)
	    {
	      invalidate ();
	      return;
	    }
	}
      entry = 0;
      table++;
    }
}

/* mark pages readonly, make this more universal... */
void
ro_pages (int v_page, int count)
{
  int table, entry;
  u32 *pt;

  if (v_page < USER_AREA)
    {
      printf ("ro_pages: invalid starting page given (0x%x)\n", v_page);
      return;
    }

  table = PAGE2TABLE (v_page);
  entry = PAGE2ENTRY (v_page);

  for (;;)
    {
      if (!(pt = PT (table)))
	{
	  printf ("ro_pages: fatal, pt 0x%x does not exist\n", table);
	  return;
	}

      while (entry < 1024)
	{
	  pt[entry++] &= ~PGS_WRITEABLE;	/* free status for user */
	  count--;
	  if (!count)
	    {
	      invalidate ();
	      return;
	    }
	}
      entry = 0;
      table++;
    }
}

/* Get continuous line of pages, return address of first of them.
   Continuous means both virtual and physical pages (1:1) mapping.
   Later we could use space behind 1:1 map if unable to find 1:1
   long enough page range. Both kernel and user environment should
   not use this for more than one page. In boot time is all memory free
   and 'fragmentation' is null so it does not matter at all...
   (I think only /dev/tty.c requests 2 or more pages) */
/* FIXME: add nr_free_dma_pages decreasing... */
int
__get_free_pages (int count, int flags)
{
  CODE_LOCK (a);
  u32 *pt;
  int highest_page, free_pages;
  int i, line = 0, nr = 0;

  if (count <= 0)
    return 0;

  SINGLE_CODE (a);

  if (flags == PG_DMA)
    {
      highest_page = IS_DMA_ABLE (nr_pages) ? nr_pages : nr_dma_pages;
      free_pages = nr_free_dma_pages;
    }
  else
    {
      highest_page = nr_pages;
      free_pages = nr_free_pages;
    }

  if (count > free_pages)
    {
      MULTI_CODE (a);

      printf ("get_free_pages: oom, count=%d  free_pages=%d\n",
	      count, free_pages);
      return 0;
    }

  for (i = 0; i < highest_page; i++)
    {
      if (!mem_map[i])
	{
	  line++;
	  if (line == count)
	    {
	      nr_free_pages -= count;
	      i -= count - 1;

	      set_map (i, count, PG_USED);

	      MULTI_CODE (a);
/*
   #ifdef MMDEBUG
   printk ("gfp: count=%d  ret=0x%x\n", count, nr);
   #endif
 */
	      return i;
	    }
	}
      else
	line = 0;
    }

  MULTI_CODE (a);

  printk
    ("get_free_pages: strange, free page no more, count=%d  free_pages=%d\n",
     count, free_pages);
  for (;;);			/* FIXME: send signal here... */
}

/* FIXME: add free_dma_pages increasing... */
void
free_pages (int page, int count)
{
  int i;
  u32 *pt;
  u32 flg;
  int table, entry;
  int low, high;

  get_flags (flg);
  cli ();

  if (page < USER_AREA)		/* kernel area */
    {
      set_flags (flg);

#ifdef MMDEBUG
      printk ("free_pages: ka, page=0x%x  count=%d\n", page, count);
#endif
      set_map (page, count, PG_FREE);
      return;
    }

/* user area */
#ifdef MMDEBUG
  printk ("free_pages: ua, page=0x%x  count=%d\n", page, count);
#endif

  table = PAGE2TABLE (page);
  entry = PAGE2ENTRY (page);

  for (;;)
    {
      pt = PT (table);
      if (!pt)
	printk ("free_pages: pid=%d, table 0x%x not defined\n",
		ctask->pid, table);

      low = entry;
      while (count && entry < 1024)
	{
	  set_map (TOPAGE (pt[entry]), 1, PG_FREE);
	  pt[entry] = 0;	/* unlink */
	  entry++;
	  count--;
	}

      high = entry;
/* free table check */
      for (entry = 0; entry < low; entry++)
	if (pt[entry] & ADDR_MASK)
	  break;

      if (entry == low)		/* scan high area, too */
	{
	  for (entry = high; entry < 1024; entry++)
	    if (pt[entry] & ADDR_MASK)
	      break;

	  if (entry == 1024)
	    {
	      printk ("free_tables: table 0x%x found free, chopping...\n",
		      table);
	      set_map (TOPAGE (PDE (table)), 1, PG_FREE);
	      PDE (table) = 0;
	    }
	}

      if (!count)
	{
	  set_flags (flg);
	  return;
	}

      table++;
      entry = 0;
    }
}

int
cow_page (int page)
{
  int tbl, ent, i;
  u32 *pt, *pg;

  tbl = PAGE2TABLE (page);
  ent = PAGE2ENTRY (page);
  pt = PT (tbl);

  if (S_ISCOW (pt[ent]))
    {
#ifdef COWDEBUG
      printk ("  pid %d  clonning cowish page 0x%x... (0x%p)\n",
	      ctask->pid, page, pt[ent]);
#endif

      if (mem_map[TOPAGE (pt[ent])] == 1)
	{
#ifdef COWDEBUG
	  printk ("  no need to copy, uncowing only...\n");
#endif
	  pt[ent] =
	    (pt[ent] & ADDR_MASK) | PGS_PRESENT | PGS_USER | PGS_WRITEABLE;
	}
      else
	{
	  if (!mem_map[TOPAGE (pt[ent])])
	    {
	      printk ("cow_page: will not cow free page! page=0x%x\n", page);
	      return -1;
	    }
	  pg = (u32 *) page_alloc ();
	  copy_page (pg, TOADDR (page));
	  pt[ent] = (u32) pg | PGS_PRESENT | PGS_USER | PGS_WRITEABLE;
	}

      invalidate ();
      return 0;
    }

  printk ("  ent 0x%x is not cow, too bad.\nSystem halted.\n", ent);
  return -1;
}

void
page_fault (struct regdump *regs, int dumb)
{
  u32 addr;
  int page;
__asm__ ("movl %%cr2,%0":"=r" (addr));
  page = TOPAGE (addr);

  if (!page)
    {
      cli();
     
      printk ("\nZERO page hit at 0x%x\n", addr);
      printregdump (NULL, regs);

      send_sig (SIGKILL, ctask);
    }

  if (page < 0x100)		/* kernel area */
    {
      cli ();			/* keep this message complete... */

      printk ("\nUnable to handle paging request in kernel area\n" \
	      "address 0x%x  page 0x%x\n", addr, page);
      printregdump (NULL, regs);

      send_sig (SIGKILL, ctask);
    }

  if (regs->error_code & 1)
    {
      if (!cow_page (page))
	return;

      cli ();

/* it was not a cowish page, possible code lock trap..., kill the task */
      printk ("\nRead only page hit, address 0x%x  page 0x%x\n", addr, page);

      printregdump (NULL, regs);

      send_sig (SIGKILL, ctask);
    }
  else
    {
      cli ();

      printk ("\nPage not alive... 0x%p\n", addr);
      printregdump (NULL, regs);

      send_sig (SIGKILL, ctask);
    }
}

void *
virt2phys (void *addr)
{
  return addr;			/* asuming kernel area only, (kernel/dma.c) */
}

void
mm_init ()
{
  u32 *pt;
  u32 flg;
  int i, j, nr_tables, mmap_pages, first_meg_gap, from;
  int low_mem_start, high_mem_start;

  setattr (ATTR (GREEN, BLACK));
  printf ("Ultra Memory Manager\n");
  setattr (DEFAULT_ATTR);

  nr_pages = (sysinfo.memory + 1024) >> 2;
  if (IS_DMA_ABLE (nr_pages))
    {				/* low memory system */
      nr_dma_pages = nr_pages;
      non_dma_start = 0;
    }
  else
    {
      nr_dma_pages = DMA_ABLE_PAGES;
      non_dma_start = DMA_ABLE_PAGES;
    }

/* 1st 4 megs are mapped by head.S:
   all except kernel_pages and 4 blocked entries (0.,1.,2.,3.)
   as NULL trap, root_page_dir, and first page table
   and table page
   note: NULL trap page is not physically used so it
   is marked as free */

  first_meg_gap = 0xA0 - sysinfo.kernel_pages - FIRST_KERNEL_PAGE
    - NR_KERNEL_STACK_PAGES;

/* 96 pages for vga (to first meg), 3 pages for NULL trap, pd and 1. pt */
  mmap_pages = TOPAGE((nr_pages << 1) + PAGE_SIZE - 1);

  if (mmap_pages < first_meg_gap)
    {
      mem_map = (u16 *) TOADDR (sysinfo.kernel_pages + FIRST_KERNEL_PAGE);
      first_meg_gap -= mmap_pages;
      low_mem_start = sysinfo.kernel_pages + FIRST_KERNEL_PAGE + mmap_pages;
      high_mem_start = 0x100;
    }
  else
    {
      mem_map = (u16 *) TOADDR (0x100);
      low_mem_start = sysinfo.kernel_pages + FIRST_KERNEL_PAGE;
      high_mem_start = 0x100 + mmap_pages;
    }

  printk ("  mem_map at 0x%x, mmap_pages=%d\n", mem_map, mmap_pages);

  bzero (mem_map, mmap_pages << PAGE_BITS);

/* protect mem_map! */
  set_map (TOPAGE ((u32) mem_map), mmap_pages, PG_USED);

  nr_free_pages = nr_pages - sysinfo.kernel_pages - 3 - 96 - mmap_pages
    - NR_KERNEL_STACK_PAGES;

  nr_free_dma_pages = nr_dma_pages - sysinfo.kernel_pages
    - 3 - 96 - mmap_pages - NR_KERNEL_STACK_PAGES;

/* Unmap zero page for trapping purposes. */
  pt = PT (0);
  pt[0] = 0;
  mem_map[0] = 1;
/* Block 1. and 2. page -- pd & 1. pt */
  mem_map[1] = 1;
  mem_map[2] = 1;
/* Lock zero task root stack. */
  set_map (FIRST_KERNEL_PAGE - NR_KERNEL_STACK_PAGES,
	   NR_KERNEL_STACK_PAGES, PG_USED);
/* Lock whole kernel. */
  set_map (FIRST_KERNEL_PAGE, sysinfo.kernel_pages, PG_USED);
/* Lock ramdisk if already loaded. Unfortunately it will most probably reside
in DMA-able memory. */
  if(sysinfo.ramdisk_base)
    set_map (TOADDR(sysinfo.ramdisk_base), 
    (sysinfo.ramdisk_blocks + 7) >> 3, PG_USED);

/* Lock VGA - 96 (0x60) pages. */
  set_map (0xA0, 0x60, PG_USED);
/* Map all physical memory one to one. (Without first pt.) */
/* Reserve memory for tables, start at first meg. (Without first pt.) */
/* We use this condition for low level systems. Phantom should be able
   to run even on one meg machines. */
/* FIXME: Machine should not own too much mem. Following things we do
   should get under 4. meg. So machines with more than 3 gigs could crash
   here :) */
  nr_tables = ((nr_pages + 1023) >> 10) - 1;
  from = (nr_tables > first_meg_gap) ? high_mem_start : low_mem_start;

  set_map (from, nr_tables, PG_USED);

  for (i = 0; i < nr_tables; i++)
    {
      pt = (u32 *) TOADDR (from + i);
      PDE (i + 1) = (u32) pt | PGS_PRESENT;
      for (j = 0; j < 1024; j++)
	pt[j] = TOADDR (TE2PAGE (i + 1, j)) | PGS_PRESENT;
    }

  invalidate ();

  printf ("  %d kB memory  %d pages  %d free pages  %d dma pages\n"
          "  %d free dma pages  %d kB occupied by kernel\n",
	  nr_pages << 2,
	  nr_pages, nr_free_pages, nr_dma_pages,
	  nr_free_dma_pages, sysinfo.kernel_pages << 3);

  printf("  kernel: %d kB text  %d kB bss\n", sysinfo.kernel_text_pages,
  sysinfo.kernel_pages - sysinfo.kernel_text_pages);
  
  if(sysinfo.ramdisk_base)
    printf("%d KB occupied by ramdisk\n", sysinfo.ramdisk_blocks >> 1);
}
