/*
   mm/task.c
   Task's specific mm operations and stuff around...

   TODO: dynamic stack...
 */

#include <sys/types.h>
#include <phantom/mm.h>
#include <phantom/sched.h>
#include <asm/system.h>

/* free all task structures (except task struct itself in some cases...) */
void
task_free (TASK * task)
{
  int table, entry, freed = 0;
  u32 *pt;

#ifdef TASKMMDEBUG
  printk ("task_free: pid=%d\n", task->pid);
#endif

/* cleaning all the user area memory */
  for (table = 0x100; table < 0x400; table++)
    {
      pt = CPT (task, table);
      if (pt == NULL)
	continue;

      for (entry = 0; entry < 1024; entry++)
	{
	  if (!pt[entry])
	    continue;

	  set_map (TOPAGE (pt[entry]), 1, PG_FREE);
	  freed++;
	}

      set_map (TOPAGE ((u32) pt), 1, PG_FREE);
      CPDE (task, table) = 0;
    }

  set_map (task->kernel_stack_page - NR_KERNEL_STACK_PAGES,
	   NR_KERNEL_STACK_PAGES, PG_FREE);

  freed += NR_KERNEL_STACK_PAGES;
  page_free (TOPAGE ((u32) task->tss.cr3));

#ifdef TASKMMDEBUG
  printk ("task_free: %d pages freed\n", freed);
#endif
/* page dir and struct task will be freed by sys_exit & co. */
}

/* kill everything but stack */
void 
ctask_cleanup ()
{
  int table, entry, freed = 0;
  u32 *pt;

#ifdef TASKMMDEBUG
  printk ("ctask_cleanup: pid=%d\n", ctask->pid);
#endif

/* cleaning all the user area memory */
  for (table = 0x100; table < 0x300; table++)
    {
      pt = PT (table);
      if (pt == NULL)
	continue;

      for (entry = 0; entry < 1024; entry++)
	{
	  if (!pt[entry])
	    continue;

	  set_map (TOPAGE (pt[entry]), 1, PG_FREE);
	  freed++;
	}

      set_map (TOPAGE ((u32) pt), 1, PG_FREE);
      PDE (table) = 0;
    }

#ifdef TASKMMDEBUG
  printk ("ctask_cleanup: %d pages freed\n", freed);
#endif
/* page dir and struct task will be freed by sys_exit & co. */
}
