#include <phantom/protocol.h>
#include <phantom/time.h>

#define CACHELEN	32
struct arp_cache arp_cache[CACHELEN];



int
arp_cacheupdate (u8 * ip, u8 * hw) 
{
  
int i, free = -1;

  printk ("arp: updating " STR_DUMP_IP " <==> " STR_DUMP_HW "\n",
	  VAR_DUMP_IP (ip), VAR_DUMP_HW (hw));
  
/* FIXME: make this faster.. */
    for (i = 0; i < CACHELEN; i++)
    
    {
      if (!memcmp (ip, arp_cache[i].ip, 4))
	break;
      if (free == -1 && !arp_cache[i].fl_used)
	free = i;
    }

  if (i == CACHELEN)
    {
      if (free == -1)
	{
	  printk ("arp: fatal, free entry not available...\n");
	  return -1;
	}
      i = free;
    }
  
/*
 do the update to the cache
 */ 
    memcpy (arp_cache[i].hw, hw, HWADDR_LEN);
  
memcpy (arp_cache[i].ip, ip, 4);

  arp_cache[i].fl_used = 1;

  
return i;

}

void
arp_request (struct device *dev, u8 * ip) 
{
  struct pkt_head *pkt;
  struct arp_pkt *arp;

  pkt = pkt_alloc (dev, sizeof (struct arp_pkt));
  arp = (struct arp_pkt *) pkt->data;

  pkt_hw_header (pkt, dev->hw_addr, dev->broad_addr, EARP);

  arp->hrd = HTYPE;		/* ethernet */
  arp->pro = ARPPRO;		/* ip protocol */
  arp->hln = HWADDR_LEN;
  arp->pln = IPADDR_LEN;
  
arp->op = ARPREQ;		/* request packet */

  
memcpy (arp->d_hw, dev->broad_addr, HWADDR_LEN);
  
memcpy (arp->d_ip, ip, IPADDR_LEN);
  
memcpy (arp->s_hw, dev->hw_addr, IPADDR_LEN);
  
memcpy (arp->s_ip, dev->ip_addr, IPADDR_LEN);

  pkt_send (pkt);		/* transmit */

}




struct arp_cache *
arp_lookup (struct device *dev, u8 * ip)
{
  int i;
  

for (i = 0; i < CACHELEN; i++)
    
if (!memcmp (arp_cache[i].ip, ip, IPADDR_LEN))
      
return &arp_cache[i];
  
printk ("arp_lookup: no valid entry for " STR_DUMP_IP
	   ", sending arp request\n", VAR_DUMP_IP (ip));

  arp_request (dev, ip);

  sys_sleep (500);		/* half sec. */

  
for (i = 0; i < CACHELEN; i++)
    
if (!memcmp (arp_cache[i].ip, ip, IPADDR_LEN))
      
return &arp_cache[i];

  
printk ("arp_lookup: no valid entry after request...\n");
  
return NULL;

}

/* destination hw stored in pkt (s_ip) */
void
arp_reply (struct pkt_head *pkt)
{
  struct device *dev = pkt->dev;
  struct arp_pkt *arp;

  arp = (struct arp_pkt *) pkt->data;

  arp->hrd = HTYPE;		/* ethernet */
  arp->pro = ARPPRO;		/* ip protocol */
  arp->hln = HWADDR_LEN;
  arp->pln = IPADDR_LEN;
  
arp->op = ARPREP;		/* replying */

  pkt_hw_header (pkt, dev->hw_addr, arp->s_hw, EARP);

  

memcpy (arp->d_hw, arp->s_hw, HWADDR_LEN);
  memcpy (arp->s_hw, dev->hw_addr, HWADDR_LEN);

  
memcpy (arp->d_ip, arp->s_ip, 4);
  
memcpy (arp->s_ip, dev->ip_addr, 4);

  pkt_send (pkt);		/* enqueue pkt on pkt_tx_queue */

}

void
arp_interpret (struct pkt_head *pkt) 
{
  struct arp_pkt *p = (struct arp_pkt *) pkt->data;
  struct device *dev = pkt->dev;

  
printk ("arp: op=0x%x, src="
	   STR_DUMP_HW "  dst=" STR_DUMP_HW "\n"
	   "src=" STR_DUMP_IP "  dst=" STR_DUMP_IP "\n",
	   p->op,
	   VAR_DUMP_HW (p->s_hw), VAR_DUMP_HW (p->d_hw),
	   VAR_DUMP_IP (p->s_ip), VAR_DUMP_IP (p->d_ip));


  if (memcmp (p->d_ip, dev->ip_addr, IPADDR_LEN))
    {
      pkt_free (pkt);
      return;
    }

/* arp request for me */
  
if (p->op == ARPREQ)
    {
      printk ("arp: sending reply\n");
      
arp_cacheupdate (p->s_ip, p->s_hw);	/* keep her address for me */
      
arp_reply (pkt);
/* no packet freeing here */
      return;
    
}
  
/* rarp response */
    
if (p->op == RARPR)
    {
      printk ("rarp response, ip %d.%d.%d.%d obtained\n",
	      p->d_ip[0], p->d_ip[1], p->d_ip[2], p->d_ip[3]);
      memcpy (dev->ip_addr, p->d_ip, IPADDR_LEN);
      pkt_free (pkt);
      
return;
    
}
  
/*
 Check for a reply that I probably asked for.
 */ 
    if (p->op == ARPREP && p->hrd == HTYPE &&
	p->hln == HWADDR_LEN && p->pln == 4)
    {
      
printk ("arp: reply i asked for\n");
      
arp_cacheupdate (p->s_ip, p->s_hw);
      pkt_free (pkt);
      
return;
    }

  
printk ("arp: invalid packet");

  pkt_free (pkt);
}






