#include <sys/types.h>
#include <phantom/netdev.h>
#include <phantom/protocol.h>

u16 ip_chksum (u16 * buff, int len)
{
  int i;
  u32 sum = 0;

  if (len & 1)
    {
      ((char *) buff)[len] = 0;
      len++;
    }

  len >>= 1;

  for (i = 0; i < len; i++)
    sum += buff[i];

  sum = (sum >> 16) + (sum & 0xffff);	/* add in carry */
  sum += sum >> 16;		/* maybe one more */

  return ~sum;
}

int
ip_send (struct pkt_head *pkt, u8 * ip_dst)
{
  struct iph *i = (struct iph *) pkt->data;
  struct device *dev = pkt->dev;
  char ip_src[4];
  int ret;

  printk ("ip_send: sending to " STR_DUMP_IP "\n", VAR_DUMP_IP (ip_dst));

  i->ihl = 5;			/* 5 << 2 = 20 */
  i->version = 4;
  i->id = 0;
  i->frags = 0;
  i->ttl = 128;

  if (ret = route_find (pkt, ip_src, ip_dst))
    return ret;

  memcpy (i->dst, ip_dst, 4);
  memcpy (i->src, ip_src, 4);

  i->tot_len = wswap (pkt->data_len);

  i->check = 0;
  i->check = ip_chksum ((u16 *) i, 20);

  pkt_send (pkt);
}

void
icmp_interpret (struct pkt_head *pkt)
{
  int len;
  struct iph *iph = (struct iph *) pkt->data;
  struct icmp *i = (struct icmp *) iph->data;

  switch (i->type)
    {
    case ICMP_ECHO:
      len = wswap (iph->tot_len) - sizeof (struct iph) - sizeof (struct icmp);

      if (len < 1)
	len = 1;
      if (len > 128)
	len = 128;

      pkt->data_len = sizeof (struct iph) + sizeof (struct icmp) + len;

      i->code = 0;
      i->type = ICMP_ECHOREPLY;	/* echo reply type */
      i->check = 0;
      i->check = ip_chksum ((u16 *) i, len);

      ip_send (pkt, iph->src);
      break;

    default:
      printf ("icmp: unknown packet type 0x%x\n", i->type);
      pkt_free (pkt);
      break;
    }
}

/* TODO: add ip checksumming */
void
ip_interpret (struct pkt_head *pkt)
{
  struct iph *iph = (struct iph *) pkt->data;
  struct device *dev = pkt->dev;

  if (memcmp (dev->ip_addr, &iph->dst, IPADDR_LEN))
    {
      pkt_free (pkt);
      return;
    }

  switch (iph->protocol)
    {
    case PROTICMP:
      icmp_interpret (pkt);
      break;

    case PROTUDP:
      udp_interpret (pkt, (struct iph *)pkt,
      (struct udph *)((void *)pkt + sizeof(struct iph)));
      break;

/*    case PROTTCP:
       tcp_interpret (pkt);
       break;
       */

    default:
      pkt_free (pkt);		/* drop unknown packets */
      break;
    }
}
