#include <sys/types.h>
#include <asm/system.h>
#include <phantom/netdev.h>
#include <phantom/protocol.h>
#include <conio.h>
#include <phantom/ext.h>
#include <phantom/mm.h>

/* TODO: make counters of packets in queues */

struct pkt_head *pkt_rx_queue = NULL;	/* incoming packet queue */
struct pkt_head *pkt_rx_tail = NULL;	/* last member of pkt_queue */
struct pkt_head *pkt_tx_queue = NULL;	/* outcoming packet queue */
struct pkt_head *pkt_tx_tail = NULL;	/* last member of tx queue */

/* remove packet from a pkt queue */
void
pkt_dequeue (struct pkt_head *pkt)
{
  u32 flg;

  if (!pkt->l_head || !pkt->l_tail)
    {
      printk ("pkt_dequeue: improperly queued packed\n");
      return;
    }

  get_flags (flg);
  cli ();

  if (pkt->next)
    pkt->next->prev = pkt->prev;

  if (pkt->prev)
    pkt->prev->next = pkt->next;

  if (pkt == *pkt->l_head)
    *pkt->l_head = pkt->next;

  if (pkt == *pkt->l_tail)
    *pkt->l_tail = pkt->prev;

  pkt->l_head = NULL;
  pkt->l_tail = NULL;
  pkt->next = NULL;		/* just to be sure... */
  pkt->prev = NULL;

  set_flags (flg);
}

/* add packet to a pkt queue */
void
pkt_enqueue (struct pkt_head *pkt, struct pkt_head **head,
	     struct pkt_head **tail)
{
  u32 flg;

  if (!head || !tail)
    {
      printk ("pkt_enqueue: can't link, invalid queue\n");
      return;
    }

  get_flags (flg);
  cli ();

  if (*tail)
    {
      (*tail)->next = pkt;
      pkt->prev = *tail;
      pkt->next = NULL;

      (*tail) = pkt;
    }
  else
    {
      pkt->next = pkt->prev = NULL;
      *head = pkt;
      *tail = pkt;
    }

  pkt->l_head = head;
  pkt->l_tail = tail;

  set_flags (flg);
}

/* called whenever any packet has been received by net driver */
void
pkt_rx (struct pkt_head *pkt)
{
  pkt_enqueue (pkt, &pkt_rx_queue, &pkt_rx_tail);
  ext_run (PACKET_EXT);
}

/* called as interrupt 'extension' -- only one copy running */
void
process_pkt_queue ()
{
  struct eth *eth;
  struct pkt_head *pkt;

  while (pkt_rx_queue || pkt_tx_queue)
    {
/* something for tx? */
      if (pkt_tx_queue)
	{
	  pkt = pkt_tx_queue;
	  pkt->dev->pkt_tx (pkt);	/* this must dequeue the packet */
	  pkt_dequeue (pkt);
	}

/* something for rx? */
      if (pkt_rx_queue)
	{
	  pkt = pkt_rx_queue;
	  pkt_dequeue (pkt);
	  eth = (struct eth *) pkt->hw_data;
	  switch (eth->type)
	    {
	    case EARP:		/* arp packet */
	      arp_interpret (pkt);
	      break;
	    case EIP:		/* ip packet */
	      ip_interpret (pkt);
	      break;
	    default:		/* unknown packet */
	      pkt_free (pkt);
	      break;
	    }
	}
    }
}

/* ethernet only now... */
struct pkt_head *
pkt_alloc (struct device *dev, int len)
{
  struct pkt_head *tmp;

  tmp =
    (struct pkt_head *) kmalloc (sizeof (struct pkt_head) + len +
				 sizeof (struct eth), 0);
  if (!tmp)
    {
      printk ("pkt_alloc: failed to gain memory for packet...\n");
      return NULL;
    }

  tmp->dev = dev;
  tmp->data_len = len;
  tmp->data = tmp->hw_data + dev->hw_hdr_len;

  tmp->l_head = tmp->l_tail = NULL;

  tmp->l_head = NULL;
  tmp->l_tail = NULL;

  return tmp;
}

/* only this routine is allowed to free packets! */
void
pkt_free (struct pkt_head *pkt)
{
  if (!pkt)
    {
      printk ("pkt_free: NULL pkt given!\n");
      return;
    }

  if (pkt->l_head || pkt->l_tail)
    {
      printk ("pkt_free: warning, packet queued, dequeueing...\n");
      pkt_dequeue (pkt);
    }

  kfree (pkt);
}

void
pkt_dump (struct pkt_head *pkt)
{
  u8 *b;
  int i, len, ofs = 0;
  struct eth *eth = (struct eth *) pkt->hw_data;

  b = pkt->data;
  len = pkt->data_len;

  printk ("from=" STR_DUMP_HW "  to=" STR_DUMP_HW "  type=0x%04X  len=%d\n",
	  VAR_DUMP_HW (eth->src), VAR_DUMP_HW (eth->dst), eth->type, len);

  b = pkt->data;
  len = pkt->data_len;

  do
    {
      printk
	("%08x  %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x  ",
	 ofs, b[0], b[1], b[2], b[3], b[4], b[5], b[6], b[7], b[8],
	 b[9], b[10], b[11], b[12], b[13], b[14], b[15]);

      for (i = 0; i < 0x10; i++)
	{
	  printk ("\033[T");	/* typechar prefix */
	  putchark (*b++);
	}

      ofs += 0x10;

      putchark ('\n');
      len -= 0x10;
    }
  while (len > 0);

  printk ("\n");

  b = pkt->data;
  len = pkt->data_len;

  do
    {
      printk
	("%08x  %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x  ",
	 ofs, b[0], b[1], b[2], b[3], b[4], b[5], b[6], b[7], b[8],
	 b[9], b[10], b[11], b[12], b[13], b[14], b[15]);

      for (i = 0; i < 0x10; i++)
	{
	  printk ("\033[T");	/* typechar prefix */
	  putchark (*b++);
	}

      ofs += 0x10;

      putchark ('\n');
      len -= 0x10;
    }
  while (len > 0);

  printk ("\n");



  do
    {
      printk
	("%08x  %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x  ",
	 ofs, b[0], b[1], b[2], b[3], b[4], b[5], b[6], b[7], b[8],
	 b[9], b[10], b[11], b[12], b[13], b[14], b[15]);

      for (i = 0; i < 0x10; i++)
	{
	  printk ("\033[T");	/* typechar prefix */
	  putchark (*b++);
	}

      ofs += 0x10;

      putchark ('\n');
      len -= 0x10;
    }
  while (len > 0);

  printk ("\n");
}

/* routine for sending / queueing tx packet */
void
pkt_send (struct pkt_head *pkt)
{
/* virgin packets only (l_head and l_tail must be NULLs) */
  pkt_enqueue (pkt, &pkt_tx_queue, &pkt_tx_tail);
  ext_run (PACKET_EXT);
/* packet is freed by net device driver */
}

/* ethernet header filler */
void
pkt_hw_header (struct pkt_head *pkt, u8 * src, u8 * dst, u16 type)
{
  struct eth *eth = (struct eth *) pkt->hw_data;

  memcpy (eth->src, src, HWADDR_LEN);
  memcpy (eth->dst, dst, HWADDR_LEN);
  eth->type = type;
}

void
net_init ()
{
  printf("Phantom NET Core...\n");

  ext_table[PACKET_EXT] = &process_pkt_queue;
  ext_enable (PACKET_EXT);

  ne_init ();

  proto_init();
}
