#include <phantom/protocol.h>
#include <phantom/socket.h>
#include <sys/types.h>
#include <phantom/lock.h>

#define NR_PROTO	8

struct proto *proto_table[NR_PROTO] = { &udp_proto, NULL, };

struct proto *
get_proto (int domain, int type, int protocol)
{
  int i;

  single_code ();		/* just to be sure.. */

  i = 0;
  while (proto_table[i] && !(proto_table[i]->domain == domain &&
			     proto_table[i]->type == type
			     && proto_table[i]->protocol == protocol))
    i++;


  if (proto_table[i] == NULL)
    {
      printk ("get_proto: protocol not found\n");
      multi_code ();

      return NULL;
    }

  multi_code ();

  return proto_table[i];
}

int
proto_register (struct proto *proto)
{
  int i;

  single_code ();
  for (i = 0; i < NR_PROTO; i++)
    {
      if (proto == proto_table[i] ||
	  (proto_table[i]->domain == proto->domain &&
	   proto_table[i]->type == proto->type &&
	   proto_table[i]->protocol == proto->protocol))
	{
	  printk
	    ("proto_register: proto %d already registered or has duplicate\n",
	     proto->name);
	  multi_code ();
	  return -1;
	}
      else if (proto_table[i] == NULL)
	{
	  proto_table[i] = proto;
	  multi_code ();
	  return 0;
	}
    }

  multi_code ();
  printk ("proto_register: too many registered protocols (NR_PROTO=%d)\n",
	  NR_PROTO);

  return -1;
}

/* unregistering function */
void
proto_free (struct proto *proto)
{
  int i;

  single_code ();
  for (i = 0; i < NR_PROTO; i++)
    if (proto_table[i] == proto)
      break;

  if (i == NR_PROTO)
    {
      printk ("proto_free: proto not registered\n");
      multi_code ();

      return;
    }

  proto_table[i] = NULL;

  multi_code ();
  return;
}

/* typeout built-in protocols */
void
proto_init ()
{
  int count = 0;
  int i;

  for (i = 0; i < NR_PROTO; i++)
    if (proto_table[i] != NULL)
      count++;

  if (count)
    {
      printf ("Implemented protocols: ");

      for (i = 0; i < NR_PROTO; i++)
	if (proto_table[i] != NULL)
	  printf ("%s ", proto_table[i]->name);

      printf ("\n");
    }
}
