#include <sys/types.h>
#include <phantom/route.h>
#include <phantom/lock.h>
#include <phantom/list.h>
#include <phantom/protocol.h>		/* arp_lookup */
#include <errno.h>

struct route
{
  LIST_HEAD;
  struct device *dev;
  u32 nmask;
  u32 ip;
  int fl_gateway;
};

struct route *route_queue = NULL;	/* empty routing */
struct route *route_tail = NULL;

int
route_add (struct device *dev, u8 *nmask)
{
  u32 tmp;
  struct route *rt = NULL;

  single_code ();
  rt = route_queue;
  
  while (rt != NULL)
    {
      if (rt->dev == dev)
	{
	  printk ("route_add: device %s already registered\n", dev->name);
	  multi_code ();
	  return -1;
	}
      rt = (struct route *)rt->next;
    }

  rt = (struct route *) kmalloc (sizeof (struct route), 0);

  rt->dev = dev;
  rt->nmask = *((u32 *)nmask);
  rt->ip = *((u32 *) dev->ip_addr);

  list_add ((LIST *)rt, (LIST **)&route_queue, (LIST **)&route_tail);
  multi_code ();

#ifdef ROUTEDEBUG
  printk ("route_add: device %s added, ip=" STR_DUMP_IP "  nm="
	  STR_DUMP_IP "\n", dev->name,
	  VAR_DUMP_IP (rt->ip), VAR_DUMP_IP (rt->mask));
#endif

  return 0;
}

int
route_del (struct device *dev)
{
  struct route *rt;

  single_code ();
  rt = route_queue;
  while (rt != NULL && rt->dev != dev)
    rt = (struct route *)rt->next;

  if (rt == NULL)
    {
      multi_code ();
      printk ("route_del: device not found in route queue\n");
      return -1;
    }

  list_add ((LIST *)rt, (LIST **)&route_queue, (LIST **)&route_tail);
  multi_code ();

  return 0;
}

/* find optimal device/route here... */
int route_find(struct pkt_head *pkt, u8 *src_ip, u8 *dst_ip)
{
u32 dst;
struct route *rt=route_queue;
struct arp_cache *ac;

  dst=*((u32 *)dst_ip);
  while(rt!=NULL && (dst & rt->nmask)!=rt->nmask)
    rt=(struct route *)rt->next;

  if(rt==NULL)
    {
    printk("route_find: cannot find route for " STR_DUMP_IP "\n",
    VAR_DUMP_IP(dst_ip));
    
/*    return -EHOSTUNREACH;*/
return -EINVAL;
    }

/* FIXME: 'modularize' :) */
  ac = arp_lookup (rt->dev, dst_ip);
  if (!ac)
    {
      printk ("ip_send: arp_lookup fails! can't go on..\n");
      pkt_free (pkt);
      return;
    }
}

