#include <phantom/protocol.h>
#include <phantom/socket.h>
#include <errno.h>
#include <phantom/lock.h>
#include <phantom/fs.h>
#include <phantom/sched.h>

struct socket *socket_queue = NULL;	/* socket 'port numbering' queue */

/* assign/evaluate port */
int
socket_getport (struct socket *skt)
{
  int port, aport;
  struct socket *tmp;
  u32 flg;

  if (port = skt->src.sin_port)
    {
      aport = tmp->src.sin_port;

      printk ("socket_getport: checking port %d\n", port);

      get_flags (flg);
      cli ();

      tmp = socket_queue;
      while (tmp && aport < port)
	tmp = tmp->next;

      if (aport == port)
	{
	  set_flags (flg);
	  printk ("socket_getport: port collision\n");
	  return -EACCES;
	}

/* ok, port accepted */
      skt->prev = tmp;
      skt->next = tmp->next;
      tmp->next = skt;
      if (tmp->next)
	tmp->next->prev = skt;

      set_flags (flg);

      return 0;
    }

  port = 1024;			/* 'allowed' ports */

  get_flags (flg);
  cli ();

  tmp = socket_queue;
  if (!tmp)
    {
      socket_queue = skt;
      skt->src.sin_port = 1024;
      set_flags (flg);
      return 0;
    }

/* find a gap */
  while (tmp->next && tmp->next->src.sin_port - tmp->src.sin_port == 1)
    tmp = tmp->next;

  if (tmp->next)
    {
      printk ("socket_getport: gap found\n");
      skt->prev = tmp;
      skt->next = tmp->next;
      tmp->next = skt;
      tmp->next->prev = skt;
      skt->src.sin_port = tmp->src.sin_port + 1;

      set_flags (flg);
      return 0;
    }

  printk ("socket_getport: tailing\n");
  tmp->next = skt;
  skt->prev = tmp;
  skt->src.sin_port = tmp->src.sin_port + 1;

  set_flags (flg);
  return 0;
}

void
socket_clearport (struct socket *skt)
{
  struct socket *tmp;

  single_code ();
  tmp = socket_queue;
  while (tmp && tmp != skt)
    tmp = tmp->next;

  if (tmp != skt)
    {
      multi_code ();
      printk ("socket_clearport: skt not found!\n");
      return;
    }

  if (tmp == socket_queue)
    {
      socket_queue = tmp->next;
      if (socket_queue)
	socket_queue->prev = NULL;
      multi_code ();
      return;
    }

  if (tmp->prev)
    tmp->prev->next = tmp->next;

  if (tmp->next)
    tmp->next->prev = tmp->prev;

  multi_code ();
}

int
socket_close (struct socket *skt)
{
  socket_clearport (skt);
  kfree (skt);
}

int
sys_socket (int domain, int type, int protocol)
{
  int fd;
  struct proto *proto;
  struct socket *tmp;

#ifdef SOCKETDEBUG
  printk ("sys_socket: d=0x%x t=0x%x p=0x%x\n", domain, type, protocol);
#endif

  proto=get_proto(domain, type, protocol);

  if(proto==NULL)
    {
    printk("sys_socket: protocol not found, domain=%d  type=%d  protocol=%d\n",
      domain, type, protocol);

    return -EINVAL;
    }
        
  tmp = (struct socket *) kmalloc (sizeof (struct socket), 0);
  if (!tmp)
    {
      printk ("sys_socket: free socket no more\n");
      return -EAGAIN;
    }

  fd = get_fd();  
  ctask->filp[fd]->f_socket = tmp;
  
  if(fd==-1)
    {
    kfree(tmp);
    return -EMFILE;
    }
  
  bzero(tmp, sizeof(struct socket));
  tmp->proto = proto;
    
  return fd;
}

int
sys_bind (int sockfd, struct sockaddr *my_addr, socklen_t addr_len)
{
  int ret;
  struct socket *socket;

#ifdef SOCKETDEBUG
  prinkt ("sys_bind: sockfd=%d  myaddr=0x%p\n", sockfd, my_addr);
#endif

  socket = get_socket (sockfd);
  if (socket == NULL)
    return -EBADF;

  memcpy (&socket->src, my_addr, sizeof (socket->src));

  if (ret = socket_getport (socket))
    return ret;

  return 0;
}

int
sys_connect (int s, struct sockaddr *serv_addr, int addrlen)
{
  struct socket *skt;

  skt = get_socket (s);
  if (!skt)
    {
      printk ("sys_connect: invalid socket handle..\n");
      return -EBADF;
    }

/*  skt->dst_port = serv_addr->sin_port;
  memcpy (skt->dst_addr, serv_addr->sin_addr, 4);*/

/* skt->p_ops->connect.... */

  return 0;
}

int
sys_shutdown (int s, int how)
{
#ifdef SOCKETDEBUG
  printk ("sys_shutdown: s=%d  how=%d\n", s, how);
#endif

}

int
socket_read (struct file *file, char *buff, size_t count)
{
  struct socket *skt;
#ifdef SOCKETDEBUG
  printk("socket_read:\n");
#endif

if(file->f_socket == NULL)
  {
  printk("socket_read: file->socket == NULL!\n");
  return -EBADF;
  }  

skt=file->f_socket;

return skt->proto->recv(skt, buff, count, 0);
}

int
socket_write (struct file *file, char *buff, size_t count)
{
struct socket *skt;

if(file->f_socket == NULL)
  {
  printk("socket_read: file->socket == NULL!\n");
  return -EBADF;
  }  

skt=file->f_socket;

return skt->proto->recv(skt, buff, count, 0);
}

int socket_lseek(struct file *file, off_t off, int whence)
{
  return -ESPIPE;
}

struct file_operations socket_file_operations = {
  socket_lseek,			/* lseek */
  socket_read,			/* read */
  socket_write,			/* write */
  NULL,				/* readdir */
  NULL,				/* ioctl */
  NULL,				/* open */
  NULL				/* close */
};
