#include <phantom/protocol.h>
#include <phantom/socket.h>
#include <sys/types.h>

u16
tcp_chksum (struct ptcph *ptcph, u16 * buff, int len)
{
  int i;
  u32 sum = 0;
  u16 *ptr;

  ptr = (u16 *) ptcph;

  for (i = 0; i < 6; i++)	/* sizeof(struct ptcph)=12 */
    sum += ptr[i];

  if (len & 1)
    {
      ((char *) buff)[len] = 0;
      len++;
    }
  len >>= 1;

  for (i = 0; i < len; ++i)
    sum += buff[i];

  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);

  return ~sum;
}

void
tcp_reject (struct tcph *t) 
{
  int i;
  struct ptcph ptcph;

  
if (t->flags & TRESET)	/* don't reject a reset */
    
return;
  

/*
 swap TCP layer portions for sending back
 */ 
    if (t->t.flags & TACK)
    {
      
t->t.seq = t->t.ack;	/* ack becomes next seq # */
      
t->t.ack = 0;		/* ack # is 0 */
      
t->t.flags = TRESET;
    
}
  else
    {
      
t->t.ack = t->t.seq;
      
t->t.seq = 0;
      
if (t->t.flags & TSYN)
	t->t.ack = longswap (longswap (t->t.ack) + 1L);
      
t->t.flags = TRESET | TACK;
    
}
  
i = t->t.source;		/* swap port #'s */
  
t->t.source = t->t.dest;
  
t->t.dest = i;
  
t->t.hlen = 20 << 2;		/* header len */
  
/*
 create pseudo header for checksum
 */ 
    ptcph.zero = 0;
  ptcph.protocol = i->protocol;
  
ptcph.len = intswap (20);
  
memcpy (ptcph.src, i->src, 4);
  
movebytes (ptcph.dst, i->dst, 4);
  
t->t.check = 0;
  
t->t.check = tcp_chksum (&ptcph, (u16 *) t, sizeof (struct tcph));

  i->len = intswap (sizeof (struct iph) + sizeof (struct tcph));
  
i->ttl = 30;

/* FIXME: find route... */
  pkt_send (pkt);

}

void
tcp_interpret (struct pkt_head *pkt) 
{
  struct iph *i = (struct iph *) pkt->data;
  struct tcph *t = (struct tcph *) (pkt->data + sizeof (struct iph));
  struct socket *skt;
  struct ptcph ptcph;
  
int i, myport, hlen, hisport;

  

if (t->check)
    {
      
memcpy (ptcph.src, i->src, 4);
      
memcpy (ptcph.dst, i->dst, 4);
      
ptcph.zero = 0;
      
ptcph.proto = i->protocol;
      
ptcph.len = t->len;

      if (tcp_chksum (&ptcph, (u16 *) t, t->len))
	{
	  
printk ("tcp_interpret: invalid chksum\n");
	  return;
	}
    }

/*
 *  find the port which is associated with the incoming packet
 *  First try open connections, then try listeners
 */ 
    myport = ntohs (t->dst_port);
  
hisport = ntohs (t->src_port);
  
hlen = GET_TCPH_LEN (t->hlen);

  skt = socket_find (hisport, myport);
  if (skt == NULL)
    {
#ifdef TCPDEBUG
      printk ("tcp_interpret: no socket associated with incoming packet\n");
      printk ("  myport=%d  hisport=%d\n", myport, hisport);
#endif
      return;
    }

  if (hisport || (!hisport && (t->flags & TSYN)))
    {
      
switch (skt->state)
	{
	
case SLISTEN:		/* waiting for remote connection */
	  
if (t->flags & TRESET)	/* ignore Resets on a connection in LISTEN state */
	    return;

	  

if (t->flags & TSYN)	/* receive SYN */
	    {
/*
*   remember anything important from the incoming TCP header 
*/ 
		prt->out.size = intswap (p->t.window);	/* credit window */
	      
prt->out.port = intswap (p->t.source);
	      
prt->in.nxt = longswap (p->t.seq) + 1;
	      
/*
*  set the necessary fields in the outgoing TCP packet
*/ 
		prt->tcpout.t.dest = p->t.source;
	      
prt->tcpout.t.ack = longswap (prt->in.nxt);
	      
prt->tcpout.t.flags = TSYN | TACK;
	      
prt->tcpout.t.hlen = 24 << 2;
	      
/*
*  note that the maxmimum segment size is installed by 'netlisten()'
*  hence the header length is 24, not 20
*/ 
/*
*  initialize all of the low-level transmission stuff(IP and lower)
*/ 
/* acknowledge a connection */
		
tcpsend (prt, 4);

	      
skt->state = SSYNR;	/* syn received */
	    
}			/* end if */
	  
break;

	

case SSYNR:		/* syn received */
	  
if (!(t->flags & TACK))
	    {
	      
tcpsend (prt, 4);
	      
break;		/* not the right one */
	    
}			/* end if */

	  
prt->tcpout.t.hlen = 20 << 2;
	  
prt->out.lasttime = n_clicks ();	/* don't need response */
	  
prt->out.nxt++;	/* count SYN as sent */
	  
prt->out.ack = longswap (p->t.ack);	/* starting ACK value */
	  
prt->out.size = intswap (p->t.window);	/* allowed window */
	  
prt->tcpout.t.flags = TACK;	/* starting ACK flag */
	  
prt->state = SEST;	/* drop through to established */
	  
checkmss (prt, p, hlen);	/* see if MSS option is there */
	  
netputevent (CONCLASS, CONOPEN, pnum);
	  
	    /* fall through */ 
	
case SEST:		/* normal data transmission */
	  
/*
*  check and accept a possible piggybacked ack
*/ 
	    ackcheck (prt, p, pnum);

	  
int dlen;
	  
uint32 sq, want;
	  

dlen = tlen - hlen;
	  
/*
 see if we want this packet, or is it a duplicate?
 */ 
	    sq = longswap (pkt->t.seq);
	  

want = prt->in.nxt;
	  
if (sq != want)
	    {			/* we may want it, may not */
	      
if (sq < want && sq + dlen >= want)
		{		/* overlap */
		  
hlen += (int) (want - sq);	/* offset desired */
		  
dlen -= (int) (want - sq);	/* skip this much */
		
}		/* end if */
	      
	      else
		{		/* tough it */
		  
		    /* Insert real cool stream reassembly to fix FTP <here>  RMG */ 
		    prt->out.lasttime = 0L;	/* make the ACK time out */
		  
return (-1);
		
}		/* end else */
	    
}			/* end if */
	  
	  else
	  
if (dlen <= 0)
	    {			/* only an ACK packet */
	      
if (pkt->t.flags & TFIN)
		{		/* fin bit found */
		  
prt->in.nxt++;	/* count the FIN byte */
		  
prt->state = SCWAIT;	/* close-wait */
		  
prt->tcpout.t.ack = longswap (prt->in.nxt);	/* set ACK in packet */
		  
prt->credit = 0;
		  
prt->out.lasttime = 0L;	/* cause ACK to be sent */
		  
netputuev (CONCLASS, CONCLOSE, pnum);
		  
/*
*   At this point, we know that we have received all data that the other
*   side is allowed to send.  Some of that data may still be in the 
*   incoming queue.  As soon as that queue empties, finish off the TCP
*   close sequence.  We are not allowing the user to utilize a half-open
*   connection, but we cannot close before the user has received all of
*   the data from the incoming queue.
*/ 
		    if (!prt->in.contain)
		    {		/* data remaining? */
		      
prt->tcpout.t.flags = TFIN | TACK;
		      
tcpsend (prt, 0);
		      
prt->state = SLAST;
		    
}
		
}

	      
return (0);
	    
}			/* end if */
	  
/*
*  If we have room in the window, update the ACK field values
*/ 
	    if ((prt->in.size) >= (uint) dlen)
	    {
	      
prt->in.nxt += dlen;	/* new ack value */
	      
prt->in.size -= dlen;	/* new window size */
	      
prt->out.lasttime = 0L;	/* force timeout for ACK */
	      
enqueue (&prt->in, pkt->x.data + hlen - 20, dlen);
	      
netputuev (CONCLASS, CONDATA, pnum);	/* tell user about it */
	      
prt->tcpout.t.ack = longswap (prt->in.nxt);
	      
prt->in.lasttime = n_clicks ();
	    
}			/* end if */
	  
	  else
	    {			/* no room in input buffer */
	      
prt->out.lasttime = 0L;	/* re-ack old sequence value */
	    
}			/* end else */
	  
return;

	  
return;

	

case SSYNS:		/* check to see if it ACKS correctly */
	  
	    /* remember that tcpout is pre-set-up */ 
	    if (p->t.flags & TACK)
	    {			/* It is ACKING us */
	      
if ((uint32) longswap (p->t.ack) != (prt->out.nxt))
		{
		  
netposterr (401);
		  
return (1);
		
}		/* end if */
	    
}			/* end if */
	  
if (p->t.flags & TRESET)
	    {
	      
netposterr (507);
	      
prt->state = SCLOSED;
	      
netputuev (CONCLASS, CONCLOSE, pnum);
	      
return (1);
	    
}			/* end if */
	  
if (p->t.flags & TSYN)
	    {			/* need to send ACK */
	      
prt->tcpout.t.flags = TACK;
	      
prt->in.nxt = longswap (p->t.seq) + 1;
	      
prt->tcpout.t.ack = longswap (prt->in.nxt);
	      
prt->out.ack = longswap (p->t.ack);
	      
prt->out.size = intswap (p->t.window);	/* credit window */
	      
prt->out.lasttime = 0L;
	      
if (p->t.flags & TACK)
		{
		  
prt->state = SEST;
		  
netputevent (CONCLASS, CONOPEN, pnum);
		  
checkmss (prt, p, hlen);
		
}		/* end if */
	      
	      else
		
prt->state = SSYNR;	/* syn received */
	    
}			/* end if */
	  
break;

	

case SCWAIT:
	  
ackcheck (prt, p, pnum);
	  
if (!prt->in.contain)
	    {
	      
prt->tcpout.t.flags = TFIN | TACK;
	      
prt->out.lasttime = 0L;
	      
prt->state = SLAST;
	    
}			/* end if */
	  
break;

	

case SLAST:		/* check ack of FIN, or reset to see if we are done */
	  
if ((p->t.flags & TRESET)
	       || ((uint32) longswap (p->t.ack) == (prt->out.nxt + 1)))
	    
prt->state = SCLOSED;
	  
break;

	

case SFW1:		/* waiting for ACK of FIN */
	  
	    /* throw away data */ 
	    prt->in.nxt = longswap (p->t.seq) + tlen - hlen;
	  
if (p->t.flags & TRESET)
	    
prt->state = SCLOSED;
	  
	  else
	  
if ((uint32) longswap (p->t.ack) != (prt->out.nxt + 1))
	    {
	      
if (p->t.flags & TFIN)
		{		/* got FIN, no ACK for mine */
		  
prt->in.nxt++;	/* account for FIN byte */
		  
prt->tcpout.t.ack = longswap (prt->in.nxt);
		  
prt->tcpout.t.flags = TACK;	/* final byte has no FIN flag */
		  
prt->out.lasttime = 0L;	/* cause last ACK to be sent */
		  
prt->state = SCLOSING;
		
}		/* end if */
	      
	      else
		{
		  
prt->tcpout.t.ack = longswap (prt->in.nxt);
		  
prt->tcpout.t.flags = TACK | TFIN;
		  
prt->out.lasttime = 0L;
		
}		/* end else */
	    
}			/* end if */
	  
	  else
	  
if (p->t.flags & TFIN)
	    {			/* ACK and FIN */
	      
prt->in.nxt++;	/* account for his FIN flag */
	      
prt->out.nxt++;	/* account for my FIN */
	      
prt->tcpout.t.ack = longswap (prt->in.nxt);
	      
prt->tcpout.t.flags = TACK;	/* final byte has no FIN flag */
	      
prt->out.lasttime = 0L;	/* cause last ACK to be sent */
	      
prt->state = STWAIT;	/* we are done */
	    
}			/* end if */
	  
	  else
	    {			/* got ACK, no FIN */
	      
prt->out.nxt++;	/* account for my FIN byte */
	      
prt->tcpout.t.flags = TACK;	/* final pkt has no FIN flag */
	      
prt->state = SFW2;
	    
}			/* end else */
	  
break;
	

case SFW2:		/* want FIN */
	  
prt->in.nxt = longswap (p->t.seq) + tlen - hlen;
	  
if (p->t.flags & TRESET)
	    
prt->state = SCLOSED;
	  
	  else
	  
if (p->t.flags & TFIN)
	    {			/* we got FIN */
	      
prt->in.nxt++;	/* count his FIN byte */
	      
prt->tcpout.t.ack = longswap (prt->in.nxt);
	      
prt->out.lasttime = 0L;	/* cause last ACK to be sent */
	      
prt->state = STWAIT;
	    
}			/* end if */
	  
break;

	

case SCLOSING:	/* want ACK of FIN */
	  
if (p->t.flags & TRESET)
	    
prt->state = SCLOSED;
	  
	  else
	  
if (!ackcheck (prt, p, pnum))
	    {
	      
prt->out.nxt++;	/* account for my FIN byte */
	      
prt->state = STWAIT;	/* time-wait state next */
	    
}			/* end if */
	  
break;

	

case STWAIT:		/* ack FIN again? */
	  
if (p->t.flags & TRESET)
	    
prt->state = SCLOSED;
	  
if (p->t.flags & TFIN)	/* only if he wants it */
	    
prt->out.lasttime = 0L;
	  
if (prt->out.lasttime
	       && (prt->out.lasttime + WAITTIME < n_clicks ()))
	    
prt->state = SCLOSED;
	  
break;

	

case SCLOSED:
	  
prt->in.port = prt->out.port = 0;
	  
break;
	
}

      return;
    }

  

tcp_reject (pkt);
  
if (!(t->flags & TSYN))
    {
      printk ("tcp_interpret: invalid port for incoming syn packet\n");
      return;
    }
}



void
checkmss (struct port *prt, TCPKT * p, int hlen) 
{
  
unsigned int i;
  
/*
*  check header for maximum segment size option
*/ 
    if (hlen > 20 && p->x.options[0] == 2 && p->x.options[1] == 4)
    {
      
movebytes ((char *) &i, (char *) &p->x.options[2], 2);	/* swapped value of maxseg */
      
i = intswap (i);
      
if (i < (unsigned int) (prt->sendsize))	/* we have our own limits too */
	
prt->sendsize = i;
    
}				/* end if */

}				/* end checkmss() */



int
tcpsend (struct port *pport, int dlen) 
{
  
struct port *p;
  

p = pport;
  
/*
*  do IP header first
*/ 
    p->tcpout.i.ident = intswap (nnipident++);
  
p->tcpout.i.tlen =
    intswap (sizeof (struct iph) + sizeof (struct tcph) + dlen);
  
p->tcpout.i.check = 0;	/* install checksum */
  
p->tcpout.i.check = ipcheck ((char *) &p->tcpout.i, 10);
  
/*
*  do TCP header
*/ 
    p->tcpout.t.seq = longswap (p->out.nxt);	/* bytes swapped */
  
/*
*  if the port has some credit limit, use it instead of large
*  window buffer.  Generally demanded by hardware limitations.
*/ 
    if ((uint) (p->credit) < (p->in.size))
    
p->tcpout.t.window = intswap (p->credit);
  
  else
    
p->tcpout.t.window = intswap (p->in.size);	/* window size */
  
/*
*  prepare pseudo-header for checksum
*/ 
    p->tcps.tcplen = intswap (dlen + sizeof (TCPLAYER));
  
p->tcpout.t.check = 0;
  
p->tcpout.t.check =
    tcpcheck ((char *) &p->tcps, (char *) &p->tcpout.t,
	      dlen + sizeof (struct tcph)); 
p->out.lasttime = n_clicks ();

  
return (dlayersend
	   ((DLAYER *) & p->tcpout,
	    sizeof (DLAYER) + sizeof (IPLAYER) + sizeof (TCPLAYER) + dlen));

}				/* end tcpsend() */


/***************************************************************************/ 
/*  ackcheck
*   take an incoming packet and see if there is an ACK for the outgoing
*   side.  Use that ACK to dequeue outgoing data.
*/ 
  int
ackcheck (struct port *p, TCPKT * t, int pnum) 
{
  
uint32 ak;
  
int32 rttl;
  
int i;
  

if ((t->t.flags & TRESET) && (t->t.seq == p->tcpout.t.ack))
    {
      
netposterr (405);
      
p->state = SCLOSED;
      
netputuev (CONCLASS, CONCLOSE, pnum);
      
return (1);
    
}				/* end if */
  
if (!(t->t.flags & TACK))	/* check ACK flag */
    
return (1);		/* if no ACK, no go */
  
p->out.size = intswap (t->t.window);	/* allowable transmission size */
  
/*
*  rmqueue any bytes which have been ACKed, update p->out.nxt to the
*  new next seq number for outgoing.  Update send window.
*
*/ 
    ak = longswap (t->t.ack);	/* other side's ACK */
  
/*
*  Need to add code to check for wrap-around of sequence space
*  for ak.  ak - p->out.ack may be affected by sequence wraparound.
*  If you have good, efficient code for this, please send it to me.
*
*  If ak is not increasing (above p->out.nxt) then we should assume
*  that it is a duplicate packet or one of those stupid keepalive
*  packets that 4.2 sends out.
*/ 
    if (ak > p->out.nxt)
    {
      
rmqueue (&p->out, (int) (ak - p->out.ack));	/* take off of queue */
      
p->out.nxt = ak;
      
p->out.ack = ak;
      
/*
*  Check to see if this acked our most recent transmission.  If so, adjust
*  the RTO value to reflect the newly measured RTT.  This formula reduces
*  the RTO value so that it gradually approaches the most recent round
*  trip measurement.  When a packet is retransmitted, this value is
*  doubled (exponential backoff).
*/ 
	rttl = n_clicks () - p->out.lasttime;
      
if (!p->out.contain && rttl < (long) (MAXRTO) && p->rto >= MINRTO)
	{			/* just now emptied queue */
	  
i = (int) (rttl);
	  
i = ((p->rto - MINRTO) * 3 + i + 1) >> 2;	/* smoothing function */
	  
p->rto = i + MINRTO;
	
}
      
if (!p->out.contain && p->out.push)	/* if the queue emptied and push was set, clear push */
	
p->out.push = 0;
      
if (p->out.contain > 0)
	
p->out.lasttime = 0L;	/* forces xmit */
      
return (0);
    
}
  

/* the following line was added by QAK in an attempt to get ftpbin working again. */ 
#ifdef NOT
    if (p->out.size > 0)
    
p->out.lasttime = 0L;	/* forces xmit */
  
#endif /* 
 */
    return (1);

}				/* end ackcheck() */







