#include <phantom/protocol.h>
#include <phantom/socket.h>
#include <phantom/netdev.h>
#include <sys/types.h>
#include <errno.h>
#include <phantom/in.h>

void udp_interpret(struct pkt_head *pkt, struct iph *i, struct udph *u)
{
  int orig_check, check;
  int udpindex;
  struct ptcph ptcph;
  int hisport, myport;

/* first compute the checksum to see if it is a valid packet */
  orig_check = u->check;
  if (orig_check)
    {
      u->check = 0;

      memcpy (ptcph.src, i->src, 4);

      ptcph.zero = 0;
      ptcph.proto = i->protocol;
      ptcph.len = ntohs (pkt->data_len);

      if (orig_check != tcp_chksum (&ptcph, (u16 *)pkt->data, pkt->data_len))
	{
	  printk ("udp_interpret: chksum mismatch\n");
	  return;
	}
    }

  hisport=ntohs(u->src_port);
  myport=ntohs(u->dst_port);

  
}

int
udp_send (struct socket *skt, char *buff, int count, int flags)
{
  struct pkt_head *pkt;
  struct udph *u;
  struct ptcph ptcph;		/* for checksum */
  struct device *dev = skt->dev;

/*#define UDP_MAX		dev->mtu-dev->hw_hdr_len-sizeof(iph)-sizeof(udph)*/
#define UDP_MAX			576

  if (count > UDP_MAX)
    {
      count = UDP_MAX;
      printk ("udp_sent: packet too long, only %d bytes will be sent\n",
	      count);
    }
#undef UDP_MAX

  count += sizeof (struct udph);

  pkt = pkt_alloc (NULL, count);
  if (pkt == NULL)
    {
      printk ("udp_send: cant gain pkt\n");
      return -ENOMEM;
    }

  u = (struct udph *) (pkt->data + sizeof (struct iph));

  memcpy (pkt->data + sizeof (struct iph) + sizeof (struct udph),
	  buff, sizeof (struct udph));

  ptcph.zero = 0;
  ptcph.proto = PROTUDP;
  ptcph.len = ntohs (count);

  memcpy (ptcph.src, dev->ip_addr, 4);
  memcpy (ptcph.dst, skt->dst.sin_addr, 4);

  u->src_port = ntohs (skt->src.sin_port);
  u->dst_port = ntohs (skt->dst.sin_port);
  u->length = ntohs (count);
  u->check = 0;
  u->check = tcp_chksum (&ptcph, (u16 *) u, count);

  ip_send (pkt, skt->dst.sin_addr);
}


struct proto udp_proto = {
  "udp",

  PF_INET,
  SOCK_DGRAM,
  IPPROTO_UDP,

  NULL,				/* interpret function */

/* socket functions */
  NULL,				/* create */
  NULL,				/* release */
  NULL,				/* connect */
  &udp_send,			/* send */
  NULL,			/* recv */
  NULL				/* ioctl */
};
