#include <rg/kernel.h>
#include <rg/event.h>
#include <rg/core.h>
#include <sys/types.h>
#include <phantom/kb.h>
#include <phantom/tty.h>

/* make this blocking */
int
rg_getevent (WINDOW * window, EVENT * event)
{
  int xkey;

#ifdef RGDEBUG
  printk ("rg_getevent: window=0x%x  event=0x%x\n", window, event);
#endif

  if (window == NULL || event == NULL)
    return -1;

  if (window != window_list)
    {
      event->type = EVENT_NONE;
      return 0;
    }

  for (;;)
    {
      xkey = _readkey (rg_tty);
      if (getkey (K_SHIFT) && getkey (K_ALT))
	{
	  switch (TOK (xkey))	/* window movement */
	    {
	    case K_UP:
	      if (!window->posy)
		break;

	      move_window (window, window->posx, window->posy - 1);
	      break;

	    case K_DOWN:
	      if (window->posy + window->sizey >= screen.sizey)
		break;

	      move_window (window, window->posx, window->posy + 1);
	      break;

	    case K_LEFT:
	      if (window->posx <= 0)
		break;

	      move_window (window, window->posx - 1, window->posy);
	      break;

	    case K_RIGHT:
	      if (window->posx + window->sizex >= screen.sizex)
		break;

	      move_window (window, window->posx + 1, window->posy);
	      break;

	    case K_Q:		/* upper left corner */
	      move_window (window, 0, 0);
	      break;

	    case K_W:		/* upper right corner */
	      if (window->sizex > screen.sizex)
		break;

	      move_window (window, screen.sizex - window->sizex, 0);
	      break;

	    case K_A:		/* lower left corner */
	      if (window->sizey > screen.sizey)
		break;

	      move_window (window, 0, screen.sizey - window->sizey);
	      break;

	    case K_S:		/* lower right */
	      if (window->sizex > screen.sizex)
		break;
	      if (window->sizey > screen.sizey)
		break;

	      move_window (window, screen.sizex - window->sizex,
			   screen.sizey - window->sizey);
	      break;

	    case K_R:		/* refresh */
	      refresh_window (window);
	      break;

	    case K_N:		/* switch next window */
	      switch_window (+1);
	      break;

	    case K_P:		/* switch previous window */
	      switch_window (-1);
	      break;
	    }
	}
      else
	{
	  event->key = TOKB (xkey);
	  event->xkey = TOK (xkey);
	  event->type = EVENT_ONKEY;
	  return 0;
	}
    }
}
