/*
   init.c
   Phantom init task.

   1.0: 04062001mo1916   
 */

//#define DEBUG

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>

#define INITTAB		"/etc/init"
#define FILEMAX		32
#define CMDMAX		64
#define ARGMAX		16

#define LINE 		struct line

#define F_RAN		1
#define F_RUNNING	2
#define F_ONCE		4
#define F_RESPAWN	8

LINE
{
  int pid;
  int flags;
  char io_file[FILEMAX];
  char cmd[CMDMAX];		/* command with parameters */
  char *argv[ARGMAX];
  int argc;
  LINE *next;
};

LINE *inittab = NULL;
LINE *tabtail = NULL;

void
add_line (LINE * l)
{
  if (l == NULL)
    return;

  l->next = NULL;

  if (tabtail == NULL)
    {
      inittab = l;
      tabtail = l;
      return;
    }

  tabtail->next = l;
  tabtail = l;
}

int
getline (char *str, int limit, FILE * f)
{
  int c, cnt = 0;

  while ((c = fgetc (f)) != EOF)
    {
      cnt++;
      if (c == '\n')
	break;

      limit--;
      if (!limit)
	break;

      *str++ = c;
    }

  *str = 0;

  return cnt;
}

void
read_inittab ()
{
  LINE *tmp;
  FILE *f;
  int line = 0, i;		/* counting from zero ;) */
  int argc;
  int flags;
  char *argv[ARGMAX];
  char *s, *io_file, buff[128];

  if ((f = fopen (INITTAB, "r")) == NULL)
    {
      printf ("init: cannot open %s\n", INITTAB);
      return;
    }

  while (getline (buff, sizeof (buff), f))
    {
/* skip newline and comment */
      line++;

      if (!*buff || (*buff == '\n') || (*buff == '#') || (*buff == ';'))
	continue;

#ifdef DEBUG
      printf ("%d: '%s'\n", line, buff);
#endif

      s = buff;

/* read flags */
      if (s[1] != ':')
	{
	  printf ("init: syntax error on line %d\n", line);
	  continue;
	}

      flags = 0;
      switch (*s)
	{
	case 'o':		/* run only once */
	  flags = F_ONCE;
	  break;

	case 'r':		/* respawning */
	  flags = F_RESPAWN;
	  break;

	default:
	  printf ("init: syntax error, invalid flags on line %d\n", line);
	  continue;
	}
      s += 2;			/* skip flag and ':' */

/* read io_file */
      io_file = s;		/* save offset */

      i = 0;
      while (*s && *s != ':')
	s++;

      if (!*s)
	{
	  printf ("init: error on line %d, skipping\n", line);
	  continue;
	}

      *s++ = 0;			/* terminate and continue... */

/* parse command and arguments */
      i = 0;
      argc = 0;
      while (*s)
	{
	  while (*s && isspace (*s))	/* trim spaces */
	    s++;

	  argv[argc++] = s;

	  while (*s && !isspace (*s))
	    s++;

	  if (!*s)
	    break;

	  *s++ = 0;

	  if (argc == ARGMAX)
	    {
	      printf ("init: warning, maximal count of arguments exceeded on line %d\n", line);
	      break;
	    }
	}

      argv[argc] = NULL;	/* last one */

      if ((tmp = (LINE *) malloc (sizeof (LINE))) == NULL)
	{
	  printf ("init: out of memory on line %d, giving up...\n", line);
	  fclose (f);
	  return;
	}

#ifdef DEBUG
      printf ("  tmp=0x%x\n", tmp);
#endif

      bzero (tmp, sizeof (LINE));

      tmp->flags = flags;
      strcpy (tmp->io_file, io_file);
      tmp->argc = argc;
      s = tmp->cmd;
      i = 0;
      while (argv[i] != NULL)
	{
#ifdef DEBUG
	  printf ("%d: argv[%d]=%s\n", line, i, argv[i]);
#endif
	  strcpy (s, argv[i]);
	  tmp->argv[i] = s;
	  s += strlen (s) + 1;
	  i++;
	}
      tmp->argv[i] = NULL;
      add_line (tmp);
    }

#ifdef DEBUG
  printf ("final: %d\n", ftell (f));
#endif

  fclose (f);

#ifdef DEBUG
  printf ("read_inittab: all done\n");
#endif
}

void
do_run (LINE * l)
{
  int pid;

  if (l == NULL)
    return;

  if ((l->flags & F_ONCE) && (l->flags & F_RAN))
    return;

  pid = fork ();
  if (pid == -1)
    {
      printf ("init: fork failed, cmd=%s\n", l->cmd);
      return;
    }

  if (!pid)			/* child */
    {
#ifdef DEBUG
  printf ("init: child running %s\n", l->cmd);
#endif

      if (l->io_file[0])
	{
#ifdef DEBUG
	  printf ("using io %s\n", l->io_file);
#endif

	  close (0);
	  close (1);
	  close (2);
	  if (open (l->io_file, O_RDWR))	/* this should be 0 */
	    printf ("init: cannot open io_file\n");
	  else
	    {
	      dup (0);		/* stdout, stderr */
	      dup (0);
	    }
	}

      execve (l->cmd, l->argv);
      printf ("init: failed to exec %s\n", l->cmd);
      exit (1);
    }

  l->pid = pid;
  l->flags |= F_RAN | F_RUNNING;
}

void
do_wait ()
{
  LINE *l;
  int pid;

  pid = wait (NULL);
  if (pid == -1)
    {
      perror("INIT wait");
      return;
    }

  l = inittab;
  while (l != NULL)
    {
      if (l->pid == pid)
	{
	  l->flags &= ~F_RUNNING;
#ifdef DEBUG
	  printf ("init: %s/%s died\n", l->io_file, l->cmd);
#endif
	}
      l = l->next;
    }

  printf ("init: strange, unknown child pid %d\n", pid);
}

void
main ()
{
  LINE *l;
  int respawners;

  printf ("Init 1.0\n");

#ifdef DEBUG
  printf ("sizeof(LINE)=%d\n", sizeof (LINE));
#endif

  read_inittab ();

#ifdef DEBUG
  {
    int j, i;

    j = 0;
    l = inittab;
    while (l != NULL)
      {
	j++;
	printf ("%d.) '%s' ", j, l->io_file);
	i = 0;
	while (l->argv[i] != NULL)
	  printf ("'%s' ", l->argv[i++]);

	putchar ('\n');
	if (i != l->argc)
	  printf ("argument count mismatch i=%d vs. %d\n", i, l->argc);
	l = l->next;
      }
  }
#endif

 for (;;)
   {
  respawners = 0;
  l = inittab;

  while (l != NULL)
    {
      do_run (l);
      if (l->flags & F_RESPAWN)
	respawners++;
      l = l->next;
    }

  if (!respawners)
     {
     printf("INIT: no task to respawn, evrything done.\n");
     for(;;)
       sleep(100);
     }

  do_wait ();
  } 
}
