#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdlib.h>
#include "karzyk.h"

struct panel left_p, right_p;

int
name_cmp (const void *d1, const void *d2)
{
  return strcmp (((struct dirent *) d1)->d_name,
		 ((struct dirent *) d2)->d_name);
}

void
scandir (char *name, struct panel *p)
{
  DIR *d;
  struct dirent *dent;
  int i;

  d = opendir (name);
  if (d == NULL)
    return;

  i = 0;
  while (i < NR_MAXF && (dent = readdir (d)) != NULL)
    {
      if (dent->d_name[0] != '.' || dent->d_name[1])	/* skip dot */
	{
	  memcpy (&p->dent[i], dent, sizeof (struct dirent));
	  i++;
	}
    }

  closedir (d);

  p->count = i;

//  if(i>1)       /* skip dots, they must be first all time... */
//    qsort(p->dent + 1, i - 1, sizeof(struct dirent), name_cmp);
}

char *file_gather (TABLE * t, int row, int col);

FORM form1 = DEFAULT_FORM;
MEMO sbar = { 80, 1,
  "F1Help F2ForEach F3View F4Edit F5Copy F6Move F7MkDir F8Delete F9Menu F10Exit"
};

TABLE left_table = {
  37, 20,
  30, 1,
  &file_gather, 0,
};

TABLE right_table = {
  37, 20,
  30, 1,
  &file_gather, 0,
};

struct panel *
get_panel (TABLE * t)
{
  struct panel *p;

  if (t == &left_table)
    return &left_p;
  if (t == &right_table)
    return &right_p;

  return NULL;
}

char *
file_gather (TABLE * t, int row, int col)
{
  struct panel *p;
  p = get_panel (t);
  if (p == NULL)
    return NULL;
  return p->dent[row].d_name;
};

int
change_dir (TABLE * t, struct panel *p)
{
  if (chdir (p->cur_dir))
    return -1;

  if (chdir (p->dent[t->arow].d_name))
    return -1;

  getcwd (p->cur_dir, PATH_MAX);
  scandir (".", p);
  t->rows = p->count;
  return 0;
}

void
event_handler (CELL * cell, unsigned int event_type, unsigned int arg)
{
  char *args[3];
  int pid;
  struct panel *p;

  switch (event_type)
    {
    case EVENT_ONKEY:
      p = get_panel (cell->u.table);
      if (p == NULL)
	break;

      switch (event.xkey)
	{
	case K_F1:
	  messagebox ("About", "Karzyk Commander 1.0", OK_BUTTON);
	  return;
	case K_F4:		/* edit */
	  args[0] = "q";
	  args[1] = p->dent[cell->u.table->arow].d_name;
	  args[2] = NULL;

	  pid = fork ();
	  if (!pid)
	    execve (args[0], args);

	  if (pid != -1)
	    wait (pid);
	  return;
	case K_F5:		/* copy */
	  copy (cell, cell->u.table, p, 0);
	  return;
	case K_F6:
	  copy (cell, cell->u.table, p, 1);
	  break;
	case K_F7:		/* mkdir */
	  make_dir (cell, cell->u.table, p);
	  break;
	case K_F8:		/* delete */
	  delete_obj (cell, cell->u.table, p);
	  return;
	case K_F10:		/* exit */
	  if (messagebox
	      ("Question", "Exit Karzyk Commander ?",
	       OK_BUTTON | CANCEL_BUTTON) == OK_BUTTON)
	    default_event_handler (NULL, EVENT_CLOSEAPP, 0);
	  return;
	}
      break;

    case EVENT_ONCLICK:
      p = get_panel (cell->u.table);
      if (p != NULL)
	{			/* change directory */
	  if (!change_dir (cell->u.table, p))
	    {
	      default_event_handler (cell, EVENT_RESET, 0);
	      default_event_handler (cell, EVENT_DRAW, 0);
	    }
	}
      break;
    }

  default_event_handler (cell, event_type, arg);
}

int
main ()
{
  CELL *left_cell, *right_cell;

  form1.x = 0;
  form1.y = 0;
  form1.lx = 80;
  form1.ly = 25;
  form1.winframe = WINFRAME_DOUBLE;
  form1.event_handler = event_handler;
  form1.cell_queue =
    new_memo (0, 24, &sbar,
	      (left_cell = new_table (1, 1, &left_table,
				      (right_cell = form1.cell_tail =
				       new_table (40, 1, &right_table,
						  NULL)))));
  form1.caption = "";
  form1.current_cell = form1.cell_queue;

  getcwd (left_p.cur_dir, PATH_MAX);
  getcwd (right_p.cur_dir, PATH_MAX);

  left_p.mate = &right_p;
  right_p.mate = &left_p;
  left_p.table = &left_table;
  right_p.table = &right_table;

  left_p.cell = left_cell;
  right_p.cell = right_cell;

  scandir (".", &left_p);
  scandir (".", &right_p);
  left_table.rows = left_p.count;
  right_table.rows = right_p.count;

  showmodal (&form1);

  return 0;
}
