#include "q.h"

#define DEBUG

LINE *
getenthline (EWIN * ew, int nr)
{
  LINE *tmp = ew->first_line;

  while (nr && tmp != NULL)
    {
      tmp = tmp->next;
      nr--;
    }

  return tmp;
}

void
block_cursor_alignment (EWIN * ew)
{
#ifdef DEBUG
  printf ("--cursor alignment stage--\n");
#endif

/* cursor alignment */
  if (ew->block_y0 >= ew->begy)	/* marked 'down' */
    {
#ifdef DEBUG
      printf ("Marked down\n");
#endif

      /* marked area is visible */
      if (ew->block_y0 - ew->begy < ew->ly)
	{
#ifdef DEBUG
	  printf ("Marked area visible\n");
#endif

	  ew->curx = 0;
	  ew->begx = 0;

	  if (ew->begy + ew->cury >= ew->block_y0)	/* and cursor placed in it */
	    {
#ifdef DEBUG
	      printf ("Cursor placed in marked block\n");
#endif

	      ew->cury = ew->block_y0 - ew->begy;
	    }
	  else if (ew->begy + ew->cury >= ew->block_y1)	/* placed under it */
	    {
#ifdef DEBUG
	      printf ("Cursor place under marked block\n");
#endif

	      ew->cury -= (ew->block_y1 - ew->block_y0);
	    }

	  ew->current_line = getenthline (ew, ew->begy + ew->cury);
	}
    }
  else
    {				/* marked 'up' */
      if (ew->block_y1 >= ew->begy)
	{
/* more than one screen from top of the file */
	  if (ew->block_y0 >= ew->ly)
	    {
	      ew->begy = ew->block_y0 - ew->ly;
	      ew->cury = ew->ly - 1;
	      ew->display_line = getenthline (ew, ew->begx);
	      ew->current_line = getenthline (ew, ew->begx + ew->curx);
	    }
	  else			/* less than one screen... */
	    {
	      ew->begy = 0;
	      ew->cury = MAX (ew->block_y0 - 1, 0);
	      ew->display_line = ew->first_line;
	      ew->current_line = getenthline (ew, ew->cury);
	    }

	  ew->begx = 0;
	  ew->curx = 0;
	}
    }


#ifdef DEBUG
  printf ("--cursor alignment ended--\n");
#endif
}

void
block_remove (EWIN * ew)
{
  int i, todo;
  LINE *tmp, *l;

  if (ew->block_mode == BLOCK_NONE)
    return;

  if ((l = getenthline (ew, ew->block_y0)) == NULL)
    return;

#ifdef DEBUG
  printf ("block_remove: x0=%d x1=%d y0=%d y1=%d\n",
	  ew->block_x0, ew->block_x1, ew->block_y0, ew->block_y1);
#endif

  switch (ew->block_mode)
    {
    case BLOCK_LINE:

      if (ew->block_y0 == ew->block_y1)	/* single-line mark */
	{
#ifdef DEBUG
	  printf ("single line mark\n");
#endif

	  if (!ew->block_x0 && ew->block_x1 == l->length)	/* full line */
	    {
#ifdef DEBUG
	      printf ("removing whole line\n");
#endif
	      if (l == ew->current_line)
		{
		  if (ew->current_line->next != NULL)
		    ew->current_line = ew->current_line->next;
		  else if (ew->current_line->prev != NULL)
		    {
		      if (ew->current_line == ew->display_line)
			{
			  ew->display_line = ew->display_line->prev;
			  ew->begy--;
			  ew->begx = 0;
			  ew->curx = 0;
			}
		      ew->current_line = ew->current_line->prev;
		    }
		}

	      kill_line (ew, l);
	    }
	  else
	    {			/* shorten only */
#ifdef DEBUG
	      printf ("shortening only\n");
#endif

	      for (i = 0; i < l->length - ew->block_x1; i++)
		l->data[ew->block_x0 + i] = l->data[ew->block_x1 + i];

	      l->length -= ew->block_x1 - ew->block_x0;

	      {
		int x = ew->curx + ew->begx, new_x;

		if (x > ew->block_x0)
		  {
		    if (x > ew->block_x1)
		      {
			new_x = x - (ew->block_x1 - ew->block_x0);
		      }
		    else
		      {		/* cursor inside marked area */
			new_x = ew->block_x0;
		      }

		    if (new_x - ew->begx >= ew->lx)
		      {
			ew->begx = new_x - ew->lx;
			ew->curx = ew->lx - 1;
		      }
		    else
		      ew->curx = new_x - ew->begx;
		  }
	      }
	    }

	  ew->block_mode = BLOCK_NONE;
	  ew->block_x0 = 0;
	  ew->block_x1 = 0;
	  ew->block_y0 = 0;
	  ew->block_y1 = 0;

	  return;
	}

#ifdef DEBUG
      printf ("larger block, block_y0 = %d, block_y1 = %d\n",
	      ew->block_y0, ew->block_y1);
#endif

/* larger block */
      if (!ew->block_x0)	/* whole first line ? */
	{
	  tmp = l->next;
	  kill_line (ew, l);
	  l = tmp;
	}
      else
	{			/* cut only part of the first line */
	  l->length -= ew->block_x0;
	  l->data[l->length] = 0;	/* should not be needed */
	  l = l->next;
	}

      i = 1;
      todo = ew->block_y1 - ew->block_y0;
      while (l != NULL && i < todo)
	{
	  tmp = l->next;
	  kill_line (ew, l);
	  l = tmp;
	  i++;
	}

      if (l->length == ew->block_x1)	/* whole last line ? */
	kill_line (ew, l);
      else
	{			/* no, only part of the last line... */
	  memcpy (l->data, l->data + ew->block_x1, l->length - ew->block_x1);
	  l->length -= ew->block_x1;
	}
      break;
    }

  block_cursor_alignment (ew);

  ew->block_mode = BLOCK_NONE;
  ew->block_x0 = 0;
  ew->block_x1 = 0;
  ew->block_y0 = 0;
  ew->block_y1 = 0;
}

void
clip_paste (EWIN * ew)
{
  LINE *l, *c, *k;

#ifdef DEBUG
  printf ("clip_paste:\n");
#endif

  if (ew->clip == NULL)
    {
#ifdef DEBUG
      printf ("clip empty, nothing to paste\n");
#endif
      return;
    }

  if (ew->clip->next == NULL)	/* single line, no newline */
    {
      if (ew->current_line == NULL)	/* empty document.. ? */
	{
	  ew->current_line = add_line (ew, NULL);
	  if (ew->current_line == NULL)
	    return;
	}

      l = ew->current_line;
      c = ew->clip;

      if (c->length + l->length >= MAXLINE)
	{
	  error ("line too long, cannot paste\n");
	  return;
	}

/* make space for clip data */
      memmove (l->data + CURX(ew) + c->length, l->data + CURX(ew),
              l->length - CURX(ew));

      l->length += c->length;

/* copy clip data */
      memcpy (l->data + CURX(ew), c->data, c->length);

    if(c->length)
      ew->fl_dirt = 1;

#ifdef DEBUG
  printf("single line paste done, %d characters added\n", c->length);
#endif

      return;
    }

/* two lines or more */
  if (ew->current_line == NULL)	/* empty, strange... */
    {
      ew->current_line = add_line (ew, NULL);
      if (ew->current_line == NULL)
	return;
    }

  l = ew->current_line;
  c = ew->clip;

  if (ew->begx + ew->curx + c->length >= MAXLINE)
    {
      error ("cannot paste, line too long\n");
      return;
    }

  /* need to split this line */
  if (l->length > ew->begx + ew->curx)
    {
      if ((k = add_line (ew, l)) == NULL)
	return;

      k->length = l->length - CURX (ew);
      memcpy (k->data, l->data + CURX (ew), k->length);

      l->length = CURX (ew);
    }

  if (l->length + c->length > MAXLINE)
    {
      error ("cannot paste, line too long\n");
      return;
    }

  memcpy (&l->data[l->length], c->data, c->length);
  l->length += c->length;

  l = l->next;

  while (c->next != NULL)
    {
      l = add_line (ew, l);
      if (l == NULL)
	return;

      memcpy (l->data, c->data, c->length);
      l->length = c->length;
      c = c->next;
    }

  if (l->next == NULL)		/* no next line, at the end.. */
    {
      l = add_line (ew, l);
      if (l == NULL)
	return;
    }

  if (l->length + c->length >= MAXLINE)
    {
      error ("cannot paste, line too long");
      return;
    }

  if (l->length)		/* shift data */
    memcpy (l->data + c->length, l->data, c->length);

  memcpy (l->data, c->data, c->length);
  l->length += c->length;
}

/* store block to clipboard */
void
clip_copy (EWIN * ew)
{
  int i, todo;
  LINE *tmp, *tmp1, *l;

#ifdef DEBUG
  printf ("clip_copy: started\n");
#endif

  switch (ew->block_mode)
    {
    case BLOCK_LINE:
      clip_flush (ew);

      if ((l = getenthline (ew, ew->block_y0)) == NULL)
	return;

      if ((tmp = new_line ()) == NULL)
	return;

      ew->clip = tmp;

      if (ew->block_y0 == ew->block_y1)	/* single line */
	{
	  tmp->length = ew->block_x1 - ew->block_x0;
	  strncpy (tmp->data, &l->data[ew->block_x0], tmp->length);
	  return;
	}

      i = MIN (l->length - ew->block_x0, MAXLINE);
      tmp->length = i;
      memcpy (tmp->data, l->data + ew->block_x0, i);

      l = l->next;

      todo = ew->block_y1 - ew->block_y0;

      while (l != NULL && todo > 1)
	{
	  tmp1 = new_line ();
	  if (tmp1 == NULL)
	    return;

	  tmp->next = tmp1;
	  tmp1->prev = tmp;

	  i = MAX (MAXLINE, l->length);
	  memcpy (tmp1->data, l->data, i);
	  tmp1->length = l->length;

	  tmp = tmp1;
	  l = l->next;

	  todo--;
	}

/* do the last line */
      if ((tmp = new_line ()) == NULL)
	return;

      i = MIN (MAXLINE, ew->block_x1);
      tmp->length = i;
      memcpy (tmp->data, l->data, i);
      tmp->prev = tmp1;

      break;
    }
}

/* flush the clipboard */
void
clip_flush (EWIN * ew)
{
  kill_lines (ew->clip);
  ew->clip = NULL;
}

int				/* 0 .. do not redraw, 1 .. do redraw */
do_block (EWIN * ew, int origx, int origy)
{
  int dx, dy;
  int absx, absy;

  if (!rg_getkey (K_SHIFT) && !rg_getkey (K_RSHIFT))
    return 0;

  absx = ew->begx + ew->curx;
  absy = ew->begy + ew->cury;

  dy = absy - origy;
  dx = absx - origx;

  if (ew->block_mode == BLOCK_NONE ||
      !((origx == ew->block_x0 && origy == ew->block_y0)
	|| (origx == ew->block_x1 && origy == ew->block_y0)
	|| (origx == ew->block_x0 && origy == ew->block_y1)
	|| (origx == ew->block_x1 && origy == ew->block_y1)))
    {
      ew->block_mode = BLOCK_LINE;

      ew->block_x0 = MIN (origx, absx);
      ew->block_x1 = MAX (origx, absx);
      ew->block_y0 = MIN (origy, absy);
      ew->block_y1 = MAX (origy, absy);
      return 1;
    }

/* x movement */
  if (dx)
    {
      if (origy == ew->block_y0 && origy == ew->block_y1)	/* single line */
	{
	  if (origx == ew->block_x0)
	    {
	      if (origx + dx > ew->block_x1)
		{
		  ew->block_x0 = ew->block_x1;
		  ew->block_x1 = origx + dx;
		}
	      else
		ew->block_x0 += dx;
	    }
	  else if (origx == ew->block_x1)	/* just to be sure... */
	    {
	      if (origx + dx < ew->block_x0)
		{
		  ew->block_x1 = ew->block_x0;
		  ew->block_x0 = origx + dx;
		}
	      else
		ew->block_x1 += dx;
	    }
	}
      else
	{			/* larger mark */
	  if (origy == ew->block_y0)
	    {
	      ew->block_x0 += dx;
	    }
	  else if (origy == ew->block_y1)
	    {
	      ew->block_x1 += dx;
	    }
	}
    }

/* y movement */
  if (dy)
    {
      if (ew->block_y0 == origy && ew->block_y1 == origy)
	{
	  if (dy < 0)
	    ew->block_y0 += dy;
	  else
	    ew->block_y1 += dy;
	}
      else if (ew->block_y0 == origy)
	{
	  if (origy + dy > ew->block_y1)
	    {
	      SWAPIT (ew->block_x0, ew->block_x1);
	      ew->block_y0 = ew->block_y1;
	      ew->block_y1 = origy + dy;
	    }
	  else
	    ew->block_y0 += dy;
	}
      else if (ew->block_y1 == origy)
	{
	  if (origy + dy < ew->block_y0)
	    {
	      SWAPIT (ew->block_x0, ew->block_x1);
	      ew->block_y1 = ew->block_y0;
	      ew->block_y0 = origy + dy;
	    }
	  else
	    ew->block_y1 += dy;
	}

      if (ew->block_x0 == ew->block_x1 && ew->block_y0 == ew->block_y1)
	ew->block_mode = BLOCK_NONE;
    }

  return 1;
}
