#include "q.h"

char info_memo_str[64] = "[----] r,c=0/0,0/0";

MEMO info_memo =
{
  64, 1, info_memo_str
};

CELL *info_cell;

#define BG_ATTR		FBPAIR(WHITE, BLACK)
#define TEXT_ATTR	FBPAIR(WHITE, BLUE)
#define BLOCK_ATTR	FBPAIR(BLACK, GREEN)
#define FAKEBLOCK_ATTR	FBPAIR(BLACK, RED)

void
do_draw (CELL * cell, CANVAS * can, EWIN * ew)
{
  int line = 0;
  int len, i, j;
  LINE *l;
  WINDOW *w = cell->window;

  rg_box (cell->window, cell->x, cell->y, can->lx, can->ly, BG_ATTR);

  l = ew->display_line;
  while (l != NULL && line < can->ly)
    {
      len = l->length - ew->begx;
      if (len < 0)
	len = 0;
      if (len > can->lx)
	len = can->lx;

      switch (ew->block_mode)
	{
	case BLOCK_NONE:
	  if (len)
	    rg_strecho (cell->window, cell->x, cell->y + line,
			TEXT_ATTR, l->data + ew->begx, len);
	  break;

	case BLOCK_LINE:
	  if (!len)
	    break;

	  i = 0;		/* chars actually printed */
	  j = 0;

/* single line, both ends critical */
	  if (ew->begy + line == ew->block_y0 && ew->begy + line == ew->block_y1)
	    {
	      if (ew->block_x0 > ew->begx)
		{
		  i = MIN (ew->block_x0 - ew->begx, len);
		  rg_strecho (cell->window, cell->x, cell->y + line,
			      TEXT_ATTR, l->data + ew->begx, i);
		}

	      if (i < len)	/* ok, now print mark */
		{
		  j = MIN (ew->block_x1 - ew->block_x0, len - i);
		  rg_strecho (cell->window, cell->x + i, cell->y + line,
			      BLOCK_ATTR, l->data + ew->begx + i, j);
		}

	      if (i + j < len)	/* unmarked rest */
		rg_strecho (cell->window, cell->x + i + j, cell->y + line,
			TEXT_ATTR, l->data + ew->begx + i + j, len - i - j);
	    }
	  else if (ew->begy + line == ew->block_y0)	/* upper critical line */
	    {
	      if (ew->block_x0 > ew->begx)
		{
		  i = MIN (ew->block_x0 - ew->begx, len);
		  rg_strecho (cell->window, cell->x, cell->y + line,
			      TEXT_ATTR, l->data + ew->begx, i);
		}

	      if (i < len)	/* ok, now print mark */
		rg_strecho (cell->window, cell->x + i, cell->y + line,
			    BLOCK_ATTR, l->data + ew->begx + i, len - i);
	    }
	  else if (ew->begy + line == ew->block_y1)	/* lower critical line */
	    {
	      if (ew->block_x1 > ew->begx)
		{
		  i = MIN (ew->block_x1 - ew->begx, len);
		  rg_strecho (cell->window, cell->x, cell->y + line,
			      BLOCK_ATTR, l->data + ew->begx, i);
		}

	      if (i < len)
		rg_strecho (cell->window, cell->x + i, cell->y + line,
			    TEXT_ATTR, l->data + ew->begx + i, len - i);
	    }
	  else
	    {			/* full-mark */
	      if (ew->begy + line > ew->block_y0 &&
		  ew->begy + line < ew->block_y1)
		i = BLOCK_ATTR;
	      else
		i = TEXT_ATTR;

	      rg_strecho (cell->window, cell->x, cell->y + line,
			  i, l->data + ew->begx, len);
	    }
	  break;
	}

      line++;
      l = l->next;
    }
  rg_setcursor (w, cell->x + ew->curx, cell->y + ew->cury);
  if (ew->fl_overwrite)
    rg_showbigcursor (w);
  else
    rg_showcursor (w);
}

void
align_cursor (CELL * cell, CANVAS * can, EWIN * ew)
{
  LINE *l = ew->current_line;

  if (ew->begx < l->length)	/* ok, do nothing, only set cursor */
    {
      if (ew->begx + ew->curx > l->length)
	ew->curx = l->length - ew->begx;

      rg_setcursor (cell->window, cell->x + ew->curx, cell->y + ew->cury);
      return;
    }

/* show end */
  if (l->length >= ew->lx)
    {
      ew->curx = ew->lx - 1;
      ew->begx = l->length - ew->lx + 1;
    }
  else
    {
      ew->begx = 0;
      ew->curx = l->length;
    }

  do_draw (cell, can, ew);
}

void
update_line_info (EWIN * ew)
{
  if (ew != NULL)
    sprintf (info_memo.text, "[%c%c%c%c] r,c=%d/%d,%d/%d %d;%d %d;%d      ",
	     ew->block_mode == BLOCK_LINE ? 'l' : '-',	/* block selected */
	     ew->fl_dirt ? 'M' : '-',	/* modified */
	     ew->fl_overwrite ? 'O' : '-',	/* overwrite */
	     ew->clip != NULL ? 'c' : '-',	/* clip */
	     ew->curx + ew->begx,
	     (ew->current_line != NULL) ? ew->current_line->length : 0,
	     ew->cury + ew->begy,
	     ew->lines,
	     ew->block_x0,
	     ew->block_y0,
	     ew->block_x1,
	     ew->block_y1);
  else
    sprintf (info_memo.text, "[----] r,c=0/0,0/0          ");

  default_event_handler (info_cell, EVENT_DRAW, 0);
}
