#include "q.h"

void
kill_line (EWIN * ew, LINE * l)
{
  if (l == ew->first_line)
    ew->first_line = l->next;
  if (l == ew->current_line)
    ew->current_line = l->prev != NULL ? l->prev : l->next;
  if (l == ew->display_line)
    ew->display_line = l->prev != NULL ? l->prev : l->next;
  if (l == ew->last_line)
    ew->last_line = l->prev;

  if (l->prev)
    l->prev->next = l->next;
  if (l->next)
    l->next->prev = l->prev;

  if (l == ew->first_line)
    ew->first_line = l->next;

  if (l == ew->last_line)
    ew->last_line = l->prev;

  if (ew->lines > 0)
    ew->lines--;
}

int
line_merge (EWIN * ew, LINE * first, LINE * second)
{
  if (first->length + second->length >= MAXLINE)
    {
      error ("line too long\n");
      return -1;
    }

  strcat (first->data + first->length, second->data);
  first->length += second->length;

  kill_line (ew, second);
  return 0;
}

LINE *
add_line (EWIN * ew, LINE * line)
{
  LINE *tmp;

  tmp = (LINE *) malloc (sizeof (LINE));
  if (tmp == NULL)
    {
      error ("OOM!\n");
      return;
    }

  tmp->length = 0;
  tmp->data[0] = 0;

  tmp->prev = NULL;
  tmp->next = NULL;

  if (line == NULL)
    {
      if (ew->first_line != NULL)
	{
	  tmp->next = ew->first_line;
	  ew->first_line->prev = tmp;
	}
      else
	{
	  tmp->next = NULL;
	  ew->last_line = tmp;
	  ew->current_line = tmp;
	  ew->display_line = tmp;
	}

      ew->first_line = tmp;
    }
  else
    {
      tmp->next = line->next;
      tmp->prev = line;
      line->next = tmp;
      if (tmp->next == NULL)
	ew->last_line = tmp;
    }

  ew->lines++;

#ifdef DEBUG
  printf ("add_line: tmp=0x%x  fl=0x%x  ll=0x%x\n", tmp, ew->first_line,
	  ew->last_line);
#endif

  return tmp;
}

LINE *
new_line ()
{
  LINE *tmp;

  tmp = (LINE *) malloc (sizeof (LINE));
  if (tmp == NULL)
    {
      error ("new_line: OOM!\n");
      return NULL;
    }

  tmp->length = 0;
  tmp->data[0] = 0;

  tmp->prev = NULL;
  tmp->next = NULL;

  return tmp;
}

void
kill_lines (LINE * l0)
{
  LINE *l1;

  while (l0 != NULL)
    {
      l1 = l0->next;
      free (l0);
      l0 = l1;
    }
}

int
flush_ewin (EWIN * ew)
{
#ifdef DEBUG
  printf ("flush_ewin: ");
#endif

  kill_lines (ew->first_line);
  kill_lines (ew->clip);

  free (ew);
  return 0;
}

void
do_delete (CELL * cell, CANVAS * canvas, EWIN * ew)
{
  LINE *line = ew->current_line;

  ew->block_mode = BLOCK_NONE;

  if (line->length - ew->curx + 1 > 1)
    {
      line = ew->current_line;
      memmove (line->data + ew->curx + ew->begx,
	       line->data + ew->curx + ew->begx + 1,
	       line->length - ew->curx - ew->begx - 1);

      ew->fl_dirt = 1;

      do_draw (cell, canvas, ew);
      rg_wrefresh (cell->window);
    }
  else
    {
      if (!line_merge (ew, line, line->next))
	{
	  ew->fl_dirt = 1;

	  ew->current_line = line;
	  do_draw (cell, canvas, ew);
	  rg_wrefresh (cell->window);
	}
    }
}

void
do_backspace (CELL * cell, CANVAS * canvas, EWIN * ew)
{
  int olen;
  LINE *line = ew->current_line, *tmp;

  ew->block_mode = BLOCK_NONE;

  if (ew->current_line == ew->first_line && ew->curx + ew->begx == 0)
    return;			/* can't touch this */

  if (ew->curx + ew->begx > 0)
    {
      memmove (line->data + ew->curx + ew->begx - 1,
	       line->data + ew->curx + ew->begx,
	       line->length - ew->curx - ew->begx);
      line->length--;

      if (ew->curx)
	ew->curx--;
      else
	ew->begx--;

      ew->fl_dirt = 1;

      do_draw (cell, canvas, ew);
      rg_wrefresh (cell->window);
    }
  else
    {
      tmp = line->prev;
      olen = tmp->length;

      if (!line_merge (ew, tmp, line))
	{
	  ew->current_line = tmp;

	  if (ew->cury)
	    ew->cury--;
	  else
	    ew->begy--;

	  ew->fl_dirt = 1;

	  if (olen > canvas->lx - 1)
	    {
	      ew->curx = canvas->lx - 1;
	      ew->begx = olen - ew->curx;
	    }
	  else
	    ew->curx = olen;

	  do_draw (cell, canvas, ew);
	  rg_wrefresh (cell->window);
	}
    }
}

void
do_newline (CELL * cell, CANVAS * can, EWIN * ew)
{
  LINE *line = ew->current_line;
  LINE *tmp;
  int size;

  ew->block_mode = BLOCK_NONE;

  tmp = add_line (ew, line);
  size = line->length - ew->curx - ew->begx;
  if (size)
    {
      tmp->length = size;
      strcpy (tmp->data, line->data + ew->curx + ew->begx);
      line->data[ew->curx + ew->begx] = 0;
      line->length = ew->curx + ew->begx;
    }

  ew->curx = 0;
  if (ew->cury < ew->ly - 1)
    ew->cury++;
  else
    {
      ew->display_line = ew->display_line->next;
      ew->begy++;
    }

  ew->fl_dirt = 1;
  ew->current_line = tmp;

  do_draw (cell, can, ew);
  update_line_info (ew);

  rg_wrefresh (cell->window);
}

void
insert_char (CELL * cell, CANVAS * canvas, EWIN * ew, int key)
{
  LINE *line = ew->current_line;

  if (line == NULL)		/* should not happen... */
    return;

  ew->block_mode = BLOCK_NONE;

  if (line->length <= MAXLINE - 1)
    {
      if (!ew->fl_overwrite)
	{
	  memmove (line->data + ew->begx + ew->curx + 1,
		   line->data + ew->begx + ew->curx,
		   line->length - ew->curx - ew->begx);
	  line->length++;
	}
      else if (ew->begx + ew->curx == line->length)
	line->length++;

      line->data[ew->curx + ew->begx] = key;
      if (ew->curx < canvas->lx - 1)
	ew->curx++;
      else
	ew->begx++;

      ew->fl_dirt = 1;

      do_draw (cell, canvas, ew);
      update_line_info (ew);
      rg_wrefresh (cell->window);
    }
}
