/*
   Sokoban 1.0, 28012000fr1620

   21112000tu2010: phantom version spawned
*/

#include <string.h>
#include <stdio.h>
/*#include <time.h>*/
#include <unistd.h>
#include <conio.h>
#include <stdlib.h>
/*#include <pc.h>*/

#define LEVEL_FILE		"level.dta"
#define MAP_X		        0x20	/* map ranges */
#define MAP_Y		        0x20

#define COLOR_ZERO              0x00
#define COLOR_TEXT              242
#define MENU_TEXT_COLOR         239

#define CHAR_GROUND		' '
#define CHAR_WALL		'#'
#define CHAR_STORE		'.'
#define CHAR_BOX		'$'
#define CHAR_STORED_BOX		'*'
#define CHAR_MINER		'@'
#define CHAR_STORED_MINER	'!'

#define LMAP(lmap,a,b)          ((lmap)->data[(a)+(b)*MAP_X])
#define HISTORY_SIZE		1000

#define NR_CELL                 5
#define C_BOX                   0x00
#define C_STORED_BOX            0x01
#define C_MINER                 0x02
#define C_STORE                 0x03
#define C_WALL                  0x04

#ifndef ATTR
#define ATTR(f,b)               (((b)<<4)|(f))
#endif

#define A_MINER                 ATTR(CYAN,BLACK)
#define A_STOREDMINER           ATTR(CYAN,BROWN)
#define A_WALL                  ATTR(RED,BLACK)
#define A_BOX                   ATTR(YELLOW,BLACK)
#define A_STOREDBOX             ATTR(YELLOW,BROWN)
#define A_STORE                 ATTR(BLACK,BROWN)
#define A_NOTHING               ATTR(LIGHTGRAY,BLACK)

#define A_DEFAULT               ATTR(LIGHTGRAY,BLACK)
#define A_HILITE                ATTR(LIGHTBLUE,BLACK)

#define S_MINER                 "@@"
#define S_STORE                 "<>"
#define S_WALL                  ""
#define S_BOX                   "[]"
#define S_NOTHING               "  "

#define SCREEN_X       80
#define SCREEN_Y       25

struct lmap_st
{
  int nr;
  int size_x, size_y;
  int miner_x, miner_y;
  int boxes, stored_boxes;
  char *data;
};

struct lmap_st *
new_lmap ()
{
  struct lmap_st *tmp;

  tmp = (struct lmap_st *) malloc (sizeof (struct lmap_st));
  if (!tmp)
    return NULL;

  bzero (tmp, sizeof (*tmp));
  tmp->data = (char *) malloc (MAP_X * MAP_Y);
  if (!tmp->data)
    {
      free (tmp);
      return NULL;
    }
  return tmp;
}

void
kill_lmap (struct lmap_st *l_map)
{
  free (l_map->data);
  free (l_map);
}

void
draw_cell (struct lmap_st *lmap, int x, int y)
{
  int c = LMAP (lmap, x, y);

  x *= 2;
  x += (SCREEN_X - (lmap->size_x << 1)) >> 1;
  y += (SCREEN_Y - lmap->size_y) >> 1;

  gotoxy (x, y);
  textattr (A_NOTHING);
  cprintf (S_NOTHING);
  gotoxy (x, y);

  switch (c)
    {
    case CHAR_WALL:
      textattr (A_WALL);
      cprintf (S_WALL);
      break;
    case CHAR_STORED_MINER:
      textattr (A_STOREDMINER);
      cprintf (S_MINER);
      break;
    case CHAR_MINER:
      textattr (A_MINER);
      cprintf (S_MINER);
      break;
    case CHAR_STORE:
      textattr (A_STORE);
      cprintf (S_STORE);
      break;
    case CHAR_BOX:
      textattr (A_BOX);
      cprintf (S_BOX);
      break;
    case CHAR_STORED_BOX:
      textattr (A_STOREDBOX);
      cprintf (S_BOX);
      break;
    }
  textattr (A_DEFAULT);
}

void
draw_level (struct lmap_st *lmap)
{
  int i, j;

  for (j = 0; j < lmap->size_y; j++)
    for (i = 0; i < lmap->size_x; i++)
      draw_cell (lmap, i, j);
}

#define CMD_NOTHING     0x00
#define CMD_UP		0x01
#define CMD_DOWN	0x02
#define CMD_LEFT	0x03
#define CMD_RIGHT	0x04
#define CMD_PUSH        0x10
#define CMD_PULL        0x20
#define CMD_UNDO        0x40

/* returns CMD of performed action */
int
do_move (struct lmap_st *lmap, int cmd)
{
  int p, p1, p2, x, y, tx, ty;

  x = lmap->miner_x;
  y = lmap->miner_y;

  tx = 0;
  ty = 0;

  switch (cmd & ~(CMD_UNDO | CMD_PUSH | CMD_PULL))
    {
    case CMD_UP:
      ty = -1;
      break;
    case CMD_DOWN:
      ty = 1;
      break;
    case CMD_LEFT:
      tx = -1;
      break;
    case CMD_RIGHT:
      tx = 1;
      break;
    }

  if (!(cmd & CMD_UNDO))
    {
      p = LMAP (lmap, x, y);
      p1 = LMAP (lmap, x + tx, y + ty);
      if (p1 == CHAR_GROUND || p1 == CHAR_STORE)
	{
	  p = LMAP (lmap, x, y) =
	    (p == CHAR_MINER) ? CHAR_GROUND : CHAR_STORE;
	  p1 = LMAP (lmap, x + tx, y + ty) =
	    (p1 == CHAR_GROUND) ? CHAR_MINER : CHAR_STORED_MINER;
	  draw_cell (lmap, x, y);
	  draw_cell (lmap, x + tx, y + ty);
	  lmap->miner_x += tx;
	  lmap->miner_y += ty;
	  return cmd;
	}

      p = LMAP (lmap, x, y);
      p1 = LMAP (lmap, tx + x, ty + y);
      p2 = LMAP (lmap, 2 * tx + x, 2 * ty + y);

      if ((p1 == CHAR_STORED_BOX || p1 == CHAR_BOX) &&
	  (p2 == CHAR_GROUND || p2 == CHAR_STORE))
	{
	  p2 = LMAP (lmap, x + 2 * tx, y + 2 * ty) =
	    (p2 == CHAR_STORE) ? CHAR_STORED_BOX : CHAR_BOX;
	  p1 = LMAP (lmap, x + tx, y + ty) =
	    (p1 == CHAR_BOX) ? CHAR_MINER : CHAR_STORED_MINER;
	  p = LMAP (lmap, x, y) =
	    (p == CHAR_MINER) ? CHAR_GROUND : CHAR_STORE;
	  if ((p1 == CHAR_STORED_MINER) ^ (p2 == CHAR_STORED_BOX))
	    if (p2 == CHAR_STORED_BOX)
	      lmap->stored_boxes++;
	    else
	      lmap->stored_boxes--;
	  draw_cell (lmap, x, y);
	  draw_cell (lmap, x + tx, y + ty);
	  draw_cell (lmap, x + 2 * tx, y + 2 * ty);
	  lmap->miner_x += tx;
	  lmap->miner_y += ty;
	  return cmd | CMD_PUSH;
	}
      return CMD_NOTHING;
    }

/* CMD_UNDO */
  tx *= -1;
  ty *= -1;

  p2 = LMAP (lmap, x - tx, y - ty);
  p = LMAP (lmap, x, y);
  p1 = LMAP (lmap, x + tx, y + ty);

  if (p1 != CHAR_GROUND && p1 != CHAR_STORE)
    return CMD_NOTHING;

  if ((p2 == CHAR_BOX || p2 == CHAR_STORED_BOX) && (cmd & CMD_PUSH))
    {
      p2 = LMAP (lmap, x - tx, y - ty) =
	(p2 == CHAR_BOX) ? CHAR_GROUND : CHAR_STORE;
      p = LMAP (lmap, x, y) = (p == CHAR_MINER) ? CHAR_BOX : CHAR_STORED_BOX;
      p1 = LMAP (lmap, x + tx, y + ty) =
	(p1 == CHAR_GROUND) ? CHAR_MINER : CHAR_STORED_MINER;
      if ((p == CHAR_STORED_BOX) ^ (p2 == CHAR_STORE))
	if (p2 == CHAR_STORE)
	  lmap->stored_boxes--;
	else
	  lmap->stored_boxes++;
      draw_cell (lmap, x - tx, y - ty);
      draw_cell (lmap, x, y);
      draw_cell (lmap, x + tx, y + ty);
      lmap->miner_x += tx;
      lmap->miner_y += ty;
      return (cmd | CMD_PULL) & ~CMD_UNDO;
    }

  p = LMAP (lmap, x, y) = (p == CHAR_MINER) ? CHAR_GROUND : CHAR_STORE;
  p1 = LMAP (lmap, x + tx, y + ty) =
    (p1 == CHAR_GROUND) ? CHAR_MINER : CHAR_STORED_MINER;
  draw_cell (lmap, x, y);
  draw_cell (lmap, x + tx, y + ty);
  lmap->miner_x += tx;
  lmap->miner_y += ty;
  return cmd & ~CMD_UNDO;
}

void
restore_level (struct lmap_st *dst, struct lmap_st *src)
{
  dst->boxes = src->boxes;
  dst->stored_boxes = src->stored_boxes;
  dst->miner_x = src->miner_x;
  dst->miner_y = src->miner_y;
  dst->size_x = src->size_x;
  dst->size_y = src->size_y;
  memcpy (dst->data, src->data, MAP_X * MAP_Y);
}

int
clone_level (struct lmap_st **dst, struct lmap_st *src)
{
  *dst = new_lmap ();
  if (!dst)
    return -1;
  restore_level (*dst, src);
  return 0;
}

int
run_game (int c_level)
{
  struct lmap_st *backup, *lmap;
  char history[HISTORY_SIZE];
  int x, y, key, history_index, undo_index, i, cmd;
  int nr_level, nr_moves = 0, nr_pushes = 0, d_level;
  FILE *level_file;
  /*  time_t stime; */

  level_file = fopen (LEVEL_FILE, "r");
  if (!level_file)
    {
      printf ("Cannot open %s. (errno=%d)\n", LEVEL_FILE, errno);
      return;
    }

/*  nr_level = count_levels (level_file);*/
  nr_level = 358;

  clrscr ();

  lmap = new_lmap ();
  if (!lmap)
    {
      printf ("Not enough memory!\n");
      return;
    }

  if (c_level >= nr_level)
    c_level = nr_level - 1;

  d_level = c_level + 1;
  if (d_level >= nr_level)
    d_level = 0;

  load_level (lmap, level_file, c_level);
  clone_level (&backup, lmap);

  history_index = 0;
  undo_index = 0;

  draw_level (lmap);

/*  time (&stime);      *//* phantom cannot handle time yet */

  do
    {
/* draw info */
      gotoxy (0, 0);		/* wow, [1,1] for dos... */
      textattr (A_DEFAULT);
      cprintf ("   lv: ");
      textattr (A_HILITE);
      cprintf ("%02d", c_level);
      textattr (A_DEFAULT);
      cprintf ("   dlv: ");
      textattr (A_HILITE);
      cprintf ("%02d", d_level);
      textattr (A_DEFAULT);
      cprintf ("   mv: ");
      textattr (A_HILITE);
      cprintf ("%04d", nr_moves);
      textattr (A_DEFAULT);
      cprintf ("   ps: ");
      textattr (A_HILITE);
      cprintf ("%04d", nr_pushes);
      textattr (A_DEFAULT);
      cprintf ("   tm: ");
      textattr (A_HILITE);
      cprintf ( /*"%06d",time (NULL)- stime */ "N/A");
      textattr (A_DEFAULT);

      x = 0;
      y = 0;

      key = getch ();

      switch (key)
	{
	case 0x1B:		/* again */
	  restore_level (lmap, backup);
	  draw_level (lmap);
	  nr_moves = 0;
	  history_index = 0;
	  undo_index = 0;
	  break;
	case ',':		/* previuos level */
	  d_level--;
	  if (d_level < 0)
	    d_level = 0;
	  break;
	case '.':		/* next level */
	  d_level++;
	  if (d_level >= nr_level)
	    d_level = nr_level - 1;
	  break;
	case 'g':		/* go to desired level */
	  c_level = d_level;
	  d_level++;
	  if (d_level >= nr_level)
	    d_level = 0;

	reload:
	  load_level (lmap, level_file, c_level);
	  restore_level (backup, lmap);
	  history_index = 0;
	  undo_index = 0;
	  nr_moves = 0;
	  nr_pushes = 0;
	  clrscr ();
	  draw_level (lmap);
	  break;
	case 3:
	case 'o':
	  x--;
	  break;
	case 4:
	case 'p':
	  x++;
	  break;
	case 1:
	case 'q':
	  y--;
	  break;
	case 2:
	case 'a':
	  y++;
	  break;
	case 'f':		/* fork game :) */
	  if(!fork())
	    {
	      close(0);
	      close(1);
	      close(2);
	      open("/dev/tty1", 0);
	      dup(0);	    	    
	      dup(0);	    	    	      
	      clrscr ();
	      draw_level (lmap);
	    }
	  break;	
	case 'u':		/* undo */
	  if (undo_index)
	    {
	      undo_index--;
	      i = do_move (lmap, history[undo_index] | CMD_UNDO);
	      if (i != CMD_NOTHING)
		{
		  nr_moves--;
		  if (i & CMD_PULL)
		    nr_pushes--;
		}
	    }
	  break;

	case 'r':		/* redo */
	  if (undo_index < history_index)
	    {
	      i = do_move (lmap, history[undo_index]);
	      undo_index++;
	      if (i != CMD_NOTHING)
		{
		  nr_moves++;
		  if (i & CMD_PUSH)
		    nr_pushes++;
		}
	    }
	}

      if (x == 1 || x == -1 || y == 1 || y == -1)
	{
	  history_index = undo_index;

	  if (x == 1)
	    cmd = CMD_RIGHT;
	  else if (x == -1)
	    cmd = CMD_LEFT;
	  else if (y == 1)
	    cmd = CMD_DOWN;
	  else if (y == -1)
	    cmd = CMD_UP;

	  i = do_move (lmap, cmd);

	  if (i != CMD_NOTHING)
	    {
	      history[undo_index] = i;
	      undo_index++;
	      nr_moves++;
	      if (i & CMD_PUSH)
		nr_pushes++;
	    }

	  if (undo_index >= HISTORY_SIZE)
	    {
	      memmove ((void *) history + (HISTORY_SIZE >> 1),
		       (void *) history, HISTORY_SIZE >> 1);
	      undo_index = history_index = (HISTORY_SIZE >> 1);
	    }
	}

      if (lmap->boxes == lmap->stored_boxes)
	{
	  c_level = d_level;
	  if (c_level >= nr_level)
	    c_level = 0;
	  if (d_level >= nr_level)
	    d_level = 0;
	  goto reload;
	}
    }
  while (key != 'x');

  kill_lmap (backup);
}
/* this is too slow with phantom and floppy... */
/*
int
count_levels (FILE * f)
{
  int i, nr = 0;

  rewind (f);
  while ((i = fgetc (f)) != EOF)
    if (!i)
      nr++;
  rewind (f);

  return nr;
}
*/

int
load_level (struct lmap_st *lmap, FILE * f, int skip)
{
  int x, y, c, fl_clear = 1;

  rewind (f);
  lmap->nr = 0;

  while (skip && (c = fgetc (f)) != EOF)
    if (!c)
      {
	skip--;
	lmap->nr++;
      }

  if (c == EOF)
    return -1;

  bzero ((void *) lmap->data, MAP_X * MAP_Y);
  lmap->miner_x = -1;
  lmap->miner_y = -1;
  lmap->size_x = 0;
  lmap->size_y = 0;
  lmap->boxes = 0;
  lmap->stored_boxes = 0;
  x = 0;
  y = 0;
  while ((c = fgetc (f)) != EOF)
    {
      if (!c)
	break;

      switch (c)
	{
	case 0x0A:
	  if (x > lmap->size_x)
	    lmap->size_x = x;
	  x = 0;
	  if (!fl_clear)
	    {
	      y++;
	      lmap->size_y++;
	    }
	  break;

	case CHAR_BOX:
	  lmap->boxes++;
	  goto add_byte;

	case CHAR_STORED_BOX:
	  lmap->boxes++;
	  lmap->stored_boxes++;
	  goto add_byte;

	case CHAR_MINER:
	case CHAR_STORED_MINER:
	  lmap->miner_x = x;
	  lmap->miner_y = y;
	  goto add_byte;

	case CHAR_WALL:
	case CHAR_GROUND:
	case CHAR_STORE:

	add_byte:
	  fl_clear = 0;
	  lmap->data[x + y * MAP_X] = c;
	  x++;
	  break;
	}
    }
  lmap->size_x++;
  return 0;
}

void
die (int eco, char *str)
{
  printf (str);
  exit (eco);
}

int
main ()
{
  int level_nr = 0;

  run_game (level_nr);
  printf ("\nThanx for playing.\n");

  return 0;
}
