/*
   TODO:
   add alphabetical sorting to var_queue
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#define DEBUG

#define NR_ARG		0x10
#define VARNAMELEN	0x10

char current_path[PATH_MAX];

struct command
{
  char *name;
  int (*cmd) (int, char **);
};

int cmd_echo (int argc, char **argv);
int cmd_cat (int argc, char **argv);
int cmd_ls (int argc, char **argv);
int cmd_pwd (int argc, char **argv);
int cmd_remdir (int, char **);
int cmd_makedir (int, char **);
int cmd_rm (int, char **);
int cmd_cd (int, char **);

/*extern char **environ; */

struct var
{
  char name[VARNAMELEN];
  int a;
  struct var *next;
  struct var *prev;
};

int fl_quit = 0;
struct var *var_queue = NULL;

struct var *
var_gain (char *name)
{
  struct var *tmp = var_queue;

  printf ("gaining %s\n", name);

  while (tmp && strncmp (tmp->name, name, VARNAMELEN))
    tmp = tmp->next;

  if (!tmp)			/* var does not exist, create... */
    {
      printf ("spawning %s\n", name);

      tmp = (struct var *) malloc (sizeof (struct var));
      bzero (tmp, sizeof (struct var));

      strncpy (tmp->name, name, VARNAMELEN);
      if (var_queue)
	{
	  tmp->next = var_queue;
	  var_queue->prev = tmp;
	}
      var_queue = tmp;
    }

  return tmp;
}

/* kill all variables /if any/ */
void
var_killall ()
{
  struct var *tmp0, *tmp = var_queue;

  while (tmp)
    {
      tmp0 = tmp->next;
      free (tmp);
      tmp = tmp0;
    }
}

int
cmd_set (int argc, char **argv)
{
  struct var *xvar;

  switch (argc)
    {
    case 1:			/* only typeout */
      xvar = var_queue;
      while (xvar)
	{
	  printf ("%s=%d\n", xvar->name, xvar->a);
	  xvar = xvar->next;
	}
      break;

    case 3:			/* a = b ; set a 3 */
      if (argv[1][0] == '=' && !argv[1][1])
	xvar = var_gain (argv[0]);

      if (!xvar)
	{
	  printf ("var_gain fails!\n");
	  return;
	}

      xvar->a = atoi (argv[2]);
    }

  return 0;
}

int
cmd_unset (int argc, char **argv)
{
  struct var *xvar = var_queue;

  while (xvar && strcmp (xvar->name, argv[1]))
    xvar = xvar->next;

  if (!xvar)
    return;

  if (xvar->prev)
    xvar->prev->next = xvar->next;
  if (xvar->next)
    xvar->next->prev = xvar->prev;

  return 0;
}

int
cmd_bye (int argc, char **argv)
{
  fl_quit = 1;
  return 0;
}

int
cmd_cd (int argc, char **argv)
{
  if (argc == 2 && !chdir (argv[1]))
    {
      getcwd (current_path, PATH_MAX);
      return 0;
    }

  printf ("cannot chdir, errno=%d\n", errno);

  return 1;
}

struct command commands[] = {
  {"set", cmd_set},
  {"echo", cmd_echo},
  {"unset", cmd_unset},
  {"cat", cmd_cat},
  {"ls", cmd_ls},
  {"bye", cmd_bye},
  {"pwd", cmd_pwd},
  {"rmdir", cmd_remdir},
  {"mkdir", cmd_makedir},
  {"rm", cmd_rm},
  {"cd", cmd_cd},
  {NULL, NULL}
};

void
find_command (int argc, char **argv)
{
  int i, pid;
  struct command *cmd = commands;

/*  for (i = 0; i < argc; i++)
   printf ("\"%s\" ", argv[i]);

   printf ("\n"); */

  if (argc == 3 && argv[1][0] == '=' && !argv[1][1])
    {
      cmd_set (argc, argv);
      return;
    }

  while (cmd->name)
    {
      if (!strcmp (cmd->name, argv[0]))
	{
	  cmd->cmd (argc, argv);
	  return;
	}
      cmd++;
    }

/* command not found, try to execute... */
  switch (pid = fork ())
    {
    case 0:			/* child */
      if (execve (argv[0], argv, /*environ */ NULL) == -1)
	{
	  printf ("Bad command or file name (errno=%d)\n", errno);
	  exit (0);
	}
      break;

    case -1:
      printf ("sh: unable to execute\n");
      break;

    default:
      printf ("sh: waiting for pid %d to die\n", pid);
      waitpid (pid, NULL, 0);
      printf ("sh: child pid %d, already dead, can continue now\n", pid);
      break;
    }
}

int
do_command (char *str)
{
  char orig_cmd_str[1024];
  char *heads[NR_ARG] = { NULL, };
  char *cmd_str = orig_cmd_str;
  int i, arg_nr, fl_trail;

  arg_nr = 0;
  fl_trail = 1;

  for (;;)
    {
      switch (*str)
	{
	case ';':
	  str++;		/* skip me */

	  arg_nr++;
	  *cmd_str++ = 0;	/* terminate last arg */

	  find_command (arg_nr, heads);

	  cmd_str = orig_cmd_str;
	  heads[0] = NULL;

	  if (fl_quit)
	    return;

	  arg_nr = 0;
	  fl_trail = 1;
	  break;

	case ':':		/* end of label */
/*        if (!arg_nr) */

	  break;

	case '=':
	  if (arg_nr)
	    goto ordinary;

	  str++;
	  *cmd_str++ = 0;	/* next command */
	  arg_nr++;
	  heads[arg_nr] = cmd_str;
	  *cmd_str++ = '=';
	  *cmd_str++ = 0;
	  arg_nr++;
	  heads[arg_nr] = NULL;
	  break;

	case ' ':		/* next argument */
	case 0xA:
	  str++;
	  if (!fl_trail)
	    {
	      *cmd_str++ = 0;
	      arg_nr++;
	      heads[arg_nr] = NULL;
	    }

	  fl_trail = 1;
	  break;

	case 0:
	  if (heads[0])
	    {
	      if (heads[arg_nr])
		{
		  arg_nr++;
		  *cmd_str++ = 0;
		}
	      find_command (arg_nr, heads);
	    }
	  return;

	default:
	ordinary:
	  fl_trail = 0;
	  if (!heads[arg_nr])
	    heads[arg_nr] = cmd_str;
	  *cmd_str++ = *str++;
	  break;
	}
    }
}

int
main ()
{
  char prompt[128];

  printf ("phantom Z shell 1.0\n");

  getcwd (current_path, PATH_MAX);

  while (!fl_quit)
    {
      printf ("%s>", current_path);
      gets (prompt);
      do_command (prompt);
    }

  var_killall ();

  return 0;
}
