#include <stdio.h>
#include <stdarg.h>

#define is_digit(c)	((c)>='0'&&(c)<='9')
/*
   skip_atoi
   converts string to integer (unsigned), updates position
 */
int 
skip_atoi (const char **s)
{
  int i = 0;
  while (is_digit (**s))
    i = i * 10 + *((*s)++) - '0';
  return i;
}

#define ZEROPAD	1		/* pad with zero */
#define SIGN	2		/* unsigned/signed long */
#define PLUS	4		/* show plus */
#define SPACE	8		/* space if plus */
#define LEFT	16		/* left justified */
#define SPECIAL	32		/* 0x */
#define LARGE	64		/* use 'ABCDEF' instead of 'abcdef' */

char *
number (char *str, long num, int base, int size, int precision, int type)
{
  char c, sign, tmp[66];
  const char *digits = "0123456789abcdefghijklmnopqrstuvwxyz";
  int i;
  if (type & LARGE)
    digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  if (type & LEFT)
    type &= ~ZEROPAD;
  if (base < 2 || base > 36)
    return(0);
  c = (type & ZEROPAD) ? '0' : ' ';
  sign = 0;
  if (type & SIGN)
    if (num < 0)
      {
      sign = '-';
      num= -num;
      size--;
      }
    else if (type & PLUS)
      sign = '+';
    else if (type & SPACE)
      sign = ' ';
  if (type & SPECIAL)
    {
      if (base == 16)
	size -= 2;
      else if (base == 8)
	size--;
    }

  i = 0;
  if (num == 0)
    tmp[i++] = '0';
  else
    while (num != 0)
      {
          tmp[i++] = digits[(unsigned long) num % (unsigned) base];
	            (unsigned long) num /= (unsigned) base;
      }
      
  if (i > precision)
    precision = i;
  size -= precision;
  if (!(type & (ZEROPAD + LEFT)))
    while ((size--) > 0)
      *str++ = ' ';
  if (sign)
    *str++ = sign;
  if (type & SPECIAL)
    if (base == 8)
      *str++ = '0';
    else if (base == 16)
      {
	*str++ = '0';
	*str++ = digits[33];
      }
  if (!(type & LEFT))
    while ((size--) > 0)
      *str++ = c;
  while (i < precision--)
    *str++ = '0';
  while ((i--) > 0)
    *str++ = tmp[i];
  while ((size--) > 0)
    *str++ = ' ';
  return str;
}

int 
vsprintf (char *buf, const char *fmt, va_list args)
{
  int len;
  unsigned long num;
  int i, base;
  char *str;
  const char *s;

  int flags;			/* flags to number() */
  int field_width;		/* width of output field */
  int precision;		/* min. # of digits for integers; max
				   number of chars for from string */
  int qualifier;		/* 'h', 'l', or 'L' for integer fields */

  for (str = buf; *fmt; ++fmt)
    {
      if (*fmt != '%')
	{
	  *str++ = *fmt;
	  continue;
	}

/* process flags */
      flags = 0;
    repeat:
      ++fmt;			/* this also skips first '%' */
      switch (*fmt)
	{
	case '-':
	  flags |= LEFT;
	  goto repeat;
	case '+':
	  flags |= PLUS;
	  goto repeat;
	case ' ':
	  flags |= SPACE;
	  goto repeat;
	case '#':
	  flags |= SPECIAL;
	  goto repeat;
	case '0':
	  flags |= ZEROPAD;
	  goto repeat;
	}

/* get field width */
      field_width = -1;
      if (is_digit (*fmt))
	field_width = skip_atoi (&fmt);
      else if (*fmt == '*')
	{
	  ++fmt;
	  field_width = (int) va_arg (args, int);
	  if (field_width < 0)
	    {
	      field_width *= -1;
	      flags |= LEFT;
	    }
	}
/* get the precision */
      precision = -1;
      if (*fmt == '.')
	{
	  ++fmt;
	  if (is_digit (*fmt))
	    precision = skip_atoi (&fmt);
	  else
	    {
	      ++fmt;
	      precision = va_arg (args, int);
	    }
	  if (precision < 0)
	    precision = 0;
	}
      /* get the conversion qualifier */
      qualifier = -1;
      if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L')
	{
	  qualifier = *fmt;
	  ++fmt;
	}

      base = 10;

      switch (*fmt)		/* determine argument's type */
	{
	case 'c':		/* single character */
	  if (!(flags & LEFT))
	    while ((--field_width) > 0)
	      *str++ = ' ';
	  *str++ = (unsigned char) va_arg (args, unsigned int/*char*/);
	  while ((--field_width) > 0)
	    *str++ = ' ';
	  continue;

	case 's':		/* string */
	  s = va_arg (args, char *);
	  if (!s)
	    s = "<NULL>";
	  len = strnlen (s, precision);
	  if (!(flags & LEFT))
	    while (len < field_width--)
	      *str++ = ' ';
	  for (i = 0; i < len; ++i)
	    *str++ = *s++;
	  while (len < field_width--)
	    *str++ = ' ';
	  continue;

	case 'p':		/* pointer */
	  if (field_width == -1)
	    {
	      field_width = 2 * sizeof (void *);
	      flags |= ZEROPAD;
	    }
	  str = number (str, (unsigned long) va_arg (args, void *), 16, field_width, precision, flags);
	  continue;
	case 'n':
	  if (qualifier == 'l')
	    {
	      long *ip = va_arg (args, long *);
	      *ip = (str - buf);
	    }
	  else
	    {
	      int *ip = va_arg (args, int *);
	      *ip = (str - buf);
	    }
	  continue;
	case 'o':		/* octal value */
	  base = 8;
	  break;
	case 'X':		/* HEXA value */
	  flags |= LARGE;
	case 'x':		/* hexa value */
	  base = 16;
	  break;
	case 'd':		/* decimal signed */
	case 'i':
	  flags |= SIGN;
	case 'u':		/* decimal unsigned */
	  break;
	default:		/* standard character */
	  if (*fmt != '%')
	    *str++ = '%';
	  if (*fmt)
	    *str++ = *fmt;
	  else
	    --fmt;
	  continue;
	}
      if (qualifier == 'l')
	num = va_arg (args, unsigned long);
      else if (qualifier == 'h')
	{
	  if (flags & SIGN)
	    num = va_arg (args, /*short*/int);
	  else
	    num = va_arg (args, /*unsigned short*/ unsigned int);
	}
      else if (flags & SIGN)
	num = va_arg (args, int);
      else
	num = va_arg (args, unsigned int);
      str = number (str, num, base, field_width, precision, flags);
    }
  *str = '\0';
  return str - buf;
}
