#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>

#include "malloc.h"

//#define DEBUG

struct chunk_info *chunk_head = NULL;

#ifdef DEBUG
void
type_list ()
{
  struct chunk_info *ci;

  printf("type_list\n");
  ci = chunk_head;
  while (ci != NULL)
    {
      printf ("ci=0x%x  sz=%d  f=0x%x  p=0x%x  n=0x%x\n",
	      ci, ci->bytes, ci->flags, ci->prev, ci->next);
      ci = ci->next;
    }
  printf("EOF\n");

}
#endif

void
link_in (struct chunk_info *ci)
{
  struct chunk_info *ci2;

  if (chunk_head == NULL || ci->bytes < chunk_head->bytes)
    {				/* first position ? */
#ifdef DEBUG
      printf ("chunk 0x%x makes head\n", ci);
#endif
      ci->prev = NULL;
      ci->next = chunk_head;

      if (chunk_head != NULL)
	chunk_head->prev = ci;

      chunk_head = ci;

#ifdef DEBUG
      type_list ();
#endif

      return;
    }

  ci2 = chunk_head;
  while (ci2->next != NULL && ci->bytes > ci2->next->bytes)
    ci2 = ci2->next;

  ci->next = ci2->next;
  ci->prev = ci2;

  ci2->next = ci;

  if (ci->next != NULL)
    ci->next->prev = ci;

#ifdef DEBUG
  printf ("chunk 0x%x placed in queue\n", ci);
  type_list ();
#endif
}

void
unlink (struct chunk_info *ci)
{
  if (ci == chunk_head)
    chunk_head = ci->next;

  if (ci->prev != NULL)
    ci->prev->next = ci->next;

  if (ci->next != NULL)
    ci->next->prev = ci->prev;

#ifdef DEBUG
  type_list ();
#endif
}

void *
malloc (size_t size)
{
  int i, pages;
  void *ptr;
  struct page_info *pi;
  struct chunk_info *ci, *ci2;

  size = CHUNKIFY (size);

#ifdef DEBUG
  printf ("malloc: %d bts to find (chunkified), chunk_head=0x%x\n",
	  size, chunk_head);
#endif

  if (size <= MAXPERPAGEALLOC)
    {
      ci = chunk_head;

      while (ci != NULL && ci->bytes < size)
	ci = ci->next;

/* Desired free chunk was not found. */
      if (ci == NULL)
	{
	  if ((ptr = user_get_pages (1)) == NULL)
	    {
#ifdef DEBUG
	      printf ("malloc: oom\n");
#endif
	      return NULL;
	    }

#ifdef DEBUG
	  printf ("user_get_pages(1)=0x%x\n", ptr);
#endif

	  pi = (struct page_info *) ptr;
	  ci = (struct chunk_info *) (ptr + PIS);

	  pi->pages = 1;
	  pi->bytes = PAGESIZE - PIS - CIS - size;

/* page's second chunk splitter */
	  ci->bytes = size;
	  ci->flags = CHUNK_USED;

	  /* set second chunk */
	  ci2 = NEXT_CHUNK (ci, size);
	  ci2->bytes = pi->bytes - CIS;
	  ci2->flags = CHUNK_LAST | CHUNK_FREE;

	  link_in (ci2);

	  return ci + 1;
	}

/* Free chunk 'ci' was found. */
      unlink (ci);

      if (ci->bytes - size >= CIS + CHUNKSIZE)
	{
#ifdef DEBUG
	  printf ("chunk splitter..., ci=0x%x\n", ci);
#endif
	  ci2 = NEXT_CHUNK (ci, size);
	  ci2->bytes = ci->bytes - size - CIS;
	  ci2->flags = (ci->flags & LAST_MASK) | CHUNK_FREE;

	  link_in (ci2);

	  ci->flags = CHUNK_USED;
	}
      else
	ci->flags = (ci->flags & LAST_MASK) | CHUNK_USED;

      pi = GETPI (ci);

      if (pi->bytes < size + CIS)
	{
	  printf ("malloc: corruption\n");
#ifdef DEBUG
	printf("pi->bytes=%d, size+CIS=%d\n", pi->bytes, size+CIS);
#endif
	  return NULL;
	}

      pi->bytes -= size + CIS;
      ci->bytes = size;

#ifdef DEBUG
      printf ("malloc: (2) returning 0x%p\n", ci + 1);
#endif
      return ci + 1;
    }

/* Desired size > MAXPERPAGEALLOC, we have to alloc whole pages. */
  pages = PAGIFY (size);
  if ((pi = (struct page_info *) user_get_pages (pages)) == NULL)
    {
#ifdef DEBUG
      printf ("malloc: cannot get %d pages from kernel...\n", pages);
#endif
      return NULL;
    }

  pi->pages = pages;
  pi->bytes = 0;

#ifdef DEBUG
  printf ("malloc: big size, returning 0x%p\n", pi + 1);
#endif

  return pi + 1;
}
