#include <string.h>
#include <errno.h>

#define RET(err_nr, err_str)	case err_nr: return err_str;

char *
strerror (int errnum)
{
  switch (errnum)
    {
      RET (EPERM, "Operation not permitted");
      RET (ENOENT, "No such file or directory");
      RET (ESRCH, "No such process");
      RET (EINTR, "Interrupted system call");
      RET (EIO, "I/O Error");
      RET (ENXIO, "No such device or address");
      RET (E2BIG, "Arg list too long");
      RET (ENOEXEC, "Exec format error");
      RET (EBADF, "Bad file number");
      RET (ECHILD, "No child process");
      RET (EAGAIN, "Try again");
      RET (ENOMEM, "Out of memory");
      RET (EACCES, "Permisson denied");
      RET (EFAULT, "Bad address");
      RET (ENOTBLK, "Block device not required");
      RET (EBUSY, "Device or resource busy");
      RET (EEXIST, "File exists");
      RET (EXDEV, "Cross-device link");
      RET (ENODEV, "No such device");
      RET (ENOTDIR, "Not a directory");
      RET (EISDIR, "Is a directory");
      RET (EINVAL, "Invalid argument");
      RET (ENFILE, "File table overflow");
      RET (EMFILE, "Too many open files");
      RET (ENOTTY, "Not a typewriter");
      RET (ETXTBSY, "Text file busy");
      RET (EFBIG, "File too large");
      RET (ENOSPC, "No space left on device");
      RET (ESPIPE, "Illegal seek");
      RET (EROFS, "Read-only file system");
      RET (EMLINK, "Too many links");
      RET (EPIPE, "Broken pipe");
      RET (EDOM, "Math argument out of domain of func");
      RET (ERANGE, "Math result not representable");
      RET (EDEADLK, "Resource deadlock would occur");
      RET (ENAMETOOLONG, "File name too long");
      RET (ENOLCK, "No record locks available");
      RET (ENOSYS, "Function not implemented");
      RET (ENOTEMPTY, "Directory not empty");
      RET (ELOOP, "Too many symbolic links encountered");
/*      RET (EWOULDBLOCK, "Operation would block");*/ /* same as EAGAIN */
      RET (ENOMSG, "No message of desired type");
      RET (EIDRM, "Identifier removed");
      RET (ECHRNG, "Channel number out of range");
      RET (EL2NSYNC, "Level 2 not synchronized");
      RET (EL3HLT, "Level 3 halted");
      RET (EL3RST, "Level 3 reset");
      RET (ELNRNG, "Link number out of range");
      RET (EUNATCH, "Protocol driver not attached");
      RET (ENOCSI, "No CSI structure available");
      RET (EL2HLT, "Level 2 halted");
      RET (EBADE, "Invalid exchange");
      RET (EBADR, "Invalid request descriptor");
      RET (EXFULL, "Exchange full");
      RET (ENOANO, "No anode");
      RET (EBADRQC, "Invalid request code");
      RET (EBADSLT, "Invalid slot");
      RET (EBFONT, "Bad font file format");
      RET (ENOSTR, "Device ot a stream");
      RET (ENODATA, "No data available");
      RET (ETIME, "Timer expired");
      RET (ENOSR, "Out of stream resources");
      RET (ENONET, "Machine is not on the network");
      RET (ENOPKG, "Package is not installed");
      RET (EREMOTE, "Object is remote");
      RET (ENOLINK, "Link has been severed");
      RET (EADV, "Advertise error");
      RET (ESRMNT, "Srmount error");
      RET (ECOMM, "Communication error on send");
      RET (EPROTO, "Protocol error");
      RET (EMULTIHOP, "Multihop attempted");
      RET (EDOTDOT, "RFS specific error");
      RET (EBADMSG, "Not a data message");
      RET (EOVERFLOW, "Value too large for defined data type");
      RET (ENOTUNIQ, "Name not unique on network");
      RET (EBADFD, "File descriptor in bad state");
      RET (EREMCHG, "Remote access changed");
      RET (ELIBACC, "Can not access a needed shared library");
      RET (ELIBBAD, "Accessing a corrupted shared library");
      RET (ELIBMAX, "Attempting to link in too many shared libraries");
      RET (ELIBEXEC, "Cannot exec a shared library directly");
      RET (EILSEQ, "Illegal byte sequence");
      RET (ERESTART, "Interrupted system call should be restarted");
      RET (ESTRPIPE, "Streams pipe error");
      RET (EUSERS, "Too many users");
      RET (ENOTSOCK, "Socket operation on non-socket");
      RET (EDESTADDRREQ, "Destination address required");
      RET (EMSGSIZE, "Message too long");
      RET (EPROTOTYPE, "Protocol wrong type for socket");
      RET (ENOPROTOOPT, "Protocol not available");
      RET (EPROTONOSUPPORT, "Protocol not supported");
      RET (ESOCKTNOSUPPORT, "Socket type not supported");
      RET (EOPNOTSUPP, "Operation not supported on transport endpoint");
      RET (EPFNOSUPPORT, "Protocol family not supported");
      RET (EAFNOSUPPORT, "Address family not supported by protocol");
      RET (EADDRINUSE, "Address already in use");
      RET (EADDRNOTAVAIL, "Cannot assign requested address");
      RET (ENETDOWN, "Network is down");
      RET (ENETUNREACH, "Network unreachable");
      RET (ENETRESET, "Network dropped connection because of reset");
      RET (ECONNABORTED, "Software caused connection abort");
      RET (ECONNRESET, "Connection reset by peer");
      RET (ENOBUFS, "No buffer space available");
      RET (EISCONN, "Transport endpoint is already connected");
      RET (ENOTCONN, "Transport endpoint is not connected");
      RET (ESHUTDOWN, "Cannot send after transport endpoint shutdown");
      RET (ETOOMANYREFS, "Too man references: cannot splice");
      RET (ETIMEDOUT, "Connection timed out");
      RET (ECONNREFUSED, "Connection refused");
      RET (EHOSTDOWN, "Host is down");
      RET (EHOSTUNREACH, "No route to host");
      RET (EALREADY, "Operation already in progress");
      RET (EINPROGRESS, "Operation now in progress");
      RET (ESTALE, "Stale NFS file handle");
      RET (EUCLEAN, "Structure needs cleaning");
      RET (EREMOTEIO, "Remote I/O error");
      RET (EDQUOT, "Quota exceeded");
      RET (ENOMEDIUM, "No medium found");
      RET (EMEDIUMTYPE, "Wrong medium type");

    default:
      return "Unknown error";
    }
}
