#include <string.h>
#include <twins/twins.h>

void
edit_draw (CELL * cell, EDIT * e)
{
  FORM *f = cell->form;
  WINDOW *w = cell->window;

  rg_box (w, cell->x, cell->y, e->lx, e->ly, ATTR_EDIT);

  rg_boxprintf (w, cell->x, cell->y, e->lx, e->ly,
		e->fl_virgin ? ATTR_VIRGINEDIT : ATTR_EDIT, e->text);

  if (cell->activated)
    {
      rg_setcursor (w, cell->x + (e->pos % e->lx), cell->y + e->pos / e->lx);
      if (e->overwrite)
	rg_showbigcursor (w);
      else
	rg_showcursor (w);
    }
  else
    rg_hidecursor (w);
}

void
edit_dokey (CELL * cell, EDIT * e, int xkey, int key)
{
  int i, length;
  FORM *f = cell->form;
  WINDOW *w = cell->window;

  switch (xkey)
    {
    case K_INSERT:

      if (e->overwrite)
	{
	  e->overwrite = 0;
	  rg_showcursor (w);
	}
      else
	{
	  e->overwrite = 1;
	  rg_showbigcursor (w);
	}
      break;

    case K_BACKSPACE:
      if (e->pos <= 0)
	break;
      e->pos--;

    case K_DELETE:
      i = e->pos;
      length = strlen (e->text);

      while (i < length)
	{
	  e->text[i] = e->text[i + 1];
	  i++;
	}

      length--;
      break;

    case K_UP:			/* arrows destroy virgin */
    case K_DOWN:
      e->fl_virgin = 0;
      break;

    case K_LEFT:
      e->fl_virgin = 0;
      if (e->pos > 0)
	e->pos--;
      break;

    case K_RIGHT:
      e->fl_virgin = 0;
      if (e->pos < strlen (e->text))
	e->pos++;
      break;

    case K_ENTER:
      do_event (f->default_cell, EVENT_ONCLICK, 0);
      break;

    default:
      if (!key)
	break;

      if (e->fl_virgin)
	{
	  e->pos = 0;
	  e->text[0] = 0;	/* destroy all the text */
	  e->fl_virgin = 0;
	}

      length = strlen (e->text);
      if (!e->overwrite)
	{
	  if (length < e->maxlength)
	    {
	      i = length + 1;	/* include zero */
	      while (i > e->pos)
		{
		  e->text[i] = e->text[i - 1];
		  i--;
		}

	      e->text[e->pos++] = key;
	    }
	}
      else
	{
	  if (e->pos < e->maxlength)
	    e->text[e->pos++] = key;
	  else if (e->pos < e->maxlength)
	    {
	      e->text[e->pos++] = key;
	      e->text[e->pos] = 0;
	    }
	}
      break;
    }
  edit_draw (cell, e);
}

void
edit_handler (CELL * cell, unsigned int event_type, unsigned int arg)
{
  EDIT *e = cell->u.edit;
  WINDOW *w = cell->window;

  switch (event_type)
    {
    case EVENT_DRAW:
      edit_draw (cell, e);
      rg_wrefresh (w);
      break;

    case EVENT_ONKEY:
      edit_dokey (cell, e, event.xkey, event.key);
      rg_wrefresh (w);
      break;

    case EVENT_ACTIVATE:
      e->fl_virgin = 1;
      edit_draw (cell, e);
      rg_wrefresh (w);
      break;

    case EVENT_DEACTIVATE:
      if (e->fl_virgin)
	{
	  e->fl_virgin = 0;
	  edit_draw (cell, e);
	  rg_wrefresh (w);
	}
    }
}
