#include <twins/twins.h>
#include <string.h>
#include <stdlib.h>

EVENT event = { 0, };		/* global program event */

u8 * double_winframe = "ɻȼ";
int fl_close = 0;

void switchcell (FORM * f, int dir);

void
form_draw (FORM * f)
{
  int i, j;
  WINDOW *w = f->window;

  if (w == NULL)
    return;

  if (f->winframe != WINFRAME_NONE)
    rg_box (w, 1, 1, f->lx - 2, f->ly - 2, ATTR_FORM);
  else
    rg_box (w, 0, 0, f->lx, f->ly, ATTR_FORM);

  if (f->winframe != WINFRAME_NONE)
    {
      rg_putch (w, 0, 0, ATTR_FORM, double_winframe[2]);
      rg_putch (w, f->lx - 1, 0, ATTR_FORM, double_winframe[3]);
      rg_putch (w, f->lx - 1, f->ly - 1, ATTR_FORM, double_winframe[5]);
      rg_putch (w, 0, f->ly - 1, ATTR_FORM, double_winframe[4]);

      for (i = 1; i < f->lx - 1; i++)
	{
	  rg_putch (w, i, 0, ATTR_FORM, double_winframe[1]);
	  rg_putch (w, i, f->ly - 1, ATTR_FORM, double_winframe[1]);
	}

      for (i = 1; i < f->ly - 1; i++)
	{
	  rg_putch (w, 0, i, ATTR_FORM, double_winframe[0]);
	  rg_putch (w, f->lx - 1, i, ATTR_FORM, double_winframe[0]);
	}
    }

  if (f->caption)
    rg_printf (w, ((f->lx - strlen (f->caption) - 1) >> 1), 0, ATTR_FORM,
	       f->caption);
}

void
default_event_handler (CELL * cell, unsigned int cmd, unsigned int arg)
{
  switch (cmd)
    {
    case EVENT_SWITCHCELL:
      if (cell->form != NULL)
	switchcell (cell->form, arg < 0 ? -1 : 1);
      break;

    case EVENT_CLOSEAPP:
      fl_close = 1;
      break;

    default:
      if (cell != NULL && cell->event_handler)
	cell->event_handler (cell, cmd, arg);
      break;
    }
}

void
switchcell (FORM * f, int dir)
{
  CELL *cell0, *cell1;

  if (!dir)
    {				/* zero -- initial switch (do not draw, only set up) */

      cell0 = f->cell_queue;

      while (cell0 && (cell0->passive || cell0->disabled))
	cell0 = cell0->next;

      f->current_cell = cell0;

      if (cell0 != NULL)
	cell0->activated = 1;

      return;
    }

  cell0 = f->current_cell;
  cell1 = cell0;

  if (!cell0)
    return;

  if (dir > 0)			/* forward */
    {
      do
	{
	  cell0 = cell0->next;
	}
      while (cell0 != NULL && (cell0->passive || cell0->disabled));
      if (cell0 == NULL)
	{
	  cell0 = f->cell_queue;
	  while (cell0 != NULL && cell0 != cell1
		 && (cell0->passive || cell0->disabled))
	    cell0 = cell0->next;
	}

      if (cell0 == NULL || cell0 == cell1)
	return;
    }
  else
    {				/* backward */
      do
	{
	  cell0 = cell0->prev;
	}
      while (cell0 != NULL && (cell0->passive || cell0->disabled));
      if (cell0 == NULL)
	{
	  cell0 = f->cell_tail;

	  while (cell0 != NULL && cell0 != cell1
		 && (cell0->passive || cell0->disabled))
	    cell0 = cell0->prev;
	}

      if (cell0 == NULL || cell0 == cell1)
	return;
    }
/* ok, new cell in cell0 */
  cell1->form = f;
  cell0->form = f;		/* just to be sure */
  cell1->activated = 0;
  cell0->activated = 1;
  do_event (cell1, EVENT_DRAW, 0);
  do_event (cell0, EVENT_DRAW, 0);

  f->current_cell = cell0;
}

void
do_event (CELL * c, int cmd, int arg)
{
  if (c != NULL && c->form != NULL)
    c->form->event_handler (c, cmd, arg);
}

void
form_handler (FORM * f, unsigned int event_type)
{
  CELL *tmp;
  WINDOW *w;

  switch (event_type)
    {
    case EVENT_REDRAW:
    case EVENT_DRAW:
      form_draw (f);
      w = f->window;

      switchcell (f, 0);	/* find first cell... */

      tmp = f->cell_queue;

      while (tmp)
	{
	  tmp->form = f;	/* important! */
	  tmp->window = w;	/* cell->form could be removed */
	  f->event_handler (tmp, EVENT_DRAW, 0);
	  tmp = tmp->next;
	}

      break;

    case EVENT_INIT:
      if (f->x == -1 || f->y == -1)	/* undefined placing */
	{
	  f->x = (SCREEN_X - f->lx) / 2;
	  f->y = (SCREEN_Y - f->ly) / 2;
	}
      f->window = rg_newwin (f->x, f->y, f->lx, f->ly);
      break;

    case EVENT_RELEASE:
/*      if (f->screen)
	free (f->screen);*/
      break;

    case EVENT_HIDE:		/* restore original screen */
/*       if(f->screen)
        restorescreen(f->screen, f->x, f->y, f->lx, f->ly);*/
      rg_hidewin (f->window);
      break;

    case EVENT_ONKEY:
      switch (event.xkey)
	{
	case K_ESC:		/* esc -- cancel */
	  do_event (f->current_cell, EVENT_ONKEY, 0);
	  do_event (f->cancel_cell, EVENT_ONCLICK, 0);
	  break;

	case K_TAB:		/* next cell */
	  switchcell (f, event.key == KB_RTAB ? -1 : 1);
	  rg_wrefresh (f->window);
	  break;

	case K_ENTER:		/* enter */
	  do_event (f->current_cell, EVENT_ONKEY, 0);
	  do_event (f->current_cell, EVENT_ONCLICK, 0);
	  break;

	default:
	  if (f->current_cell != NULL)
	    f->event_handler (f->current_cell, EVENT_ONKEY, 0);
	}
      break;
    }
}

void
showmodal (FORM * f)
{
  int ret;

  form_handler (f, EVENT_INIT);
  rg_showwin (f->window);

  form_handler (f, EVENT_DRAW);

  while (!f->fl_close && !fl_close)
    {
      if ((ret = rg_getevent (f->window, &event)) != EVENT_NONE)
	form_handler (f, ret);
    }

  form_handler (f, EVENT_HIDE);
  rg_freewin (f->window);
}
