#include <twins/twins.h>

void
table_draw (CELL * cell, TABLE * t)
{
  FORM *f = cell->form;
  WINDOW *w = cell->window;
  char *str;
  int i, j, todo, attr;

  rg_box (w, cell->x, cell->y, t->lx, t->ly, ATTR_TABLE);

  if (t->gather_data == NULL)
    return;

  todo = MIN (t->rows - t->frow, t->ly);

  for (j = 0; j < todo; j++)
    for (i = 0; i < t->cols; i++)
      {
	if (t->arow == t->frow + j)
	  {
	    if (cell->activated)
	      attr = ATTR_AHITABLE;
	    else
	      attr = ATTR_HITABLE;
	  }
	else
	  attr = ATTR_TABLE;

	str = t->gather_data (t, j + t->frow, i);
	rg_boxprintf (w, cell->x, cell->y + j, t->lx, 1, attr, str);
      }
}

void
table_handler (CELL * cell, unsigned int cmd, unsigned int arg)
{
  TABLE *t = cell->u.table;
  WINDOW *w = cell->window;

  switch (cmd)
    {
    case EVENT_ONKEY:
      switch (event.key)
	{
	case KB_UP:
	  if (t->arow > 0)
	    {
	      t->arow--;
	      if (t->frow > t->arow)
		t->frow = t->arow;
	      table_draw (cell, t);
	      rg_wrefresh (w);
	    }
	  break;

	case KB_DOWN:
	  if (t->arow < t->rows - 1)
	    {
	      t->arow++;
	      if (t->arow >= t->frow + t->ly)
		t->frow = t->arow - t->ly + 1;
	      table_draw (cell, t);
	      rg_wrefresh (w);
	    }
	  break;

	case KB_PGUP:
	  if (t->arow > 0)
	    {
	      if (t->arow - t->ly < 0)
		{
		  t->arow = 0;
		  t->frow = 0;
		}
	      else
		{
		  t->arow -= t->ly;
		  t->frow = t->arow;
		}
	      table_draw (cell, t);
	      rg_wrefresh (w);
	    }
	  break;

	case KB_PGDOWN:
	  if (t->arow < t->rows - 1)
	    {
	      if (t->arow + t->ly > t->rows - 1)
		{
		  t->arow = t->rows - 1;
		  t->frow = t->arow - t->ly + 1;
		}
	      else
		{
		  t->arow += t->ly;
		  t->frow = t->arow - t->ly + 1;
		}
	      if (t->frow < 0)
		t->frow = 0;

	      table_draw (cell, t);
	      rg_wrefresh (w);
	    }
	  break;

	case KB_HOME:
	  t->arow = 0;
	  t->frow = 0;
	  table_draw (cell, t);
	  rg_wrefresh (w);
	  break;

	case KB_END:
	  t->arow = t->rows - 1;
	  if (t->rows >= t->ly)
	    t->frow = t->arow - t->ly + 1;

	  table_draw (cell, t);
	  rg_wrefresh (w);
	  break;
	}
      break;

    case EVENT_DRAW:
      table_draw (cell, t);
      rg_wrefresh (w);
      break;

    case EVENT_RESET:
      t->arow = 0;
      t->acol = 0;
      t->frow = 0;
      t->fcol = 0;
      break;
    }
}
